Package: spanishoddata
Title: Get Spanish Origin-Destination Data
Version: 0.2.6
Authors@R: c(
    person("Egor", "Kotov", , "kotov.egor@gmail.com", role = c("aut", "cre"),
           comment = c(ORCID = "0000-0001-6690-5345")),
    person("Robin", "Lovelace", , "rob00x@gmail.com", role = "aut",
           comment = c(ORCID = "0000-0001-5679-6536")),
    person("Eugeni", "Vidal-Tortosa", role = "ctb",
           comment = c(ORCID = "0000-0001-5199-4103"))
  )
Description: Gain seamless access to origin-destination (OD) data from the
    Spanish Ministry of Transport, hosted at
    <https://www.transportes.gob.es/ministerio/proyectos-singulares/estudios-de-movilidad-con-big-data/opendata-movilidad>.
    This package simplifies the management of these large datasets by
    providing tools to download zone boundaries, handle associated
    origin-destination data, and process it efficiently with the 'duckdb'
    database interface.  Local caching minimizes repeated downloads,
    streamlining workflows for researchers and analysts.  Methods
    described in Kotov et al. (2026) <doi:10.1177/23998083251415040>.
    Extensive documentation is available at
    <https://ropenspain.github.io/spanishoddata/index.html>, offering
    guides on creating static and dynamic mobility flow visualizations and
    transforming large datasets into analysis-ready formats.
License: MIT + file LICENSE
URL: https://rOpenSpain.github.io/spanishoddata/,
        https://github.com/rOpenSpain/spanishoddata,
        https://doi.org/10.1177/23998083251415040
BugReports: https://github.com/rOpenSpain/spanishoddata/issues
Depends: R (>= 4.1.0)
Imports: checkmate, DBI, digest, dplyr, duckdb (>= 0.5.0), fs, glue,
        httr2, jsonlite, lifecycle, lubridate, memoise, openssl,
        parallelly, paws.storage (>= 0.4.0), purrr, Rdpack, readr,
        rlang, sf, stats, stringr, tibble, xml2
Suggests: cli, flowmapblue, flowmapper (>= 0.1.2), furrr, future,
        future.mirai, ggiraph, hexSticker, htmlwidgets, mapSpain,
        quarto, remotes, scales, testthat (>= 3.0.0), tidyverse, withr
VignetteBuilder: quarto
RdMacros: Rdpack
Config/Needs/website: rmarkdown
Config/testthat/edition: 3
Encoding: UTF-8
RoxygenNote: 7.3.3
NeedsCompilation: no
Packaged: 2026-06-06 20:49:48 UTC; ek
Author: Egor Kotov [aut, cre] (ORCID: <https://orcid.org/0000-0001-6690-5345>),
  Robin Lovelace [aut] (ORCID: <https://orcid.org/0000-0001-5679-6536>),
  Eugeni Vidal-Tortosa [ctb] (ORCID:
    <https://orcid.org/0000-0001-5199-4103>)
Maintainer: Egor Kotov <kotov.egor@gmail.com>
Repository: CRAN
Date/Publication: 2026-06-07 06:30:02 UTC
Built: R 4.4.3; ; 2026-06-07 10:25:05 UTC; unix

Package: spanishoddata
Title: Get Spanish Origin-Destination Data
Version: 0.2.6
Authors@R: c(
    person("Egor", "Kotov", , "kotov.egor@gmail.com", role = c("aut", "cre"),
           comment = c(ORCID = "0000-0001-6690-5345")),
    person("Robin", "Lovelace", , "rob00x@gmail.com", role = "aut",
           comment = c(ORCID = "0000-0001-5679-6536")),
    person("Eugeni", "Vidal-Tortosa", role = "ctb",
           comment = c(ORCID = "0000-0001-5199-4103"))
  )
Description: Gain seamless access to origin-destination (OD) data from the
    Spanish Ministry of Transport, hosted at
    <https://www.transportes.gob.es/ministerio/proyectos-singulares/estudios-de-movilidad-con-big-data/opendata-movilidad>.
    This package simplifies the management of these large datasets by
    providing tools to download zone boundaries, handle associated
    origin-destination data, and process it efficiently with the 'duckdb'
    database interface.  Local caching minimizes repeated downloads,
    streamlining workflows for researchers and analysts.  Methods
    described in Kotov et al. (2026) <doi:10.1177/23998083251415040>.
    Extensive documentation is available at
    <https://ropenspain.github.io/spanishoddata/index.html>, offering
    guides on creating static and dynamic mobility flow visualizations and
    transforming large datasets into analysis-ready formats.
License: MIT + file LICENSE
URL: https://rOpenSpain.github.io/spanishoddata/,
        https://github.com/rOpenSpain/spanishoddata,
        https://doi.org/10.1177/23998083251415040
BugReports: https://github.com/rOpenSpain/spanishoddata/issues
Depends: R (>= 4.1.0)
Imports: checkmate, DBI, digest, dplyr, duckdb (>= 0.5.0), fs, glue,
        httr2, jsonlite, lifecycle, lubridate, memoise, openssl,
        parallelly, paws.storage (>= 0.4.0), purrr, Rdpack, readr,
        rlang, sf, stats, stringr, tibble, xml2
Suggests: cli, flowmapblue, flowmapper (>= 0.1.2), furrr, future,
        future.mirai, ggiraph, hexSticker, htmlwidgets, mapSpain,
        quarto, remotes, scales, testthat (>= 3.0.0), tidyverse, withr
VignetteBuilder: quarto
RdMacros: Rdpack
Config/Needs/website: rmarkdown
Config/testthat/edition: 3
Encoding: UTF-8
RoxygenNote: 7.3.3
NeedsCompilation: no
Packaged: 2026-06-06 20:49:48 UTC; ek
Author: Egor Kotov [aut, cre] (ORCID: <https://orcid.org/0000-0001-6690-5345>),
  Robin Lovelace [aut] (ORCID: <https://orcid.org/0000-0001-5679-6536>),
  Eugeni Vidal-Tortosa [ctb] (ORCID:
    <https://orcid.org/0000-0001-5199-4103>)
Maintainer: Egor Kotov <kotov.egor@gmail.com>
Repository: CRAN
Date/Publication: 2026-06-07 06:30:02 UTC
Built: R 4.5.2; ; 2026-06-07 07:23:57 UTC; unix

# Generated by roxygen2: do not edit by hand

export(spod_available_data)
export(spod_check_files)
export(spod_cite)
export(spod_codebook)
export(spod_connect)
export(spod_convert)
export(spod_disconnect)
export(spod_download)
export(spod_get)
export(spod_get_data_dir)
export(spod_get_valid_dates)
export(spod_get_zones)
export(spod_quick_get_od)
export(spod_quick_get_zones)
export(spod_set_data_dir)
importFrom(Rdpack,reprompt)
importFrom(lifecycle,deprecated)
importFrom(memoise,memoise)
importFrom(rlang,.data)
importFrom(stats,end)
importFrom(stats,median)
importFrom(stats,start)
importFrom(utils,capture.output)
importFrom(utils,vignette)

# spanishoddata: Get Spanish Origin-Destination Data <a href="https://rOpenSpain.github.io/spanishoddata/"><img src="man/figures/logo.png" align="right" width="200" alt="spanishoddata website" /></a>

<!-- badges: start -->

[![Project Status:
Active](https://www.repostatus.org/badges/latest/active.svg)](https://www.repostatus.org/#active)
<a href="https://lifecycle.r-lib.org/articles/stages.html#stable"
target="_blank"><img
src="https://img.shields.io/badge/lifecycle-stable-brightgreen.svg"
alt="Lifecycle: stable" /></a>
<a href="https://CRAN.R-project.org/package=spanishoddata"
target="_blank"><img
src="https://www.r-pkg.org/badges/version/spanishoddata"
alt="CRAN status" /></a>
<a href="https://CRAN.R-project.org/package=spanishoddata"
target="_blank"><img
src="https://cranlogs.r-pkg.org/badges/grand-total/spanishoddata?color=blue"
alt="CRAN/METACRAN Total downloads" /></a>
<a href="https://CRAN.R-project.org/package=spanishoddata"
target="_blank"><img
src="https://cranlogs.r-pkg.org/badges/spanishoddata?color=blue"
alt="CRAN/METACRAN Downloads per month" /></a>
[![R-CMD-check](https://github.com/rOpenSpain/spanishoddata/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/rOpenSpain/spanishoddata/actions/workflows/R-CMD-check.yaml)
[![DOI](https://zenodo.org/badge/DOI/10.32614/CRAN.package.spanishoddata.svg)](https://doi.org/10.32614/CRAN.package.spanishoddata)
[![DOI](https://zenodo.org/badge/DOI/10.5281/zenodo.14516104.svg)](https://doi.org/10.5281/zenodo.14516104)
<!-- 10.5281/zenodo.14516104 --> <!-- badges: end -->

**spanishoddata** is an R package that provides functions for
downloading and formatting Spanish open mobility data released by the
Spanish government (Ministerio de Transportes y Movilidad Sostenible
(MITMS) 2024).

It supports the two versions of the Spanish mobility data. [The first
version (2020 to
2021)](https://www.transportes.gob.es/ministerio/proyectos-singulares/estudios-de-movilidad-con-big-data/estudios-de-movilidad-anteriores/covid-19/opendata-movilidad),
covering the period of the COVID-19 pandemic, contains tables detailing
trip numbers and distances, broken down by origin, destination,
activity, residence province, time interval, distance interval, and
date. It also provides tables of individual counts by location and trip
frequency. [The second version (2022
onwards)](https://www.transportes.gob.es/ministerio/proyectos-singulares/estudios-de-movilidad-con-big-data/opendata-movilidad)
improves spatial resolution, adds trips to and from Portugal and France,
and introduces new fields for study-related activities and
sociodemographic factors (income, age, and sex) in the
origin-destination tables, along with additional tables showing
individual counts by overnight stay location, residence, and date. See
the [package website](https://rOpenSpain.github.io/spanishoddata/) and
vignettes for
[v1](https://rOpenSpain.github.io/spanishoddata/articles/v1-2020-2021-mitma-data-codebook)
and
[v2](https://rOpenSpain.github.io/spanishoddata/articles/v2-2022-onwards-mitma-data-codebook)
data for more details.

**spanishoddata** is designed to save time by providing the data in
analysis-ready formats. Automating the process of downloading, cleaning,
and importing the data can also reduce the risk of errors in the
laborious process of data preparation. It also reduces computational
resources by using computationally efficient packages behind the scenes.
To effectively work with multiple data files, it’s recommended you set
up a data directory where the package can search for the data and
download only the files that are not already present.

Under the hood, **spanishoddata** uses the efficient
[DuckDB](https://duckdb.org/) database engine via
<a href="https://r.duckdb.org/"
target="_blank"><code>{duckdb}</code></a> R package and is featured in
the <a href="https://duckdb.org/library/spanishoddata-r-package/"
target="_blank">DuckDB Library</a>.

## Examples of available data

<div id="fig-barcelona-flows">

<img src="vignettes/media/flows_plot_barcelona.svg"
style="width:100.0%" />

Figure 1: Example of the data available through the package: daily flows
in Barcelona on 7 April 2021

</div>

To create static maps like that see our vignette
[here](https://ropenspain.github.io/spanishoddata/articles/flowmaps-static.html).

------------------------------------------------------------------------

<div id="fig-spain-flows">

![](https://ropenspain.github.io/spanishoddata/media/spain-folding-flows.gif)

Figure 2: Example of the data available through the package: interactive
daily flows in Spain

</div>

<div id="fig-spain-flows">

![](https://ropenspain.github.io/spanishoddata/media/barcelona-time.gif)

Figure 3: Example of the data available through the package: interactive
daily flows in Barcelona with time filter

</div>

To create interactive maps see our vignette
[here](https://ropenspain.github.io/spanishoddata/articles/flowmaps-interactive.html).

### Data availability

The following plot shows the dates for which origin-destination data is
currently available. It is updated weekly by a [GitHub
Action](https://github.com/rOpenSpain/spanishoddata/actions/workflows/data-availability-plot.yaml).

Latest Data Availability:

- **v1 (2020–2021)**: Data available from **14 February 2020** to **09
  May 2021** (Districts and Municipalities).
- **v2 (2022 onwards)**: Data available from **01 January 2022** onwards
  (Districts, Municipalities, and Large Urban Areas).

| Version | Live Status Badge |
|----|----|
| **v1** | ![Latest v1 data](https://img.shields.io/endpoint?url=https://rOpenSpain.github.io/spanishoddata/data-availability/latest_v1.json) |
| **v2** | ![Latest v2 data](https://img.shields.io/endpoint?url=https://rOpenSpain.github.io/spanishoddata/data-availability/latest_v2.json) |

[Open the interactive data availability
plot](https://rOpenSpain.github.io/spanishoddata/data-availability/data_availability.html).

![Dates for which origin-destination data is
available](https://rOpenSpain.github.io/spanishoddata/data-availability/data_availability.png)

## Install the package

Install from CRAN:

``` r
install.packages("spanishoddata")
```

<details>

<summary>

Alternative installation and developemnt
</summary>

You can also install the latest development version of the package from
rOpenSpain R universe:

``` r
install.packages("spanishoddata",
  repos = c("https://ropenspain.r-universe.dev",
    "https://cloud.r-project.org"))
```

Alternative way to install the development version from GitHub:

``` r
if (!require("remotes")) install.packages("remotes")

remotes::install_github("rOpenSpain/spanishoddata",
  force = TRUE, dependencies = TRUE)
```

**For Developers**

To load the package locally, clone it and navigate to the root of the
package in the terminal, e.g. with the following:

``` bash
gh repo clone rOpenSpain/spanishoddata
code spanishoddata
# with rstudio:
rstudio spanishoddata/spanishoddata.Rproj
```

Then run the following command from the R console:

``` r
devtools::load_all()
```

</details>

You can also explore the package and the data in an interactive RStudio
container right in your web browser thanks to Binder, just click [the
link](https://mybinder.org/v2/gh/e-kotov/spanishoddata-playground/HEAD?urlpath=rstudio)
or the button:
[![Binder](https://mybinder.org/badge_logo.svg)](https://mybinder.org/v2/gh/e-kotov/spanishoddata-playground/HEAD?urlpath=rstudio).
Note that the session will be limited by memory and you will only be
able to work with one full day of data.

Load it as follows:

``` r
library(spanishoddata)
```

## Set the data directory

Choose where `{spanishoddata}` should download (and convert) the data by
setting the data directory following command:

``` r
spod_set_data_dir(data_dir = "~/spanish_od_data")
```

The function above will also ensure that the directory is created and
that you have sufficient permissions to write to it.

<details>

<summary>

Setting data directory for advanced users
</summary>

You can also set the data directory with an environment variable:

``` r
Sys.setenv(SPANISH_OD_DATA_DIR = "~/spanish_od_data")
```

The package will create this directory if it does not exist on the first
run of any function that downloads the data.

To permanently set the directory for all projects, you can specify the
data directory globally by setting the `SPANISH_OD_DATA_DIR` environment
variable, e.g. with the following command:

``` r
usethis::edit_r_environ()
# Then set the data directory globally, by typing this line in the file:
```

    SPANISH_OD_DATA_DIR = "~/spanish_od_data"

You can also set the data directory locally, just for the current
project. Set the ‘envar’ in the working directory by editing `.Renviron`
file in the root of the project:

``` r
file.edit(".Renviron")
```

</details>

## Overall approach to accessing the data

If you only need flows data aggregated by day at municipal level, you
can use the `spod_quick_get_od()` function. This will download the data
directly from the web API and let you analyse it in-memory. More on this
in the [Quickly get daily
data](https://ropenspain.github.io/spanishoddata/articles/quick-get.html)
vignette.

If you only want to analyse the data for a few days, you can use the
`spod_get()` function. It will download the raw data in CSV format and
let you analyse it in-memory. This is what we cover in the steps on this
page.

If you need longer periods (several months or years), you should use the
`spod_convert()` and `spod_connect()` functions, which will convert the
data into special format which is much faster for analysis, for this see
the [Download and convert OD
datasets](https://ropenspain.github.io/spanishoddata/articles/convert.html)
vignette. `spod_get_zones()` will give you spatial data with zones that
can be matched with the origin-destination flows from the functions
above using zones ’id’s. Please see a simple example below, and also
consult the vignettes with detailed data description and instructions in
the package vignettes with `spod_codebook(ver = 1)` and
`spod_codebook(ver = 2)`, or simply visit the package website at
<https://ropenspain.github.io/spanishoddata/>. The
<a href="#fig-overall-flow" class="quarto-xref">Figure 4</a> presents
the overall approach to accessing the data in the `spanishoddata`
package.

<div id="fig-overall-flow">

<img src="man/figures/package-functions-overview.svg"
style="width:78.0%" />

Figure 4: The overview of package functions to get the data

</div>

## Showcase

To run the code in this README we will use the following setup:

``` r
library(tidyverse)
theme_set(theme_minimal())
sf::sf_use_s2(FALSE)
```

Get metadata for the datasets as follows (we are using version 2 data
covering years 2022 and onwards):

``` r
metadata <- spod_available_data(ver = 2) # for version 2 of the data
metadata
```

    # A tibble: 9,442 × 6
       target_url           pub_ts              file_extension data_ym data_ymd  
       <chr>                <dttm>              <chr>          <date>  <date>    
     1 https://movilidad-o… 2024-07-30 10:54:08 gz             NA      2022-10-23
     2 https://movilidad-o… 2024-07-30 10:51:07 gz             NA      2022-10-22
     3 https://movilidad-o… 2024-07-30 10:47:52 gz             NA      2022-10-20
     4 https://movilidad-o… 2024-07-30 10:14:55 gz             NA      2022-10-18
     5 https://movilidad-o… 2024-07-30 10:11:58 gz             NA      2022-10-17
     6 https://movilidad-o… 2024-07-30 10:09:03 gz             NA      2022-10-12
     7 https://movilidad-o… 2024-07-30 10:05:57 gz             NA      2022-10-07
     8 https://movilidad-o… 2024-07-30 10:02:12 gz             NA      2022-08-07
     9 https://movilidad-o… 2024-07-30 09:58:34 gz             NA      2022-08-06
    10 https://movilidad-o… 2024-07-30 09:54:30 gz             NA      2022-08-05
    # ℹ 9,432 more rows
    # ℹ 1 more variable: local_path <chr>

### Zones

Zones can be downloaded as follows:

``` r
distritos <- spod_get_zones("distritos", ver = 2)
distritos_wgs84 <- distritos |>
  sf::st_simplify(dTolerance = 200) |>
  sf::st_transform(4326)
plot(sf::st_geometry(distritos_wgs84), lwd = 0.2)
```

![](man/figures/README-distritos-1.png)

### OD data

``` r
od_db <- spod_get(
  type = "origin-destination",
  zones = "districts",
  dates = c(start = "2024-03-01", end = "2024-03-07")
)
class(od_db)
```

    [1] "tbl_duckdb_connection" "tbl_dbi"               "tbl_sql"              
    [4] "tbl_lazy"              "tbl"                  

``` r
colnames(od_db)
```

     [1] "full_date"                   "hour"                  
     [3] "id_origin"                   "id_destination"             
     [5] "distance"                    "activity_origin"            
     [7] "activity_destination"        "study_possible_origin"      
     [9] "study_possible_destination"  "residence_province_ine_code"
    [11] "residence_province"          "income"                     
    [13] "age"                         "sex"                        
    [15] "n_trips"                     "trips_total_length_km"      
    [17] "year"                        "month"                      
    [19] "day"                        

The result is an R database interface object (`tbl_dbi`) that can be
used with dplyr functions and SQL queries ‘lazily’, meaning that the
data is not loaded into memory until it is needed. Let’s do an
aggregation to find the total number trips per hour over the 7 days:

``` r
n_per_hour <- od_db |>
  group_by(date, hour) |>
  summarise(n = n(), Trips = sum(n_trips)) |>
  collect() |>
  mutate(Time = lubridate::ymd_h(paste0(date, hour, sep = " "))) |>
  mutate(Day = lubridate::wday(Time, label = TRUE))
n_per_hour |>
  ggplot(aes(x = Time, y = Trips)) +
  geom_line(aes(colour = Day)) +
  labs(title = "Number of trips per hour over 7 days")
```

![](man/figures/README-trips-per-hour-1.png)

The figure above summarises 925,874,012 trips over the 7 days associated
with 135,866,524 records.

### `spanishoddata` advantage over accessing the data yourself

As we demonstrated above, you can perform very quick analysis using just
a few lines of code.

To highlight the benefits of the package, here is how you would do this
manually:

- download the [xml](https://movilidad-opendata.mitma.es/RSS.xml) file
  with the download links

- parse this xml to extract the download links

- write a script to download the files and locate them on disk in a
  logical manner

- figure out the data structure of the downloaded files, read the
  codebook

- translate the data (columns and values) into English, if you are not
  familiar with Spanish

- write a script to load the data into the database or figure out a way
  to claculate summaries on multiple files

- and much more…

We did all of that for you and present you with a few simple functions
that get you straight to the data in one line of code, and you are ready
to run any analysis on it.

## Desire lines

We’ll use the same input data to pick-out the most important flows in
Spain, with a focus on longer trips for visualisation:

``` r
od_national_aggregated <- od_db |>
  group_by(id_origin, id_destination) |>
  summarise(Trips = sum(n_trips), .groups = "drop") |>
  filter(Trips > 500) |>
  collect() |>
  arrange(desc(Trips))
od_national_aggregated
```

    # A tibble: 96,404 × 3
       id_origin id_destination    Trips
       <fct>     <fct>             <dbl>
     1 2807908   2807908        2441404.
     2 0801910   0801910        2112188.
     3 0801902   0801902        2013618.
     4 2807916   2807916        1821504.
     5 2807911   2807911        1785981.
     6 04902     04902          1690606.
     7 2807913   2807913        1504484.
     8 2807910   2807910        1299586.
     9 0704004   0704004        1287122.
    10 28106     28106          1286058.
    # ℹ 96,394 more rows

The results show that the largest flows are intra-zonal. Let’s keep only
the inter-zonal flows:

``` r
od_national_interzonal <- od_national_aggregated |>
  filter(id_origin != id_destination)
```

We can convert these to geographic data with the {od} package (Lovelace
and Morgan 2024):

``` r
od_national_sf <- od::od_to_sf(
  od_national_interzonal,
  z = distritos_wgs84
)
distritos_wgs84 |>
  ggplot() +
  geom_sf(fill = "grey") +
  geom_sf(data = spData::world, fill = NA, colour = "black") +
  geom_sf(aes(linewidth = Trips), colour = "blue", data = od_national_sf) +
  coord_sf(xlim = c(-10, 5), ylim = c(35, 45)) +
  theme_void() +
  scale_linewidth_continuous(range = c(0.2, 3))
```

![](man/figures/README-desire-lines-1.png)

Let’s focus on trips in and around a particular area (Salamanca):

``` r
salamanca_zones <- zonebuilder::zb_zone("Salamanca")
distritos_salamanca <- distritos_wgs84[salamanca_zones, ]
plot(distritos_salamanca)
```

![](man/figures/README-salamanca-zones-1.png)

We will use this information to subset the rows, to capture all movement
within the study area:

``` r
ids_salamanca <- distritos_salamanca$id
od_salamanca <- od_national_sf |>
  filter(id_origin %in% ids_salamanca) |>
  filter(id_destination %in% ids_salamanca) |>
  arrange(Trips)
```

Let’s plot the results:

``` r
od_salamanca_sf <- od::od_to_sf(
  od_salamanca,
  z = distritos_salamanca
)
ggplot() +
  geom_sf(fill = "grey", data = distritos_salamanca) +
  geom_sf(aes(colour = Trips), size = 1, data = od_salamanca_sf) +
  scale_colour_viridis_c() +
  theme_void()
```

![](man/figures/README-salamanca-plot-1.png)

## Further information

For more information on the package, see:

- The [pkgdown site](https://rOpenSpain.github.io/spanishoddata/)

  - [Functions
    reference](https://rOpenSpain.github.io/spanishoddata/reference/index.html)
  - [v1 data (2020-2021)
    codebook](https://rOpenSpain.github.io/spanishoddata/articles/v1-2020-2021-mitma-data-codebook.html)
  - [v2 data (2022 onwards) codebook (work in
    progress)](https://rOpenSpain.github.io/spanishoddata/articles/v2-2022-onwards-mitma-data-codebook.html)
  - [Download and convert
    data](https://rOpenSpain.github.io/spanishoddata/articles/convert.html)
  - The [OD disaggregation
    vignette](https://rOpenSpain.github.io/spanishoddata/articles/disaggregation.html)
    showcases flows disaggregation
  - [Making static
    flowmaps](https://rOpenSpain.github.io/spanishoddata/articles/flowmaps-static.html)
    vignette shows how to create flowmaps using the data acquired with
    `{spanishoddata}`
  - [Making interactive
    flowmaps](https://rOpenSpain.github.io/spanishoddata/articles/flowmaps-interactive.html)
    shows how to create an interactive flowmap using the data acquired
    with `{spanishoddata}`
  - [Quickly getting daily aggregated 2022+ data at municipality
    level](https://ropenspain.github.io/spanishoddata/articles/quick-get.html)

- Teaching materials that use `spanishoddata`:

  - Tutorial/workshop [“Analysing massive open human mobility data using
    spanishoddata, duckdb and
    flowmaps”](https://www.ekotov.pro/agit-2025-spanishoddata/) by [Egor
    Kotov](https://github.com/e-kotov) (held at [Applied Geoinformatics
    (AGIT) Conference 2025](https://agit.at/en/program/), Salzburg,
    Austria)
    [![DOI:10.5281/zenodo.15794849](https://zenodo.org/badge/DOI/10.5281/zenodo.15794849.svg)](https://doi.org/10.5281/zenodo.15794849)

  - Tutorial [“Mobility Flows and Accessibility Using R and Big Open
    Data”](https://www.ekotov.pro/spanish-open-mobility-workshop-ic2s2-2025/)
    by [Egor Kotov](https://github.com/e-kotov) and [Johannes
    Mast](https://github.com/JohMast) (held at [IC2S2
    2025](https://web.archive.org/web/20260113100927/https://ic2s2-2025.org/)
    (11th International Conference on Computational Social Science),
    Norrköping, Sweden)
    [![DOI:10.5281/zenodo.16238597](https://zenodo.org/badge/DOI/10.5281/zenodo.16238597.svg)](https://doi.org/10.5281/zenodo.16238597)

  - [Data Science for Transport
    Planning](https://tdscience.github.io/dstp/) course by [Robin
    Lovelace](https://github.com/robinlovelace), [Juan P.
    Fonseca-Zamora](https://github.com/juanfonsecaLS1), and [Yuanxuan
    Yang](https://github.com/yyuanxuan) (held at the [Institute for
    Transport Studies, University of
    Leeds](https://environment.leeds.ac.uk/transport))

  - ["Data Science for Transport Planning" course in
    Tartu](https://tdscience.github.io/tartu26/) by [Robin
    Lovelace](https://github.com/robinlovelace) and [Juan P.
    Fonseca-Zamora](https://github.com/juanfonsecaLS1) (held at the
    [Mobile Tartu 2026 conference](https://mobiletartu.ut.ee/))

### Citation

To cite the `spanishoddata` R package in publications use:

Kotov E, Vidal-Tortosa E, Cantú-Ros O, Burrieza-Galán J, Herranz R,
Gullón Muñoz-Repiso T, Lovelace R (2026). “spanishoddata: A package for
accessing and working with Spanish Open Mobility Big Data.” *Environment
and Planning B: Urban Analytics and City Science*. ISSN 2399-8083.
doi:10.1177/23998083251415040
<https://doi.org/10.1177/23998083251415040>.

To cite the official website of the mobility study use:

Ministerio de Transportes y Movilidad Sostenible (MITMS) (2024).
“Estudio de la movilidad con Big Data (Study of mobility with Big
Data).”
<https://www.transportes.gob.es/ministerio/proyectos-singulares/estudio-de-movilidad-con-big-data>.

To cite the methodology for 2022 and onwards data use:

Ministerio de Transportes y Movilidad Sostenible (MITMS) (2024).
*Estudio de movilidad de viajeros de ámbito nacional aplicando la
tecnología Big Data. Informe metodológico (Study of National Traveler
mobility Using Big Data Technology. Methodological Report)*.
<https://www.transportes.gob.es/recursos_mfom/paginabasica/recursos/a3_informe_metodologico_estudio_movilidad_mitms_v8.pdf>.

To cite the methodology for 2020-2021 data use:

Ministerio de Transportes, Movilidad y Agenda Urbana (MITMA) (2021).
*Análisis de la movilidad en España con tecnología Big Data durante el
estado de alarma para la gestión de la crisis del COVID-19 (Analysis of
mobility in Spain with Big Data technology during the state of alarm for
COVID-19 crisis management)*.
<https://cdn.mitma.gob.es/portal-web-drupal/covid-19/bigdata/mitma_-_estudio_movilidad_covid-19_informe_metodologico_v3.pdf>.

See package website for more details:
https://ropenspain.github.io/spanishoddata/

BibTeX:


    BibTeX citations:
    -----------------
    %% To cite the spanishoddata package
    @Article{spanishoddata-r-pkg,
      author = {Kotov, Egor and {Vidal-Tortosa}, Eugeni and {Cant{\'u}-Ros}, Oliva G. and {Burrieza-Gal{\'a}n}, Javier and Herranz, Ricardo and {Gull{\'o}n Mu{\~n}oz-Repiso}, Tania and Lovelace, Robin},
      title = {spanishoddata: {{A}} package for accessing and working with {{Spanish Open Mobility Big Data}}},
      issn = {2399-8083},
      doi = {10.1177/23998083251415040},
      url = {https://doi.org/10.1177/23998083251415040},
      shorttitle = {spanishoddata},
      year = {2026},
      journal = {Environment and Planning B: Urban Analytics and City Science},
    }


    %% To cite the Ministry's mobility study website
    @Misc{mitms_mobility_web,
      author = {{Ministerio de Transportes y Movilidad Sostenible (MITMS)}},
      title = {Estudio de la movilidad con Big Data (Study of mobility with Big Data)},
      url = {https://www.transportes.gob.es/ministerio/proyectos-singulares/estudio-de-movilidad-con-big-data},
      year = {2024},
      note = {Data License: \url{https://www.transportes.gob.es/el-ministerio/buen-gobierno/licencia_datos}},
    }


    %% To cite the methodology for 2020-2021 data
    @Manual{mitma_methodology_2020_v3,
      author = {{Ministerio de Transportes, Movilidad y Agenda Urbana (MITMA)}},
      title = {An{\'a}lisis de la movilidad en Espa{\~n}a con tecnolog{\'\i}a Big Data durante el estado de alarma para la gesti{\'o}n de la crisis del COVID-19 (Analysis of mobility in Spain with Big Data technology during the state of alarm for COVID-19 crisis management)},
      url = {https://cdn.mitma.gob.es/portal-web-drupal/covid-19/bigdata/mitma_-_estudio_movilidad_covid-19_informe_metodologico_v3.pdf},
      year = {2021},
    }


    %% To cite the methodology for 2022 and onwards data
    @Manual{mitms_methodology_2022_v8,
      author = {{Ministerio de Transportes y Movilidad Sostenible (MITMS)}},
      title = {Estudio de movilidad de viajeros de {\'a}mbito nacional aplicando la tecnolog{\'\i}a Big Data. Informe metodol{\'o}gico (Study of National Traveler mobility Using Big Data Technology. Methodological Report)},
      url = {https://www.transportes.gob.es/recursos_mfom/paginabasica/recursos/a3_informe_metodologico_estudio_movilidad_mitms_v8.pdf},
      year = {2024},
    }


    %% Note: A more up-to-date methodology document may be available at https://www.transportes.gob.es/ministerio/proyectos-singulares/estudios-de-movilidad-con-big-data/metodologia-del-estudio-de-movilidad-con-bigdata

### Working with Python?

Try the new work-in-progress package:
<https://github.com/pySpainMobility/pySpainMobility>.

## References

<!-- metadata for better search engine indexing -->

<!-- should be picked up by pkgdown -->

<!-- update metadata before release with  -->

<!-- cffr::cff_write() -->

<!-- codemetar::write_codemeta(write_minimeta = T) -->

<div id="refs" class="references csl-bib-body hanging-indent">

<div id="ref-lovelace_od_2024" class="csl-entry">

Lovelace, Robin, and Malcolm Morgan. 2024. *Od: Manipulate and Map
Origin-Destination Data*. <https://doi.org/10.32614/CRAN.package.od>.

</div>

<div id="ref-mitms_mobility_web" class="csl-entry">

Ministerio de Transportes y Movilidad Sostenible (MITMS). 2024. *Estudio
de La Movilidad Con Big Data (Study of Mobility with Big Data)*.
<https://www.transportes.gob.es/ministerio/proyectos-singulares/estudio-de-movilidad-con-big-data>.

</div>

</div>

#' Get metadata cache subfolder name
#'
#' Change subfolder name in the code of this function for metadata cache here to apply globally, as all functions in the package should use this function to get the metadata cache path.
#' @return A `character` string with the subfolder name for the raw data cache.
#' @keywords internal
spod_subfolder_metadata_cache <- function() {
  "metadata_cache"
}

#' Get raw data cache subfolder name
#'
#' Change subfolder name in the code of this function for raw data cache here to apply globally, as all functions in the package should use this function to get the raw data cache path.
#' @inheritParams spod_available_data
#' @return A `character` string with the subfolder name for the raw data cache.
#' @keywords internal
spod_subfolder_raw_data_cache <- function(ver = 1) {
  ver <- as.integer(ver)
  if (!ver %in% c(1, 2)) {
    stop("Invalid version number. Must be 1 or 2.")
  }
  base_subdir_name <- "raw_data_cache"
  return(paste0(base_subdir_name, "/v", ver))
}

#' Get clean data subfolder name
#'
#' Change subfolder name in the code of this function for clean data cache here to apply globally, as all functions in the package should use this function to get the clean data cache path.
#' @inheritParams spod_available_data
#' @return A `character` string with the subfolder name for the clean data cache.
#' @keywords internal
spod_subfolder_clean_data_cache <- function(ver = 1) {
  ver <- as.integer(ver)
  if (!ver %in% c(1, 2)) {
    stop("Invalid version number. Must be 1 or 2.")
  }
  base_subdir_name <- "clean_data"
  return(paste0(base_subdir_name, "/v", ver))
}

#' Get temporary directory for DuckDB intermediate spilling
#' 
#' @description
#' Get the The path to the temp folder for DuckDB for \href{https://duckdb.org/2024/07/09/memory-management.html#intermediate-spilling}{intermediate spilling} in case the set memory limit and/or physical memory of the computer is too low to perform the query.
#' @param data_dir The directory where the data is stored. Defaults to the value returned by `spod_get_data_dir()`.
#' @return A `character` string with the path to the temp folder for `DuckDB` for \href{https://duckdb.org/2024/07/09/memory-management.html#intermediate-spilling}{intermediate spilling}.
#' @keywords internal
spod_get_temp_dir <- function(
  data_dir = spod_get_data_dir()
) {
  temp_dir <- fs::path(data_dir, "temp")
  if (!dir.exists(temp_dir)) {
    fs::dir_create(temp_dir)
  }
  return(temp_dir)
}

# Prepare the testing environment using bundled xml files to avoid downloading data from the internet

extdata_path <- system.file("extdata", package = "spanishoddata")
gz_files <- list.files(
  extdata_path,
  pattern = "(data_links_.*\\.xml\\.gz)|(url_file_sizes_v[1-2]\\.csv\\.gz)",
  full.names = TRUE
)

if (length(gz_files) == 0) {
  stop("No gzipped XML files found.")
}

# Create a temporary directory
test_data_dir <- tempfile()
dir.create(test_data_dir, recursive = TRUE)
# Create metadata directory
# Create metadata directory
metadata_dir <- paste0(test_data_dir, "/", spod_subfolder_metadata_cache())
dir.create(metadata_dir, recursive = TRUE)

current_date <- format(Sys.time(), format = "%Y-%m-%d", usetz = FALSE)

# Generate comprehensive fake XML metadata
v1_dates <- seq.Date(as.Date("2020-01-01"), as.Date("2021-12-31"), by = "day")
v2_dates <- seq.Date(as.Date("2022-01-01"), as.Date("2024-12-31"), by = "day")

generate_xml_content <- function(dates, version) {
  base_url <- if (version == 1) {
    "https://example.com/maestra1-mitma-distritos/ficheros-diarios/"
  } else {
    "https://example.com/viajes/ficheros-diarios/"
  }
  items <- paste0(
    "<item><link>",
    base_url,
    format(dates, "%Y%m%d"),
    "_test.csv.gz</link>",
    "<pubDate>",
    format(dates, "%a, %d %b %Y %H:%M:%S GMT"),
    "</pubDate></item>",
    collapse = "\n"
  )
  paste0("<rss><channel>", items, "</channel></rss>")
}

v1_xml_path <- file.path(
  metadata_dir,
  paste0("data_links_v1_", current_date, ".xml.gz")
)
v1_con <- gzfile(v1_xml_path, "w")
writeLines(generate_xml_content(v1_dates, 1), v1_con)
close(v1_con)

v2_xml_path <- file.path(
  metadata_dir,
  paste0("data_links_v2_", current_date, ".xml.gz")
)
v2_con <- gzfile(v2_xml_path, "w")
writeLines(generate_xml_content(v2_dates, 2), v2_con)
close(v2_con)

# Set the environment variable to the test directory
Sys.setenv(SPANISH_OD_DATA_DIR = test_data_dir)
# Sys.getenv("SPANISH_OD_DATA_DIR")

# Global setup: clear cache and mock S3 once
memoise::forget(spod_get_valid_dates_memoised)
if (exists("read_data_links_memoised", envir = asNamespace("spanishoddata"))) {
  memoise::forget(read_data_links_memoised)
}
# Mock S3 globally for this file
local_mocked_bindings(
  spod_available_data_s3 = function(...) stop("S3 disabled"),
  .package = "spanishoddata"
)

test_that("single ISO date input", {
  dates <- "2023-07-01"
  result <- spod_dates_argument_to_dates_seq(dates)
  expect_equal(result, as.Date("2023-07-01"))
})

test_that("single YYYYMMDD date input", {
  dates <- "20230701"
  result <- spod_dates_argument_to_dates_seq(dates)
  expect_equal(result, as.Date("2023-07-01"))
})

test_that("vector of ISO dates", {
  dates <- c("2023-07-01", "2023-07-03", "2023-07-05")
  result <- spod_dates_argument_to_dates_seq(dates)
  expect_equal(result, as.Date(c("2023-07-01", "2023-07-03", "2023-07-05")))
})

test_that("vector of YYYYMMDD dates", {
  dates <- c("20230701", "20230703", "20230705")
  result <- spod_dates_argument_to_dates_seq(dates)
  expect_equal(result, as.Date(c("2023-07-01", "2023-07-03", "2023-07-05")))
})

test_that("date range in ISO format", {
  dates <- "2023-07-01_2023-07-05"
  result <- spod_dates_argument_to_dates_seq(dates)
  expect_equal(
    result,
    seq.Date(
      from = as.Date("2023-07-01"),
      to = as.Date("2023-07-05"),
      by = "day"
    )
  )
})

test_that("date range in YYYYMMDD format", {
  dates <- "20230701_20230705"
  result <- spod_dates_argument_to_dates_seq(dates)
  expect_equal(
    result,
    seq.Date(
      from = as.Date("2023-07-01"),
      to = as.Date("2023-07-05"),
      by = "day"
    )
  )
})

test_that("named vector date range in ISO format", {
  dates <- c(start = "2023-07-01", end = "2023-07-05")
  result <- spod_dates_argument_to_dates_seq(dates)
  expect_equal(
    result,
    seq.Date(
      from = as.Date("2023-07-01"),
      to = as.Date("2023-07-05"),
      by = "day"
    )
  )
})

test_that("named vector date range in YYYYMMDD format", {
  dates <- c(start = "20230701", end = "20230705")
  result <- spod_dates_argument_to_dates_seq(dates)
  expect_equal(
    result,
    seq.Date(
      from = as.Date("2023-07-01"),
      to = as.Date("2023-07-05"),
      by = "day"
    )
  )
})

test_that("regex pattern matching dates", {
  dates <- "^202307"
  result <- spod_dates_argument_to_dates_seq(dates)
  expected_dates <- seq.Date(
    from = as.Date("2023-07-01"),
    to = as.Date("2023-07-31"),
    by = "day"
  )
  expect_equal(result, expected_dates)
})

test_that("invalid input type", {
  dates <- 20230701
  expect_error(
    spod_dates_argument_to_dates_seq(dates),
    "Invalid date input format. Please provide a character vector or Date object."
  )
})

test_that("dates span both v1 and v2 data", {
  dates <- c("2021-05-09", "2022-01-01")
  expect_error(
    spod_dates_argument_to_dates_seq(dates),
    "Dates found in both v1 and v2 data."
  )
})

# clean up
unlink(test_data_dir, recursive = TRUE)

# nocov start
.onLoad <- function(libname, pkgname) {
  op <- options()
  op.spanishoddata <- list(
    spanishoddata.graphql_api_endpoint = "https://mapas-movilidad.transportes.gob.es/api/graphql",
    spanishoddata.user_agent = "spanishoddata R package, https://github.com/rOpenSpain/spanishoddata/"
  )
  toset <- !(names(op.spanishoddata) %in% names(op))
  if (any(toset)) {
    options(op.spanishoddata[toset])
  }

  invisible()
}
# nocov end

#' Filter available data by version, zones, type and dates
#'
#' @description
#' Internal helper to filter available_data tibble based on version-specific
#' grepl patterns. Extracted for testability.
#'
#' @param available_data Tibble from spod_available_data
#' @param ver Integer. Data version (1 or 2)
#' @param zones Character. Zone name in Spanish (e.g., "distritos")
#' @param matched_type Character. Data type in Spanish (e.g., "viajes")
#' @param dates_to_use Date vector. Dates to filter by
#' @return Filtered tibble
#' @keywords internal
spod_filter_available_files <- function(
  available_data,
  ver,
  zones,
  matched_type,
  dates_to_use
) {
  if (ver == 1) {
    # For v1, always use distritos due to data quality issues
    # See: https://www.ekotov.pro/mitma-data-issues/issues/011-v1-tpp-mismatch-zone-ids-in-table-and-spatial-data.html
    available_data[
      grepl(
        glue::glue("v{ver}.*{matched_type}.*distritos"),
        available_data$local_path
      ) &
        available_data$data_ymd %in% dates_to_use,
    ]
  } else {
    available_data[
      grepl(
        glue::glue("v{ver}.*{zones}.*{matched_type}"),
        available_data$local_path
      ) &
        available_data$data_ymd %in% dates_to_use,
    ]
  }
}

#' Check cached files consistency against checksums from S3
#'
#' @description
#'
#' `r lifecycle::badge("experimental")`
#'
#' **WARNING: The checks may fail for May 2022 data and for some 2025 data, as the remote cheksums that are used for checking the file consistency are incorrect. We are working on solving this in future updates, for now, kindly rely on the built-in file size checks of \code{\link{spod_download}}, \code{\link{spod_get}}, and \code{\link{spod_convert}}.** This function checks downloaded data files whether they are consistent with their checksums in Amazon S3 by computing ETag for each file. This involves computing MD5 for each part of the file and concatenating them and computing MD5 again on the resulting concatenated MD5s. This may take very long time if you check all files, so use with caution.
#' @inheritParams spod_get
#' @inheritParams spod_download
#' @inheritParams global_quiet_param
#' @param n_threads Numeric. Number of threads to use for file verificaiton. Defaults to 1. When set to 2 or more threads, uses `future.mirai` as a backend for parallelization, resulting in significant (~4x) speedup, unless disk read speed is a bottleneck.
#'
#' @return A tibble similar to the output of `spod_available_data`, but with an extra column `local_file_consistent`, where `TRUE` indicates that the file cheksum matches the expected checksums in Amazon S3. Note: some v1 (2020-2021) files were not stored correctly on S3 and their ETag checksums are incorrectly reported by Amazon S3, so their true file sizes and ETag checksums were cached inside the `spanishoddata` package.
#'
#' @export
#'
#' @examplesIf interactive()
#' \donttest{
#' spod_set_data_dir(tempdir())
#' spod_download(
#'  type = "number_of_trips",
#'  zones = "distr",
#'  dates = "2020-03-14"
#' )
#'
#' # now check the consistency
#' check_results <- spod_check_files(
#'   type = "number_of_trips",
#'   zones = "distr",
#'   dates = "2020-03-14"
#' )
#' all(check_results$local_file_consistent)
#' }
spod_check_files <- function(
  type = c(
    "od",
    "origin-destination",
    "os",
    "overnight_stays",
    "nt",
    "number_of_trips"
  ),
  zones = c(
    "districts",
    "dist",
    "distr",
    "distritos",
    "municipalities",
    "muni",
    "municip",
    "municipios",
    "lua",
    "large_urban_areas",
    "gau",
    "grandes_areas_urbanas"
  ),
  dates = NULL,
  data_dir = spod_get_data_dir(),
  quiet = FALSE,
  ignore_missing_dates = FALSE,
  n_threads = 1
) {
  # Validate inputs
  checkmate::assert_choice(
    type,
    choices = c(
      "od",
      "origin-destination",
      "os",
      "overnight_stays",
      "nt",
      "number_of_trips"
    )
  )
  checkmate::assert_choice(
    zones,
    choices = c(
      "districts",
      "dist",
      "distr",
      "distritos",
      "municipalities",
      "muni",
      "municip",
      "municipios",
      "lua",
      "large_urban_areas",
      "gau",
      "grandes_areas_urbanas"
    )
  )
  checkmate::assert_directory_exists(data_dir, access = "r")
  checkmate::assert_flag(quiet)
  checkmate::assertNumber(n_threads, lower = 1)

  # normalise zones
  zones <- spod_zone_names_en2es(zones)

  # simple null check is enough here, as spod_dates_arugument_to_dates_seq will do additional checks anyway
  if (is.null(dates)) {
    message(
      "`dates` argument is undefined. Please set `dates='cached_v1'` or `dates='cached_v2'` to convert all data that was previously downloaded. Alternatively, specify at least one date between 2020-02-14 and 2021-05-09 (for v1 data) or between 2022-01-01 onwards (for v2). Any missing data will be downloaded before conversion. For more details on the dates argument, see ?spod_download."
    )
  }

  dates_to_use <- spod_dates_argument_to_dates_seq(dates = dates)

  # check version
  ver <- spod_infer_data_v_from_dates(
    dates = dates_to_use,
    ignore_missing_dates = ignore_missing_dates
  )

  if (isFALSE(quiet)) {
    message("Data version detected from dates: ", ver)
  }

  # convert english data type names to spanish words used in the default data paths
  type <- match.arg(type)
  matched_type <- spod_match_data_type_for_local_folders(type = type, ver = ver)

  available_data <- spod_available_data(
    ver = ver,
    check_local_files = TRUE,
    data_dir = data_dir,
    quiet = quiet,
    use_s3 = TRUE
  )

  if (all(as.character(dates) %in% c("cached_v1", "cached_v2"))) {
    dates_to_use <- available_data |>
      dplyr::filter(.data$downloaded == TRUE & !is.na(.data$data_ymd)) |>
      dplyr::pull("data_ymd")
    available_data <- available_data |>
      dplyr::filter(.data$downloaded == TRUE & !is.na(.data$data_ymd))
  }

  # match the available_data to type, zones, version and dates
  requested_files <- spod_filter_available_files(
    available_data = available_data,
    ver = ver,
    zones = zones,
    matched_type = matched_type,
    dates_to_use = dates_to_use
  )

  # if some requested files are missing issue a warning
  if (!all(requested_files$downloaded)) {
    dates_missing_downloads <- requested_files |>
      dplyr::filter(.data$downloaded == FALSE) |>
      dplyr::filter(.data$data_ymd %in% dates_to_use) |>
      dplyr::pull("data_ymd")
    warning(glue::glue(
      'Some files for the requested dates are missing ({paste(spod_convert_dates_to_ranges(dates_missing_downloads), collapse = ", ")}). Make sure you have downloaded all files requested to be checked for consistency with `spod_download()`. For now, `spod_check_files()` will only check the files that were previously downloaded and currently exist on disk.',
    ))
    requested_files <- requested_files |>
      dplyr::filter(.data$downloaded == TRUE)
  }

  # compute ETag for each file
  if (n_threads == 1) {
    local_etags <- requested_files$local_path |>
      purrr::map_chr(~ spod_compute_s3_etag(.x), .progress = TRUE)
  } else if (n_threads > 1) {
    spod_assert_package(c("future", "furrr", "future.mirai"))
    with(
      future::plan(future.mirai::mirai_multisession, workers = n_threads),
      local = TRUE
    )
    local_etags <- requested_files$local_path |>
      furrr::future_map_chr(
        ~ spod_compute_s3_etag(.x),
        .progress = TRUE,
        future.seed = TRUE
      )
  }

  # compare ETags
  requested_files <- requested_files |>
    dplyr::mutate(
      local_etag = local_etags,
    ) |>
    dplyr::mutate(
      local_file_consistent = dplyr::if_else(
        condition = .data$local_etag == .data$etag,
        true = TRUE,
        false = FALSE,
        missing = FALSE
      )
    )

  # issue a warning if there are mismatches or inform that everything is ok
  if (isFALSE(quiet)) {
    if (!all(requested_files$local_file_consistent)) {
      broken_dates <- requested_files |>
        dplyr::filter(.data$local_file_consistent == FALSE) |>
        dplyr::pull("data_ymd")
      warning(glue::glue(
        'Some files are inconsistent with the reference data ({paste(spod_convert_dates_to_ranges(broken_dates), collapse = ", ")}). Please inspect the returned table by filtering the output table by "local_file_consistent == FALSE". To re-download the inconsistent files, use `spod_download()` and specify the missing dates.'
      ))
    } else {
      message("All checked files are consistent.")
    }
  }

  return(requested_files)
}

#' Compute ETag for a file
#' @param file_path Character. The path to the file.
#' @param part_size Numeric. The size of each part in bytes. Do not change, as this is a default for S3 Etag.
#' @return Character. The ETag for the file.
#' @keywords internal
spod_compute_s3_etag <- function(file_path, part_size = 8 * 1024^2) {
  con <- file(file_path, "rb")
  on.exit(close(con))

  # MD5 each part
  part_md5s <- list()
  repeat {
    buf <- readBin(con, "raw", n = part_size)
    if (length(buf) == 0) {
      break
    }
    part_md5s[[length(part_md5s) + 1]] <- digest::digest(
      buf,
      algo = "md5",
      serialize = FALSE,
      raw = TRUE
    )
  }

  # Single‐part fallback
  if (length(part_md5s) == 1) {
    return(digest::digest(file = file_path, algo = "md5", serialize = FALSE))
  }

  # Concatenate raw MD5s and MD5 again
  combined <- do.call(c, part_md5s)
  final_raw <- digest::digest(
    combined,
    algo = "md5",
    serialize = FALSE,
    raw = TRUE
  )
  raw_to_hex <- function(r) paste(sprintf("%02x", as.integer(r)), collapse = "")
  etag_hex <- raw_to_hex(final_raw)
  paste0(etag_hex, "-", length(part_md5s))
}

library(testthat)
library(spanishoddata)

# Tests using real fixtures to verify ETag computation and consistency checking logic

test_that("spod_check_files detects inconsistency in fixtures", {
  test_dir <- setup_test_data_dir()
  withr::defer(unlink(test_dir, recursive = TRUE))
  withr::local_envvar(c("SPANISH_OD_DATA_DIR" = test_dir))

  mock_meta_path <- system.file("testdata", "metadata", "available_data_s3_v2_mock.rds", package = "spanishoddata")
  if (mock_meta_path == "") mock_meta_path <- testthat::test_path("../../inst/testdata/metadata/available_data_s3_v2_mock.rds")
  mock_meta <- readRDS(mock_meta_path)

  testthat::local_mocked_bindings(
    spod_available_data_s3 = function(...) mock_meta,
    read_data_links_memoised = function(...) tibble::tibble(),
    spod_get_valid_dates = function(ver, ...) {
      if (ver == 1) {
        return(as.Date(character(0)))
      }
      if (ver == 2) {
        return(as.Date("2022-02-01"))
      }
      return(as.Date(character(0)))
    },
    .package = "spanishoddata"
  )

  expect_warning(
    res <- suppressMessages(spod_check_files(
      type = "od",
      zones = "dist",
      dates = "2022-02-01",
      quiet = FALSE
    )),
    "Some files are inconsistent"
  )

  expect_s3_class(res, "data.frame")
  expect_true(nrow(res) > 0)
  expect_false(all(res$local_file_consistent))

  expect_true(nchar(res$local_etag[1]) >= 32)
  expect_true(res$local_etag[1] != res$etag[1])
})

test_that("spod_check_files reports consistency when metadata matches", {
  test_dir <- setup_test_data_dir()
  withr::defer(unlink(test_dir, recursive = TRUE))
  withr::local_envvar(c("SPANISH_OD_DATA_DIR" = test_dir))

  target_file <- file.path(
    test_dir,
    "raw_data_cache/v2/estudios_basicos/por-distritos/viajes/ficheros-diarios/year=2022/month=2/day=1/data.csv.gz"
  )
  real_etag <- spod_compute_s3_etag(target_file)

  meta_cache_path <- file.path(test_dir, spod_subfolder_metadata_cache())
  rds_files <- list.files(
    meta_cache_path,
    full.names = TRUE,
    pattern = "metadata_s3_v2_.*\\.rds"
  )
  expect_true(length(rds_files) > 0)

  meta <- readRDS(rds_files[1])
  meta$etag <- real_etag
  saveRDS(meta, rds_files[1])

  mock_meta_path <- system.file("testdata", "metadata", "available_data_s3_v2_mock.rds", package = "spanishoddata")
  if (mock_meta_path == "") mock_meta_path <- testthat::test_path("../../inst/testdata/metadata/available_data_s3_v2_mock.rds")
  mock_meta <- readRDS(mock_meta_path)
  mock_meta$etag <- real_etag

  testthat::local_mocked_bindings(
    spod_available_data_s3 = function(...) mock_meta,
    read_data_links_memoised = function(...) tibble::tibble(),
    spod_get_valid_dates = function(ver, ...) {
      if (ver == 1) {
        return(as.Date(character(0)))
      }
      if (ver == 2) {
        return(as.Date("2022-02-01"))
      }
      return(as.Date(character(0)))
    },
    .package = "spanishoddata"
  )

  msgs <- capture_messages(
    res <- spod_check_files(
      type = "od",
      zones = "dist",
      dates = "2022-02-01",
      quiet = FALSE
    )
  )

  expect_match(msgs, "All checked files are consistent", all = FALSE)
  expect_true(all(res$local_file_consistent))
})

test_that("spod_check_files detects missing files in fixtures", {
  test_dir <- setup_test_data_dir()
  withr::defer(unlink(test_dir, recursive = TRUE))
  withr::local_envvar(c("SPANISH_OD_DATA_DIR" = test_dir))

  target_file <- file.path(
    test_dir,
    "raw_data_cache/v2/estudios_basicos/por-distritos/viajes/ficheros-diarios/year=2022/month=2/day=1/data.csv.gz"
  )
  unlink(target_file)

  mock_meta_path <- system.file("testdata", "metadata", "available_data_s3_v2_mock.rds", package = "spanishoddata")
  if (mock_meta_path == "") mock_meta_path <- testthat::test_path("../../inst/testdata/metadata/available_data_s3_v2_mock.rds")
  mock_meta <- readRDS(mock_meta_path)

  testthat::local_mocked_bindings(
    spod_available_data_s3 = function(...) mock_meta,
    read_data_links_memoised = function(...) tibble::tibble(),
    spod_get_valid_dates = function(ver, ...) {
      if (ver == 1) {
        return(as.Date(character(0)))
      }
      if (ver == 2) {
        return(as.Date("2022-02-01"))
      }
      return(as.Date(character(0)))
    },
    .package = "spanishoddata"
  )

  expect_warning(
    res <- suppressMessages(spod_check_files(
      type = "od",
      zones = "dist",
      dates = "2022-02-01",
      quiet = FALSE
    )),
    "Some files.*are missing"
  )

  expect_equal(nrow(res), 0)
})

#' Get tabular mobility data
#'
#'
#' @description
#'
#' `r lifecycle::badge("stable")`
#'
#' This function creates a DuckDB lazy table connection object from the specified type and zones. It checks for missing data and downloads it if necessary. The connnection is made to the raw CSV files in gzip archives, so analysing the data through this connection may be slow if you select more than a few days. You can manipulate this object using `dplyr` functions such as \code{\link[dplyr]{select}}, \code{\link[dplyr]{filter}}, \code{\link[dplyr]{mutate}}, \code{\link[dplyr]{group_by}}, \code{\link[dplyr]{summarise}}, etc. In the end of any sequence of commands you will need to add \code{\link[dplyr]{collect}} to execute the whole chain of data manipulations and load the results into memory in an R `data.frame`/`tibble`. See codebooks for v1 and v2 data in vignettes with \code{\link{spod_codebook}(1)} and \code{\link{spod_codebook}(2)}.
#'
#' If you want to analyse longer periods of time (especiially several months or even the whole data over several years), consider using the \code{\link{spod_convert}} and then \code{\link{spod_connect}}.
#'
#' If you want to quickly get the origin-destination data with flows aggregated for a single day at municipal level and without any extra socio-economic variables, consider using the \code{\link{spod_quick_get_od}} function.
#'
#' For detailed data descriptions, see package vignettes using [`spod_codebook(ver = 1)`][spod_codebook] and [`spod_codebook(ver = 2)`][spod_codebook] and official methodology documents in **References** section.
#'
#' @template references
#'
#' @param duckdb_target (Optional) The path to the duckdb file to save the data to, if a convertation from CSV is reuqested by the `spod_convert` function. If not specified, it will be set to ":memory:" and the data will be stored in memory.
#' @inheritParams spod_download
#' @inheritParams spod_duckdb_limit_resources
#' @inheritParams spod_duckdb_set_temp
#' @inheritParams global_quiet_param
#' @return A DuckDB lazy table connection object of class `tbl_duckdb_connection`.
#' @export
#' @examplesIf interactive()
#' \donttest{
#'
#' # create a connection to the v1 data
#' spod_set_data_dir(tempdir())
#' dates <- c("2020-02-14", "2020-03-14", "2021-02-14", "2021-02-14", "2021-02-15")
#' nt_dist <- spod_get(type = "number_of_trips", zones = "distr", dates = dates)
#'
#' # nt_dist is a table view filtered to the specified dates
#'
#' # for advanced users only
#' # access the source connection with all dates
#' # list tables
#' DBI::dbListTables(nt_dist$src$con)
#'
#' # disconnect
#' spod_disconnect(nt_dist)
#' }
#'
spod_get <- function(
  type = c(
    "od",
    "origin-destination",
    "os",
    "overnight_stays",
    "nt",
    "number_of_trips"
  ),
  zones = c(
    "districts",
    "dist",
    "distr",
    "distritos",
    "municipalities",
    "muni",
    "municip",
    "municipios",
    "lua",
    "large_urban_areas",
    "gau",
    "grandes_areas_urbanas"
  ),
  dates = NULL,
  data_dir = spod_get_data_dir(),
  quiet = FALSE,
  max_mem_gb = NULL,
  max_n_cpu = max(1, parallelly::availableCores() - 1),
  max_download_size_gb = 1,
  duckdb_target = ":memory:",
  temp_path = spod_get_temp_dir(),
  ignore_missing_dates = FALSE
) {
  # Validate inputs
  checkmate::assert_choice(
    type,
    choices = c(
      "od",
      "origin-destination",
      "os",
      "overnight_stays",
      "nt",
      "number_of_trips"
    )
  )
  checkmate::assert_choice(
    zones,
    choices = c(
      "districts",
      "dist",
      "distr",
      "distritos",
      "municipalities",
      "muni",
      "municip",
      "municipios",
      "lua",
      "large_urban_areas",
      "gau",
      "grandes_areas_urbanas"
    )
  )
  checkmate::assert_flag(quiet)
  checkmate::assert_number(max_mem_gb, lower = 1, null.ok = TRUE)
  checkmate::assert_integerish(max_n_cpu, lower = 1)
  checkmate::assert_number(max_download_size_gb, lower = 0.1)
  checkmate::assert_string(duckdb_target)
  checkmate::assert_directory_exists(data_dir, access = "rw")
  checkmate::assert_directory_exists(temp_path, access = "rw")
  checkmate::assert_flag(ignore_missing_dates)

  # simple null check is enough here, as spod_dates_arugument_to_dates_seq will do additional checks anyway
  if (is.null(dates)) {
    message(
      "`dates` argument is undefined. Please set `dates='cached_v1'` or `dates='cached_v2'` to convert all data that was previously downloaded. Alternatively, specify at least one date between 2020-02-14 and 2021-05-09 (for v1 data) or between 2022-01-01 onwards (for v2). Any missing data will be downloaded before conversion. For more details on the dates argument, see ?spod_get."
    )
  }

  # normalise type
  type <- spod_match_data_type(type = type)

  # normalise zones
  zones <- spod_zone_names_en2es(zones)

  # check if user is requesting to just get all cached data
  cached_data_requested <- length(dates) == 1 &&
    all(as.character(dates) %in% c("cached_v1", "cached_v2"))

  if (isFALSE(cached_data_requested)) {
    dates <- spod_dates_argument_to_dates_seq(dates = dates)
    ver <- spod_infer_data_v_from_dates(
      dates = dates,
      ignore_missing_dates = ignore_missing_dates
    )
    # use the spot_download_data() function to download any missing data
    spod_download(
      type = type,
      zones = zones,
      dates = dates,
      max_download_size_gb = max_download_size_gb,
      data_dir = data_dir,
      quiet = quiet,
      return_local_file_paths = FALSE,
      ignore_missing_dates = ignore_missing_dates
    )
  } else if (isTRUE(cached_data_requested)) {
    ver <- as.numeric(stringr::str_extract(dates, "(1|2)$"))
  }

  # create in memory duckdb connection
  drv <- duckdb::duckdb()
  con <- DBI::dbConnect(drv, dbdir = duckdb_target, read_only = FALSE)

  # define memory and threads limits
  con <- spod_duckdb_limit_resources(
    con = con,
    max_mem_gb = max_mem_gb,
    max_n_cpu = max_n_cpu
  )

  # attach the folder with csv.gz files with predefined and cleaned up data types
  if (type == "od") {
    con <- spod_duckdb_od(
      con = con,
      zones = zones,
      ver = ver,
      data_dir = data_dir
    )
  } else if (type == "nt") {
    con <- spod_duckdb_number_of_trips(
      con = con,
      zones = zones,
      ver = ver,
      data_dir = data_dir
    )
  } else if (type == "os") {
    con <- spod_duckdb_overnight_stays(
      con = con,
      zones = zones,
      ver = ver,
      data_dir = data_dir
    )
  }

  clean_csv_view_name <- glue::glue("{type}_csv_clean")
  clean_filtered_csv_view_name <- glue::glue("{type}_csv_clean_filtered")

  # filter by date, unless cached data requested
  if (isFALSE(cached_data_requested)) {
    con <- spod_duckdb_filter_by_dates(
      con = con,
      source_view_name = clean_csv_view_name,
      new_view_name = clean_filtered_csv_view_name,
      dates = dates
    )
  }

  # if working with in-memory database
  # set temp path for intermediate spilling
  # https://duckdb.org/2024/07/09/memory-management.html#intermediate-spilling
  # if target were set as a database file, temp would be created at the same path
  # however, when the working in-memory on folder of CSV files, temp is created in the root of R working directory, which may be undesirable
  if (duckdb_target == ":memory:") {
    con <- spod_duckdb_set_temp(con, temp_path = temp_path)
  }

  # return either a full view of all available data (dates = "cached") or a view filtered to the specified dates
  if (isFALSE(cached_data_requested)) {
    return(dplyr::tbl(con, clean_filtered_csv_view_name))
  } else if (isTRUE(cached_data_requested)) {
    return(dplyr::tbl(con, clean_csv_view_name))
  }
}

library(testthat)
library(spanishoddata)

# This test file is designed for scheduled live integration tests on CI.
# It hits real APIs and S3 buckets to ensure the data provider infrastructure
# and the package's parsing logic remain compatible.

test_that("live integration: v1 and v2 data validation with numerical expectations", {
  skip_on_cran()
  skip_if_not(
    Sys.getenv("RUN_LIVE_TESTS") == "TRUE",
    "Skipping live integration tests. Set RUN_LIVE_TESTS='TRUE' to run."
  )
  
  # Setup temporary directory for isolation
  test_dir <- tempfile("spod_live_")
  dir.create(test_dir)
  withr::defer(unlink(test_dir, recursive = TRUE))
  withr::local_envvar(c("SPANISH_OD_DATA_DIR" = test_dir))
  
  v1_date <- "2020-02-14"
  v2_date <- "2023-11-05"
  
  # --- V1 INTEGRATION TESTS ---
  cli::cli_h1("Testing V1 Integration (2020-2021)")
  
  # 1. Get Zones V1
  cli::cli_h2("Testing spod_get_zones (V1)")
  v1_zones <- spod_get_zones("distr", ver = 1, quiet = TRUE)
  expect_s3_class(v1_zones, "sf")
  expect_equal(nrow(v1_zones), 2850) # Exact match for districts
  
  # 2. Get OD V1
  cli::cli_h2("Testing spod_get OD (V1)")
  v1_od <- spod_get(type = "od", zones = "dist", dates = v1_date, quiet = TRUE)
  expect_s3_class(v1_od, "tbl_duckdb_connection")
  
  v1_summary <- v1_od |> 
    dplyr::summarise(
      total_trips = sum(n_trips, na.rm = TRUE),
      total_rows = dplyr::n()
    ) |> 
    dplyr::collect()
  
  # Verify summary statistics with small tolerance (0.1% for trips)
  expect_equal(v1_summary$total_rows, 7232427)
  expect_equal(v1_summary$total_trips, 145131837, tolerance = 0.001)
  
  
  # --- V2 INTEGRATION TESTS ---
  cli::cli_h1("Testing V2 Integration (2022 onwards)")
  
  # 1. Get Zones V2
  cli::cli_h2("Testing spod_get_zones (V2)")
  v2_zones <- spod_get_zones("distr", ver = 2, quiet = TRUE)
  expect_s3_class(v2_zones, "sf")
  expect_equal(nrow(v2_zones), 3909) # Exact match for V2 districts
  
  # 2. Get OD V2
  cli::cli_h2("Testing spod_get OD (V2)")
  v2_od <- spod_get(type = "od", zones = "dist", dates = v2_date, quiet = TRUE)
  expect_s3_class(v2_od, "tbl_duckdb_connection")
  
  v2_od_summary <- v2_od |> 
    dplyr::summarise(
      total_trips = sum(n_trips, na.rm = TRUE),
      total_rows = dplyr::n()
    ) |> 
    dplyr::collect()
  
  expect_equal(v2_od_summary$total_rows, 15257106)
  expect_equal(v2_od_summary$total_trips, 104758528, tolerance = 0.001)
  
  # 3. Get NT V2
  cli::cli_h2("Testing spod_get NT (V2)")
  v2_nt <- spod_get(type = "nt", zones = "dist", dates = v2_date, quiet = TRUE)
  
  v2_nt_summary <- v2_nt |> 
    dplyr::summarise(
      total_persons = sum(n_persons, na.rm = TRUE),
      total_rows = dplyr::n()
    ) |> 
    dplyr::collect()
  
  expect_equal(v2_nt_summary$total_rows, 118717)
  expect_equal(v2_nt_summary$total_persons, 47399843, tolerance = 0.001)
  
  # 4. Quick OD (GraphQL)
  cli::cli_h2("Testing spod_quick_get_od (GraphQL)")
  qod <- spod_quick_get_od(date = v2_date, id_origin = "01001")
  expect_s3_class(qod, "tbl_df")
  expect_equal(nrow(qod), 4)
  expect_equal(sum(qod$n_trips), 3590, tolerance = 0.001)
  
  cli::cli_alert_success("Live integration tests for v1 and v2 completed successfully.")
})

#' Creates a duckdb connection to origin-destination data
#'
#' This function creates a duckdb connection to the origin-destination data stored in CSV.gz files.
#'
#' @param con A duckdb connection object. If not specified, a new in-memory connection will be created.
#' @inheritParams spod_available_data
#' @inheritParams spod_download
#' @return A `duckdb` connection object with 2 views:
#'
#'  * `od_csv_raw` - a raw table view of all cached CSV files with the origin-destination data that has been previously cached in $SPANISH_OD_DATA_DIR
#'
#'  * `od_csv_clean` - a cleaned-up table view of `od_csv_raw` with column names and values translated and mapped to English. This still includes all cached data.
#'
#' The structure of the cleaned-up views `od_csv_clean` is as follows:
#'
#' \describe{
#'   \item{date}{\code{Date}. The full date of the trip, including year, month, and day.}
#'   \item{id_origin}{\code{factor}. The identifier for the origin location of the trip, formatted as a code (e.g., '01001_AM').}
#'   \item{id_destination}{\code{factor}. The identifier for the destination location of the trip, formatted as a code (e.g., '01001_AM').}
#'   \item{activity_origin}{\code{factor}. The type of activity at the origin location (e.g., 'home', 'work'). \strong{Note:} Only available for district level data.}
#'   \item{activity_destination}{\code{factor}. The type of activity at the destination location (e.g., 'home', 'other'). \strong{Note:} Only available for district level data.}
#'   \item{residence_province_ine_code}{\code{factor}. The province of residence for the group of individual making the trip, encoded according to the INE classification. \strong{Note:} Only available for district level data.}
#'   \item{residence_province_name}{\code{factor}. The province of residence for the group of individuals making the trip (e.g., 'Cuenca', 'Girona'). \strong{Note:} Only available for district level data.}
#'   \item{hour}{\code{integer}. The time slot (the hour of the day) during which the trip started, represented as an integer (e.g., 0, 1, 2).}
#'   \item{distance}{\code{factor}. The distance category of the trip, represented as a code (e.g., '002-005' for 2-5 km).}
#'   \item{n_trips}{\code{double}. The number of trips taken within the specified time slot and distance.}
#'   \item{trips_total_length_km}{\code{double}. The total length of all trips in kilometers for the specified time slot and distance.}
#'   \item{year}{\code{double}. The year of the trip.}
#'   \item{month}{\code{double}. The month of the trip.}
#'   \item{day}{\code{double}. The day of the trip.}
#' }
#'
#' The structure of the original data in `od_csv_raw` is as follows:
#'
#' \describe{
#'   \item{fecha}{\code{Date}. The date of the trip, including year, month, and day.}
#'   \item{origen}{\code{character}. The identifier for the origin location of the trip, formatted as a character string (e.g., '01001_AM').}
#'   \item{destino}{\code{character}. The identifier for the destination location of the trip, formatted as a character string (e.g., '01001_AM').}
#'   \item{actividad_origen}{\code{character}. The type of activity at the origin location (e.g., 'casa', 'trabajo').}
#'   \item{actividad_destino}{\code{character}. The type of activity at the destination location (e.g., 'otros', 'trabajo').}
#'   \item{residencia}{\code{character}. The code representing the residence of the individual making the trip (e.g., '01') according to the official INE classification.}
#'   \item{edad}{\code{character}. The age of the individual making the trip. This data is actaully filled with 'NA' values, which is why this column is removed in the cleaned-up and translated view described above.}
#'   \item{periodo}{\code{integer}. The time period during which the trip started, represented as an integer (e.g., 0, 1, 2).}
#'   \item{distancia}{\code{character}. The distance category of the trip, represented as a character string (e.g., '002-005' for 2-5 km).}
#'   \item{viajes}{\code{double}. The number of trips taken within the specified time period and distance.}
#'   \item{viajes_km}{\code{double}. The total length of all trips in kilometers for the specified time period and distance.}
#'   \item{day}{\code{double}. The day of the trip.}
#'   \item{month}{\code{double}. The month of the trip.}
#'   \item{year}{\code{double}. The year of the trip.}
#' }
#' @keywords internal
spod_duckdb_od <- function(
  con = DBI::dbConnect(duckdb::duckdb(), dbdir = ":memory:", read_only = FALSE),
  zones = c(
    "districts",
    "dist",
    "distr",
    "distritos",
    "municipalities",
    "muni",
    "municip",
    "municipios",
    "lua",
    "large_urban_areas",
    "gau",
    "grandes_areas_urbanas"
  ),
  ver = NULL,
  data_dir = spod_get_data_dir()
) {
  locale <- "en" # hardcode locale for now, but be ready to make it a function argument across the package #TODO

  ver <- as.integer(ver)
  if (!ver %in% c(1, 2)) {
    stop("Invalid version number. Must be 1 or 2.")
  }

  zones <- match.arg(zones)
  zones <- spod_zone_names_en2es(zones)

  # check that gau is not selected with ver = 1
  if (ver == 1 && zones == "gau") {
    stop("The 'gau' zones are not available in v1 data.")
  }

  if (ver == 1) {
    # selecting districts files for v1 to avoid issues with municipalities # this is to address the bugs described in detail in:
    # https://www.ekotov.pro/mitma-data-issues/issues/011-v1-tpp-mismatch-zone-ids-in-table-and-spatial-data.html
    # https://www.ekotov.pro/mitma-data-issues/issues/012-v1-tpp-district-files-in-municipality-folders.html
    # the decision was to use distrcit data and aggregate it to replicate municipal data
    csv_folder <- fs::path(
      data_dir,
      spod_subfolder_raw_data_cache(ver = ver),
      "maestra1-mitma-distritos",
      "ficheros-diarios"
    )
  } else if (ver == 2) {
    csv_folder <- fs::path(
      data_dir,
      spod_subfolder_raw_data_cache(ver = ver),
      "estudios_basicos",
      paste0("por-", spod_zone_names_en2es(zones)),
      "viajes",
      "ficheros-diarios"
    )
  }

  # Ensure it ends with a separator for the glob pattern if needed,
  # but DuckDB's read_csv_auto works with either.
  # Given the SQL template has '{csv_folder}**/*.csv.gz', we need to ensure
  # csv_folder ends with a separator.
  csv_folder <- paste0(csv_folder, "/")

  # create view to the raw TXT/CSV.gz files
  DBI::dbExecute(
    con,
    spod_read_sql(glue::glue("v{ver}-od-{zones}-raw-csv-view.sql"))
  )

  # create ENUMs

  # zones ENUMs from uniqe ids of relevant zones
  spatial_data <- spod_get_zones(
    zones,
    ver = ver,
    data_dir = data_dir,
    quiet = TRUE
  )

  if (ver == 1) {
    unique_ids <- unique(spatial_data$id)
  } else if (ver == 2) {
    if (locale == "en") {
      unique_ids <- c("external", unique(spatial_data$id))
    } else if (locale == "es") {
      unique_ids <- c("externo", unique(spatial_data$id))
    }
  }

  DBI::dbExecute(
    con,
    dplyr::sql(
      paste0(
        "CREATE TYPE ZONES_ENUM AS ENUM ('",
        paste0(unique_ids, collapse = "','"),
        "');"
      )
    )
  )

  # create ACTIV_ENUM (for all except for municipalities in v1 data)
  DBI::dbExecute(
    con,
    spod_read_sql(glue::glue("v{ver}-od-enum-activity-{locale}.sql"))
  )

  # create DISTANCE_ENUM
  DBI::dbExecute(
    con,
    spod_read_sql(glue::glue("v{ver}-od-enum-distance.sql"))
  )

  con <- spod_duckdb_create_province_enum(con)

  # v2 only enums
  if (ver == 2) {
    # income ENUM
    DBI::dbExecute(
      con,
      spod_read_sql(glue::glue("v{ver}-od-enum-income.sql"))
    )

    # age ENUM
    DBI::dbExecute(
      con,
      spod_read_sql(glue::glue("v{ver}-od-enum-age.sql"))
    )

    # sex ENUM
    DBI::dbExecute(
      con,
      spod_read_sql(glue::glue("v{ver}-od-enum-sex-{locale}.sql"))
    )
  }

  # create od_csv_clean view
  if (ver == 1 && zones == "municipios") {
    # this will be picked up by the sql loaded below if neccessary
    relations_districts_municipalities <- fs::path(
      data_dir,
      spod_subfolder_raw_data_cache(1),
      "relaciones_distrito_mitma.csv"
    )
  }
  DBI::dbExecute(
    con,
    spod_read_sql(glue::glue("v{ver}-od-{zones}-clean-csv-view-{locale}.sql"))
  )

  # return the connection as duckdb object
  return(con)
}

#' Create a duckdb number of trips table
#'
#' @description
#' This function creates a duckdb connection to the number of trips data stored in a folder of CSV.gz files.
#' @inheritParams spod_duckdb_od
#' @inheritParams spod_available_data
#' @inheritParams spod_download
#'
#' @return A `duckdb` connection object with 2 views:
#'
#'  * `od_csv_raw` - a raw table view of all cached CSV files with the origin-destination data that has been previously cached in $SPANISH_OD_DATA_DIR
#'
#'  * `od_csv_clean` - a cleaned-up table view of `od_csv_raw` with column names and values translated and mapped to English. This still includes all cached data.
#'
#' @keywords internal
spod_duckdb_number_of_trips <- function(
  con = DBI::dbConnect(duckdb::duckdb(), dbdir = ":memory:", read_only = FALSE),
  zones = c(
    "districts",
    "dist",
    "distr",
    "distritos",
    "municipalities",
    "muni",
    "municip",
    "municipios",
    "lua",
    "large_urban_areas",
    "gau",
    "grandes_areas_urbanas"
  ),
  ver = NULL,
  data_dir = spod_get_data_dir()
) {
  locale <- "en" # TODO: add support for Spanish, hardcode for now

  ver <- as.integer(ver)
  if (!ver %in% c(1, 2)) {
    stop("Invalid version number. Must be 1 or 2.")
  }

  zones <- match.arg(zones)
  zones <- spod_zone_names_en2es(zones)

  # check that gau is not selected with ver = 1
  if (ver == 1 && zones == "gau") {
    stop("The 'gau' zones are not available in v1 data.")
  }

  if (ver == 1) {
    # selecting districts files for v1 to avoid issues with municipalities # this is to address the bugs described in detail in:
    # https://www.ekotov.pro/mitma-data-issues/issues/011-v1-tpp-mismatch-zone-ids-in-table-and-spatial-data.html
    # https://www.ekotov.pro/mitma-data-issues/issues/012-v1-tpp-district-files-in-municipality-folders.html
    # the decision was to use distrcit data and aggregate it to replicate municipal data
    csv_folder <- fs::path(
      data_dir,
      spod_subfolder_raw_data_cache(ver = ver),
      "maestra2-mitma-distritos",
      "ficheros-diarios"
    )
  } else if (ver == 2) {
    csv_folder <- fs::path(
      data_dir,
      spod_subfolder_raw_data_cache(ver = ver),
      "estudios_basicos",
      paste0("por-", spod_zone_names_en2es(zones)),
      "personas",
      "ficheros-diarios"
    )
  }

  csv_folder <- paste0(csv_folder, "/")

  # create view of csv files and preset variable types

  # create view to the raw TXT/CSV.gz files
  DBI::dbExecute(
    con,
    spod_read_sql(glue::glue("v{ver}-nt-{zones}-raw-csv-view.sql"))
  )

  # create ENUMs

  # zones ENUMs from uniqe ids of relevant zones
  spatial_data <- spod_get_zones(
    zones,
    ver = ver,
    data_dir = data_dir,
    quiet = TRUE
  )

  unique_ids <- unique(spatial_data$id)

  DBI::dbExecute(
    con,
    dplyr::sql(
      paste0(
        "CREATE TYPE ZONES_ENUM AS ENUM ('",
        paste0(unique_ids, collapse = "','"),
        "');"
      )
    )
  )

  # create N_TRIPS_ENUM
  DBI::dbExecute(
    con,
    spod_read_sql(glue::glue("v{ver}-nt-enum-ntrips.sql"))
  )

  # v2 only enums
  # v2 only enums
  if (ver == 2) {
    # age ENUM
    DBI::dbExecute(
      con,
      spod_read_sql(glue::glue("v{ver}-nt-enum-age.sql"))
    )

    # sex ENUM
    DBI::dbExecute(
      con,
      spod_read_sql(glue::glue("v{ver}-nt-enum-sex-{locale}.sql"))
    )
  }

  # create od_csv_clean view
  if (ver == 1 && zones == "municipios") {
    # this will be picked up by the sql loaded below if neccessary
    relations_districts_municipalities <- fs::path(
      data_dir,
      spod_subfolder_raw_data_cache(1),
      "relaciones_distrito_mitma.csv"
    )
  }
  DBI::dbExecute(
    con,
    spod_read_sql(glue::glue("v{ver}-nt-{zones}-clean-csv-view-{locale}.sql"))
  )

  return(con)
}

#' Create a duckdb overnight stays table
#'
#' @description
#' This function creates a duckdb connection to the overnight stays data stored in a folder of CSV.gz files.
#' @inheritParams spod_duckdb_od
#' @inheritParams spod_available_data
#' @inheritParams spod_download
#'
#' @return A `duckdb` connection object with 2 views:
#'
#'  * `od_csv_raw` - a raw table view of all cached CSV files with the origin-destination data that has been previously cached in $SPANISH_OD_DATA_DIR
#'
#'  * `od_csv_clean` - a cleaned-up table view of `od_csv_raw` with column names and values translated and mapped to English. This still includes all cached data.
#'
#' @keywords internal
spod_duckdb_overnight_stays <- function(
  con = DBI::dbConnect(duckdb::duckdb(), dbdir = ":memory:", read_only = FALSE),
  zones = c(
    "districts",
    "dist",
    "distr",
    "distritos",
    "municipalities",
    "muni",
    "municip",
    "municipios",
    "lua",
    "large_urban_areas",
    "gau",
    "grandes_areas_urbanas"
  ),
  ver = NULL,
  data_dir = spod_get_data_dir()
) {
  locale <- "en" # TODO: add support for Spanish, hardcode for now

  ver <- as.integer(ver)
  if (ver == 1) {
    stop(
      "Overnight stays data is only available in v2 data (2022-01-01 onwards)."
    )
  }

  zones <- match.arg(zones)
  zones <- spod_zone_names_en2es(zones)

  csv_folder <- fs::path(
    data_dir,
    spod_subfolder_raw_data_cache(ver = ver),
    "estudios_basicos",
    paste0("por-", spod_zone_names_en2es(zones)),
    "pernoctaciones",
    "ficheros-diarios"
  )

  csv_folder <- paste0(csv_folder, "/")

  # create view of csv files and preset variable types

  # create view to the raw TXT/CSV.gz files
  DBI::dbExecute(
    con,
    spod_read_sql(glue::glue("v{ver}-os-{zones}-raw-csv-view.sql"))
  )

  # create ENUMs

  # zones ENUMs for residence, these are always detailed down to "districts", despite the selected zones
  spatial_data_residence <- spod_get_zones(
    "distr",
    ver = ver,
    data_dir = data_dir,
    quiet = TRUE
  )

  unique_ids_residence <- unique(spatial_data_residence$id)

  DBI::dbExecute(
    con,
    dplyr::sql(
      paste0(
        "CREATE TYPE RESID_ZONES_ENUM AS ENUM ('",
        paste0(unique_ids_residence, collapse = "','"),
        "');"
      )
    )
  )

  # zones ENUMs for overnight stays, these always match the selected zones
  spatial_data_overnight <- spod_get_zones(
    zones,
    ver = ver,
    data_dir = data_dir,
    quiet = TRUE
  )

  unique_ids_overnight <- unique(spatial_data_overnight$id)

  DBI::dbExecute(
    con,
    dplyr::sql(
      paste0(
        "CREATE TYPE OVERNIGHT_ZONES_ENUM AS ENUM ('",
        paste0(unique_ids_overnight, collapse = "','"),
        "');"
      )
    )
  )

  DBI::dbExecute(
    con,
    spod_read_sql(glue::glue("v{ver}-os-{zones}-clean-csv-view-{locale}.sql"))
  )

  return(con)
}

#' Filter a duckdb conenction by dates
#'
#' @description
#' IMPORTANT: This function assumes that the table or view that is being filtered has separate `year`, `month` and `day` columns with integer values. This is done so that the filtering is faster on CSV files that are stored in a folder structure with hive-style `/year=2020/month=2/day=14/`.
#'
#'
#' @param con A duckdb connection
#' @param source_view_name The name of the source duckdb "view" (the virtual table, in the context of current package likely connected to a folder of CSV files).
#' @param new_view_name The name of the new duckdb "view" (the virtual table, in the context of current package likely connected to a folder of CSV files).
#' @inheritParams spod_dates_argument_to_dates_seq
#' @param source_view_name The name of the source duckdb "view" (the virtual table, in the context of current package likely connected to a folder of CSV files)
#' @keywords internal
#' @return A `duckdb` connection with original views and a new filtered view.
#'
spod_duckdb_filter_by_dates <- function(
  con,
  source_view_name,
  new_view_name,
  dates
) {
  # grab the clean(source)-view sql query
  clean_sql <- DBI::dbGetQuery(
    con,
    sprintf(
      "SELECT sql
         FROM duckdb_views()
        WHERE view_name = '%s';",
      source_view_name
    )
  )$sql[1]

  # identify the raw view behind the source view
  all_views <- DBI::dbGetQuery(
    con,
    "SELECT view_name FROM duckdb_views();"
  )$view_name
  candidates <- setdiff(all_views, source_view_name)
  raw_view <- Filter(
    function(v) grepl(paste0("\\b", v, "\\b"), clean_sql, ignore.case = TRUE),
    candidates
  )[[1]]
  if (is.null(raw_view)) {
    stop("Could not identify the raw view behind ", source_view_name)
  }

  # build the fast WHERE clause that relies on duckdb selecting files using hive style partitioning
  where_clause <- spod_sql_where_dates(dates)

  # create the raw-filtered view
  filtered_raw <- paste0(source_view_name, "_raw_filtered")
  DBI::dbExecute(
    con,
    glue::glue(
      "CREATE OR REPLACE VIEW {filtered_raw} AS
         SELECT *
           FROM {raw_view}
         {where_clause};"
    )
  )

  # strip off the CREATE...AS header so we only have the SELECT body
  select_body <- sub(
    '(?i)^CREATE(?:\\s+OR\\s+REPLACE)?\\s+VIEW\\s+"?[A-Za-z_][A-Za-z0-9_]*"?\\s+AS\\s+',
    "",
    clean_sql,
    perl = TRUE
  )

  # point that body at our filtered-raw view
  new_body <- gsub(
    paste0("\\b", raw_view, "\\b"),
    filtered_raw,
    select_body
  )

  # create the final clean-filtered view
  DBI::dbExecute(
    con,
    glue::glue(
      "CREATE OR REPLACE VIEW {new_view_name} AS
       {new_body};"
    )
  )

  return(con)
}


#' Create province names ENUM in a duckdb connection
#' @param con A `duckdb` connection.
#' @return A `duckdb` connection with `INE_PROV_NAME_ENUM` and `INE_PROV_CODE_ENUM` created.
#' @keywords internal
#'
spod_duckdb_create_province_enum <- function(con) {
  # LOAD SQL STATEMENT to create province names ENUM
  province_names_enum_sql <- readLines(
    system.file(
      "extdata/sql-queries/province_names_enum.sql",
      package = "spanishoddata"
    )
  ) |>
    paste(collapse = "\n") |>
    dplyr::sql()

  # create INE_PROV_NAME_ENUM
  DBI::dbSendStatement(
    con,
    province_names_enum_sql
  )

  # also create INE encoded INE_PROV_CODE_ENUM

  # format codes 1:52 as 2 digits with leading zeros
  ine_codes <- sprintf("%02d", 1:52)

  # create INE_PROV_CODE_ENUM
  DBI::dbSendStatement(
    con,
    dplyr::sql(
      paste0(
        "CREATE TYPE INE_PROV_CODE_ENUM AS ENUM ('",
        paste0(ine_codes, collapse = "','"),
        "');"
      )
    )
  )

  return(con)
}

#' Generate a WHERE part of an SQL query from a sequence of dates
#' @param dates A Dates vector of dates to process.
#' @return A character vector of the SQL query.
#' @keywords internal
spod_sql_where_dates <- function(dates) {
  # Extract unique year, month, and day combinations from the dates
  date_parts <- data.frame(
    year = format(dates, "%Y"),
    month = format(dates, "%m"),
    day = format(dates, "%d")
  )

  # Get distinct rows and sort them by year, month, and day
  date_parts <- date_parts[!duplicated(date_parts), ]
  date_parts <- date_parts[
    order(date_parts$year, date_parts$month, date_parts$day),
  ]

  # Create the WHERE conditions for each unique date
  where_conditions <- stats::aggregate(
    day ~ year + month,
    data = date_parts,
    FUN = function(x) paste(x, collapse = ", ")
  )
  where_conditions$condition <- paste0(
    "(year = ",
    where_conditions$year,
    " AND month = ",
    where_conditions$month,
    " AND day IN (",
    where_conditions$day,
    "))"
  )

  # Combine all conditions into a single WHERE clause
  sql_query <- paste0(
    "WHERE ",
    paste(where_conditions$condition, collapse = " OR ")
  )

  return(sql_query)
}

#' Set maximum memory and number of threads for a `DuckDB` connection
#' @param con A `duckdb` connection
#' @param max_mem_gb `integer` value of the maximum operating memory to use in GB. `NULL` by default, delegates the choice to the `DuckDB` engine which usually sets it to 80% of available memory. Caution, in HPC use, the amount of memory available to your job may be determined incorrectly by the `DuckDB` engine, so it is recommended to set this parameter explicitly according to your job's memory limits.
#' @param max_n_cpu The maximum number of threads to use. Defaults to the number of available cores minus 1.
#' @return A `duckdb` connection.
#' @keywords internal
spod_duckdb_limit_resources <- function(
  con,
  max_mem_gb = NULL,
  max_n_cpu = max(1, parallelly::availableCores() - 1)
) {
  if (!is.null(max_mem_gb)) {
    DBI::dbExecute(
      con,
      dplyr::sql(
        glue::glue("SET max_memory='{max_mem_gb}GB';")
      )
    )
  }

  DBI::dbExecute(
    con,
    dplyr::sql(
      glue::glue("SET threads='{max_n_cpu}';")
    )
  )

  return(con)
}

#' Load an SQL query, glue it, dplyr::sql it
#' @description
#' Load an SQL query from a specified file in package installation directory, glue::collapse it, glue::glue it in case of any variables that need to be replaced, and dplyr::sql it for additional safety.
#'
#' @return Text of the SQL query of class `sql`/`character`.
#' @param sql_file_name The name of the SQL file to load from the package installation directory.
#' @keywords internal
spod_read_sql <- function(sql_file_name) {
  sql_file_path <- glue::glue("extdata/sql-queries/{sql_file_name}")

  sql_query <- readLines(
    system.file(sql_file_path, package = "spanishoddata")
  ) |>
    glue::glue_collapse(sep = "\n") |>
    glue::glue(.envir = parent.frame()) |>
    dplyr::sql()

  return(sql_query)
}

#' Set temp file for DuckDB connection
#' @param con A duckdb connection
#' @param temp_path The path to the temp folder for DuckDB for \href{https://duckdb.org/2024/07/09/memory-management.html#intermediate-spilling}{intermediate spilling} in case the set memory limit and/or physical memory of the computer is too low to perform the query. By default this is set to the `temp` directory in the data folder defined by SPANISH_OD_DATA_DIR environment variable (set by [spod_set_data_dir()])). Otherwise, for queries on folders of CSV files or parquet files, the temporary path would be set to the current R working directory, which probably is undesirable, as the current working directory can be on a slow storage, or storage that may have limited space, compared to the data folder.
#' @return A `duckdb` connection.
#' @keywords internal
spod_duckdb_set_temp <- function(
  con,
  temp_path = spod_get_temp_dir()
) {
  DBI::dbExecute(
    con,
    dplyr::sql(
      glue::glue("SET temp_directory='{temp_path}';")
    )
  )

  return(con)
}

library(testthat)
library(spanishoddata)

# Use setup_test_data_dir helper instead of mocks
# This ensures we test against the real file structure and logic

test_that("spod_duckdb_od validates version", {
  expect_error(spod_duckdb_od(ver = 3), "Invalid version number")
})

test_that("spod_duckdb_od handles gau zones in v1", {
  expect_error(
    spod_duckdb_od(zones = "gau", ver = 1),
    "not available in v1 data"
  )
})

test_that("spod_duckdb_number_of_trips handles gau zones in v1", {
  expect_error(
    spod_duckdb_number_of_trips(zones = "gau", ver = 1),
    "not available in v1 data"
  )
})

test_that("spod_duckdb_od works with real fixtures (v2 dist)", {
  test_dir <- setup_test_data_dir()
  withr::defer(unlink(test_dir, recursive = TRUE))
  withr::local_envvar(c("SPANISH_OD_DATA_DIR" = test_dir))

  con <- DBI::dbConnect(duckdb::duckdb(), ":memory:")
  on.exit(DBI::dbDisconnect(con, shutdown = TRUE))

  res <- spod_duckdb_od(con = con, ver = 2, zones = "dist")

  expect_identical(res, con)

  views <- DBI::dbGetQuery(
    con,
    "SELECT view_name FROM duckdb_views()"
  )$view_name
  expect_true("od_csv_raw" %in% views)
  expect_true("od_csv_clean" %in% views)

  count <- DBI::dbGetQuery(con, "SELECT count(*) as n FROM od_csv_raw")$n
  expect_true(count >= 0)
})

test_that("spod_duckdb_number_of_trips works with real fixtures (v2 dist)", {
  test_dir <- setup_test_data_dir()
  withr::defer(unlink(test_dir, recursive = TRUE))
  withr::local_envvar(c("SPANISH_OD_DATA_DIR" = test_dir))

  nt_path <- file.path(
    test_dir,
    "raw_data_cache/v2/estudios_basicos/por-distritos/personas/ficheros-diarios/year=2022/month=2/day=1"
  )
  dir.create(nt_path, recursive = TRUE)

  gz_file <- file.path(nt_path, "dummy.csv.gz")
  gz_con <- gzfile(gz_file, "w")
  writeLines(
    "fecha|zona_pernoctacion|edad|sexo|numero_viajes|personas\n20220201|01|1|1|2|10.5",
    gz_con
  )
  close(gz_con)

  con <- DBI::dbConnect(duckdb::duckdb(), ":memory:")
  on.exit(DBI::dbDisconnect(con, shutdown = TRUE))

  res <- spod_duckdb_number_of_trips(con = con, ver = 2, zones = "dist")

  expect_identical(res, con)

  views <- DBI::dbGetQuery(
    con,
    "SELECT view_name FROM duckdb_views()"
  )$view_name
  # The view names might be different for NT
  expect_true(any(grepl("_csv_raw", views)))
  expect_true(any(grepl("_csv_clean", views)))

  # Verify schema of the clean view
  clean_view <- views[grepl("_csv_clean", views)][1]
  schema <- DBI::dbGetQuery(con, paste("DESCRIBE", clean_view))
  expect_true("n_trips" %in% schema$column_name)
})

test_that("spod_duckdb_overnight_stays fails for v1", {
  expect_error(spod_duckdb_overnight_stays(ver = 1), "only available in v2")
})

test_that("spod_duckdb_overnight_stays works with real fixtures (v2 dist)", {
  test_dir <- setup_test_data_dir()
  withr::defer(unlink(test_dir, recursive = TRUE))
  withr::local_envvar(c("SPANISH_OD_DATA_DIR" = test_dir))

  # Create Hive directory structure for 'pernoctaciones'
  os_path <- file.path(
    test_dir,
    "raw_data_cache/v2/estudios_basicos/por-distritos/pernoctaciones/ficheros-diarios/year=2022/month=2/day=1"
  )
  dir.create(os_path, recursive = TRUE)

  # Create valid GZIP file
  gz_file <- file.path(os_path, "dummy.csv.gz")
  gz_con <- gzfile(gz_file, "w")
  writeLines(
    "fecha|zona_residencia|zona_pernoctacion|personas\n20220201|01|02|10",
    gz_con
  )
  close(gz_con)

  con <- DBI::dbConnect(duckdb::duckdb(), ":memory:")
  on.exit(DBI::dbDisconnect(con, shutdown = TRUE))

  res <- spod_duckdb_overnight_stays(con = con, ver = 2, zones = "dist")
  expect_identical(res, con)

  views <- DBI::dbGetQuery(
    con,
    "SELECT view_name FROM duckdb_views()"
  )$view_name
  expect_true(any(grepl("_csv_raw", views)))
})

test_that("spod_duckdb_filter_by_dates logic works with real SQL generation", {
  con <- DBI::dbConnect(duckdb::duckdb(), ":memory:")
  on.exit(DBI::dbDisconnect(con, shutdown = TRUE))

  DBI::dbExecute(
    con,
    "CREATE VIEW raw_view AS SELECT 2020 as year, 1 as month, 1 as day, 'a' as val"
  )
  DBI::dbExecute(con, "CREATE VIEW source_view AS SELECT * FROM raw_view")

  dates <- as.Date("2020-01-01")

  res <- spod_duckdb_filter_by_dates(
    con = con,
    source_view_name = "source_view",
    new_view_name = "filtered_view",
    dates = dates
  )

  expect_identical(res, con)

  count <- DBI::dbGetQuery(con, "SELECT count(*) as n FROM filtered_view")$n
  expect_equal(count, 1)

  dates_exclude <- as.Date("2021-01-01")
  res <- spod_duckdb_filter_by_dates(
    con = con,
    source_view_name = "source_view",
    new_view_name = "filtered_view_empty",
    dates = dates_exclude
  )
  count_empty <- DBI::dbGetQuery(
    con,
    "SELECT count(*) as n FROM filtered_view_empty"
  )$n
  expect_equal(count_empty, 0)
})

test_that("spod_sql_where_dates generates correct WHERE clause", {
  dates <- as.Date(c("2020-01-01", "2020-01-02", "2020-02-15"))
  res <- spod_sql_where_dates(dates)

  expect_match(res, "WHERE")
  expect_match(res, "year = 2020 AND month = 01 AND day IN \\(01, 02\\)")
  expect_match(res, "year = 2020 AND month = 02 AND day IN \\(15\\)")
})

test_that("spod_duckdb_limit_resources executes valid SQL", {
  con <- DBI::dbConnect(duckdb::duckdb(), ":memory:")
  on.exit(DBI::dbDisconnect(con, shutdown = TRUE))

  spod_duckdb_limit_resources(con, max_mem_gb = 1, max_n_cpu = 1)

  threads <- DBI::dbGetQuery(
    con,
    "SELECT value FROM duckdb_settings() WHERE name = 'threads'"
  )$value
  expect_equal(as.integer(threads), 1)
})

test_that("spod_duckdb_set_temp executes valid SQL", {
  con <- DBI::dbConnect(duckdb::duckdb(), ":memory:")
  on.exit(DBI::dbDisconnect(con, shutdown = TRUE))

  tmp <- tempdir()
  spod_duckdb_set_temp(con, temp_path = tmp)

  res <- DBI::dbGetQuery(
    con,
    "SELECT value FROM duckdb_settings() WHERE name = 'temp_directory'"
  )$value
  expect_true(grepl(basename(tmp), res))
})

#' View codebooks for v1 and v2 open mobility data
#' 
#' @description
#' 
#' `r lifecycle::badge("stable")`
#' 
#' Opens relevant vignette with a codebook for v1 (2020-2021) and v2 (2022 onwards) data or provide a webpage if vignette is missing.
#' 
#' 
#' @param ver An `integer` or `numeric` value. The version of the data. Defaults to 1. Can be `1` for v1 (2020-2021) data and 2 for v2 (2022 onwards) data.
#' @return Nothing, opens vignette if it is installed. If vignette is missing, prints a message with a link to a webpage with the codebook.
#' @export
#' @examples
#' 
#' # View codebook for v1 (2020-2021) data
#' spod_codebook(ver = 1)
#' 
#' # View codebook for v2 (2022 onwards) data
#' spod_codebook(ver = 2)
#' 
#' @importFrom utils vignette
spod_codebook <- function(ver = 1) {
  # Validate input
  checkmate::assertIntegerish(ver, max.len = 1)
  if (!ver %in% c(1, 2)) {
    stop("Invalid version number. Must be 1 (for v1 2020-2021 data) or 2 (for v2 2022 onwards).")
  }

  if (ver == 1){
    help <- vignette(
      topic = "v1-2020-2021-mitma-data-codebook",
      package = "spanishoddata"
    )
    if( inherits(help, what = "vignette") ){
      return(help)
    } else {
      message("For some reason the codebook was not installed with the package. Please refer to the online version at: https://ropenspain.github.io/spanishoddata/articles/v1-2020-2021-mitma-data-codebook.html")
    }
  } else if (ver == 2) {
    help <- vignette(
      topic = "v2-2022-onwards-mitma-data-codebook",
      package = "spanishoddata"
    ) 
    if( inherits(help, what = "vignette") ){
      return(help)
    } else {
      message("For some reason the codebook was not installed with the package. Please refer to the online version at: https://ropenspain.github.io/spanishoddata/articles/v2-2022-onwards-mitma-data-codebook.html")
    }
  }
}

test_that("spod_codebook works for v1 and matches vignette topic", {
  
  # Mock utils::vignette to return a fake vignette object or NULL
  local_mock_vignette <- function(topic_to_find, env = parent.frame()) {
    mock_vignette <- function(topic, package) {
      if (topic == topic_to_find && package == "spanishoddata") {
        # Return a structure that mimics a vignette object
        structure(list(topic = topic, package = package), class = "vignette")
      } else {
        # Return something that doesn't inherit "vignette" (e.g. NULL or empty list)
        return(NULL)
      }
    }
    
    # We use local_mocked_bindings for utils::vignette if possible.
    # But since utils is a base package, testthat might restrict it.
    # However, testthat::local_mocked_bindings supports package namespaces.
    # Let's try mocking in the 'spanishoddata' namespace where it's imported/used.
    # Since spod_codebook uses `vignette(...)`, it's resolved from `utils` unless imported.
    # The file has `@importFrom utils vignette`, so it is in the package imports.
    # We can mock it in the package namespace.
    
    testthat::local_mocked_bindings(
      vignette = mock_vignette,
      .package = "spanishoddata",
      .env = env
    )
  }
  
  local_mock_vignette("v1-2020-2021-mitma-data-codebook")
  
  # Success case
  res <- spod_codebook(ver = 1)
  expect_s3_class(res, "vignette")
  expect_equal(res$topic, "v1-2020-2021-mitma-data-codebook")
  
  # Fallback case (mock failure for v2 in this scope)
  # This relies on the mock NOT matching v2
  expect_message(
    spod_codebook(ver = 2),
    "codebook was not installed.*https://ropenspain.github.io/spanishoddata/articles/v2"
  )
})

test_that("spod_codebook works for v2", {
  local_mock_vignette <- function(topic_to_find, env = parent.frame()) {
    mock_vignette <- function(topic, package) {
      if (topic == topic_to_find && package == "spanishoddata") {
        structure(list(topic = topic, package = package), class = "vignette")
      } else {
        return(NULL)
      }
    }
    testthat::local_mocked_bindings(
      vignette = mock_vignette,
      .package = "spanishoddata",
      .env = env
    )
  }
  
  local_mock_vignette("v2-2022-onwards-mitma-data-codebook")
  
  res <- spod_codebook(ver = 2)
  expect_s3_class(res, "vignette")
  expect_equal(res$topic, "v2-2022-onwards-mitma-data-codebook")
})

test_that("spod_codebook handles invalid inputs", {
  expect_error(spod_codebook(ver = 3), "Invalid version number")
  expect_error(spod_codebook(ver = "a"), "Must be of type 'integerish'")
})

# Helper functions for setting up test data fixtures

#' Setup a temporary data directory with bundled fixtures
#'
#' This function creates a temporary directory and populates it with
#' the bundled test fixtures (metadata, zones, OD data) to mimic
#' a populated cache directory.
#'
#' @return The path to the temporary directory
setup_test_data_dir <- function() {
  test_dir <- tempfile("spod_test_")
  dir.create(test_dir, recursive = TRUE)

  fixture_path <- system.file("testdata", package = "spanishoddata")
  if (fixture_path == "") {
    # Fallback for when running interactively without package load?
    fixture_path <- testthat::test_path("../../inst/testdata")
  }

  if (!dir.exists(fixture_path)) {
    # Try another fallback for R CMD check where inst/ is gone
    fixture_path <- testthat::test_path("../..", "testdata")
  }

  if (!dir.exists(fixture_path)) {
    stop("Could not find test fixtures at: ", fixture_path)
  }

  # 1. Setup Metadata Cache
  # -----------------------
  # Copy mock metadata to expected cache location so spod_available_data finds it
  meta_cache <- file.path(test_dir, spod_subfolder_metadata_cache())
  dir.create(meta_cache, recursive = TRUE)

  # The cache logic looks for pattern: metadata_s3_v{ver}_YYYY-MM-DD.rds
  # We copy our mock file to a name that guarantees it's picked up as "latest"
  mock_meta_v2 <- file.path(
    fixture_path,
    "metadata/available_data_s3_v2_mock.rds"
  )
  if (file.exists(mock_meta_v2)) {
    # Use a future date to ensure it's NEVER considered stale by spod_available_data_s3
    target_name <- paste0("metadata_s3_v2_", Sys.Date() + 30, ".rds")
    if (!file.copy(mock_meta_v2, file.path(meta_cache, target_name))) {
      stop("Failed to copy v2 mock metadata")
    }
  }

  mock_meta_v1 <- file.path(
    fixture_path,
    "metadata/available_data_s3_v1_mock.rds"
  )
  if (file.exists(mock_meta_v1)) {
    target_name <- paste0("metadata_s3_v1_", Sys.Date() + 30, ".rds")
    if (!file.copy(mock_meta_v1, file.path(meta_cache, target_name))) {
      stop("Failed to copy v1 mock metadata")
    }
  }

  # 2. Setup Zone Cache
  # -------------------
  # spod_get_zones checks 'clean_data/v{ver}/zones'
  clean_zones <- file.path(test_dir, "clean_data/v2/zones")
  dir.create(clean_zones, recursive = TRUE)

  # Copy all bundled zone GPKGs
  zone_fixtures <- list.files(
    file.path(fixture_path, "clean_data/v2/zones"),
    full.names = TRUE
  )
  if (length(zone_fixtures) > 0) {
    if (!all(file.copy(zone_fixtures, clean_zones))) {
      stop("Failed to copy zone fixtures")
    }
  }

  # 3. Setup Raw OD Data
  # --------------------
  # spod_duckdb_od expects 'raw_data_cache/v{ver}/estudios_basicos/por-distritos/viajes/ficheros-diarios/'
  # and 'raw_data_cache/v{ver}/distritos/viajes/ficheros-diarios/'

  # Recreate Hive structure
  raw_od_hive <- file.path(
    test_dir,
    "raw_data_cache/v2/estudios_basicos/por-distritos/viajes/ficheros-diarios/year=2022/month=2/day=1"
  )
  dir.create(raw_od_hive, recursive = TRUE)
  hive_src <- file.path(fixture_path, "raw_data/v2/od_hive_data.csv.gz")
  if (!file.exists(hive_src)) {
    stop("Hive OD fixture missing at: ", hive_src)
  }
  if (!file.copy(hive_src, file.path(raw_od_hive, "data.csv.gz"))) {
    stop("Failed to copy Hive OD fixture")
  }

  # Recreate standard daily structure
  raw_od_day <- file.path(
    test_dir,
    "raw_data_cache/v2/distritos/viajes/ficheros-diarios"
  )
  dir.create(raw_od_day, recursive = TRUE)
  day_src <- file.path(fixture_path, "raw_data/v2/od_dist_day.csv.gz")
  if (!file.exists(day_src)) {
    stop("Daily OD fixture missing at: ", day_src)
  }
  if (
    !file.copy(
      day_src,
      file.path(raw_od_day, "viajes_distrito_2022-02-01.csv.gz")
    )
  ) {
    stop("Failed to copy daily OD fixture")
  }

  return(test_dir)
}

#' Context manager for mocked downloads
#'
#' Prevents actual downloads during tests.
#'
#' @param code The code to run
local_mock_download <- function(env = parent.frame()) {
  # Mock spod_download to do nothing (files are already in test_dir)
  testthat::local_mocked_bindings(
    spod_download = function(...) invisible(NULL),
    .package = "spanishoddata",
    .env = env
  )
}

#' Setup a temporary directory with a sample parquet file
#'
#' Creates a parquet file using arrow (if available) or duckdb
#' to test spod_connect with parquet folders.
#'
#' @return Path to the parquet folder
setup_parquet_dir <- function() {
  parquet_dir <- tempfile("spod_parquet_")
  dir.create(parquet_dir)

  # Create a simple data frame
  df <- data.frame(
    id = 1:10,
    val = letters[1:10]
  )

  parquet_file <- file.path(parquet_dir, "test.parquet")

  # Try using arrow if installed, otherwise skip (or fail if this is essential)
  # But for testing `spod_connect` which connects to parquet, we need a parquet file.
  # We can use duckdb to write it since we depend on duckdb.

  con <- DBI::dbConnect(duckdb::duckdb())
  on.exit(DBI::dbDisconnect(con, shutdown = TRUE))

  DBI::dbWriteTable(con, "temp_df", df)
  DBI::dbExecute(
    con,
    paste0("COPY temp_df TO '", parquet_file, "' (FORMAT PARQUET)")
  )

  return(parquet_dir)
}

#' Setup a temporary duckdb database file
#'
#' @return Path to the .duckdb file
setup_duckdb_file <- function() {
  db_file <- tempfile("spod_db_", fileext = ".duckdb")
  con <- DBI::dbConnect(duckdb::duckdb(), dbdir = db_file)

  df <- data.frame(id = 1:5, val = LETTERS[1:5])
  DBI::dbWriteTable(con, "my_table", df)

  DBI::dbDisconnect(con, shutdown = TRUE)
  return(db_file)
}

library(testthat)
library(spanishoddata)

# Use real data directory fixtures
# Tests integration of metadata cache -> decision logic -> download call

test_that("spod_download validates inputs", {
  expect_error(spod_download(type = "invalid"), "Must be element of set")
  expect_error(spod_download(zones = "invalid"), "Must be element of set")
})

test_that("spod_download uses local metadata fixture (no mock on available_data)", {
  test_dir <- setup_test_data_dir()
  withr::defer(unlink(test_dir, recursive = TRUE))
  withr::local_envvar(c("SPANISH_OD_DATA_DIR" = test_dir))

  mock_meta_path <- system.file("testdata", "metadata", "available_data_s3_v2_mock.rds", package = "spanishoddata")
  if (mock_meta_path == "") mock_meta_path <- testthat::test_path("../../inst/testdata/metadata/available_data_s3_v2_mock.rds")
  mock_meta <- readRDS(mock_meta_path)

  testthat::local_mocked_bindings(
    spod_available_data_s3 = function(...) mock_meta,
    read_data_links_memoised = function(...) tibble::tibble(),
    spod_get_valid_dates = function(ver, ...) {
      if (ver == 1) {
        return(as.Date(character(0)))
      }
      if (ver == 2) {
        return(as.Date("2022-02-01"))
      }
      return(as.Date(character(0)))
    },
    .package = "spanishoddata"
  )

  called <- FALSE
  testthat::local_mocked_bindings(
    spod_download_in_batches = function(files) {
      called <<- TRUE
      return(files)
    },
    .package = "spanishoddata"
  )

  spod_download(type = "od", zones = "dist", dates = "2022-02-01", quiet = TRUE)

  expect_false(called)
})

test_that("spod_download triggers download when local file is missing", {
  test_dir <- setup_test_data_dir()
  withr::defer(unlink(test_dir, recursive = TRUE))
  withr::local_envvar(c("SPANISH_OD_DATA_DIR" = test_dir))

  target_file <- file.path(
    test_dir,
    "raw_data_cache/v2/estudios_basicos/por-distritos/viajes/ficheros-diarios/year=2022/month=2/day=1/data.csv.gz"
  )
  if (file.exists(target_file)) {
    unlink(target_file)
  }

  mock_meta_path <- system.file("testdata", "metadata", "available_data_s3_v2_mock.rds", package = "spanishoddata")
  if (mock_meta_path == "") mock_meta_path <- testthat::test_path("../../inst/testdata/metadata/available_data_s3_v2_mock.rds")
  mock_meta <- readRDS(mock_meta_path)

  testthat::local_mocked_bindings(
    spod_available_data_s3 = function(...) mock_meta,
    read_data_links_memoised = function(...) tibble::tibble(),
    spod_get_valid_dates = function(ver, ...) {
      if (ver == 1) {
        return(as.Date(character(0)))
      }
      if (ver == 2) {
        return(as.Date("2022-02-01"))
      }
      return(as.Date(character(0)))
    },
    .package = "spanishoddata"
  )

  download_files <- NULL
  testthat::local_mocked_bindings(
    spod_download_in_batches = function(files) {
      download_files <<- files
      files$downloaded <- TRUE
      files$complete_download <- TRUE
      files$local_file_size <- files$file_size_bytes
      return(files)
    },
    .package = "spanishoddata"
  )

  spod_download(type = "od", zones = "dist", dates = "2022-02-01", quiet = TRUE)

  expect_true(!is.null(download_files))
  expect_equal(nrow(download_files), 1)
  expect_equal(download_files$data_ymd[1], as.Date("2022-02-01"))
})

test_that("spod_download_in_batches executes download logic", {
  test_dir <- withr::local_tempdir()
  dest_file <- file.path(test_dir, "test.gz")

  files <- tibble::tibble(
    target_url = "https://example.com/file",
    local_path = dest_file,
    file_size_bytes = 10L,
    data_ymd = as.Date("2022-01-01")
  )

  testthat::local_mocked_bindings(
    spod_download_file = function(url, destfile, ...) {
      # Fixed typo: dest -> destfile
      writeBin(raw(10), destfile)
      return(0L)
    },
    spod_interactive = function() FALSE,
    .package = "spanishoddata"
  )

  res <- spod_download_in_batches(files, show_progress = FALSE)

  expect_true(res$complete_download)
  expect_true(file.exists(dest_file))
  expect_equal(file.info(dest_file)$size, 10L)
})

test_that("spod_download integration with file size mismatch", {
  test_dir <- setup_test_data_dir()
  withr::defer(unlink(test_dir, recursive = TRUE))
  withr::local_envvar(c("SPANISH_OD_DATA_DIR" = test_dir))

  target_file <- file.path(
    test_dir,
    "raw_data_cache/v2/estudios_basicos/por-distritos/viajes/ficheros-diarios/year=2022/month=2/day=1/data.csv.gz"
  )
  writeLines("short", target_file)

  mock_meta_path <- system.file("testdata", "metadata", "available_data_s3_v2_mock.rds", package = "spanishoddata")
  if (mock_meta_path == "") mock_meta_path <- testthat::test_path("../../inst/testdata/metadata/available_data_s3_v2_mock.rds")
  mock_meta <- readRDS(mock_meta_path)

  testthat::local_mocked_bindings(
    spod_available_data_s3 = function(...) mock_meta,
    read_data_links_memoised = function(...) tibble::tibble(),
    spod_get_valid_dates = function(ver, ...) {
      if (ver == 1) {
        return(as.Date(character(0)))
      }
      if (ver == 2) {
        return(as.Date("2022-02-01"))
      }
      return(as.Date(character(0)))
    },
    .package = "spanishoddata"
  )

  download_triggered <- FALSE
  testthat::local_mocked_bindings(
    spod_download_in_batches = function(files) {
      download_triggered <<- TRUE
      files$downloaded <- TRUE
      files$complete_download <- TRUE
      return(files)
    },
    .package = "spanishoddata"
  )

  spod_download(
    type = "od",
    zones = "dist",
    dates = "2022-02-01",
    check_local_files = TRUE,
    quiet = TRUE
  )

  expect_true(download_triggered)
})

test_that("spod_get_zones returns cached zones", {
  test_dir <- setup_test_data_dir()
  withr::defer(unlink(test_dir, recursive = TRUE))
  withr::local_envvar(c("SPANISH_OD_DATA_DIR" = test_dir))

  # Mock download to ensure no network calls happen
  local_mock_download()

  # This should pick up "distritos_mitma.gpkg" from clean_data/v2/zones/
  zones <- spod_get_zones("dist", ver = 2, quiet = TRUE)

  expect_s3_class(zones, "sf")
  expect_true("id" %in% names(zones))
  
  # Check for our specific filtered filtered IDs (Alava)
  # fixtures generated in data-raw/generate_fixtures.R filtered for 01001, 01002
  expect_true("01001" %in% zones$id)
  expect_true("01002" %in% zones$id)
  
  # Ensure we only have the sampled data (2 rows) to confirm it used our fixture
  expect_equal(nrow(zones), 2)
  
  # Check geometry column
  expect_s3_class(sf::st_geometry(zones), "sfc_MULTIPOLYGON")
})

test_that("spod_get_zones returns municipalities", {
  test_dir <- setup_test_data_dir()
  withr::defer(unlink(test_dir, recursive = TRUE))
  withr::local_envvar(c("SPANISH_OD_DATA_DIR" = test_dir))
  local_mock_download()
  
  zones <- spod_get_zones("muni", ver = 2, quiet = TRUE)
  expect_s3_class(zones, "sf")
  expect_true(nrow(zones) > 0)
})

test_that("spod_get_zones returns GAUs", {
  test_dir <- setup_test_data_dir()
  withr::defer(unlink(test_dir, recursive = TRUE))
  withr::local_envvar(c("SPANISH_OD_DATA_DIR" = test_dir))
  local_mock_download()
  
  zones <- spod_get_zones("gau", ver = 2, quiet = TRUE)
  expect_s3_class(zones, "sf")
  expect_true(nrow(zones) > 0)
})

test_that("spod_get_zones validates version", {
  expect_error(spod_get_zones("dist", ver = 3), "Invalid version number")
})

test_that("spod_get returns DuckDB connection with cached data", {
  # Mock valid dates to include our fixture date
  orig_valid_dates <- spod_get_valid_dates
  mock_valid_dates <- function(ver, ...) {
    if (ver == 2) {
      return(as.Date("2022-02-01"))
    }
    return(as.Date(character(0)))
  }

  assignInNamespace(
    "spod_get_valid_dates",
    mock_valid_dates,
    ns = "spanishoddata"
  )
  on.exit(
    {
      assignInNamespace(
        "spod_get_valid_dates",
        orig_valid_dates,
        ns = "spanishoddata"
      )
    },
    add = TRUE
  )

  # Setup test data
  test_dir <- setup_test_data_dir()
  withr::defer(unlink(test_dir, recursive = TRUE))
  withr::local_envvar(c("SPANISH_OD_DATA_DIR" = test_dir))

  # Mock download (should not be triggered if cache hits, but good safety)
  local_mock_download()

  # Use assignInNamespace to robustly mock internal function
  # This works even for internal calls within the package
  mock_func <- function(...) {
    # Find the mock file we copied in setup_test_data_dir
    meta_dir <- file.path(test_dir, spod_subfolder_metadata_cache())
    files <- list.files(meta_dir, pattern = "metadata_s3_v2", full.names = TRUE)
    if (length(files) == 0) {
      stop("Mock metadata not found in test dir")
    }
    readRDS(files[1])
  }

  # Save original and mock
  orig_func <- get(
    "spod_available_data_v2",
    envir = asNamespace("spanishoddata")
  )
  # Note: assignInNamespace is locked for base/recommended packages but fine for us
  assignInNamespace("spod_available_data_v2", mock_func, ns = "spanishoddata")
  withr::defer(assignInNamespace(
    "spod_available_data_v2",
    orig_func,
    ns = "spanishoddata"
  ))

  # 1. Test OD data retrieval
  # Filter to the specific date we have in fixtures
  dates <- "2022-02-01"

  # "dist" zones + "od" type
  # This triggers:
  # - spod_available_data (mocked above)
  # - spod_get_zones (hits mock GPKG)
  # - spod_duckdb_od (hits mock CSV.gz)

  res <- spod_get(type = "od", zones = "dist", dates = dates)

  expect_s3_class(res, "tbl_duckdb_connection")

  # Collect to verify content
  df <- res |> dplyr::collect()

  expect_true(nrow(df) > 0)
  expect_true("n_trips" %in% names(df))
  expect_true("id_origin" %in% names(df))

  # Verify we got our specific filtered data (Alava IDs)
  expect_true(any(df$id_origin %in% c("01001", "01002")))

  # Clean up connection
  spod_disconnect(res)
})

test_that("spod_get handles missing dates gracefully (mocked)", {
  test_dir <- setup_test_data_dir()
  withr::defer(unlink(test_dir, recursive = TRUE))
  withr::local_envvar(c("SPANISH_OD_DATA_DIR" = test_dir))
  local_mock_download()

  # Mock again for this test block (assignInNamespace persists per session but we deferred revert)
  # But we defer happens at end of test block.
  # So we re-apply here.

  mock_func <- function(...) {
    meta_dir <- file.path(test_dir, spod_subfolder_metadata_cache())
    files <- list.files(meta_dir, pattern = "metadata_s3_v2", full.names = TRUE)
    if (length(files) == 0) {
      stop("Mock metadata not found in test dir")
    }
    readRDS(files[1])
  }

  orig_func <- get(
    "spod_available_data_v2",
    envir = asNamespace("spanishoddata")
  )
  # Note: assignInNamespace is locked for base/recommended packages but fine for us
  assignInNamespace("spod_available_data_v2", mock_func, ns = "spanishoddata")
  withr::defer(assignInNamespace(
    "spod_available_data_v2",
    orig_func,
    ns = "spanishoddata"
  ))

  # Request a date we don't have
  dates <- "2025-01-01"

  # Since our mock metadata ONLY has 2022-02-01 (or whatever we put in fixture),
  # 2025-01-01 will be considered "not valid" or "not available".

  expect_error(
    spod_get(type = "od", zones = "dist", dates = dates),
    "All requested dates are missing"
  )
})

test_that("spod_convert works for duckdb output", {
  skip_if_not(requireNamespace("duckdb", quietly = TRUE))
  
  # Setup fixture env
  test_dir <- setup_test_data_dir()
  on.exit(unlink(test_dir, recursive = TRUE))
  spod_set_data_dir(test_dir, quiet = TRUE)
  local_mock_download()
  
  # Mock valid dates to ensure version inference works
  local_mock_valid_dates <- function(env = parent.frame()) {
    mock_dates <- seq.Date(as.Date("2022-02-01"), as.Date("2022-02-10"), by = "day")
    mock_func <- function(ver = NULL) {
      if (ver == 2) return(mock_dates)
      if (ver == 1) return(as.Date(character(0))) # V1 empty
      stop("Invalid version")
    }
    orig <- get("spod_get_valid_dates", envir = asNamespace("spanishoddata"))
    assignInNamespace("spod_get_valid_dates", mock_func, ns = "spanishoddata")
    withr::defer(assignInNamespace("spod_get_valid_dates", orig, ns = "spanishoddata"), envir = env)
  }
  local_mock_valid_dates()
  
  # Use a date range covered by our mocked fixtures 
  # Fixtures: v2/zones/distritos_mitma.gpkg and v2/.../viajes_distrito_2022-02-01.csv.gz
  # Note: The fixture CSV is 2022-02-01 in Hive structure.
  

  
  # Define output path
  out_file <- file.path(test_dir, "output.duckdb")
  
  # Note: spod_convert calls spod_get which triggers the SQL logic.
  # We know spod_get currently fails on SQL view ordering in v2-od-dist-clean-csv-view-en.sql
  # However, let's write the test assuming it might work or to confirm failure mode directly.
  # If it fails, that confirms we really need to fix the internal SQL bug.
  
  expect_error(
    spod_convert(
      type = "od",
      zones = "distr",
      dates = "2022-02-01",
      save_format = "duckdb",
      save_path = out_file,
      overwrite = TRUE,
      quiet = TRUE
    ),
    NA # Expect no error
  )
  
  expect_true(file.exists(out_file))
  con <- spod_connect(out_file)
  expect_s3_class(con, "tbl_duckdb_connection")
  spod_disconnect(con)
})

test_that("spod_convert works for parquet output", {
  skip_if_not(requireNamespace("duckdb", quietly = TRUE))
  
  test_dir <- setup_test_data_dir()
  on.exit(unlink(test_dir, recursive = TRUE))
  spod_set_data_dir(test_dir, quiet = TRUE)
  local_mock_download()
  
  # Mock valid dates here too
  local_mock_valid_dates <- function(env = parent.frame()) {
    mock_dates <- seq.Date(as.Date("2022-02-01"), as.Date("2022-02-10"), by = "day")
    mock_func <- function(ver = NULL) {
      if (ver == 2) return(mock_dates)
      if (ver == 1) return(as.Date(character(0))) 
      stop("Invalid version")
    }
    orig <- get("spod_get_valid_dates", envir = asNamespace("spanishoddata"))
    assignInNamespace("spod_get_valid_dates", mock_func, ns = "spanishoddata")
    withr::defer(assignInNamespace("spod_get_valid_dates", orig, ns = "spanishoddata"), envir = env)
  }
  local_mock_valid_dates()
  
  out_dir <- file.path(test_dir, "output_parquet")
  
  expect_error(
    spod_convert(
      type = "od",
      zones = "distr",
      dates = "2022-02-01",
      save_format = "parquet",
      save_path = out_dir,
      overwrite = TRUE,
      quiet = TRUE
    ),
    NA
  )
  
  if (dir.exists(out_dir)) {
    # check for partition structure (year=2022...)
    files <- list.files(out_dir, recursive = TRUE)
    expect_true(any(grepl("year=2022", files)))
  }
})

test_that("spod_convert infers format from save_path (.duckdb)", {
  skip_if_not(requireNamespace("duckdb", quietly = TRUE))
  
  test_dir <- withr::local_tempdir()
  
  local_mocked_bindings(
    spod_get = function(...) {
      # Return a mock duckdb connection with src$con
      con <- DBI::dbConnect(duckdb::duckdb(), ":memory:")
      DBI::dbExecute(con, "CREATE VIEW od_csv_clean_filtered AS SELECT 1::INTEGER as id")
      tbl <- dplyr::tbl(con, "od_csv_clean_filtered")
      return(tbl)
    }
  )
  
  local_mocked_bindings(
    spod_dates_argument_to_dates_seq = function(...) as.Date("2022-01-01"),
    spod_infer_data_v_from_dates = function(...) 2
  )
  
  out_file <- file.path(test_dir, "output.duckdb")
  
  # Call with save_format=NULL (will default), should infer from path not used here
  result <- spod_convert(
    type = "od",
    zones = "dist",
    dates = "2022-01-01",
    save_format = NULL,  # Will default to duckdb
    save_path = out_file,
    overwrite = TRUE,
    quiet = TRUE
  )
  
  expect_equal(result, out_file)
  # expect_true(file.exists(out_file)) # Skipped because mock uses in-memory DB
})

test_that("spod_convert infers format from save_path (parquet folder)", {
  skip_if_not(requireNamespace("duckdb", quietly = TRUE))
  
  test_dir <- withr::local_tempdir()
  
  local_mocked_bindings(
    spod_get = function(...) {
      con <- DBI::dbConnect(duckdb::duckdb(), ":memory:")
      DBI::dbExecute(con, "CREATE VIEW od_csv_clean_filtered AS SELECT 1::INTEGER as id, 2022::INTEGER as year, 1::INTEGER as month, 1::INTEGER as day")
      tbl <- dplyr::tbl(con, "od_csv_clean_filtered")
      return(tbl)
    }
  )
  
  local_mocked_bindings(
    spod_dates_argument_to_dates_seq = function(...) as.Date("2022-01-01"),
    spod_infer_data_v_from_dates = function(...) 2
  )
  
  out_dir <- file.path(test_dir, "output.parquet")
  
  result <- spod_convert(
    type = "od",
    zones = "dist",
    dates = "2022-01-01",
    save_format = NULL,  # Will default to duckdb
    save_path = out_dir,
    overwrite = TRUE,
    quiet = TRUE
  )
  
  expect_equal(result, out_dir)
})

test_that("spod_convert prompts for duckdb overwrite when file exists", {
  skip_if_not(requireNamespace("duckdb", quietly = TRUE))
  
  test_dir <- withr::local_tempdir()
  out_file <- file.path(test_dir, "existing.duckdb")
  
  # Create an existing file
  writeLines("existing", out_file)
  
  local_mocked_bindings(
    spod_get = function(...) {
      con <- DBI::dbConnect(duckdb::duckdb(), ":memory:")
      DBI::dbExecute(con, "CREATE TABLE od_csv_clean_filtered (id INTEGER)")
      tbl <- dplyr::tbl(con, "od_csv_clean_filtered")
      return(tbl)
    }
  )
  
  local_mocked_bindings(
    spod_dates_argument_to_dates_seq = function(...) as.Date("2022-01-01"),
    spod_infer_data_v_from_dates = function(...) 2
  )
  
  # Mock spod_readline to say "no"
  local_mocked_bindings(
    spod_readline = function(...) "no"
  )
  
  result <- NULL
  msgs <- capture_messages(
    result <- spod_convert(
      type = "od",
      zones = "dist",
      dates = "2022-01-01",
      save_path = out_file,
      overwrite = FALSE,
      quiet = FALSE
    )
  )
  expect_match(msgs, "Exiting without overwriting", all = FALSE)
  
  expect_null(result)
})

test_that("spod_convert overwrites duckdb when confirmed", {
  skip_if_not(requireNamespace("duckdb", quietly = TRUE))
  
  test_dir <- withr::local_tempdir()
  out_file <- file.path(test_dir, "existing.duckdb")
  
  writeLines("existing", out_file)
  
  local_mocked_bindings(
    spod_get = function(...) {
      con <- DBI::dbConnect(duckdb::duckdb(), ":memory:")
      DBI::dbExecute(con, "CREATE VIEW od_csv_clean_filtered AS SELECT 1::INTEGER as id")
      tbl <- dplyr::tbl(con, "od_csv_clean_filtered")
      return(tbl)
    }
  )
  
  local_mocked_bindings(
    spod_dates_argument_to_dates_seq = function(...) as.Date("2022-01-01"),
    spod_infer_data_v_from_dates = function(...) 2
  )
  
  # Mock spod_readline to say "yes"
  local_mocked_bindings(
    spod_readline = function(...) "yes"
  )
  
  result <- spod_convert(
    type = "od",
    zones = "dist",
    dates = "2022-01-01",
    save_path = out_file,
    overwrite = FALSE,
    quiet = TRUE
  )
  
  expect_equal(result, out_file)
})

test_that("spod_convert errors on update mode for duckdb", {
  # Mock to prevent reaching real functions
  local_mocked_bindings(
    spod_dates_argument_to_dates_seq = function(...) as.Date("2022-01-01"),
    spod_infer_data_v_from_dates = function(...) 2
  )
  
  # Test the early validation that happens before spod_get is called
  expect_error(
    spod_convert(
      type = "od",
      zones = "dist",
      dates = "2022-01-01",
      save_format = "duckdb",
      save_path = tempfile(fileext = ".duckdb"),
      overwrite = "update",
      quiet = TRUE
    ),
    "not supported"
  )
})

test_that("spod_convert handles parquet update mode (cancel)", {
  skip_if_not(requireNamespace("duckdb", quietly = TRUE))
  
  test_dir <- withr::local_tempdir()
  out_dir <- file.path(test_dir, "existing_parquet")
  dir.create(out_dir, recursive = TRUE)
  
  # Create existing parquet file
  dir.create(file.path(out_dir, "year=2022/month=1/day=1"), recursive = TRUE)
  writeLines("existing", file.path(out_dir, "year=2022/month=1/day=1/data_0.parquet"))
  
  local_mocked_bindings(
    spod_get = function(...) {
      con <- DBI::dbConnect(duckdb::duckdb(), ":memory:")
      DBI::dbExecute(con, "CREATE VIEW od_csv_clean_filtered AS SELECT 1::INTEGER as id, 2022::INTEGER as year, 1::INTEGER as month, 1::INTEGER as day")
      tbl <- dplyr::tbl(con, "od_csv_clean_filtered")
      return(tbl)
    },
    spod_dates_argument_to_dates_seq = function(...) as.Date("2022-01-01"),
    spod_infer_data_v_from_dates = function(...) 2
  )
  
  # Mock spod_readline to say "cancel"
  local_mocked_bindings(
    spod_readline = function(...) "c"
  )
  
  result <- NULL
  msgs <- capture_messages(
    result <- spod_convert(
      type = "od",
      zones = "dist",
      dates = "2022-01-01",
      save_path = out_dir,
      save_format = "parquet",
      overwrite = FALSE,
      quiet = FALSE
    )
  )
  expect_match(msgs, "Cancelled", all = FALSE)
  
  expect_null(result)
})

test_that("spod_convert handles parquet update mode (delete)", {
  skip_if_not(requireNamespace("duckdb", quietly = TRUE))
  
  test_dir <- withr::local_tempdir()
  out_dir <- file.path(test_dir, "existing_parquet")
  dir.create(out_dir, recursive = TRUE)
  
  # Create existing parquet file
  dir.create(file.path(out_dir, "year=2022/month=1/day=1"), recursive = TRUE)
  writeLines("existing", file.path(out_dir, "year=2022/month=1/day=1/data_0.parquet"))
  
  local_mocked_bindings(
    spod_get = function(...) {
      con <- DBI::dbConnect(duckdb::duckdb(), ":memory:")
      DBI::dbExecute(con, "CREATE VIEW od_csv_clean_filtered AS SELECT 1::INTEGER as id, 2022::INTEGER as year, 1::INTEGER as month, 1::INTEGER as day")
      tbl <- dplyr::tbl(con, "od_csv_clean_filtered")
      return(tbl)
    },
    spod_dates_argument_to_dates_seq = function(...) as.Date("2022-01-01"),
    spod_infer_data_v_from_dates = function(...) 2
  )
  
  # Mock spod_readline to say "delete"
  local_mocked_bindings(
    spod_readline = function(...) "d"
  )
  
  result <- NULL
  capture_messages(
    result <- spod_convert(
      type = "od",
      zones = "dist",
      dates = "2022-01-01",
      save_path = out_dir,
      save_format = "parquet",
      overwrite = FALSE,
      quiet = FALSE
    )
  )
  
  expect_equal(result, out_dir)
})

test_that("spod_convert validates inconsistent save_format and save_path", {
  # Mock to prevent reaching real functions
  local_mocked_bindings(
    spod_dates_argument_to_dates_seq = function(...) as.Date("2022-01-01"),
    spod_infer_data_v_from_dates = function(...) 2
  )
  
  expect_error(
    spod_convert(
      type = "od",
      zones = "dist",
      dates = "2022-01-01",
      save_format = "duckdb",
      save_path = "/path/to/folder",  # No .duckdb extension
      quiet = TRUE
    ),
    "folder.*duckdb"
  )
})

test_that("spod_connect works with parquet folder", {
  skip_if_not(requireNamespace("duckdb", quietly = TRUE))
  
  # Setup parquet fixture
  parquet_dir <- setup_parquet_dir()
  on.exit(unlink(parquet_dir, recursive = TRUE))
  
  # Connect
  con <- spod_connect(parquet_dir)
  expect_s3_class(con, "tbl_duckdb_connection")
  
  # Verify data access
  df <- dplyr::collect(con)
  expect_equal(nrow(df), 10)
  expect_true(all(c("id", "val") %in% names(df)))
  
  # Disconnect
  spod_disconnect(con)
})

test_that("spod_connect works with duckdb file", {
  skip_if_not(requireNamespace("duckdb", quietly = TRUE))
  
  # Setup duckdb fixture
  db_file <- setup_duckdb_file()
  on.exit(unlink(db_file))
  
  # Connect (auto-detect table)
  con <- spod_connect(db_file)
  expect_s3_class(con, "tbl_duckdb_connection")
  
  df <- dplyr::collect(con)
  expect_equal(nrow(df), 5)
  
  spod_disconnect(con)
})

test_that("spod_connect respects resource limits", {
  skip_if_not(requireNamespace("duckdb", quietly = TRUE))
  
  parquet_dir <- setup_parquet_dir()
  on.exit(unlink(parquet_dir, recursive = TRUE))
  
  # Connect with limits
  con <- spod_connect(parquet_dir, max_mem_gb = 1, max_n_cpu = 1)
  
  # Verify limits (tricky to verify exact values without querying pragma, 
  # but we can check if it runs without error)
  expect_s3_class(con, "tbl_duckdb_connection")
  
  spod_disconnect(con)
})

test_that("spod_disconnect closes connection", {
  skip_if_not(requireNamespace("duckdb", quietly = TRUE))
  
  parquet_dir <- setup_parquet_dir()
  on.exit(unlink(parquet_dir, recursive = TRUE))
  
  con <- spod_connect(parquet_dir)
  raw_con <- con$src$con
  
  spod_disconnect(con)
  
  expect_false(DBI::dbIsValid(raw_con))
})

test_that("spod_convert handles 'NA' in renta column using mini fixtures", {
  skip_if_not(requireNamespace("duckdb", quietly = TRUE))
  
  # 1. Setup temporary data directory using the package helper
  test_dir <- setup_test_data_dir()
  on.exit(unlink(test_dir, recursive = TRUE))
  spod_set_data_dir(test_dir, quiet = TRUE)
  
  # 2. Mock valid dates and downloads
  local_mock_valid_dates <- function(env = parent.frame()) {
    mock_dates <- as.Date("2022-02-01")
    mock_func <- function(ver = NULL) {
      if (is.null(ver) || ver == 2) return(mock_dates)
      return(as.Date(character(0)))
    }
    orig <- get("spod_get_valid_dates", envir = asNamespace("spanishoddata"))
    assignInNamespace("spod_get_valid_dates", mock_func, ns = "spanishoddata")
    withr::defer(assignInNamespace("spod_get_valid_dates", orig, ns = "spanishoddata"), envir = env)
  }
  local_mock_valid_dates()
  local_mock_download()
  
  # 3. Modify one of the mini fixtures to include 'NA' in renta for testing
  raw_file <- file.path(test_dir, "raw_data_cache/v2/estudios_basicos/por-distritos/viajes/ficheros-diarios/year=2022/month=2/day=1/data.csv.gz")
  
  # Use cross-platform R functions to handle gz files
  tmp_csv <- tempfile(fileext = ".csv")
  gz_con <- gzfile(raw_file, "rt")
  lines <- readLines(gz_con)
  close(gz_con)
  
  # Change the first data row's renta to 'NA'
  parts <- strsplit(lines[2], "|", fixed = TRUE)[[1]]
  parts[11] <- "NA" # renta
  lines[2] <- paste(parts, collapse = "|")
  
  writeLines(lines, tmp_csv)
  # Re-compress
  dest_gz <- gzfile(raw_file, "wt")
  writeLines(lines, dest_gz)
  close(dest_gz)
  
  # 4. Run conversion
  # Use a name without dots other than .duckdb to avoid spod_connect guessing issues
  out_db_name <- "testrentana"
  out_db <- file.path(test_dir, paste0(out_db_name, ".duckdb"))
  expect_error(
    spod_convert(
      type = "od", zones = "distr", dates = "2022-02-01",
      save_format = "duckdb", save_path = out_db,
      overwrite = TRUE, quiet = TRUE
    ),
    NA
  )
  
  # 5. Verify data
  # Provide explicit table name to avoid guessing failures in CI
  tbl_con <- spod_connect(out_db, target_table_name = out_db_name)
  res <- tbl_con %>% dplyr::collect()
  spod_disconnect(tbl_con)
  expect_true(any(is.na(res$income)))
})

test_that("spod_convert handles 'NA' in distancia column using mini fixtures", {
  skip_if_not(requireNamespace("duckdb", quietly = TRUE))
  
  # 1. Setup temporary data directory
  test_dir <- setup_test_data_dir()
  on.exit(unlink(test_dir, recursive = TRUE))
  spod_set_data_dir(test_dir, quiet = TRUE)
  
  # 2. Mock valid dates and downloads
  local_mock_valid_dates <- function(env = parent.frame()) {
    mock_dates <- as.Date("2022-02-01")
    mock_func <- function(ver = NULL) {
      if (is.null(ver) || ver == 2) return(mock_dates)
      return(as.Date(character(0)))
    }
    orig <- get("spod_get_valid_dates", envir = asNamespace("spanishoddata"))
    assignInNamespace("spod_get_valid_dates", mock_func, ns = "spanishoddata")
    withr::defer(assignInNamespace("spod_get_valid_dates", orig, ns = "spanishoddata"), envir = env)
  }
  local_mock_valid_dates()
  local_mock_download()
  
  # 3. Modify one of the mini fixtures to include 'NA' in distancia for testing
  raw_file <- file.path(test_dir, "raw_data_cache/v2/estudios_basicos/por-distritos/viajes/ficheros-diarios/year=2022/month=2/day=1/data.csv.gz")
  
  # Use cross-platform R functions to handle gz files
  gz_con <- gzfile(raw_file, "rt")
  lines <- readLines(gz_con)
  close(gz_con)
  
  # Change the first data row's distancia to 'NA'
  parts <- strsplit(lines[2], "|", fixed = TRUE)[[1]]
  parts[5] <- "NA" # distancia
  lines[2] <- paste(parts, collapse = "|")
  
  # Re-compress
  dest_gz <- gzfile(raw_file, "wt")
  writeLines(lines, dest_gz)
  close(dest_gz)
  
  # 4. Run conversion
  out_db_name <- "testdistanciana"
  out_db <- file.path(test_dir, paste0(out_db_name, ".duckdb"))
  expect_error(
    spod_convert(
      type = "od", zones = "distr", dates = "2022-02-01",
      save_format = "duckdb", save_path = out_db,
      overwrite = TRUE, quiet = TRUE
    ),
    NA
  )
  
  # 5. Verify data
  tbl_con <- spod_connect(out_db, target_table_name = out_db_name)
  res <- tbl_con %>% dplyr::collect()
  spod_disconnect(tbl_con)
  expect_true(any(is.na(res$distance)))
})

# skip("Requires network access - skipped to prevent downloading RSS.xml")

test_that("spod_available_data retrieves metadata from cache", {
  test_dir <- setup_test_data_dir()
  withr::defer(unlink(test_dir, recursive = TRUE))
  
  # Set data dir to our test dir with fixtures
  withr::local_envvar(c("SPANISH_OD_DATA_DIR" = test_dir))
  
  mock_meta_path_v2 <- system.file("testdata", "metadata", "available_data_s3_v2_mock.rds", package = "spanishoddata")
  if (mock_meta_path_v2 == "") mock_meta_path_v2 <- testthat::test_path("../../inst/testdata/metadata/available_data_s3_v2_mock.rds")
  mock_meta_v2 <- readRDS(mock_meta_path_v2)

  # Mock to prevent real RSS.xml download
  local_mocked_bindings(
    spod_available_data_s3 = function(ver, ...) {
      if (ver == 2) return(mock_meta_v2)
      return(tibble::tibble())
    },
    read_data_links_memoised = function(...) tibble::tibble(),
    spod_get_latest_v2_file_list = function(...) NULL,
    spod_get_latest_v1_file_list = function(...) NULL
  )
  
  # Fetch data (should hit the cached RDS)
  # We suppress messages to keep test output clean
  meta <- suppressMessages(spod_available_data(ver = 2, check_local_files = FALSE))
  
  expect_s3_class(meta, "tbl_df")
  expect_true(nrow(meta) > 0)
  
  # Check key columns
  expect_true("target_url" %in% names(meta))
  expect_true("pub_ts" %in% names(meta))
  
  # Our mock fixture has exactly 1 row (from generate_fixtures.R)
  # But we might have more if we copied other things? No, script made 1 row.
  expect_equal(nrow(meta), 1)
  
  # Check if URL looks correct
  expect_true(grepl("20220201_data\\.csv\\.gz", meta$target_url[1]))
})

test_that("spod_available_data retrieves v1 metadata from cache", {
  test_dir <- setup_test_data_dir()
  withr::defer(unlink(test_dir, recursive = TRUE))
  
  withr::local_envvar(c("SPANISH_OD_DATA_DIR" = test_dir))
  
  mock_meta_path_v1 <- system.file("testdata", "metadata", "available_data_s3_v1_mock.rds", package = "spanishoddata")
  if (mock_meta_path_v1 == "") mock_meta_path_v1 <- testthat::test_path("../../inst/testdata/metadata/available_data_s3_v1_mock.rds")
  mock_meta_v1 <- readRDS(mock_meta_path_v1)

  # Mock to prevent real RSS.xml download
  local_mocked_bindings(
    spod_available_data_s3 = function(ver, ...) {
      if (ver == 1) return(mock_meta_v1)
      return(tibble::tibble())
    },
    read_data_links_memoised = function(...) tibble::tibble(),
    spod_get_latest_v2_file_list = function(...) NULL,
    spod_get_latest_v1_file_list = function(...) NULL
  )
  
  # Fetch data for v1
  meta <- suppressMessages(spod_available_data(ver = 1, check_local_files = FALSE))
  
  expect_s3_class(meta, "tbl_df")
  expect_true(nrow(meta) > 0)
  
  # Check key columns
  expect_true("target_url" %in% names(meta))
  expect_true("pub_ts" %in% names(meta))
  expect_true("local_path" %in% names(meta))
  
  # Check if it picked up our mock
  expect_true(any(grepl("20200214_maestra_1_distritos", meta$target_url)))
})

# internal wrappers to allow mocking of base/utils functions
spod_download_file <- function(...) utils::download.file(...)
spod_interactive <- function(...) base::interactive(...)
spod_readline <- function(...) base::readline(...)

#' Download the data files of specified type, zones, and dates
#' @title Download the data files of specified type, zones, and dates
#' @description
#'
#' `r lifecycle::badge("stable")`
#'
#' This function downloads the data files of the specified type, zones, dates and data version.
#'
#'
#' For detailed data descriptions, see package vignettes using [`spod_codebook(ver = 1)`][spod_codebook] and [`spod_codebook(ver = 2)`][spod_codebook] and official methodology documents in **References** section.
#'
#' @template references
#'
#' @param type The type of data to download. Can be `"origin-destination"` (or ust `"od"`), or `"number_of_trips"` (or just `"nt"`) for v1 data. For v2 data `"overnight_stays"` (or just `"os"`) is also available. More data types to be supported in the future. See codebooks for v1 and v2 data in vignettes with `spod_codebook(1)` and `spod_codebook(2)` (\link{spod_codebook}).
#' @param zones The zones for which to download the data. Can be `"districts"` (or `"dist"`, `"distr"`, or the original Spanish `"distritos"`) or `"municipalities"` (or `"muni"`, `"municip"`, or the original Spanish `"municipios"`) for both data versions. Additionaly, these can be `"large_urban_areas"` (or `"lua"`, or the original Spanish `"grandes_areas_urbanas"`, or `"gau"`) for v2 data (2022 onwards).
#' @inheritParams spod_dates_argument_to_dates_seq
#' @param data_dir The directory where the data is stored. Defaults to the value returned by `spod_get_data_dir()` which returns the value of the environment variable `SPANISH_OD_DATA_DIR` or a temporary directory if the variable is not set. To set the data directory, use \link{spod_set_data_dir}.
#' @param max_download_size_gb The maximum download size in gigabytes. Defaults to 1.
#' @param return_local_file_paths Logical. If `TRUE`, the function returns a character vector of the paths to the downloaded files. If `FALSE`, the function returns `NULL`.
#' @param ignore_missing_dates Logical. If `TRUE`, the function will not raise an error if the some of the specified dates are missing. Any dates that are missing will be skipped, however the data for any valid dates will be acquired. Defaults to `FALSE`.
#' @param check_local_files Logical. Whether to check the file size of local files against known remote file sizes on the Amazon S3 storage. Defaults to `TRUE`, which fetches the metadata from Amazon S3. This setting ensures your downloaded files are not broken, so it is recommended to keep it `TRUE`.
#' @inheritParams global_quiet_param
#'
#' @return Nothing. If `return_local_file_paths = TRUE`, a `character` vector of the paths to the downloaded files.
#'
#' @export
#' @examplesIf interactive()
#' \donttest{
#'
#' # Set data dir for file downloads
#' spod_set_data_dir(tempdir())
#'
#' # Download the number of trips on district level for the a date range in March 2020
#' spod_download(
#'   type = "number_of_trips", zones = "districts",
#'   dates = c(start = "2020-03-20", end = "2020-03-21")
#' )
#'
#' # Download the number of trips on district level for select dates in 2020 and 2021
#' spod_download(
#'   type = "number_of_trips", zones = "dist",
#'   dates = c("2020-03-20", "2020-03-24", "2021-03-20", "2021-03-24")
#' )
#'
#' # Download the number of trips on municipality level using regex for a date range in March 2020
#' # (the regex will capture the dates 2020-03-20 to 2020-03-24)
#' spod_download(
#'   type = "number_of_trips", zones = "municip",
#'   dates = "2020032[0-4]"
#' )
#' }

spod_download <- function(
  type = c(
    "od",
    "origin-destination",
    "os",
    "overnight_stays",
    "nt",
    "number_of_trips"
  ),
  zones = c(
    "districts",
    "dist",
    "distr",
    "distritos",
    "municipalities",
    "muni",
    "municip",
    "municipios",
    "lua",
    "large_urban_areas",
    "gau",
    "grandes_areas_urbanas"
  ),
  dates = NULL,
  max_download_size_gb = 1, # 1GB
  data_dir = spod_get_data_dir(),
  quiet = FALSE,
  return_local_file_paths = FALSE,
  ignore_missing_dates = FALSE,
  check_local_files = TRUE
) {
  # Validate inputs
  checkmate::assert_choice(
    type,
    choices = c(
      "od",
      "origin-destination",
      "os",
      "overnight_stays",
      "nt",
      "number_of_trips"
    )
  )
  checkmate::assert_choice(
    zones,
    choices = c(
      "districts",
      "dist",
      "distr",
      "distritos",
      "municipalities",
      "muni",
      "municip",
      "municipios",
      "lua",
      "large_urban_areas",
      "gau",
      "grandes_areas_urbanas"
    )
  )
  checkmate::assert_number(max_download_size_gb, lower = 0.1)
  checkmate::assert_directory_exists(data_dir, access = "rw")
  checkmate::assert_flag(quiet)
  checkmate::assert_flag(return_local_file_paths)
  checkmate::assert_flag(ignore_missing_dates)

  # normalise zones
  zones <- spod_zone_names_en2es(zones)

  # simple null check is enough here, as spod_dates_arugument_to_dates_seq will do additional checks anyway
  if (is.null(dates)) {
    message(
      "`dates` argument is undefined. Please set `dates='cached_v1'` or `dates='cached_v2'` to convert all data that was previously downloaded. Alternatively, specify at least one date between 2020-02-14 and 2021-05-09 (for v1 data) or between 2022-01-01 onwards (for v2). Any missing data will be downloaded before conversion. For more details on the dates argument, see ?spod_download."
    )
  }

  dates_to_use <- spod_dates_argument_to_dates_seq(dates = dates)

  # check version
  ver <- spod_infer_data_v_from_dates(
    dates = dates_to_use,
    ignore_missing_dates = ignore_missing_dates
  )
  # this leads to a second call to an internal spod_get_valid_dates() which in turn causes a second call to spod_available_data(). This results in reading xml files with metadata for the second time. This is not optimal and should be fixed.

  if (isFALSE(quiet)) {
    message("Data version detected from dates: ", ver)
  }

  # convert english data type names to spanish words used in the default data paths
  type <- match.arg(type)
  type <- spod_match_data_type_for_local_folders(type = type, ver = ver)

  # get the available  data list while checking for files already cached on disk
  # TODO: make requests faster by providing filtering prefix for Amazon S3 to only get the files we need?
  available_data <- spod_available_data(
    ver = ver,
    check_local_files = check_local_files,
    data_dir = data_dir,
    quiet = quiet,
    use_s3 = TRUE
  )

  # match the available_data to type, zones, version and dates
  if (ver == 1) {
    requested_files <- available_data[
      # selecting districts files for v1 to avoid issues with municipalities # this is to address the bugs described in detail in:
      # https://www.ekotov.pro/mitma-data-issues/issues/011-v1-tpp-mismatch-zone-ids-in-table-and-spatial-data.html
      # https://www.ekotov.pro/mitma-data-issues/issues/012-v1-tpp-district-files-in-municipality-folders.html
      # the decision was to use distrcit data and aggregate it to replicate municipal data
      grepl(
        paste0("v", ver, ".*", type, ".*distritos"),
        available_data$local_path
      ) &
        available_data$data_ymd %in% dates_to_use,
    ]
  } else if (ver == 2) {
    requested_files <- available_data[
      grepl(
        paste0("v", ver, ".*", zones, ".*", type),
        available_data$local_path
      ) &
        available_data$data_ymd %in% dates_to_use,
    ]
  }

  # compare file sizes
  requested_files <- requested_files |>
    dplyr::mutate(
      complete_download = dplyr::if_else(
        condition = .data$file_size_bytes == as.numeric(.data$local_file_size),
        true = TRUE,
        false = FALSE,
        missing = FALSE
      )
    )

  files_to_download <- requested_files |>
    dplyr::filter(.data$complete_download == FALSE)

  # only download files if some are missing
  if (nrow(files_to_download) > 0) {
    total_size_to_download_gb <- round(
      sum(files_to_download$remote_file_size_mb / 1024, na.rm = TRUE),
      4
    )
    # warn if more than 1 GB is to be downloaded
    if (total_size_to_download_gb > max_download_size_gb) {
      message(glue::glue(
        "Approximately {round(total_size_to_download_gb, 2)} GB of data will be downloaded."
      ))
      # ask for confirmation
      response <- spod_readline(
        prompt = "Are you sure you would like to continue with this download? (yes/no) "
      )
      response <- tolower(response) %in% c("y", "yes", "Yes")
      if (!response) {
        message(glue::glue(
          "Exiting without downloading missing files by user request. Requested data download size is larger than {max_download_size_gb} GB. Please increase `max_download_size_gb` parameter when running the function again."
        ))
        return()
      }
    }

    if (isFALSE(quiet)) {
      message(glue::glue(
        "Downloading approximately {round(total_size_to_download_gb, 2)} GB of data."
      ))
    }

    # pre-generate target paths for the files to download
    # and create all directories in the path
    # fs::dir_create(
    #   unique(fs::path_dir(files_to_download$local_path)),
    #   recurse = TRUE
    # )

    # download the missing files
    # TODO: disable mass curl::multi_download due to multiple failures on some connections
    # downloaded_files <- curl::multi_download(
    #   urls = files_to_download$target_url,
    #   destfiles = files_to_download$local_path,
    #   progress = TRUE,
    #   resume = TRUE,
    #   multiplex = FALSE
    # )

    # use curl::multi_download in a loop on one file at a time with manual progress bar
    downloaded_files <- spod_download_in_batches(
      files_to_download
    )
    # downloaded_files <- spod_multi_download_with_progress(
    #   files_to_download
    # )

    # set download status for downloaded files as TRUE in requested_files
    # update the columns in requested_files to have new local file size, downloaded and download complete status columns
    requested_files <- requested_files |>
      dplyr::rows_update(
        downloaded_files |>
          dplyr::select(
            "local_path",
            "local_file_size",
            "downloaded",
            "complete_download"
          ),
        by = "local_path"
      )

    if (isFALSE(quiet)) {
      message("Retrieved data for requested dates.")
    }
  }

  if (return_local_file_paths) {
    return(requested_files$local_path)
  }
}


#' Download multiple files with progress bar in parallel
#'
#' @description
#' Download multiple files with a progress bar. Retries failed downloads up to 3 times. Downloads are in parallel and in batches to show progress. First 10 Mb of a file is downloaded to check the speed.
#'
#' @param files_to_download A data frame with columns `target_url`, `local_path` and `file_size_bytes`.
#' @param batch_size Numeric. Number of files to download at a time.
#' @param bar_width Numeric. Width of the progress bar.
#' @param chunk_size Numeric. Number of bytes to download at a time for speed test.
#' @param show_progress Logical. Whether to show the progress bar.
#' @param max_retries Integer. Maximum number of retries for failed downloads.
#' @param timeout Numeric. Timeout in seconds for each download.
#' @return A data frame with columns `target_url`, `local_path`, `file_size_bytes` and `local_file_size`.
#'
#' @keywords internal
#'
spod_download_in_batches <- function(
  files_to_download,
  batch_size = 5,
  bar_width = 20,
  chunk_size = 1024 * 1024,
  test_size = 10 * 1024 * 1024, # 10 MB test
  max_retries = 3L,
  timeout = 900,
  show_progress = spod_interactive() && !isTRUE(getOption("knitr.in.progress"))
) {
  # Check interactive context
  if (!interactive() || isTRUE(getOption("knitr.in.progress"))) {
    show_progress <- FALSE
  }

  original_timeout <- getOption("timeout")
  options(timeout = timeout)
  on.exit(options(timeout = original_timeout))

  # Sort and ensure directories exist
  files_to_download <- files_to_download[order(files_to_download$data_ymd), ]
  dirs <- unique(dirname(files_to_download$local_path))
  for (d in dirs) {
    if (!dir.exists(d)) dir.create(d, recursive = TRUE, showWarnings = FALSE)
  }

  # Totals for progress
  total_files <- nrow(files_to_download)
  total_expected_bytes <- sum(files_to_download$file_size_bytes, na.rm = TRUE)
  total_gb <- total_expected_bytes / 2^30

  # Ensure tracking columns
  if (!"downloaded" %in% names(files_to_download)) {
    files_to_download$downloaded <- FALSE
  }
  if (!"local_file_size" %in% names(files_to_download)) {
    files_to_download$local_file_size <- NA_integer_
  }
  if (!"complete_download" %in% names(files_to_download)) {
    files_to_download$complete_download <- FALSE
  }

  # Pre-skip already-completed files
  cum_bytes <- 0L
  files_counted <- 0L
  to_download <- logical(total_files)
  for (i in seq_len(total_files)) {
    expb <- files_to_download$file_size_bytes[i]
    dest <- files_to_download$local_path[i]
    if (isTRUE(files_to_download$complete_download[i])) {
      actual <- if (file.exists(dest)) file.info(dest)$size else NA_integer_
      if (!is.na(actual) && actual == expb) {
        cum_bytes <- cum_bytes + expb
        files_counted <- files_counted + 1L
        to_download[i] <- FALSE
        next
      } else {
        files_to_download$complete_download[i] <- FALSE
      }
    }
    if (
      isTRUE(files_to_download$downloaded[i]) &&
        !is.na(files_to_download$local_file_size[i]) &&
        files_to_download$local_file_size[i] == expb
    ) {
      files_to_download$complete_download[i] <- TRUE
      cum_bytes <- cum_bytes + expb
      files_counted <- files_counted + 1L
      to_download[i] <- FALSE
    } else if (file.exists(dest) && file.info(dest)$size == expb) {
      files_to_download$downloaded[i] <- TRUE
      files_to_download$complete_download[i] <- TRUE
      files_to_download$local_file_size[i] <- expb
      cum_bytes <- cum_bytes + expb
      files_counted <- files_counted + 1L
      to_download[i] <- FALSE
    } else {
      to_download[i] <- TRUE
    }
  }

  # ETA formatter
  format_eta <- function(eta) {
    if (is.na(eta) || eta <= 0 || !is.finite(eta)) {
      return("--")
    }
    if (eta > 3600) {
      sprintf("%.1fh", eta / 3600)
    } else if (eta > 60) {
      sprintf("%.0fm", eta / 60)
    } else {
      sprintf("%.0fs", eta)
    }
  }

  # Progress bar redraw
  redraw_bar <- function(bytes_done, speed_bytes = NULL) {
    pct <- bytes_done / total_expected_bytes
    nfill <- max(floor(pct * bar_width), 1L)
    bar <- if (nfill < bar_width) {
      paste0(strrep("=", nfill), ">", strrep(" ", bar_width - nfill - 1))
    } else {
      strrep("=", bar_width)
    }
    elapsed <- as.numeric(Sys.time() - start_time, "secs")
    speed_bps <- if (!is.null(speed_bytes)) {
      speed_bytes
    } else {
      (bytes_done / max(elapsed, 0.1))
    }
    speed_mb <- speed_bps / 2^20
    eta_secs <- if (speed_bps > 0) {
      (total_expected_bytes - bytes_done) / speed_bps
    } else {
      NA
    }
    eta <- format_eta(eta_secs)

    msg <- sprintf(
      "Downloading: [%s] %3.0f%% (%d/%d files, %.2f/%.2f GB, %.1f MB/s, ETA: %s)",
      bar,
      pct * 100,
      files_counted,
      total_files,
      bytes_done / 2^30,
      total_gb,
      speed_mb,
      eta
    )
    cat(sprintf("\r%-120s", msg))
    utils::flush.console()
  }

  # Speed test with smallest file, read up to test_size without saving
  if (show_progress) {
    start_time <- Sys.time()
    redraw_bar(cum_bytes)
    rem_idx <- which(to_download)
    if (length(rem_idx) > 0) {
      sizes <- files_to_download$file_size_bytes[rem_idx]
      first_i <- rem_idx[which.min(sizes)]
      url1 <- files_to_download$target_url[first_i]

      bytes_read <- 0L
      t0 <- Sys.time()
      con <- url(url1, "rb")
      repeat {
        to_read <- min(chunk_size, test_size - bytes_read)
        if (to_read <= 0) {
          break
        }
        chunk <- readBin(con, "raw", n = to_read)
        if (length(chunk) == 0) {
          break
        }
        bytes_read <- bytes_read + length(chunk)
      }
      close(con)
      t1 <- Sys.time()

      dt <- as.numeric(t1 - t0, "secs")
      bps <- if (dt > 0) bytes_read / dt else NA
      redraw_bar(cum_bytes, speed_bytes = bps)
    }
  }

  # Prepare batches of all to_download (including test file)
  rem <- which(to_download)
  idx_batches <- split(rem, ceiling(seq_along(rem) / batch_size))

  # Download batches in parallel via libcurl
  for (batch in idx_batches) {
    urls <- files_to_download$target_url[batch]
    dests <- files_to_download$local_path[batch]

    res <- spod_download_file(
      url = urls,
      destfile = dests,
      method = "libcurl",
      mode = "wb",
      quiet = TRUE
    )
    if (length(res) == 1L) {
      res <- rep(res, length(batch))
    }

    # Retry on size mismatch or initial failure
    for (k in seq_along(batch)) {
      i <- batch[k]
      expected <- files_to_download$file_size_bytes[i]
      dest <- dests[k]
      # if initial status not ok or size mismatch
      if (
        res[k] != 0L || !file.exists(dest) || file.info(dest)$size != expected
      ) {
        attempts <- 1L
        while (attempts < max_retries) {
          attempts <- attempts + 1L
          status2 <- spod_download_file(
            url = urls[k],
            destfile = dest,
            method = "libcurl",
            mode = "wb",
            quiet = TRUE
          )
          actual <- if (file.exists(dest)) file.info(dest)$size else NA_integer_
          if (status2 == 0L && identical(actual, expected)) {
            res[k] <- 0L
            break
          }
        }
      }
    }

    for (k in seq_along(batch)) {
      i <- batch[k]
      expected <- files_to_download$file_size_bytes[i]
      if (
        res[k] == 0L &&
          file.exists(dests[k]) &&
          file.info(dests[k])$size == expected
      ) {
        sz <- file.info(dests[k])$size
        files_to_download$downloaded[i] <- TRUE
        files_to_download$complete_download[i] <- TRUE
        files_to_download$local_file_size[i] <- sz
        cum_bytes <- cum_bytes + sz
      } else {
        warning(sprintf(
          "Failed to download %s after %d attempts (got %s bytes, expected %s)",
          urls[k],
          max_retries,
          if (file.exists(dests[k])) file.info(dests[k])$size else NA,
          files_to_download$file_size_bytes[i]
        ))
      }
      files_counted <- files_counted + 1L
      if (show_progress) redraw_bar(cum_bytes)
    }
  }

  if (show_progress) {
    cat("\nAll downloads complete.\n")
  }
  files_to_download
}

test_that("spod_download integration flow for v1", {
  test_dir <- withr::local_tempdir()

  # Mock dependencies
  testthat::local_mocked_bindings(
    spod_available_data = function(...) {
      tibble::tibble(
        data_ymd = as.Date("2021-02-01"),
        local_path = "v1/dist/type/distritos/data.csv.gz",
        downloaded = FALSE,
        file_size_bytes = 100L,
        remote_file_size_mb = 0.0001,
        local_file_size = NA_real_,
        target_url = "https://mock/v1"
      )
    },
    spod_dates_argument_to_dates_seq = function(...) as.Date("2021-02-01"),
    spod_infer_data_v_from_dates = function(...) 1,
    spod_match_data_type_for_local_folders = function(...) "type",
    spod_zone_names_en2es = function(z) z,
    # Mock batch downloader to just write file and return success
    spod_download_in_batches = function(files) {
      writeLines(
        "test",
        file.path(test_dir, "v1/dist/type/distritos/data.csv.gz")
      )
      files$downloaded <- TRUE
      files$complete_download <- TRUE
      files$local_file_size <- 100L
      return(files)
    }
  )

  dir.create(file.path(test_dir, "v1/dist/type/distritos"), recursive = TRUE)

  msgs <- capture_messages(
    res <- spod_download(
      type = "od",
      zones = "dist",
      dates = "2021-02-01",
      data_dir = test_dir,
      quiet = FALSE,
      return_local_file_paths = TRUE
    )
  )
  expect_match(msgs, "Retrieved data", all = FALSE)

  names(res) <- NULL # strip names if any
  expect_equal(basename(res), "data.csv.gz")
})

test_that("spod_download integration flow for v2", {
  test_dir <- withr::local_tempdir()

  # Mock dependencies
  testthat::local_mocked_bindings(
    spod_available_data = function(...) {
      tibble::tibble(
        data_ymd = as.Date("2022-02-01"),
        local_path = "v2/dist/type/data.csv.gz",
        downloaded = FALSE,
        file_size_bytes = 100L,
        remote_file_size_mb = 0.0001,
        local_file_size = NA_real_,
        target_url = "https://mock/v2"
      )
    },
    spod_dates_argument_to_dates_seq = function(...) as.Date("2022-02-01"),
    spod_infer_data_v_from_dates = function(...) 2,
    spod_match_data_type_for_local_folders = function(...) "type",
    spod_zone_names_en2es = function(z) z,
    spod_download_in_batches = function(files) {
      writeLines("test", file.path(test_dir, "v2/dist/type/data.csv.gz"))
      files$downloaded <- TRUE
      files$complete_download <- TRUE
      files$local_file_size <- 100L
      return(files)
    }
  )

  dir.create(file.path(test_dir, "v2/dist/type"), recursive = TRUE)

  msgs <- capture_messages(
    res <- spod_download(
      type = "od",
      zones = "dist",
      dates = "2022-02-01",
      data_dir = test_dir,
      quiet = FALSE,
      return_local_file_paths = TRUE
    )
  )
  expect_match(msgs, "Retrieved data", all = FALSE)
  expect_equal(basename(res), "data.csv.gz")
})

test_that("spod_download input validation works", {
  test_dir <- withr::local_tempdir()

  # Type validation
  expect_error(spod_download(type = "invalid"), "Must be element of set")

  # Zones validation
  expect_error(spod_download(zones = "invalid"), "Must be element of set")

  # Max download size validation - need valid type/zones first
  expect_error(
    spod_download(type = "od", zones = "dist", max_download_size_gb = 0),
    "Assertion on 'max_download_size_gb' failed"
  )

  # Data dir validation
  expect_error(
    spod_download(type = "od", zones = "dist", data_dir = "/non/existent/dir"),
    "Assertion on 'data_dir' failed"
  )
})

test_that("spod_download shows message for NULL dates", {
  test_dir <- withr::local_tempdir()
  # Use try() to catch the error that happens later (dates validation) but check for the message first
  expect_message(
    try(
      spod_download(
        type = "od",
        zones = "dist",
        dates = NULL,
        data_dir = test_dir,
        quiet = FALSE
      ),
      silent = TRUE
    ),
    "dates.*argument is undefined"
  )
})

test_that("spod_download asks for confirmation for large downloads", {
  test_dir <- withr::local_tempdir()

  # Mock available data to return a large file
  testthat::local_mocked_bindings(
    spod_available_data = function(...) {
      tibble::tibble(
        data_ymd = as.Date("2022-02-01"),
        local_path = "v2/dist/type/large_file.csv.gz",
        downloaded = FALSE,
        file_size_bytes = 2 * 1024^3, # 2 GB
        remote_file_size_mb = 2048,
        local_file_size = NA_real_
      )
    },
    spod_dates_argument_to_dates_seq = function(...) as.Date("2022-02-01"),
    spod_infer_data_v_from_dates = function(...) 2,
    spod_match_data_type_for_local_folders = function(...) "type",
    # Mock spod_readline to say "no"
    spod_readline = function(...) "no"
  )

  # Mock checkmate to avoid complex setup
  testthat::local_mocked_bindings(
    spod_zone_names_en2es = function(z) z
  )

  # Should show message and return NULL when user says no
  msgs <- capture_messages(
    res <- spod_download(
      type = "od",
      zones = "dist",
      dates = "2022-02-01",
      data_dir = test_dir,
      max_download_size_gb = 1, # limit is 1GB, file is 2GB
      quiet = FALSE
    )
  )
  expect_match(msgs, "Exiting without downloading", all = FALSE)

  expect_null(res)
})

test_that("spod_download proceeds with large download when confirmed", {
  test_dir <- withr::local_tempdir()

  # Mock dependencies
  testthat::local_mocked_bindings(
    spod_available_data = function(...) {
      tibble::tibble(
        data_ymd = as.Date("2022-02-01"),
        local_path = "v2/dist/type/large_file.csv.gz",
        downloaded = FALSE,
        file_size_bytes = 2 * 1024^3, # 2 GB
        remote_file_size_mb = 2048,
        local_file_size = NA_real_
      )
    },
    spod_dates_argument_to_dates_seq = function(...) as.Date("2022-02-01"),
    spod_infer_data_v_from_dates = function(...) 2,
    spod_match_data_type_for_local_folders = function(...) "type",
    spod_zone_names_en2es = function(z) z,
    # Mock spod_readline to say "yes"
    spod_readline = function(...) "yes",
    # Mock actual download to do nothing but return success
    spod_download_in_batches = function(files) {
      files$downloaded <- TRUE
      files$complete_download <- TRUE
      files$local_file_size <- files$file_size_bytes
      return(files)
    }
  )

  # Should proceed and print download message
  msgs <- capture_messages(
    spod_download(
      type = "od",
      zones = "dist",
      dates = "2022-02-01",
      data_dir = test_dir,
      max_download_size_gb = 1,
      quiet = FALSE
    )
  )
  expect_match(msgs, "Downloading approximately", all = FALSE)
})

test_that("spod_download checks local files with complete_download logic", {
  test_dir <- withr::local_tempdir()

  # Mock data where file exists but size doesn't match
  testthat::local_mocked_bindings(
    spod_available_data = function(...) {
      tibble::tibble(
        data_ymd = as.Date("2022-02-01"),
        # Path must match regex in spod_download: v2.*dist.*type
        local_path = "v2/dist/type/test_file.csv.gz",
        downloaded = FALSE,
        file_size_bytes = 1000,
        remote_file_size_mb = 0.001,
        local_file_size = 500 # Size mismatch
      )
    },
    spod_dates_argument_to_dates_seq = function(...) as.Date("2022-02-01"),
    spod_infer_data_v_from_dates = function(...) 2,
    spod_match_data_type_for_local_folders = function(...) "type",
    spod_zone_names_en2es = function(z) z
  )

  # Mock downloader to assert it was called
  called <- FALSE
  testthat::local_mocked_bindings(
    spod_download_in_batches = function(files) {
      called <<- TRUE
      files$downloaded <- TRUE
      files$complete_download <- TRUE
      return(files)
    }
  )

  spod_download(
    type = "od",
    zones = "dist",
    dates = "2022-02-01",
    data_dir = test_dir,
    quiet = TRUE
  )

  expect_true(called)
})

test_that("spod_download_in_batches skips existing valid files", {
  test_dir <- withr::local_tempdir()
  dest_file <- file.path(test_dir, "existing.bin")
  writeBin(raw(10), dest_file)

  files <- tibble::tibble(
    target_url = "https://mock/file",
    local_path = dest_file,
    file_size_bytes = 10L,
    data_ymd = as.Date("2022-01-01"),
    downloaded = FALSE,
    local_file_size = NA_integer_,
    complete_download = NA
  )

  called <- FALSE
  testthat::local_mocked_bindings(
    spod_download_file = function(...) {
      called <<- TRUE
      return(0L)
    },
    spod_interactive = function() FALSE
  )

  res <- spod_download_in_batches(files, show_progress = FALSE)

  expect_false(called) # Should NOT call download
  expect_true(res$complete_download)
  expect_equal(res$local_file_size, 10L)
})

test_that("spod_download_in_batches handles max retries failure", {
  skip_on_cran()

  test_dir <- withr::local_tempdir()
  dest_file <- file.path(test_dir, "bad.bin")

  files <- tibble::tibble(
    target_url = "https://mock/fail",
    local_path = dest_file,
    file_size_bytes = 10L,
    data_ymd = as.Date("2022-01-01")
  )

  call_count <- 0
  testthat::local_mocked_bindings(
    spod_download_file = function(url, destfile, ...) {
      call_count <<- call_count + 1
      writeBin(raw(5), destfile) # Always wrong size
      return(0L)
    },
    spod_interactive = function() FALSE
  )

  expect_warning(
    res <- spod_download_in_batches(
      files,
      max_retries = 2,
      show_progress = FALSE
    ),
    "Failed to download"
  )

  expect_false(res$complete_download)
  expect_equal(call_count, 1 + 1) # Initial + 1 retry (max_retries=2 means 2 total attempts code-wise)
})

test_that("spod_download_in_batches runs speed test in interactive mode", {
  skip_on_cran()

  test_dir <- withr::local_tempdir()
  dest_file <- file.path(test_dir, "dest.bin")

  files <- tibble::tibble(
    target_url = "https://mock/speedtest",
    local_path = dest_file,
    file_size_bytes = 10L,
    data_ymd = as.Date("2022-01-01")
  )

  # Mock interactive to TRUE to trigger speed test
  testthat::local_mocked_bindings(
    spod_interactive = function() TRUE,
    spod_download_file = function(url, destfile, ...) {
      writeBin(raw(10), destfile)
      return(0L)
    }
  )

  # We need to mock url connection for the speed test (lines 626-638)
  # It uses url(url, "rb"). This is hard to mock without wrapping.
  # But we can use a local file URL!

  src_file <- file.path(test_dir, "source.bin")
  writeBin(as.raw(1:100), src_file)
  files$target_url <- paste0("file://", src_file)

  # Redirect output to capture progress bar
  output <- capture.output(
    res <- spod_download_in_batches(files, show_progress = TRUE, test_size = 10)
  )

  expect_true(res$complete_download)
  expect_true(res$complete_download)
})

test_that("spod_download_in_batches handles multiple files in matches", {
  skip_on_cran()

  test_dir <- withr::local_tempdir()
  f1 <- file.path(test_dir, "f1.bin")
  f2 <- file.path(test_dir, "f2.bin")

  files <- tibble::tibble(
    target_url = c("https://mock/1", "https://mock/2"),
    local_path = c(f1, f2),
    file_size_bytes = c(10L, 10L),
    data_ymd = as.Date(c("2022-01-01", "2022-01-02"))
  )

  # Mock download file
  testthat::local_mocked_bindings(
    spod_download_file = function(url, destfile, ...) {
      writeBin(raw(10), destfile)
      return(0L)
    },
    spod_interactive = function() FALSE
  )

  # Set batch size to 1 to force loop iterations
  res <- spod_download_in_batches(files, batch_size = 1, show_progress = FALSE)

  expect_true(all(res$complete_download))
  expect_true(file.exists(f1))
  expect_true(file.exists(f2))
})

test_that("spod_download with check_local_files=FALSE skips local file validation", {
  test_dir <- withr::local_tempdir()
  local_file <- file.path(test_dir, "v2/distritos/viajes/test.csv.gz")
  dir.create(dirname(local_file), recursive = TRUE)

  # Create a file with wrong size
  writeBin(raw(50), local_file) # File is 50 bytes

  # Mock available_data to return file info
  # When check_local_files=FALSE, local_file_size should be NA
  testthat::local_mocked_bindings(
    spod_available_data = function(...) {
      list_arg <- list(...)
      check_local <- list_arg$check_local_files

      tibble::tibble(
        data_ymd = as.Date("2022-02-01"),
        local_path = local_file,
        downloaded = if (isTRUE(check_local)) TRUE else FALSE, # Only mark as downloaded when checked
        file_size_bytes = 100L, # Expected: 100 bytes
        remote_file_size_mb = 0.0001,
        local_file_size = if (isTRUE(check_local)) 50 else NA_real_ # Size check only when enabled
      )
    },
    spod_dates_argument_to_dates_seq = function(...) as.Date("2022-02-01"),
    spod_infer_data_v_from_dates = function(...) 2,
    spod_match_data_type_for_local_folders = function(...) "viajes",
    spod_zone_names_en2es = function(z) "distritos"
  )

  download_called <- FALSE
  testthat::local_mocked_bindings(
    spod_download_in_batches = function(files) {
      download_called <<- TRUE
      files$downloaded <- TRUE
      files$complete_download <- TRUE
      return(files)
    }
  )

  # With check_local_files=FALSE, should still trigger download because local_file_size is NA
  # which means complete_download will be FALSE
  spod_download(
    type = "od",
    zones = "dist",
    dates = "2022-02-01",
    data_dir = test_dir,
    check_local_files = FALSE,
    quiet = TRUE
  )

  # This test documents current behavior: when check_local_files=FALSE,
  # file is still downloaded because size comparison fails (NA != expected)
  expect_true(download_called)
})

test_that("spod_download with check_local_files=TRUE detects size mismatch", {
  test_dir <- withr::local_tempdir()
  local_file <- file.path(test_dir, "v2/distritos/viajes/test.csv.gz")
  dir.create(dirname(local_file), recursive = TRUE)

  # Create a file with wrong size
  writeBin(raw(50), local_file) # File is 50 bytes

  # Mock available_data to return file info with different expected size
  testthat::local_mocked_bindings(
    spod_available_data = function(...) {
      tibble::tibble(
        data_ymd = as.Date("2022-02-01"),
        local_path = local_file,
        downloaded = TRUE,
        file_size_bytes = 100L, # Expected: 100 bytes
        remote_file_size_mb = 0.0001,
        local_file_size = 50 # Actual: 50 bytes (mismatch!)
      )
    },
    spod_dates_argument_to_dates_seq = function(...) as.Date("2022-02-01"),
    spod_infer_data_v_from_dates = function(...) 2,
    spod_match_data_type_for_local_folders = function(...) "viajes",
    spod_zone_names_en2es = function(z) "distritos"
  )

  download_called <- FALSE
  testthat::local_mocked_bindings(
    spod_download_in_batches = function(files) {
      download_called <<- TRUE
      files$downloaded <- TRUE
      files$complete_download <- TRUE
      files$local_file_size <- files$file_size_bytes
      return(files)
    }
  )

  # With check_local_files=TRUE (default), should trigger download due to size mismatch
  spod_download(
    type = "od",
    zones = "dist",
    dates = "2022-02-01",
    data_dir = test_dir,
    check_local_files = TRUE,
    quiet = TRUE
  )

  # SHOULD download because size mismatch detected
  expect_true(download_called)
})

#' @references
#' \itemize{
#' \item **For the official website of the mobility study**: \insertRef{mitms_mobility_web}{spanishoddata}
#' \item **For v1 data methodology**: \insertRef{mitma_methodology_2020_v3}{spanishoddata}
#' \item **For v2 data methodology**: \insertRef{mitms_methodology_2022_v8}{spanishoddata}
#' \item **For the `spanishoddata` R package**: \insertRef{spanishoddata-r-pkg}{spanishoddata}
#' }
#'
#' Use \code{\link[=spod_cite]{spod_cite()}} to cite the package and the data with correct plain text, markdown, or BibTeX formats.

#' Set the data directory
#'
#' @description
#'
#' `r lifecycle::badge("stable")`
#'
#' This function sets the data directory in the environment variable SPANISH_OD_DATA_DIR, so that all other functions in the package can access the data. It also creates the directory if it doesn't exist.
#'
#' @param data_dir The data directory to set.
#' @inheritParams global_quiet_param
#' @return Nothing. If quiet is `FALSE`, prints a message with the path and confirmation that the path exists.
#' @export
#' @examples
#' spod_set_data_dir(tempdir())
#'
spod_set_data_dir <- function(
  data_dir,
  quiet = FALSE
) {
  checkmate::assert_character(data_dir, len = 1, null.ok = FALSE)
  checkmate::assert_flag(quiet)

  data_dir_abs_path <- fs::path_abs(data_dir)

  tryCatch(
    {
      # Check if the directory exists; if not, attempt to create it
      if (!dir.exists(data_dir_abs_path)) {
        if (isFALSE(quiet)) {
          message(
            "Data directory ",
            data_dir_abs_path,
            " does not exist. Attempting to create it."
          )
        }
        fs::dir_create(data_dir_abs_path, recurse = TRUE)
      }

      # Check for write permissions
      test_file <- fs::path(data_dir_abs_path, ".test_write")
      file.create(test_file)
      fs::file_delete(test_file)
      if (isFALSE(quiet)) {
        message("Data directory is writeable.")
      }

      # Set the environment variable
      Sys.setenv(SPANISH_OD_DATA_DIR = data_dir_abs_path)

      if (isFALSE(quiet)) {
        message("Data directory successfully set to: ", data_dir_abs_path)
      }
    },
    error = function(e) {
      message(
        "Error: Unable to create or access the directory at '",
        data_dir_abs_path,
        "'."
      )
      message(
        "This may be due to write access restrictions or system permissions issues."
      )
      message(
        "Please verify that you have write permissions for the specified path and try again."
      )
      stop(e) # Re-throw the error for debugging purposes, if needed
    }
  )

  return(invisible(TRUE))
}

#' Get the data directory
#'
#' @description
#'
#' `r lifecycle::badge("stable")`
#'
#' This function retrieves the data directory from the environment variable SPANISH_OD_DATA_DIR (and previously set by [spod_set_data_dir()]).
#' If the environment variable is not set, it returns the temporary directory.
#' @inheritParams global_quiet_param
#' @return A `character` vector of length 1 containing the path to the data directory where the package will download and convert the data.
#' @export
#' @examples
#' spod_set_data_dir(tempdir())
#' spod_get_data_dir()
#'
spod_get_data_dir <- function(quiet = FALSE) {
  checkmate::assert_flag(quiet)
  data_dir_env <- Sys.getenv("SPANISH_OD_DATA_DIR")
  if (data_dir_env == "") {
    if (isFALSE(quiet)) {
      warning(
        "Warning: SPANISH_OD_DATA_DIR is not set. Using the temporary directory, which is not recommended, as the data will be deleted when the session ends.\n\n To set the data directory, use `spod_set_data_dir('/path/to/data')` or set SPANISH_OD_DATA_DIR permanently in the environment by editing the `.Renviron` file locally for current project with `usethis::edit_r_environ('project')` or `file.edit('.Renviron')` or globally for all projects with `usethis::edit_r_environ('user')` or `file.edit('~/.Renviron')`."
      )
    }
    data_dir_env <- tempdir() # if not set, use the temp directory
  }
  # check if dir exists and create it if it doesn't
  data_dir_env_abs <- fs::path_abs(data_dir_env)
  if (!dir.exists(data_dir_env_abs)) {
    fs::dir_create(data_dir_env_abs, recurse = TRUE)
  }
  data_dir_env_real <- fs::path_abs(data_dir_env_abs)
  return(data_dir_env_real)
}

test_that("spod_get_data_dir returns default or set directory", {
  # Save original env var to restore later
  old_dir <- Sys.getenv("SPANISH_OD_DATA_DIR")
  withr::defer(if (old_dir == "") Sys.unsetenv("SPANISH_OD_DATA_DIR") else Sys.setenv(SPANISH_OD_DATA_DIR = old_dir))

  # Case 1: Default (unset)
  Sys.unsetenv("SPANISH_OD_DATA_DIR")
  expect_warning(
    dir <- spod_get_data_dir(),
    "SPANISH_OD_DATA_DIR is not set"
  )
  expect_true(nchar(as.character(dir)) > 0)

  # Case 2: Set via function
  temp_dir <- tempfile()
  spod_set_data_dir(temp_dir, quiet = TRUE)
  
  # Normalize both to avoid double slash issues / trailing slash issues
  actual <- normalizePath(as.character(spod_get_data_dir()), mustWork = FALSE)
  expected <- normalizePath(temp_dir, mustWork = FALSE)
  expect_equal(actual, expected)
  
  expect_equal(normalizePath(Sys.getenv("SPANISH_OD_DATA_DIR"), mustWork = FALSE), expected)
})



test_that("spod_set_data_dir validates input", {
  expect_error(spod_set_data_dir(123), "Must be of type 'character'")
  expect_error(spod_set_data_dir(c("a", "b")), "Must have length 1")
})

test_that("spod_get_data_dir respects env var", {
  temp_dir <- tempfile()
  withr::local_envvar(c("SPANISH_OD_DATA_DIR" = temp_dir))
  
  # Check equality as character to avoid fs_path class issues
  expect_equal(normalizePath(as.character(spod_get_data_dir()), mustWork = FALSE), normalizePath(temp_dir, mustWork = FALSE))
})

test_that("spod_set_data_dir quiet argument works", {
  test_dir <- setup_test_data_dir()
  withr::defer(unlink(test_dir, recursive = TRUE))
  
  # When quiet = FALSE, it should message
  msgs <- capture_messages(
    spod_set_data_dir(test_dir, quiet = FALSE)
  )
  expect_match(msgs, "Data directory successfully set to", all = FALSE)
  
  # When quiet = TRUE, it should be silent
  expect_silent(
    spod_set_data_dir(test_dir, quiet = TRUE)
  )
})

test_that("spod_get_data_dir warns when env var not set", {
  # Unset env var
  withr::local_envvar(c("SPANISH_OD_DATA_DIR" = NA))
  
  # Should warn if quiet = FALSE (default)
  expect_warning(
    dir <- spod_get_data_dir(quiet = FALSE),
    "SPANISH_OD_DATA_DIR is not set"
  )
  expect_true(dir == fs::path_abs(tempdir()))
  
  # Should be silent if quiet = TRUE
  # Note: The warning is only emitted if quiet = FALSE explicitly?
  # The code: if (isFALSE(quiet)) warning(...)
  # If quiet = TRUE, no warning.
  expect_silent(
    spod_get_data_dir(quiet = TRUE)
  )
})

#' Connect to data converted to `DuckDB` or hive-style `parquet` files
#'
#' @description
#'
#' `r lifecycle::badge("stable")`
#'
#' This function allows the user to quickly connect to the data converted to DuckDB with the \link{spod_convert} function. This function simplifies the connection process. The user is free to use the `DBI` and `DuckDB` packages to connect to the data manually, or to use the `arrow` package to connect to the `parquet` files folder.
#'
#' @param data_path a path to the `DuckDB` database file with '.duckdb' extension, or a path to the folder with `parquet` files. Eigher one should have been created with the \link{spod_convert} function.
#' @param target_table_name Default is `NULL`. When connecting to a folder of `parquet` files, this argument is ignored. When connecting to a `DuckDB` database, a `character` vector of length 1 with the table name to open from the database file. If not specified, it will be guessed from the `data_path` argument and from table names that are available in the database. If you have not manually interfered with the database, this should be guessed automatically and you do not need to specify it.
#' @inheritParams spod_duckdb_limit_resources
#' @inheritParams spod_duckdb_set_temp
#' @inheritParams global_quiet_param
#' @export
#' @return a `DuckDB` table connection object.
#'
#' @examplesIf interactive()
#' \donttest{
#' # Set data dir for file downloads
#' spod_set_data_dir(tempdir())
#'
#' # download and convert data
#' dates_1 <- c(start = "2020-02-17", end = "2020-02-18")
#' db_2 <- spod_convert(
#'  type = "number_of_trips",
#'  zones = "distr",
#'  dates = dates_1,
#'  overwrite = TRUE
#' )
#'
#' # now connect to the converted data
#' my_od_data_2 <- spod_connect(db_2)
#'
#' # disconnect from the database
#' spod_disconnect(my_od_data_2)
#' }
#'
spod_connect <- function(
  data_path,
  target_table_name = NULL,
  quiet = FALSE,
  max_mem_gb = NULL,
  max_n_cpu = max(1, parallelly::availableCores() - 1),
  temp_path = spod_get_temp_dir()
) {
  # Validate imputs
  checkmate::assert_access(data_path, access = 'r')
  checkmate::assert_character(target_table_name, null.ok = TRUE)
  checkmate::assert_flag(quiet)
  checkmate::assert_number(max_mem_gb, lower = 1, null.ok = TRUE)
  checkmate::assert_integerish(max_n_cpu, lower = 1)
  checkmate::assert_directory_exists(temp_path, access = "rw")

  # determine if data_path is a folder or a duckdb file
  if (grepl("\\.duckdb$", data_path)) {
    duckdb_path <- data_path
    target_format <- "duckdb"
  } else {
    duckdb_path <- ":memory:"
    target_format <- "parquet"
  }

  con <- DBI::dbConnect(
    duckdb::duckdb(),
    dbdir = duckdb_path,
    read_only = TRUE
  )

  con <- spod_duckdb_limit_resources(
    con = con,
    max_mem_gb = max_mem_gb,
    max_n_cpu = max_n_cpu
  )

  if (target_format == "duckdb") {
    # try to guess the table name if not provided

    if (is.null(target_table_name)) {
      # try the same name as the file name
      target_table_name <- gsub("\\..*", "", basename(duckdb_path)) # experimental
      tables_list <- DBI::dbListTables(con)
      if (target_table_name %in% tables_list) {
        # if the table with the same name exists, use it
        target_table_name <- target_table_name
      } else {
        # pick the first table that does not contain CSV
        target_table_name <- tables_list[
          !stringr::str_detect(tables_list, "csv")
        ][1]
      }
    }
    tbl_con <- dplyr::tbl(con, target_table_name)
  }

  if (target_format == "parquet") {
    view_name <- basename(data_path)
    parquet_glob_path <- fs::path(data_path, "**", "*.parquet")

    DBI::dbExecute(
      con,
      dplyr::sql(
        glue::glue(
          "
        CREATE OR REPLACE VIEW {view_name} AS
        SELECT *
        FROM read_parquet(
          '{parquet_glob_path}',
          hive_partitioning = true
        ) ;
        "
        )
      )
    )

    # set temp path for intermediate spilling
    # https://duckdb.org/2024/07/09/memory-management.html#intermediate-spilling
    # we do not do the same for duckdb above, as there the temp is automatically created in the same folder as the database
    # however, when the target is parquet files, temp is created in the root of R working directory, which may be undesirable
    con <- spod_duckdb_set_temp(con, temp_path = temp_path)

    tbl_con <- dplyr::tbl(con, view_name)
  }

  return(tbl_con)
}

#' Convert data from plain text to duckdb or parquet format
#'
#' @description
#'
#' `r lifecycle::badge("stable")`
#'
#' Converts data for faster analysis into either `DuckDB` file or into `parquet` files in a hive-style directory structure. Running analysis on these files is sometimes 100x times faster than working with raw CSV files, espetially when these are in gzip archives. To connect to converted data, please use 'mydata <- \link{spod_connect}(data_path = path_returned_by_spod_convert)' passing the path to where the data was saved. The connected `mydata` can be analysed using `dplyr` functions such as \link[dplyr]{select}, \link[dplyr]{filter}, \link[dplyr]{mutate}, \link[dplyr]{group_by}, \link[dplyr]{summarise}, etc. In the end of any sequence of commands you will need to add \link[dplyr]{collect} to execute the whole chain of data manipulations and load the results into memory in an R `data.frame`/`tibble`. For more in-depth usage of such data, please refer to DuckDB documentation and examples at [https://duckdb.org/docs/api/r#dbplyr](https://duckdb.org/docs/api/r#dbplyr) . Some more useful examples can be found here [https://arrow-user2022.netlify.app/data-wrangling#combining-arrow-with-duckdb](https://arrow-user2022.netlify.app/data-wrangling#combining-arrow-with-duckdb) . You may also use `arrow` package to work with parquet files [https://arrow.apache.org/docs/r/](https://arrow.apache.org/docs/r/).
#'
#' For detailed data descriptions, see package vignettes using [`spod_codebook(ver = 1)`][spod_codebook] and [`spod_codebook(ver = 2)`][spod_codebook] and official methodology documents in **References** section.
#'
#' @template references
#'
#' @param save_format A `character` vector of length 1 with values "duckdb" or "parquet". Defaults to "duckdb". If `NULL` automatically inferred from the `save_path` argument. If only `save_format` is provided, `save_path` will be set to the default location set in `SPANISH_OD_DATA_DIR` environment variable using \link{spod_set_data_dir}`(path = 'path/to/your/cache/dir')`. So for v1 data that path would be `<data_dir>/clean_data/v1/tabular/duckdb/` or `<data_dir>/clean_data/v1/tabular/parquet/`.
#'
#' You can also set `save_path`. If it ends with ".duckdb", will save to `DuckDB` database format, if `save_path` does not end with ".duckdb", will save to `parquet` format and will treat the `save_path` as a path to a folder, not a file, will create necessary hive-style subdirectories in that folder. Hive style looks like `year=2020/month=2/day=14` and inside each such directory there will be a `data_0.parquet` file that contains the data for that day.
#'
#' @param save_path A `character` vector of length 1. The full (not relative) path to a `DuckDB` database file or `parquet` folder.
#'
#' * If `save_path` ends with `.duckdb`, it will be saved as a DuckDB database file. The format argument will be automatically set to `save_format='duckdb'`.
#'
#' * If `save_path` ends with a folder name (e.g. `/data_dir/clean_data/v1/tabular/parquet/od_distr` for origin-destination data for district level), the data will be saved as a collection of `parquet` files in a hive-style directory structure. So the subfolders of `od_distr` will be `year=2020/month=2/day=14` and inside each of these folders a single `parquet` file will be placed containing the data for that day.
#'
#' * If `NULL`, uses the default location in `data_dir` (set by the `SPANISH_OD_DATA_DIR` environment variable using \link{spod_set_data_dir}`(path = 'path/to/your/cache/dir')`. Therefore, the default relative path for `DuckDB` is `<data_dir>/clean_data/v1/tabular/duckdb/<type>_<zones>.duckdb` and for `parquet` files is `<data_dir>/clean_data/v1/tabular/parquet/<type>_<zones>/`, where `type` is the type of data (e.g. 'od', 'os', 'nt', that correspoind to 'origin-destination', 'overnight-stays', 'number-of-trips', etc.) and `zones` is the name of the geographic zones (e.g. 'distr', 'muni', etc.). See the details below in the function arguments description.
#'
#' @inheritParams spod_get_zones
#' @inheritParams spod_download
#' @inheritParams spod_get
#' @inheritParams spod_duckdb_limit_resources
#' @param overwrite A `logical` or a `character` vector of length 1. If `TRUE`, overwrites existing `DuckDB` or `parquet` files. Defaults to `FALSE`. For parquet files can also be set to 'update', so that only parquet files are only created for the dates that have not yet been converted.
#' @inheritParams global_quiet_param
#' @return Path to saved `DuckDB` database file or to a folder with `parquet` files in hive-style directory structure.
#' @export
#'
#' @examplesIf interactive()
#' \donttest{
#' # Set data dir for file downloads
#' spod_set_data_dir(tempdir())
#'
#' # download and convert data
#' dates_1 <- c(start = "2020-02-17", end = "2020-02-18")
#' db_2 <- spod_convert(
#'   type = "number_of_trips",
#'   zones = "distr",
#'   dates = dates_1,
#'   overwrite = TRUE
#' )
#'
#' # now connect to the converted data
#' my_od_data_2 <- spod_connect(db_2)
#'
#' # disconnect from the database
#' spod_disconnect(my_od_data_2)
#' }
#'
spod_convert <- function(
  type = c(
    "od",
    "origin-destination",
    "os",
    "overnight_stays",
    "nt",
    "number_of_trips"
  ),
  zones = c(
    "districts",
    "dist",
    "distr",
    "distritos",
    "municipalities",
    "muni",
    "municip",
    "municipios"
  ),
  dates = NULL,
  save_format = "duckdb",
  save_path = NULL,
  overwrite = FALSE,
  data_dir = spod_get_data_dir(),
  quiet = FALSE,
  max_mem_gb = NULL,
  max_n_cpu = max(1, parallelly::availableCores() - 1),
  max_download_size_gb = 1,
  ignore_missing_dates = FALSE
) {
  # Validate inputs
  checkmate::assert_choice(
    type,
    choices = c(
      "od",
      "origin-destination",
      "os",
      "overnight_stays",
      "nt",
      "number_of_trips"
    )
  )
  checkmate::assert_choice(
    zones,
    choices = c(
      "districts",
      "dist",
      "distr",
      "distritos",
      "municipalities",
      "muni",
      "municip",
      "municipios",
      "lua",
      "large_urban_areas",
      "gau",
      "grandes_areas_urbanas"
    )
  )
  checkmate::assert_choice(
    save_format,
    choices = c("duckdb", "parquet"),
    null.ok = TRUE
  )
  checkmate::assert_character(save_path, len = 1, null.ok = TRUE)
  # overwrite validation moved to after "update" check (line ~235)
  checkmate::assert_directory_exists(data_dir, access = "rw")
  checkmate::assert_flag(quiet)
  checkmate::assert_number(max_mem_gb, lower = 0.1, null.ok = TRUE)
  checkmate::assert_integerish(max_n_cpu, lower = 1)
  checkmate::assert_number(max_download_size_gb, lower = 0.1)
  checkmate::assert_flag(ignore_missing_dates)

  # simple null check is enough here, as spod_dates_arugument_to_dates_seq will do additional checks anyway
  if (is.null(dates)) {
    message(
      "No period specified in the `dates` argument. Please set `dates='cached_v1'` or `dates='cached_v2'` to convert all data that was previously downloaded. Alternatively, specify at least one date between 2020-02-14 and 2021-05-09 (for v1 data) or between 2022-01-01 onwards (for v2). Any missing data will be downloaded before conversion. For more details on the dates argument, see ?spod_convert."
    )
  }

  dates <- spod_dates_argument_to_dates_seq(dates = dates)
  # check if user is requesting to just get all cached data
  cached_data_requested <- length(dates) == 1 &&
    all(as.character(dates) %in% c("cached_v1", "cached_v2"))

  ver <- spod_infer_data_v_from_dates(
    dates = dates,
    ignore_missing_dates = ignore_missing_dates
  )

  # normalise type
  type <- spod_match_data_type(type = type)

  # normalise zones
  zones <- spod_zone_names_en2es(zones)

  # check format
  if (length(save_format) > 1) {
    stop(
      "Invalid save_format. Please select 'duckdb' or 'parquet', or leave empty to use the default, which is 'duckdb'."
    )
  }
  # Set default if NULL
  if (is.null(save_format)) {
    if (!is.null(save_path) && grepl("\\.parquet$", save_path, ignore.case = TRUE)) {
      save_format <- "parquet"
    } else {
      save_format <- "duckdb"
    }
  }
  
  if (!save_format %in% c("duckdb", "parquet")) {
    stop(
      "Invalid save_format. Please set `save_format` to 'duckdb' or 'parquet', or leave empty to use the default, which is 'duckdb'."
    )
  }

  # check for inconsistencies between save_format and save_path
  if (!is.null(save_format) & !is.null(save_path)) {
    if (!grepl("duckdb$", save_path) & save_format == "duckdb") {
      stop(
        "`save_path` is set to a folder, while `save_format` is set to 'duckdb'. Please make sure that `save_format` also ends with '.duckdb' when setting format to 'duckdb', or leave `save_format` empty, so that it is inferred from `save_path`. Perhaps you wanted to save to `parquet` files? Then please set `save_format` to 'parquet', or leave `save_format` empty."
      )
    }
  }

  # `save_path` is NULL, format 'duckdb'
  if (is.null(save_path) & save_format == "duckdb") {
    save_path <- fs::path(
      data_dir,
      spod_subfolder_clean_data_cache(ver = ver),
      "tabular",
      "duckdb",
      glue::glue("{type}_{zones}.duckdb")
    )
    if (isFALSE(quiet)) {
      message("Using default save_path: ", save_path)
    }
  }

  # `save_path` is NULL, format 'parquet'
  if (is.null(save_path) & save_format == "parquet") {
    save_path <- fs::path(
      data_dir,
      spod_subfolder_clean_data_cache(ver = ver),
      "tabular",
      "parquet",
      glue::glue("{type}_{zones}")
    )
    if (isFALSE(quiet)) {
      message("Using default save_path: ", save_path)
    }
  }

  # infer format from `save_path`
  if (is.null(save_format)) {
    if (grepl("\\.duckdb$", save_path)) {
      save_format <- "duckdb"
      save_dir <- fs::path_dir(save_path)
      duckdb_target <- save_path
      # else check if it is a path to a folder
      # basically check that the path does not end with any other file extension other than .duckdb and then consider it as path to folder
    } else if (!grepl("\\.[a-zA-Z0-9]+$", save_path)) {
      save_format <- "parquet"
    }
  }

  # make sure there is no trailing slash for parquet
  if (save_format == "parquet") {
    save_dir <- gsub("\\/$", "", save_path)
    duckdb_target <- ":memory:"
  } else if (save_format == "duckdb") {
    save_dir <- fs::path_dir(save_path)
    duckdb_target <- save_path
  }

  if (!dir.exists(save_dir)) {
    fs::dir_create(save_dir, recurse = TRUE)
  }

  # check if overwrite update conflicts with `save_format` BEFORE validating as flag
  if (identical(overwrite, "update") && save_format == "duckdb") {
    stop(
      "You are trying to save to a duckdb file with `overwrite = 'update'`. This is not supported. Please set `overwrite=TRUE` to remove the existing duckdb file and create a new one with the requested data."
    )
  }
  
  # Now validate overwrite if it's not "update"
  if (!identical(overwrite, "update")) {
    checkmate::assert_flag(overwrite)
  }

  # check if duckdb file already exists
  # for conversion to duckdb we need to initialise a writeable database file
  # therefore  this check is before we create a database with `spod_get()` below
  if (save_format == "duckdb" && fs::file_exists(save_path)) {
    if (!overwrite && isFALSE(quiet)) {
      message("Duckdb file already exists: ", save_path)
      response <- spod_readline(prompt = "Overwrite existing duckdb file? (yes/no) ")
      overwrite <- tolower(response) %in% c("y", "yes", "Yes")

      if (!overwrite) {
        message(glue::glue(
          "Exiting without overwriting existing duckdb file. You may delete it from {save_path} manually and rerun the function. Or rerun the function with `overwrite = TRUE`."
        ))
        return()
      }
    }
    if (overwrite) {
      if (isFALSE(quiet)) {
        message(glue::glue("Overwriting existing duckdb file: ", save_path))
      }
      fs::file_delete(save_path)
    }
  }

  # create duckdb view with a file target
  con <- NULL
  con <- spod_get(
    type = type,
    zones = zones,
    dates = dates,
    data_dir = data_dir,
    quiet = quiet,
    max_mem_gb = max_mem_gb,
    max_n_cpu = max_n_cpu,
    max_download_size_gb = max_download_size_gb,
    duckdb_target = duckdb_target,
    ignore_missing_dates = ignore_missing_dates
  )
  on.exit(
    {
      if (
        exists("con", inherits = FALSE) &&
          !is.null(con) &&
          inherits(con, "tbl_duckdb_connection")
      ) {
        try(spod_disconnect(con, free_mem = TRUE), silent = TRUE)
      }
    },
    add = TRUE
  )

  # resolve the actual database connection from the returned table
  db_con <- con$src$con

  if (isTRUE(cached_data_requested)) {
    table_to_convert <- glue::glue("{type}_csv_clean")
  } else {
    table_to_convert <- glue::glue("{type}_csv_clean_filtered")
  }

  if (isFALSE(quiet)) {
    if (!is.null(max_mem_gb)) {
      message(glue::glue(
        "Using {max_mem_gb} GB of memory and {max_n_cpu} threads. You may adjust this using the function arguments `max_mem_gb` and `max_n_cpu`."
      ))
    } else {
      max_mem_duckdb <- DBI::dbGetQuery(
        db_con,
        "SELECT current_setting('memory_limit');"
      )[
        1,
      ]
      message(glue::glue(
        "Using DuckDB's default of 80% of operating memory (currently detected as {max_mem_duckdb}) and {max_n_cpu} threads. You may adjust this using the function arguments `max_mem_gb` and `max_n_cpu`."
      ))
    }

    message(glue::glue(
      "Converting v{ver} {type} data for {zones} to {save_format}: {save_path} \n..."
    ))
  }
  # add some indication on how long it may take from empirical experimentation
  # hopefully, the progress_bar feature will be implemented in duckdb R package soon, bug filed here https://github.com/duckdb/duckdb-r/issues/199
  # if I change the code to import data in batches it slows down significantly

  if (save_format == "duckdb") {
    # convert to ducdkb
    # prepare SQL query for converting to duckdb
    target_table_name <- gsub("\\..*", "", basename(save_path)) # experimental
    sql_import_query <- dplyr::sql(
      # glue::glue("CREATE TABLE {type} AS SELECT * FROM {table_to_convert} ;") should be standard
      glue::glue(
        'CREATE TABLE "{target_table_name}" AS SELECT * FROM {table_to_convert} ;'
      ) # experimental - for the user friendly duckdb connection wrapper so that it can guess the table name from the file name, hopefully the user will not rename it
    )

    # import view of CSV files into duckdb
    DBI::dbExecute(
      db_con,
      statement = sql_import_query
    )
    # drop the views to csv files
    tables_list <- DBI::dbListTables(db_con)
    # find all views that contain csv in any case
    tables_list <- tables_list[stringr::str_detect(tables_list, "csv")]
    if (length(tables_list) > 0) {
      for (i in seq_along(tables_list)) {
        sql_drop_view <- dplyr::sql(
          glue::glue(
            "DROP VIEW IF EXISTS {tables_list[i]};"
          )
        )
        DBI::dbExecute(
          db_con,
          sql_drop_view
        )
      }
    }
  }

  # if save_format is parquet
  need_to_return_parquet_files <- FALSE # will change in the if below if needed
  if (save_format == "parquet") {
    if (overwrite == FALSE | overwrite == "update") {
      # check if there are any existing parquet files in the save_path
      parquet_files <- fs::dir_ls(
        save_path,
        regexp = "\\.parquet$",
        recurse = TRUE
      )
      if (overwrite == FALSE & length(parquet_files) > 0) {
        if (isFALSE(quiet)) {
          message(
            "You have set `overwrite = FALSE`, but there are already parquet files in ",
            save_path
          )
          response <- spod_readline(
            "What should be done? [D]elete all existing files in the target folder and convert all requested data from scratch? [U]pdate the folder with any new data while keeping the existing files? [C]ancel and quit? (D/U/C): "
          )
        } else {
          response <- "cancel"
        }
      }
      if (overwrite == "update" & length(parquet_files) > 0) {
        response <- overwrite
      }
      if (
        (overwrite == "update" | overwrite == FALSE) &
          length(parquet_files) == 0
      ) {
        response <- "skip"
        overwrite <- TRUE
      }

      if (tolower(response) %in% c("c", "cancel")) {
        if (isFALSE(quiet)) {
          message(
            "Cancelled by the user. Exiting without overwriting existing parquet files. You may delete them from ",
            save_path,
            " manually and rerun the function. Or rerun the function with `overwrite = TRUE` or `overwrite = 'update'`."
          )
        }
        return(invisible(NULL))
      }

      if (tolower(response) %in% c("d", "delete")) {
        # set overwrite true, so that DuckDB to parquet export will overwrite the existing files, but also just to be on the safe side delete the directory recursively
        overwrite <- TRUE
        fs::file_delete(save_path)
      }

      if (tolower(response) %in% c("u", "update")) {
        # find dates to filter out from the conversion
        file_list <- fs::dir_ls(save_path, recurse = TRUE, type = "file")
        parquet_skip_dates <- stringr::str_extract(
          file_list,
          "(?<=year=)\\d{4}(?:/month=\\d{1,2})(?:/day=\\d{1,2})"
        ) |>
          stringr::str_replace_all("year=|/month=|/day=", "-") |>
          lubridate::ymd()
        requested_dates <- spod_dates_argument_to_dates_seq(dates)
        filter_to_dates <- requested_dates[
          !(requested_dates %in% parquet_skip_dates)
        ]
        # rename/move the folder temporarily, we will then have to restore it once the DuckDB export to parquet is done
        temp_path <- paste0(save_path, "_temp")
        fs::file_move(save_path, temp_path)
        # set overwrite fales, so that if the move failed, DuckDB will fail if it sees existing directory
        overwrite <- FALSE
        need_to_return_parquet_files <- TRUE # to remember to return the temporary moved existing parquet files\
        # note: if we used `arrow` package instead to convert to parquet, it would be easier to add new parquet files without overwriting the existing ones, however, it would mean (1) we through away all the SQL and DuckDB code that cleans up the data, (2) we would not be able to limit memory usage, as `arrow` has no such setting and just consumes all available memory, and (3) we would not be able to limit the number of threads, as `arrow` has no such setting.
      }
    }

    # check if parquet_skip_dates exists and not empty note that object may not extist at all but the check should not fail
    if (exists("filter_to_dates") == TRUE) {
      if (length(filter_to_dates) > 0) {
        # reset the table_to_convert to the all view, as all missing files must have been downloaded by the call to `spod_get()` above and it is better to filter from the main view to avoid filtering on filtered view
        table_to_convert <- glue::glue("{type}_csv_clean")
        table_to_convert_update <- paste0(table_to_convert, "_update")
        db_con <- spod_duckdb_filter_by_dates(
          con = db_con,
          source_view_name = table_to_convert,
          new_view_name = table_to_convert_update,
          dates = filter_to_dates
        )
        table_to_convert <- table_to_convert_update
      }
    }

    # prepare sql query with `table_to_convert` as source table name
    sql_save_parquet_query <- dplyr::sql(
      glue::glue(
        "
      COPY {table_to_convert} TO '{save_path}'
        (
          FORMAT PARQUET,
          OVERWRITE_OR_IGNORE {overwrite},
          PARTITION_BY (year, month, day),
          -- PER_THREAD_OUTPUT true,
          COMPRESSION snappy,
          FILENAME_PATTERN 'data_{{i}}' -- this prevents the user from saving several differnt file formats into the same folder structure
        ) ;
      "
      )
    )

    DBI::dbExecute(
      db_con,
      statement = sql_save_parquet_query
    )
  }

  spod_disconnect(con, free_mem = TRUE)
  con <- NULL

  # time to move back the pre-existing parquet files if we moved them to temp folder
  if (need_to_return_parquet_files) {
    # we cannot simply merge the folders, we will have to list the files and move them one by one
    file_list <- fs::dir_ls(temp_path, recurse = TRUE, type = "file")
    new_paths <- gsub(temp_path, save_path, file_list)
    fs::dir_create(unique(fs::path_dir(new_paths)), recurse = TRUE)
    fs::file_move(file_list, new_paths)
    fs::dir_delete(temp_path)
  }

  if (isFALSE(quiet)) {
    message("Data imported into ", save_format, " at: ", save_path)
  }

  # a few instructions on how to use the duckdb file
  if (isFALSE(quiet)) {
    message(
      "You can start working with the imported data by runining:\n mydata <- spod_connect(data_path='",
      save_path,
      "')"
    )
    message(
      "You can then manipulate `mydata` using `dplyr` functions such as `select()`, `filter()`, `mutate()`, `group_by()`, `summarise()`, etc. In the end of any sequence of commands you will need to add `collect()` to execute the whole chain and load the results into memory in an R `data.frame`/`tibble`."
    )
    if (save_format == "duckdb") {
      message(
        "For more in-depth usage of such data, please refer to DuckDB documentation and examples at https://duckdb.org/docs/api/r#dbplyr . Some more useful examples can be found here https://arrow-user2022.netlify.app/data-wrangling#combining-arrow-with-duckdb ."
      )
    } else if (save_format == "parquet") {
      message(
        "For more in-depth usage of such data, please refer to DuckDB documentation and examples at https://duckdb.org/docs/api/r#dbplyr . Some more useful examples can be found here https://arrow-user2022.netlify.app/data-wrangling#combining-arrow-with-duckdb . You may also use `arrow` package to work with parquet files https://arrow.apache.org/docs/r/ ."
      )
    }
  }

  return(save_path)
}

#' Safely disconnect from data and free memory
#' 
#' @description
#' 
#' `r lifecycle::badge("stable")`
#' 
#' This function is to ensure that `DuckDB` connections to CSV.gz files (created via `spod_get()`), as well as to `DuckDB` files or folders of `parquet` files (created via `spod_convert()`) are closed properly to prevent conflicting connections. Essentially this is just a wrapper around `DBI::dbDisconnect()` that reaches out into the `.$src$con` object of the `tbl_duckdb_connection` connection object that is returned to the user via `spod_get()` and `spod_connect()`. After disonnecting the database, it also frees up memory by running `gc()`.
#' @param tbl_con A `tbl_duckdb_connection` connection object that you get from either `spod_get()` or `spod_connect()`.
#' @param free_mem A `logical`. Whether to free up memory by running `gc()`. Defaults to `TRUE`.
#' @return No return value, called for side effect of disconnecting from the database and freeing up memory.
#' @export
#' @examplesIf interactive()
#' \donttest{
#' # Set data dir for file downloads
#' spod_set_data_dir(tempdir())
#' 
#' # basic example
#' # create a connection to the v1 data without converting
#' # this creates a duckdb database connection to CSV files
#' od_distr <- spod_get(
#'  "od",
#'  zones = "distr",
#'  dates = c("2020-03-01", "2020-03-02")
#' )
#' # disconnect from the database connection
#' spod_disconnect(od_distr)
#' 
#' # Advanced example
#' # download and convert data
#' dates_1 <- c(start = "2020-02-17", end = "2020-02-19")
#' db_2 <- spod_convert(
#'  type = "od",
#'  zones = "distr",
#'  dates = dates_1,
#'  overwrite = TRUE
#' )
#' 
#' # now connect to the converted data
#' my_od_data_2 <- spod_connect(db_2)
#' 
#' # disconnect from the database
#' spod_disconnect(my_od_data_2)
#' }
#' 
spod_disconnect <- function(
  tbl_con,
  free_mem = TRUE
) {
  # Validate imputs
  checkmate::assert_class(tbl_con, "tbl_duckdb_connection")
  checkmate::assert_flag(free_mem)

  DBI::dbDisconnect(tbl_con$src$con, shutdown = TRUE)
  if (free_mem){
    gc()
  }
  return(invisible(NULL))
}

#' Get zones
#'
#' @description
#'
#' `r lifecycle::badge("stable")`
#'
#' Get spatial zones for the specified data version. Supports both v1 (2020-2021) and v2 (2022 onwards) data.
#'
#' For detailed data descriptions, see package vignettes using [`spod_codebook(ver = 1)`][spod_codebook] and [`spod_codebook(ver = 2)`][spod_codebook] and official methodology documents in **References** section.
#'
#' @template references
#'
#' @inheritParams spod_download
#' @inheritParams spod_available_data
#' @return An `sf` object (Simple Feature collection).
#'
#' The columns for v1 (2020-2021) data include:
#' \describe{
#'   \item{id}{A character vector containing the unique identifier for each district, assigned by the data provider. This `id` matches the `id_origin`, `id_destination`, and `id` in district-level origin-destination and number of trips data.}
#'   \item{census_districts}{A string with semicolon-separated identifiers of census districts classified by the Spanish Statistical Office (INE) that are spatially bound within the polygons for each `id`.}
#'   \item{municipalities_mitma}{A string with semicolon-separated municipality identifiers (as assigned by the data provider) corresponding to each district `id`.}
#'   \item{municipalities}{A string with semicolon-separated municipality identifiers classified by the Spanish Statistical Office (INE) corresponding to each `id`.}
#'   \item{district_names_in_v2/municipality_names_in_v2}{A string with semicolon-separated district names (from the v2 version of this data) corresponding to each district `id` in v1.}
#'   \item{district_ids_in_v2/municipality_ids_in_v2}{A string with semicolon-separated district identifiers (from the v2 version of this data) corresponding to each district `id` in v1.}
#'   \item{geometry}{A `MULTIPOLYGON` column containing the spatial geometry of each district, stored as an sf object. The geometry is projected in the ETRS89 / UTM zone 30N coordinate reference system (CRS), with XY dimensions.}
#' }
#'
#' The columns for v2 (2022 onwards) data include:
#' \describe{
#'   \item{id}{A character vector containing the unique identifier for each zone, assigned by the data provider.}
#'   \item{name}{A character vector with the name of each district.}
#'   \item{population}{A numeric vector representing the population of each district (as of 2022).}
#'   \item{census_sections}{A string with semicolon-separated identifiers of census sections corresponding to each district.}
#'   \item{census_districts}{A string with semicolon-separated identifiers of census districts as classified by the Spanish Statistical Office (INE) corresponding to each district.}
#'   \item{municipalities}{A string with semicolon-separated identifiers of municipalities classified by the Spanish Statistical Office (INE) corresponding to each district.}
#'   \item{municipalities_mitma}{A string with semicolon-separated identifiers of municipalities, as assigned by the data provider, that correspond to each district.}
#'   \item{luas_mitma}{A string with semicolon-separated identifiers of LUAs (Local Urban Areas) from the provider, associated with each district.}
#'   \item{district_ids_in_v1/municipality_ids_in_v1}{A string with semicolon-separated district identifiers from v1 data corresponding to each district in v2. If no match exists, it is marked as `NA`.}
#'   \item{geometry}{A `MULTIPOLYGON` column containing the spatial geometry of each district, stored as an sf object. The geometry is projected in the ETRS89 / UTM zone 30N coordinate reference system (CRS), with XY dimensions.}
#' }
#'
#'
#'
#' @export
#' @examplesIf interactive()
#' \donttest{
#' # get polygons for municipalities for the v2 data
#' municip_v2 <- spod_get_zones(zones = "municipalities", ver = 2)
#'
#' # get polygons for the districts for the v1 data
#' distr_v1 <- spod_get_zones(zones = "districts", ver = 1)
#' }
#'
spod_get_zones <- function(
  zones = c(
    "districts",
    "dist",
    "distr",
    "distritos",
    "municipalities",
    "muni",
    "municip",
    "municipios",
    "lua",
    "large_urban_areas",
    "gau",
    "grandes_areas_urbanas"
  ),
  ver = NULL,
  data_dir = spod_get_data_dir(),
  quiet = FALSE
) {
  # Validate inputs
  checkmate::assert_choice(
    zones,
    choices = c(
      "districts",
      "dist",
      "distr",
      "distritos",
      "municipalities",
      "muni",
      "municip",
      "municipios",
      "lua",
      "large_urban_areas",
      "gau",
      "grandes_areas_urbanas"
    )
  )

  checkmate::assertIntegerish(ver, max.len = 1)
  if (!ver %in% c(1, 2)) {
    stop(
      "Invalid version number. Must be 1 (for v1 2020-2021 data) or 2 (for v2 2022 onwards)."
    )
  }

  checkmate::assert_directory_exists(data_dir, access = "rw")
  checkmate::assert_flag(quiet)

  # normalise zones
  zones <- spod_zone_names_en2es(zones)

  if (ver == 1) {
    zones_sf <- spod_get_zones_v1(
      zones = zones,
      data_dir = data_dir,
      quiet = quiet
    )
  } else if (ver == 2) {
    zones_sf <- spod_get_zones_v2(
      zones = zones,
      data_dir = data_dir,
      quiet = quiet
    )
  }

  return(zones_sf)
}

#' Retrieves the zones for v1 data
#'
#' This function retrieves the zones data from the specified data directory.
#' It can retrieve either "distritos" or "municipios" zones data.
#'
#' @param data_dir The directory where the data is stored.
#' @param zones The zones for which to download the data. Can be `"districts"` (or `"dist"`, `"distr"`, or the original Spanish `"distritos"`) or `"municipalities"` (or `"muni"`, `"municip"`, or the original Spanish `"municipios"`).
#' @inheritParams global_quiet_param
#' @return An `sf` object (Simple Feature collection) with 2 fields:
#' \describe{
#'   \item{id}{A character vector containing the unique identifier for each zone, to be matched with identifiers in the tabular data.}
#'   \item{geometry}{A `MULTIPOLYGON` column containing the spatial geometry of each zone, stored as an sf object.
#'   The geometry is projected in the ETRS89 / UTM zone 30N coordinate reference system (CRS), with XY dimensions.}
#' }
#' @keywords internal
spod_get_zones_v1 <- function(
  zones = c(
    "districts",
    "dist",
    "distr",
    "distritos",
    "municipalities",
    "muni",
    "municip",
    "municipios"
  ),
  data_dir = spod_get_data_dir(),
  quiet = FALSE
) {
  zones <- spod_zone_names_en2es(zones)

  metadata <- spod_available_data(
    ver = 1,
    data_dir = data_dir,
    check_local_files = FALSE,
    quiet = quiet
  )

  # Ensure the raw data is downloaded and extracted
  spod_download_zones_v1(
    zones = zones,
    data_dir = data_dir,
    quiet = quiet,
    metadata = metadata
  )

  # check if gpkg files are already saved and load them if available
  expected_gpkg_path <- fs::path(
    data_dir,
    spod_subfolder_clean_data_cache(ver = 1),
    "zones",
    glue::glue("{zones}_mitma.gpkg")
  )
  if (fs::file_exists(expected_gpkg_path)) {
    if (isFALSE(quiet)) {
      message(
        "Loading .gpkg file that already exists in data dir: ",
        expected_gpkg_path
      )
    }
    return(sf::read_sf(expected_gpkg_path))
  }

  zones_path <- fs::dir_ls(
    path = fs::path(data_dir, spod_subfolder_raw_data_cache(ver = 1)),
    glob = glue::glue("*v1**{zones}/*.shp"),
    recurse = TRUE
  )

  zones_sf <- spod_clean_zones_v1(zones_path, zones = zones, quiet = quiet)
  fs::dir_create(fs::path_dir(expected_gpkg_path), recurse = TRUE)
  sf::st_write(
    zones_sf,
    expected_gpkg_path,
    delete_dsn = TRUE,
    delete_layer = TRUE
  )

  return(zones_sf)
}


#' Fixes common issues in the zones data and cleans up variable names
#'
#' This function fixes any invalid geometries in the zones data and renames the "ID" column to "id".
#'
#' @param zones_path The path to the zones spatial data file.
#' @inheritParams spod_get_zones
#' @return A spatial object containing the cleaned zones data.
#' @keywords internal
#' @importFrom rlang .data
#'
spod_clean_zones_v1 <- function(zones_path, zones, quiet = FALSE) {
  if (fs::file_exists(zones_path) == FALSE) {
    stop("File does not exist: ", zones_path)
  }
  suppressWarnings({
    zones_sf <- sf::read_sf(zones_path)
  })
  invalid_geometries <- !sf::st_is_valid(zones_sf)
  if (sum(invalid_geometries) > 0) {
    fixed_zones_sf <- sf::st_make_valid(zones_sf[invalid_geometries, ])
    zones_sf <- rbind(zones_sf[!invalid_geometries, ], fixed_zones_sf)
  }
  names(zones_sf)[names(zones_sf) == "ID"] <- "id"

  # load and prepare id relations for districts
  relations_districts <- readr::read_delim(
    file = fs::path(
      spod_get_data_dir(),
      spod_subfolder_raw_data_cache(1),
      "relaciones_distrito_mitma.csv"
    ),
    delim = "|",
    show_col_types = FALSE
  )
  relations_districts_col_names <- names(relations_districts)
  relations_districts_col_names <- gsub(
    "distrito",
    "district",
    relations_districts_col_names
  )
  relations_districts_col_names <- gsub(
    "municipio",
    "municipality",
    relations_districts_col_names
  )
  relations_districts_col_names <- gsub(
    "^district$",
    "census_district",
    relations_districts_col_names
  )
  names(relations_districts) <- relations_districts_col_names

  # load and prepare id relations for municipalities
  relations_municipalities <- readr::read_delim(
    file = fs::path(
      spod_get_data_dir(),
      spod_subfolder_raw_data_cache(1),
      "relaciones_municipio_mitma.csv"
    ),
    delim = "|",
    show_col_types = FALSE
  )
  relations_municipalities_col_names <- names(relations_municipalities)
  relations_municipalities_col_names <- gsub(
    "municipio",
    "municipality",
    relations_municipalities_col_names
  )
  names(relations_municipalities) <- relations_municipalities_col_names

  # summarise districts relations including municipality data
  relations_districts_aggregated <- relations_districts |>
    dplyr::left_join(
      relations_municipalities |>
        dplyr::group_by(.data$municipality_mitma) |>
        dplyr::summarize(
          municipalities = paste(.data$municipality, collapse = "; ")
        ),
      by = "municipality_mitma"
    ) |>
    dplyr::group_by(.data$district_mitma) |>
    dplyr::summarize(
      census_districts = paste(.data$census_district, collapse = "; "),
      municipalities_mitma = paste(.data$municipality_mitma, collapse = "; "),
      municipalities = paste(.data$municipalities, collapse = "; ")
    )

  # summarise municipalities relations
  relations_municipalities_aggregated <- relations_municipalities |>
    dplyr::left_join(
      relations_districts |>
        dplyr::group_by(.data$municipality_mitma) |>
        dplyr::summarize(
          census_districts = paste(.data$census_district, collapse = "; "),
          districts_mitma = paste(.data$district_mitma, collapse = "; ")
        ),
      by = "municipality_mitma"
    ) |>
    dplyr::group_by(.data$municipality_mitma) |>
    dplyr::summarize(
      municipalities = paste(.data$municipality, collapse = "; "),
      districts_mitma = paste(.data$districts_mitma, collapse = "; "),
      census_districts = paste(.data$census_districts, collapse = "; ")
    )

  # cleanup duplacate ids in municipalities
  relations_municipalities_aggregated <- relations_municipalities_aggregated |>
    dplyr::mutate(
      dplyr::across(
        c("municipalities", "districts_mitma", "census_districts"),
        spod_unique_separated_ids
      )
    )
  names(relations_municipalities_aggregated)[
    names(relations_municipalities_aggregated) == "municipality_mitma"
  ] <- "id"

  # cleanup duplicate ids in districts
  relations_districts_aggregated <- relations_districts_aggregated |>
    dplyr::mutate(
      dplyr::across(
        c("census_districts", "municipalities_mitma"),
        spod_unique_separated_ids
      )
    )
  names(relations_districts_aggregated)[
    names(relations_districts_aggregated) == "district_mitma"
  ] <- "id"

  if (zones == "distritos") {
    zones_sf <- zones_sf |>
      dplyr::left_join(relations_districts_aggregated, by = "id") |>
      dplyr::relocate("geometry", .after = dplyr::last_col())
  } else if (zones == "municipios") {
    zones_sf <- zones_sf |>
      dplyr::left_join(relations_municipalities_aggregated, by = "id") |>
      dplyr::relocate("geometry", .after = dplyr::last_col())
  }

  # add metadata from v2 zones
  zones_v2_sf <- spod_get_zones_v2(zones = zones, quiet = quiet)
  zones_v2_sf <- zones_v2_sf[, c("id", "name")]
  names(zones_v2_sf)[names(zones_v2_sf) == "id"] <- "id_in_v2"
  names(zones_v2_sf)[names(zones_v2_sf) == "name"] <- "name_in_v2"
  suppressWarnings(
    zones_v2_sf_centroids <- zones_v2_sf |> sf::st_point_on_surface()
  )
  v2_to_v1 <- sf::st_join(zones_sf, zones_v2_sf_centroids, left = TRUE) |>
    sf::st_drop_geometry()
  v2_v_1ref <- v2_to_v1 |>
    dplyr::group_by(.data$id) |>
    dplyr::summarize(
      names_in_v2_data = paste(.data$name_in_v2, collapse = "; "),
      ids_in_v2_data = paste(.data$id_in_v2, collapse = "; ")
    )
  eng_zones <- dplyr::if_else(
    zones == "distritos",
    true = "district",
    false = "municipality"
  )
  names(v2_v_1ref)[names(v2_v_1ref) == "names_in_v2_data"] <- glue::glue(
    "{eng_zones}_names_in_v2"
  )
  names(v2_v_1ref)[names(v2_v_1ref) == "ids_in_v2_data"] <- glue::glue(
    "{eng_zones}_ids_in_v2"
  )

  zones_sf <- zones_sf |>
    dplyr::left_join(v2_v_1ref, by = "id") |>
    dplyr::relocate("geometry", .after = dplyr::last_col())

  return(zones_sf)
}


#' Downloads and extracts the raw v1 zones data
#'
#' This function ensures that the necessary v1 raw data for zones files are downloaded and extracted from the specified data directory.
#'
#' @param zones The zones for which to download the data. Can be `"districts"` (or `"dist"`, `"distr"`, or the original Spanish `"distritos"`) or `"municipalities"` (or `"muni"`, `"municip"`, or the original Spanish `"municipios"`).
#' @param data_dir The directory where the data is stored.
#' @param quiet Boolean flag to control the display of messages.
#' @param metadata Optional metadata table returned by `spod_available_data()`.
#' @return A `character` string containing the path to the downloaded and extracted file.
#' @keywords internal
spod_download_zones_v1 <- function(
  zones = c(
    "districts",
    "dist",
    "distr",
    "distritos",
    "municipalities",
    "muni",
    "municip",
    "municipios"
  ),
  data_dir = spod_get_data_dir(),
  quiet = FALSE,
  metadata = NULL
) {
  zones <- match.arg(zones)
  zones <- spod_zone_names_en2es(zones)

  if (is.null(metadata)) {
    metadata <- spod_available_data(
      ver = 1,
      data_dir = data_dir,
      use_s3 = TRUE,
      check_local_files = FALSE,
      quiet = quiet
    )
  }

  # download id relation files if missing
  relation_files <- metadata[
    grepl("relaciones_(distrito|municipio)_mitma.csv", metadata$target_url),
  ]
  if (any(!fs::file_exists(relation_files$local_path))) {
    fs::dir_create(
      unique(fs::path_dir(relation_files$local_path)),
      recurse = TRUE
    )
    # disable curl::multi_download() for now
    # invisible(curl::multi_download(urls = relation_files$target_url, destfile = relation_files$local_path, resume = FALSE, progress = TRUE))
    # relation_files <- spod_multi_download_with_progress(relation_files)
    relation_files <- spod_download_in_batches(relation_files)
  }

  regex <- glue::glue("zonificacion_{zones}\\.")
  sel_zones <- stringr::str_detect(metadata$target_url, regex)
  metadata_zones <- metadata[sel_zones, ]
  dir_name <- fs::path_dir(metadata_zones$local_path[1])
  if (!dir.exists(dir_name)) {
    fs::dir_create(dir_name, recurse = TRUE)
  }

  if (!fs::file_exists(metadata_zones$local_path)) {
    if (isFALSE(quiet)) {
      message("Downloading the file to: ", metadata_zones$local_path)
    }
    downloaded_file <- spod_download_in_batches(metadata_zones)
    # downloaded_file <- spod_multi_download_with_progress(metadata_zones)
    # disable curl::multi_download() for now
    # downloaded_file <- curl::multi_download(
    #   metadata_zones$target_url,
    #   destfiles = metadata_zones$local_path,
    #   resume = TRUE,
    #   progress = TRUE
    # )
    downloaded_file <- downloaded_file$local_path
  } else {
    if (isFALSE(quiet)) {
      message("File already exists: ", metadata_zones$local_path)
    }
    downloaded_file <- metadata_zones$local_path
  }

  if (isFALSE(quiet)) {
    message("Unzipping the file: ", downloaded_file)
  }
  if (!dir.exists(fs::path_dir(downloaded_file))) {
    fs::dir_create(fs::path_dir(downloaded_file), recurse = TRUE)
  }
  utils::unzip(
    downloaded_file,
    exdir = paste0(fs::path_dir(downloaded_file), "/")
  )

  # remove artifacts (remove __MACOSX if exists)
  junk_path <- paste0(fs::path_dir(downloaded_file), "/__MACOSX")
  if (dir.exists(junk_path)) {
    fs::dir_delete(junk_path)
  }

  return(metadata_zones$local_path)
}


#' Retrieves the zones v2 data
#'
#' This function retrieves the zones data from the specified data directory.
#' It can retrieve either "distritos" or "municipios" zones data.
#'
#' @param data_dir The directory where the data is stored.
#' @param zones The zones for which to download the data. Can be `"districts"` (or `"dist"`, `"distr"`, or the original Spanish `"distritos"`) or `"municipalities"` (or `"muni"`, `"municip"`, or the original Spanish `"municipios"`).
#' @inheritParams global_quiet_param
#' @return An `sf` object (Simple Feature collection) with 4 fields:
#' \describe{
#'   \item{id}{A character vector containing the unique identifier for each zone, to be matched with identifiers in the tabular data.}
#'   \item{name}{A character vector with the name of the zone.}
#'   \item{population}{A numeric vector representing the population of each zone (as of 2022).}
#'   \item{geometry}{A `MULTIPOLYGON` column containing the spatial geometry of each zone, stored as an sf object.
#'   The geometry is projected in the ETRS89 / UTM zone 30N coordinate reference system (CRS), with XY dimensions.}
#' }
#' @keywords internal
spod_get_zones_v2 <- function(
  zones = c(
    "districts",
    "dist",
    "distr",
    "distritos",
    "municipalities",
    "muni",
    "municip",
    "municipios",
    "lua",
    "large_urban_areas",
    "gau",
    "grandes_areas_urbanas"
  ),
  data_dir = spod_get_data_dir(),
  quiet = FALSE
) {
  zones <- spod_zone_names_en2es(zones)

  # check if gpkg files are already saved and load them if available
  expected_gpkg_path <- fs::path(
    data_dir,
    spod_subfolder_clean_data_cache(ver = 2),
    "zones",
    glue::glue("{zones}_mitma.gpkg")
  )

  if (fs::file_exists(expected_gpkg_path)) {
    if (isFALSE(quiet)) {
      message(
        "Loading .gpkg file that already exists in data dir: ",
        expected_gpkg_path
      )
    }
    return(sf::read_sf(expected_gpkg_path))
  }

  # if no existing gpkg found above, continue here with download and data cleanup
  metadata <- spod_available_data_v2(data_dir, check_local_files = TRUE)
  zones_regex <- glue::glue(
    "(zonificacion_{zones}\\.*)|(poblacion\\.csv)|(relacion_ine_zonificacionMitma\\.csv)"
  )
  sel_zones <- stringr::str_detect(metadata$local_path, zones_regex)
  metadata_zones <- metadata[sel_zones, ]
  metadata_zones_for_download <- metadata_zones[
    metadata_zones$downloaded == FALSE,
  ]
  if (nrow(metadata_zones_for_download) > 0) {
    dir_names <- unique(fs::path_dir(metadata_zones_for_download$local_path))
    if (any(!dir.exists(dir_names))) {
      fs::dir_create(dir_names, recurse = TRUE)
    }
    if (isFALSE(quiet)) {
      message("Downloading missing zones data...")
    }
    # disable curl::multi_download() for now
    # curl::multi_download(
    #   urls = metadata_zones_for_download$target_url,
    #   destfiles = metadata_zones_for_download$local_path,
    #   resume = TRUE,
    #   progress = TRUE
    # )
    metadata_zones_for_download <- spod_download_in_batches(
      metadata_zones_for_download
    )
    # metadata_zones_for_download <- spod_multi_download_with_progress(
    #   metadata_zones_for_download
    # )
  }

  zones_path <- fs::dir_ls(
    path = fs::path(data_dir, spod_subfolder_raw_data_cache(ver = 2)),
    regexp = glue::glue("zonificacion_{tolower(zones)}s?\\.shp$"),
    recurse = TRUE
  )

  zones_sf <- spod_clean_zones_v2(zones_path)
  fs::dir_create(fs::path_dir(expected_gpkg_path), recurse = TRUE)
  sf::st_write(
    zones_sf,
    expected_gpkg_path,
    delete_dsn = TRUE,
    delete_layer = TRUE
  )

  return(zones_sf)
}

#' Fixes common issues in the zones data and cleans up variable names
#'
#' This function fixes any invalid geometries in the zones data and renames the "ID" column to "id". It also attacches the population counts and zone names provided in the csv files supplied by the original data provider.
#'
#' @param zones_path The path to the zones spatial data file.
#' @return A spatial `sf`` object containing the cleaned zones data.
#' @importFrom stats median
#' @keywords internal
#'
spod_clean_zones_v2 <- function(zones_path) {
  # detect what kind of zones find out if it is distritos, municipios or GAU
  zones <- stringr::str_extract(zones_path, "distritos|municipios|gaus?")

  if (!is.na(zones) && zones == "gau") {
    zones <- "gaus"
  }

  if (fs::file_exists(zones_path) == FALSE) {
    stop("File does not exist: ", zones_path)
  }
  suppressWarnings({
    zones_sf <- sf::read_sf(zones_path)
  })

  # fix geometry
  invalid_geometries <- !sf::st_is_valid(zones_sf)
  if (sum(invalid_geometries) > 0) {
    fixed_zones_sf <- sf::st_make_valid(zones_sf[invalid_geometries, ])
    zones_sf <- rbind(zones_sf[!invalid_geometries, ], fixed_zones_sf)
  }

  # lowercase id column name
  names(zones_sf)[names(zones_sf) == "ID"] <- "id"

  population <- readr::read_delim(
    glue::glue(fs::path_dir(zones_path), "/poblacion_{zones}.csv"),
    delim = "|",
    col_names = c("id", "population"),
    col_types = c("c", "i")
  )

  if (zones %in% c("distritos", "gaus")) {
    zone_names <- readr::read_delim(
      glue::glue(fs::path_dir(zones_path), "/nombres_{zones}.csv"),
      skip = 1,
      delim = "|",
      col_names = c("id", "name"),
      col_types = c("c", "i")
    )
  } else if (zones == "municipios") {
    zone_names <- readr::read_delim(
      glue::glue(fs::path_dir(zones_path), "/nombres_{zones}.csv"),
      skip = 1,
      delim = "|",
      col_names = c("row", "id", "name"),
      col_types = c("i", "c", "i")
    ) |>
      dplyr::select(-"row")
  }

  # zones reference
  zones_ref <- readr::read_delim(
    fs::path(
      spod_get_data_dir(quiet = TRUE),
      spod_subfolder_raw_data_cache(ver = 2),
      "zonificacion",
      "relacion_ine_zonificacionMitma.csv"
    ),
    delim = "|",
    col_types = rep("c", 6)
  )

  zone_mitma <- glue::glue("{gsub('s$', '', zones)}_mitma")

  zones_ref_renamed <- zones_ref |>
    dplyr::rename(
      census_sections = "seccion_ine",
      census_districts = "distrito_ine",
      municipalities = "municipio_ine",
      districts_mitma = "distrito_mitma",
      municipalities_mitma = "municipio_mitma",
      luas_mitma = "gau_mitma",
      id = dplyr::all_of(zone_mitma)
    )

  zones_ref_aggregated <- zones_ref_renamed |>
    dplyr::group_by(.data$id) |>
    dplyr::summarise(
      dplyr::across(
        .cols = dplyr::everything(),
        .fns = ~ paste(.x, collapse = "; "),
        .names = "{.col}"
      )
    )

  # cleanup duplacate ids in zones_ref_aggregated
  zones_ref_aggregated <- zones_ref_aggregated |>
    dplyr::mutate(
      dplyr::across(
        .cols = dplyr::everything(),
        .fns = spod_unique_separated_ids
      )
    )

  # combine zones with population, names, and zones reference
  zones_sf <- zones_sf |>
    dplyr::left_join(zone_names, by = "id") |>
    dplyr::left_join(population, by = "id") |>
    dplyr::left_join(zones_ref_aggregated, by = "id") |>
    dplyr::relocate("geometry", .after = dplyr::last_col())

  # load v1 zones to join ids, unless it's gau zones
  if (zones != "gaus") {
    spod_download_zones_v1(zones = zones, quiet = TRUE)
    zones_v1_path <- fs::dir_ls(
      path = fs::path(
        spod_get_data_dir(),
        spod_subfolder_raw_data_cache(ver = 1)
      ),
      glob = glue::glue("*v1**{zones}/*.shp"),
      recurse = TRUE
    )
    suppressWarnings({
      zones_v1_sf <- sf::read_sf(zones_v1_path)
    })
    invalid_geometries <- !sf::st_is_valid(zones_v1_sf)
    if (sum(invalid_geometries) > 0) {
      fixed_zones_v1_sf <- sf::st_make_valid(zones_v1_sf[invalid_geometries, ])
      zones_v1_sf <- rbind(zones_v1_sf[!invalid_geometries, ], fixed_zones_sf)
    }

    names(zones_v1_sf)[names(zones_v1_sf) == "ID"] <- "id_in_v1"

    suppressWarnings(
      zones_v2_sf_centroids <- zones_sf |> sf::st_point_on_surface()
    )
    v2_to_v1 <- sf::st_join(zones_v1_sf, zones_v2_sf_centroids, left = TRUE) |>
      sf::st_drop_geometry()
    v2_v_1ref <- v2_to_v1 |>
      dplyr::group_by(.data$id) |>
      dplyr::summarize(
        ids_in_v1_data = paste(.data$id_in_v1, collapse = "; ")
      )
    eng_zones <- dplyr::if_else(
      zones == "distritos",
      true = "district",
      false = "municipality"
    )
    names(v2_v_1ref)[names(v2_v_1ref) == "ids_in_v1_data"] <- glue::glue(
      "{eng_zones}_ids_in_v1"
    )

    zones_sf <- zones_sf |>
      dplyr::left_join(v2_v_1ref, by = "id") |>
      dplyr::relocate("geometry", .after = dplyr::last_col())
  }

  return(zones_sf)
}

test_that("spod_get_zones validation works", {
  test_dir <- setup_test_data_dir()
  withr::defer(unlink(test_dir, recursive = TRUE))
  withr::local_envvar(c("SPANISH_OD_DATA_DIR" = test_dir))

  expect_error(
    spod_get_zones(zones = "districts", ver = 3),
    "Invalid version number"
  )
  expect_error(spod_get_zones(zones = "invalid"), "Must be element of set")
})

test_that("spod_clean_zones_v1 throws error on missing file", {
  expect_error(
    spod_clean_zones_v1("non_existent_file.shp", zones = "districts"),
    "File does not exist"
  )
})

test_that("spod_clean_zones_v2 throws error on missing file", {
  expect_error(
    spod_clean_zones_v2("distritos/non_existent_file.shp"),
    "File does not exist"
  )
})

test_that("spod_get_zones dispatcher works (mocked)", {
  local_mocked_bindings(
    spod_get_zones_v1 = function(...) {
      sf::st_sf(id = "1", geometry = sf::st_sfc(sf::st_point(c(0, 0))))
    },
    spod_get_zones_v2 = function(...) {
      sf::st_sf(id = "2", geometry = sf::st_sfc(sf::st_point(c(0, 0))))
    }
  )

  v1 <- spod_get_zones(zones = "districts", ver = 1, quiet = TRUE)
  expect_equal(v1$id, "1")

  v2 <- spod_get_zones(zones = "districts", ver = 2, quiet = TRUE)
  expect_equal(v2$id, "2")
})

test_that("spod_clean_zones_v1 handles invalid geometries", {
  # Create a simple valid sf object first
  valid_poly <- sf::st_polygon(list(matrix(
    c(0, 0, 1, 0, 1, 1, 0, 1, 0, 0),
    ncol = 2,
    byrow = TRUE
  )))
  # Create an invalid one (self-intersection)
  invalid_poly <- sf::st_polygon(list(matrix(
    c(0, 0, 1, 1, 0, 1, 1, 0, 0, 0),
    ncol = 2,
    byrow = TRUE
  )))

  mock_sf <- sf::st_sf(
    ID = c("1", "2"),
    geometry = sf::st_sfc(valid_poly, invalid_poly)
  )

  # Mock sf::read_sf
  local_mocked_bindings(
    read_sf = function(...) mock_sf,
    .package = "sf"
  )

  # Mock fs::file_exists
  local_mocked_bindings(
    file_exists = function(...) TRUE,
    .package = "fs"
  )

  # Mock readr::read_delim
  local_mocked_bindings(
    read_delim = function(...) {
      # Return minimal structure expected by joins
      if (grepl("relaciones_distrito", ..1)) {
        return(data.frame(
          distrito_mitma = "1",
          municipio_mitma = "101",
          census_district = "1"
          # district_mitma removed to avoid duplication after renaming
        ))
      } else if (grepl("relaciones_municipio", ..1)) {
        return(data.frame(
          municipio_mitma = "101",
          municipality = "Muni1",
          census_district = "1"
        ))
      }
      return(data.frame())
    },
    .package = "readr"
  )

  # Also need to mock spod_get_zones_v2 as it is called inside clean_zones_v1 for metadata
  local_mocked_bindings(
    spod_get_zones_v2 = function(...) {
      sf::st_sf(
        id = "1",
        name = "Zone1",
        geometry = sf::st_sfc(valid_poly)
      )
    }
  )

  cleaned <- spod_clean_zones_v1("dummy_path.shp", "districts")

  # Check if invalid geometry was fixed (st_make_valid usually splits strict bowtie into multipolygon or fixes it)
  expect_true(all(sf::st_is_valid(cleaned)))
})

test_that("spod_clean_zones_v1 aggregates relations data correctly for districts", {
  valid_poly <- sf::st_polygon(list(matrix(
    c(0, 0, 1, 0, 1, 1, 0, 1, 0, 0),
    ncol = 2,
    byrow = TRUE
  )))
  mock_sf <- sf::st_sf(
    ID = c("D001", "D002"),
    geometry = sf::st_sfc(valid_poly, valid_poly)
  )

  local_mocked_bindings(
    read_sf = function(...) mock_sf,
    .package = "sf"
  )

  local_mocked_bindings(
    file_exists = function(...) TRUE,
    .package = "fs"
  )

  # Mock relations to test aggregation
  local_mocked_bindings(
    read_delim = function(...) {
      if (grepl("relaciones_distrito", ..1)) {
        return(data.frame(
          distrito_mitma = c("D001", "D001", "D002"),
          municipio_mitma = c("M001", "M001", "M002"),
          census_district = c("C001", "C002", "C003")
        ))
      } else if (grepl("relaciones_municipio", ..1)) {
        return(data.frame(
          municipio_mitma = c("M001", "M002"),
          municipality = c("500", "501"),
          census_district = c("C001", "C003")
        ))
      }
      return(data.frame())
    },
    .package = "readr"
  )

  local_mocked_bindings(
    spod_get_zones_v2 = function(...) {
      sf::st_sf(
        id = c("D001", "D002"),
        name = c("Name1", "Name2"),
        geometry = sf::st_sfc(valid_poly, valid_poly)
      )
    }
  )

  cleaned <- spod_clean_zones_v1("dummy.shp", "distritos")

  expect_true("census_districts" %in% names(cleaned))
  expect_true("municipalities_mitma" %in% names(cleaned))
  expect_equal(nrow(cleaned), 2)
})

test_that("spod_clean_zones_v1 aggregates relations data correctly for municipalities", {
  valid_poly <- sf::st_polygon(list(matrix(
    c(0, 0, 1, 0, 1, 1, 0, 1, 0, 0),
    ncol = 2,
    byrow = TRUE
  )))
  mock_sf <- sf::st_sf(
    ID = c("M001", "M002"),
    geometry = sf::st_sfc(valid_poly, valid_poly)
  )

  local_mocked_bindings(
    read_sf = function(...) mock_sf,
    .package = "sf"
  )

  local_mocked_bindings(
    file_exists = function(...) TRUE,
    .package = "fs"
  )

  local_mocked_bindings(
    read_delim = function(...) {
      if (grepl("relaciones_distrito", ..1)) {
        return(data.frame(
          distrito_mitma = c("D001", "D002"),
          municipio_mitma = c("M001", "M002"),
          census_district = c("C001", "C002")
        ))
      } else if (grepl("relaciones_municipio", ..1)) {
        return(data.frame(
          municipio_mitma = c("M001", "M002"),
          municipality = c("500", "501"),
          census_district = c("C001", "C002")
        ))
      }
      return(data.frame())
    },
    .package = "readr"
  )

  local_mocked_bindings(
    spod_get_zones_v2 = function(...) {
      sf::st_sf(
        id = c("M001", "M002"),
        name = c("Name1", "Name2"),
        geometry = sf::st_sfc(valid_poly, valid_poly)
      )
    }
  )

  cleaned <- spod_clean_zones_v1("dummy.shp", "municipios")

  expect_true("municipalities" %in% names(cleaned))
  expect_true("districts_mitma" %in% names(cleaned))
  expect_equal(nrow(cleaned), 2)
})

test_that("spod_clean_zones_v2 joins population and name data correctly", {
  valid_poly <- sf::st_polygon(list(matrix(
    c(0, 0, 1, 0, 1, 1, 0, 1, 0, 0),
    ncol = 2,
    byrow = TRUE
  )))
  mock_sf <- sf::st_sf(
    ID = c("01001", "01002"),
    geometry = sf::st_sfc(valid_poly, valid_poly)
  )

  local_mocked_bindings(
    read_sf = function(...) mock_sf,
    .package = "sf"
  )

  local_mocked_bindings(
    file_exists = function(...) TRUE,
    .package = "fs"
  )

  # Mock population and name files
  local_mocked_bindings(
    read_delim = function(file, ...) {
      if (grepl("poblacion", file)) {
        return(data.frame(
          id = c("01001", "01002"),
          population = c(1000L, 2000L)
        ))
      } else if (grepl("nombres", file)) {
        return(data.frame(id = c("01001", "01002"), name = c("Zone1", "Zone2")))
      } else if (grepl("relacion_ine", file)) {
        return(data.frame(
          distrito_mitma = c("01001", "01002"),
          seccion_ine = c("S1", "S2"),
          distrito_ine = c("D1", "D2"),
          municipio_ine = c("M1", "M2"),
          municipio_mitma = c("M001", "M002"),
          gau_mitma = c("G1", "G2")
        ))
      }
      return(data.frame())
    },
    .package = "readr"
  )

  # Mock download zones v1 (called for v1-v2 mapping)
  local_mocked_bindings(
    spod_download_zones_v1 = function(...) NULL
  )

  # Mock dir_ls for v1 zones
  local_mocked_bindings(
    dir_ls = function(...) "mock_v1.shp",
    .package = "fs"
  )

  cleaned <- spod_clean_zones_v2("mock_path/zonificacion_distritos.shp")

  expect_true("population" %in% names(cleaned))
  expect_true("name" %in% names(cleaned))
  expect_equal(cleaned$population, c(1000L, 2000L))
  expect_equal(cleaned$name, c("Zone1", "Zone2"))
})

test_that("spod_clean_zones_v2 handles GAU zones (no v1 mapping)", {
  valid_poly <- sf::st_polygon(list(matrix(
    c(0, 0, 1, 0, 1, 1, 0, 1, 0, 0),
    ncol = 2,
    byrow = TRUE
  )))
  mock_sf <- sf::st_sf(ID = c("G001"), geometry = sf::st_sfc(valid_poly))

  local_mocked_bindings(
    read_sf = function(...) mock_sf,
    .package = "sf"
  )

  local_mocked_bindings(
    file_exists = function(...) TRUE,
    .package = "fs"
  )

  local_mocked_bindings(
    read_delim = function(file, ...) {
      if (grepl("poblacion", file)) {
        return(data.frame(id = "G001", population = 5000L))
      } else if (grepl("nombres", file)) {
        return(data.frame(id = "G001", name = "GauZone"))
      } else if (grepl("relacion_ine", file)) {
        return(data.frame(
          gau_mitma = "G001",
          seccion_ine = "S1",
          distrito_ine = "D1",
          municipio_ine = "M1",
          municipio_mitma = "M001",
          distrito_mitma = "D001"
        ))
      }
      return(data.frame())
    },
    .package = "readr"
  )

  cleaned <- spod_clean_zones_v2("mock_path/zonificacion_gau.shp")

  expect_true("population" %in% names(cleaned))
  expect_true("name" %in% names(cleaned))
  expect_false("district_ids_in_v1" %in% names(cleaned)) # GAU should not have v1 mapping
})

test_that("spod_download_zones_v1 calls download and unzip logic", {
  test_dir <- withr::local_tempdir()

  # Mock metadata
  metadata <- tibble::tibble(
    target_url = c(
      "https://mock/relaciones_distrito_mitma.csv",
      "https://mock/relaciones_municipio_mitma.csv",
      "https://mock/zonificacion_distritos.zip"
    ),
    local_path = c(
      file.path(test_dir, "raw_data_cache/v1/relaciones_distrito_mitma.csv"),
      file.path(test_dir, "raw_data_cache/v1/relaciones_municipio_mitma.csv"),
      file.path(test_dir, "raw_data_cache/v1/zonificacion_distritos.zip")
    )
  )

  local_mocked_bindings(
    spod_available_data = function(...) metadata
  )

  # Mock file_exists to say files don't exist
  local_mocked_bindings(
    file_exists = function(path) {
      # Return FALSE for all paths to trigger download
      return(rep(FALSE, length(path)))
    },
    .package = "fs"
  )

  download_called <- FALSE
  unzip_called <- FALSE

  # Mock download
  local_mocked_bindings(
    spod_download_in_batches = function(files) {
      download_called <<- TRUE
      # Create dummy files
      for (p in files$local_path) {
        dir.create(dirname(p), recursive = TRUE, showWarnings = FALSE)
        writeLines("test", p)
      }
      return(files)
    }
  )

  # Mock unzip
  local_mocked_bindings(
    unzip = function(zipfile, exdir, ...) {
      unzip_called <<- TRUE
      # Create a dummy file in the extraction directory
      dir.create(exdir, recursive = TRUE, showWarnings = FALSE)
      writeLines("unzipped", file.path(exdir, "test.shp"))
    },
    .package = "utils"
  )

  result <- spod_download_zones_v1(
    zones = "distritos",
    data_dir = test_dir,
    quiet = TRUE,
    metadata = metadata
  )

  expect_true(download_called)
  expect_true(unzip_called)
  expect_true(length(result) > 0)
})

#' Get available data list from Amazon S3 storage
#'
#' @description
#'
#' Get a table with links to available data files for the specified data version from Amazon S3 storage.
#'
#' @inheritParams spod_available_data
#' @inheritParams global_quiet_param
#' @return A tibble with links, release dates of files in the data, dates of data coverage, local paths to files, and the download status.
#'
#' @keywords internal
spod_available_data_s3 <- function(
  ver = c(1, 2),
  force = FALSE,
  quiet = FALSE,
  data_dir = spod_get_data_dir()
) {
  ver <- as.character(ver)
  ver <- match.arg(ver)
  metadata_folder <- glue::glue("{data_dir}/{spod_subfolder_metadata_cache()}")

  # if forcing, evict the in-session cache now
  if (isTRUE(force)) {
    memoise::forget(spod_available_data_s3_memoised)
  }

  # shortcut: if we already have it memoised, return immediately
  if (!force && memoise::has_cache(spod_available_data_s3_memoised)(ver)) {
    if (!quiet) message("Using memory-cached available data from S3")
    return(spod_available_data_s3_memoised(ver))
  }

  # no in-session data, check your on-disk RDS pool
  pattern <- glue::glue("metadata_s3_v{ver}_\\d{{4}}-\\d{{2}}-\\d{{2}}\\.rds$")
  rds_files <- fs::dir_ls(
    path = metadata_folder,
    type = "file",
    regexp = pattern
  ) |>
    sort()

  latest_file <- utils::tail(rds_files, 1)
  latest_date <- if (length(latest_file) == 1) {
    stringr::str_extract(basename(latest_file), "\\d{4}-\\d{2}-\\d{2}") |>
      as.Date()
  } else {
    NA
  }

  needs_update <- isTRUE(force) ||
    length(rds_files) == 0 ||
    (!is.na(latest_date) && latest_date < Sys.Date())

  if (!needs_update) {
    if (!quiet) message("Using existing disk cache: ", latest_file)
    return(readRDS(latest_file))
  }

  # if forcing, also clear old disk files
  if (isTRUE(force) && length(rds_files) > 0) {
    fs::file_delete(rds_files)
  }

  # fetch via the memoised function (this will re-hit S3 if we forgot it)
  if (!quiet) message("Fetching latest metadata from AmazonS3 (v", ver, ")...")
  dat <- spod_available_data_s3_memoised(ver)

  # write a new RDS stamped with today's date
  file_date <- format(Sys.Date(), "%Y-%m-%d")
  out_path <- file.path(
    metadata_folder,
    glue::glue("metadata_s3_v{ver}_{file_date}.rds")
  )
  saveRDS(dat, out_path)
  if (!quiet) message("Cached new data to: ", out_path)

  dat
}


spod_available_data_s3_function <- function(
  ver = c(1, 2)
) {
  ver <- as.character(ver)
  ver <- match.arg(ver)

  bucket <- paste0("mitma-movilidad-v", ver)

  # original_aws_region <- Sys.getenv("AWS_DEFAULT_REGION")
  # original_aws_url_style <- Sys.getenv("AWS_S3_URL_STYLE")
  # on.exit({
  #   Sys.setenv(
  #     AWS_DEFAULT_REGION = original_aws_region,
  #     AWS_S3_URL_STYLE = original_aws_url_style
  #   )
  # })
  # Sys.setenv(
  #   AWS_DEFAULT_REGION = "eu-west-1",
  #   AWS_S3_URL_STYLE = "virtual"
  # )

  if (ver == 1) {
    url_prefix <- "https://opendata-movilidad.mitma.es/"
  } else {
    url_prefix <- "https://movilidad-opendata.mitma.es/"
  }

  s3 <- paws.storage::s3(
    config = list(
      credentials = list(
        anonymous = TRUE
      )
    )
  )

  all_objects <- list_objects_v2_all(s3, bucket)

  # all_objects <- aws.s3::get_bucket_df(
  #   bucket = bucket,
  #   prefix = "", # root of bucket
  #   max = Inf # fetch beyond the default 1000
  # )

  all_objects <- all_objects |>
    dplyr::as_tibble() |>
    dplyr::mutate(
      target_url = paste0(url_prefix, .data$Key),
      pub_ts = as.POSIXct(
        .data$LastModified,
        format = "%Y-%m-%dT%H:%M:%OSZ",
        tz = "UTC"
      ),
      file_size_bytes = as.numeric(.data$Size),
      etag = gsub('\\"', '', .data$ETag)
    ) |>
    dplyr::select(
      "target_url",
      "pub_ts",
      "file_size_bytes",
      "etag"
    )

  return(all_objects)
}

spod_available_data_s3_memoised <- memoise::memoise(
  spod_available_data_s3_function
)

list_objects_v2_all <- function(s3, bucket, prefix = "", max_keys = 10000) {
  pages <- paws.storage::paginate(
    s3$list_objects_v2(
      Bucket = bucket,
      Prefix = prefix,
      MaxKeys = max_keys
    ),
    PageSize = max_keys
  )

  all_objects <- unlist(
    lapply(pages, `[[`, "Contents"),
    recursive = FALSE
  )

  metadata <- dplyr::tibble(
    Key = vapply(all_objects, `[[`, character(1), "Key"),
    LastModified = as.POSIXct(
      vapply(all_objects, `[[`, numeric(1), "LastModified"),
      origin = "1970-01-01",
      tz = "UTC"
    ),
    Size = vapply(all_objects, `[[`, numeric(1), "Size"),
    ETag = vapply(all_objects, `[[`, character(1), "ETag")
  )

  # S3 generate download urls
  # urls <- metadata$Key |>
  #   purrr::map(
  #     ~ s3$generate_presigned_url(
  #       client_method = "get_object",
  #       params = list(Bucket = "mitma-movilidad-v1", Key = .x)
  #     ),
  #     .progress = TRUE
  #   )

  return(metadata)
}

#' @keywords internal
"_PACKAGE"

## usethis namespace: start
#' @importFrom lifecycle deprecated
#' @importFrom Rdpack reprompt
#' @importFrom stats end start
#' @importFrom utils capture.output
## usethis namespace: end
NULL

#' @title Global Quiet Parameter
#'
#' @description
#' Documentation for the `quiet` parameter, used globally.
#'
#' @param quiet A `logical` value indicating whether to suppress messages. Default is `FALSE`.
#' @return Nothing. This function is just a placeholder for global quiet parameter.
#' @keywords internal
global_quiet_param <- function(quiet = FALSE){
  # this is just a placeholder for global quiet parameter
}

#' Cite the package and the data
#'
#' @param what Character vector specifying what to cite.
#'   Can include "package", "data", "methodology_v1", "methodology_v2", or "all".
#'   Default is "all".
#' @param format Character vector specifying output format(s).
#'   Can include "text", "markdown", "bibtex", or "all".
#'   Default is "all".
#' @return Nothing. Prints citation in plain text, markdown, BibTeX, or all formats at once to console.
#' @export
#'
#' @examples
#' # Cite everything in all formats
#' \dontrun{
#' spod_cite()
#' }
#'
#' # Cite just the package in BibTeX format
#' \dontrun{
#' spod_cite(what = "package", format = "bibtex")
#' }
#'
#' # Cite both methodologies in plain text
#' \dontrun{
#' spod_cite(what = c("methodology_v1", "methodology_v2"), format = "text")
#' }
spod_cite <- function(
  what = "all",
  format = "all"
) {
  # 1. Define valid inputs
  valid_what <- c("all", "package", "data", "methodology_v1", "methodology_v2")
  valid_format <- c("all", "text", "markdown", "bibtex")

  # 2. Use checkmate to validate arguments
  checkmate::assertCharacter(what, any.missing = FALSE, min.len = 1)
  checkmate::assertCharacter(format, any.missing = FALSE, min.len = 1)
  checkmate::assertSubset(what, choices = valid_what)
  checkmate::assertSubset(format, choices = valid_format)

  # 3. Expand "all" options
  if ("all" %in% what) {
    what <- unique(c(what, valid_what[valid_what != "all"]))
  }
  if ("all" %in% format) {
    format <- unique(c(format, valid_format[valid_format != "all"]))
  }

  # Remove the literal "all" from each
  what <- setdiff(what, "all")
  format <- setdiff(format, "all")

  # 4. Map what options to BibTeX keys in REFERENCES.bib
  citation_keys <- list(
    package = "spanishoddata-r-pkg",
    data = "mitms_mobility_web",
    methodology_v1 = "mitma_methodology_2020_v3",
    methodology_v2 = "mitms_methodology_2022_v8"
  )

  # 4.1. Human-readable labels for each "what"
  citation_labels <- list(
    package = "To cite the spanishoddata package",
    data = "To cite the Ministry's mobility study website",
    methodology_v1 = "To cite the methodology for 2020-2021 data",
    methodology_v2 = "To cite the methodology for 2022 and onwards data"
  )

  # 5. Read REFERENCES.bib using Rdpack
  bibs <- Rdpack::get_bibentries(package = "spanishoddata")

  # 6. Collect the requested citations
  citations_to_show <- list()
  for (w in what) {
    key <- citation_keys[[w]]
    if (key %in% names(bibs)) {
      citations_to_show[[w]] <- bibs[key]
    }
  }

  # If nothing was found
  if (length(citations_to_show) == 0) {
    message("No valid citations found for the requested 'what'.")
    return(invisible(NULL))
  }

  # 7. Print the citations in requested formats
  for (f in format) {
    if (f == "text") {
      cat("\nPlain text citations:\n---------------------\n")
      for (w in names(citations_to_show)) {
        cit_entry <- citations_to_show[[w]]
        cat(paste0(citation_labels[[w]], ":\n"))
        cat(format(cit_entry, style = "text"), "\n\n")
        # Note for methodology_v2
        if (w == "methodology_v2") {
          cat(
            "Note: A more up-to-date methodology document may be available at https://www.transportes.gob.es/ministerio/proyectos-singulares/estudios-de-movilidad-con-big-data/metodologia-del-estudio-de-movilidad-con-bigdata\n\n"
          )
        }
      }
    } else if (f == "markdown") {
      cat("\nMarkdown citations:\n-------------------\n")
      for (w in names(citations_to_show)) {
        cit_entry <- citations_to_show[[w]]
        cat(paste0("**", citation_labels[[w]], ":**\n"))
        cat(format(cit_entry, style = "text"), "\n\n")
        # Note for methodology_v2
        if (w == "methodology_v2") {
          cat(
            "> **Note:** A more up-to-date methodology document may be available at https://www.transportes.gob.es/ministerio/proyectos-singulares/estudios-de-movilidad-con-big-data/metodologia-del-estudio-de-movilidad-con-bigdata\n\n"
          )
        }
      }
    } else if (f == "bibtex") {
      cat("\nBibTeX citations:\n-----------------\n")
      for (w in names(citations_to_show)) {
        cit_entry <- citations_to_show[[w]]
        cat(paste0("%% ", citation_labels[[w]], "\n"))
        
        # Capture standard BibTeX output
        bib_lines <- capture.output(print(cit_entry, style = "bibtex"))

        # CUSTOM FORMATTING: Ensure "Last, First" author format with strict brace preservation
        # 1. Generate text string for authors in "Last, First" format
        auth_obj <- cit_entry$author
        if (is.null(auth_obj)) {
          # Fallback via unclass if direct access fails (unlikely for bibentry)
          auth_obj <- unclass(cit_entry)[[1]]$author
        }

        custom_author_str <- NULL
        if (!is.null(auth_obj) && length(auth_obj) > 0) {
          custom_author_str <- paste(
            sapply(auth_obj, function(p) {
              fam <- paste(p$family, collapse = " ")
              giv <- paste(p$given, collapse = " ")
              
              # Heuristic: If family name contains hyphens, spaces, or backslashes (latex),
              # preserve it wrapped in curly braces.
              # Note: Standard names like "Kotov" (alphanumeric) are left as is.
              if (grepl("[-\\\\\\s]", fam)) {
                 fam <- paste0("{", fam, "}")
              }
              
              if (length(p$given) == 0 || nchar(giv) == 0) {
                # Organization or Mononym
                # Even if simple, organization names often benefit from braces if not broken down
                # But our heuristic handles spaces, so "Ministry of ..." gets braces.
                return(fam) 
              } else {
                return(paste0(fam, ", ", giv))
              }
            }),
            collapse = " and "
          )
        }

        # 2. Replace the 'author = { ... },' block in bib_lines
        start_idx <- grep("^\\s*author\\s*=", bib_lines)
        
        if (length(start_idx) == 1 && !is.null(custom_author_str)) {
          # Find end of the author block by brace balancing
          current_line_idx <- start_idx
          brace_balance <- 0
          
          # Iterate to find the closing brace
          for (i in start_idx:length(bib_lines)) {
            line <- bib_lines[i]
            # Simple count of { and }
            # Note: R's toBibtex output is regular and clean
            n_open <- stringr::str_count(line, "\\{")
            n_close <- stringr::str_count(line, "\\}")
            brace_balance <- brace_balance + n_open - n_close
            
            if (brace_balance == 0) {
              current_line_idx <- i
              break
            }
          }
          end_idx <- current_line_idx
          
          # Replace the block with the single formatted line
          new_line <- paste0("  author = {", custom_author_str, "},")
          
          # Construct new lines vector
          pre_block <- if (start_idx > 1) bib_lines[1:(start_idx - 1)] else character(0)
          post_block <- if (end_idx < length(bib_lines)) bib_lines[(end_idx + 1):length(bib_lines)] else character(0)
          
          bib_lines <- c(pre_block, new_line, post_block)
        }
        
        cat(bib_lines, sep = "\n")
        cat("\n\n")
        
        # Note for methodology_v2
        if (w == "methodology_v2") {
          cat(
            "%% Note: A more up-to-date methodology document may be available at https://www.transportes.gob.es/ministerio/proyectos-singulares/estudios-de-movilidad-con-big-data/metodologia-del-estudio-de-movilidad-con-bigdata\n\n"
          )
        }
      }
    }
  }

  invisible(NULL)
}

# Load testthat
library(testthat)

# --- Tests for spod_cite function ---

test_that("spod_cite with default arguments prints all citations in all formats", {
  # Capture the output of a single function call
  output <- capture.output(spod_cite())
  output_str <- paste(output, collapse = "\n")

  # Run all checks on the captured string
  expect_true(grepl("Plain text citations:", output_str))
  expect_true(grepl("Markdown citations:", output_str))
  expect_true(grepl("BibTeX citations:", output_str))
  expect_true(grepl("To cite the spanishoddata package", output_str))
  expect_true(grepl(
    "To cite the Ministry's mobility study website",
    output_str
  ))
  expect_true(grepl("To cite the methodology for 2020-2021 data", output_str))
  expect_true(grepl(
    "To cite the methodology for 2022 and onwards data",
    output_str
  ))
})

test_that("spod_cite with specific 'what' and 'format' arguments works", {
  # Capture the specific output to a string for inspection
  output <- capture.output(spod_cite(what = "package", format = "bibtex"))
  output_str <- paste(output, collapse = "\n")

  # Check for expected content
  expect_true(grepl("BibTeX citations:", output_str))
  expect_true(grepl("%% To cite the spanishoddata package", output_str))
  
  # Check for case-insensitive article type (Rdpack uses @Article)
  expect_true(grepl("@article", output_str, ignore.case = TRUE))

  # CRITICAL: Check for "Last, First" author format with BRACES
  # "Kotov" should be simple
  expect_true(grepl("Kotov, Egor", output_str))
  
  # "Vidal-Tortosa" should have braces because of hyphen
  expect_true(grepl("\\{Vidal-Tortosa\\}, Eugeni", output_str))
  
  # "Cantú-Ros" should have braces (assuming backslash makes it complex, even if not visible in grepl, 
  # we check for the brace wrapper)
  expect_true(grepl("\\{Cant.*\\-Ros\\}", output_str)) 

  # Check that plain output does NOT appear (e.g. Vidal-Tortosa without braces)
  # Actually, simple grepl for "{Vidal-Tortosa}" confirms braces are present.
  
  # Check that other content is NOT present
  expect_false(grepl("Plain text citations:", output_str))
})

test_that("spod_cite with multiple 'what' arguments works", {
  output <- capture.output(spod_cite(
    what = c("methodology_v1", "data"),
    format = "text"
  ))
  output_str <- paste(output, collapse = "\n")

  # Check for expected content
  expect_true(grepl("Plain text citations:", output_str))
  expect_true(grepl(
    "To cite the Ministry's mobility study website:",
    output_str
  ))
  expect_true(grepl("To cite the methodology for 2020-2021 data:", output_str))

  # Check that other content is NOT present
  expect_false(grepl("To cite the spanishoddata package", output_str))
  expect_false(grepl("BibTeX citations:", output_str))
})

test_that("spod_cite handles invalid inputs gracefully", {
  expect_error(spod_cite(what = "invalid_input"))
  expect_error(spod_cite(format = "invalid_format"))
})

test_that("spod_cite includes the special note for methodology_v2", {
  output_bib <- capture.output(spod_cite(
    what = "methodology_v2",
    format = "bibtex"
  ))
  expect_true(any(grepl(
    "%% Note: A more up-to-date methodology document may be available",
    output_bib
  )))
})

test_that("spod_cite expands 'all' argument correctly", {
  output_what <- capture.output(spod_cite(
    what = c("package", "all"),
    format = "text"
  ))
  expect_true(grepl("To cite the spanishoddata package", paste(output_what, collapse="\n")))
})

# This file is for internal functions that update some of the package internal packaged data. These functions are not intended neither to be used by the user nor to be used in any of the package functions.

#' Get files sizes for remote files of v1 and v2 data and save them into a csv.gz file in the inst/extdata folder.
#' @param ver The version of the data (1 or 2). Can be both. Defaults to 2, as v1 data is not being updated since 2021.
#' @return Nothing. Only saves a csv.gz file with up to date file sizes in the inst/extdata folder.
#'
#' @keywords internal
#'
spod_files_sizes <- function(ver = 2) { # nocov start
  data_dir <- spod_get_data_dir()

  if (any(ver %in% 1)) {
    v1 <- spod_available_data(1)

    # takes about 1 minute
    # tictoc::tic()
    future::plan(future::multisession, workers = 8)
    v1$true_remote_file_size_bytes <- furrr::future_map_dbl(
      .x = v1$target_url,
      .f = ~ spod_get_file_size_from_url(x_url = .x),
      .progress = TRUE
    )
    future::plan(future::sequential)
    # tictoc::toc()

    v1 <- v1 |>
      dplyr::arrange(!!!dplyr::syms(c("data_ymd", "target_url"))) |>
      dplyr::select(-"local_path")

    saveRDS(v1, file.path("inst", "extdata", "available_data_v1.rds"))
  }

  if (any(ver %in% 2)) {
    v2 <- spod_available_data(2)
    if (all(v2$size_imputed == FALSE)) {
      stop("all file sizes are known")
    }
    v2_known_size <- v2[v2$size_imputed == FALSE, ]
    v2_unknown_size <- v2[v2$size_imputed == TRUE, ]

    # takes about 5 minutes on full data set, but less when only updating the previously uknown files
    future::plan(future::multisession, workers = 6)
    v2_unknown_size$remote_file_size_mb <- furrr::future_map_dbl(
      .x = v2_unknown_size$target_url,
      .f = ~ spod_get_file_size_from_url(x_url = .x),
      .progress = TRUE
    )
    future::plan(future::sequential)

    v2_combined <- dplyr::bind_rows(v2_known_size, v2_unknown_size)
    v2_url_file_sizes <- v2_combined[, c("target_url", "remote_file_size_mb")]
    readr::write_csv(
      x = v2_url_file_sizes,
      file = "inst/extdata/url_file_sizes_v2.txt.gz"
    )
  }
}


#' Get file size from URL
#' @param x_url URL
#' @return File size in MB
#' @keywords internal
spod_get_file_size_from_url <- function(x_url) {
  url <- utils::URLencode(x_url)
  headers <- curlGetHeaders(url)
  content_length_line <- grep("Content-Length", headers, value = TRUE)
  content_length_value <- sub(
    "Content-Length:\\s*(\\d+).*",
    "\\1",
    content_length_line
  ) |>
    as.numeric()

  return(content_length_value)
}

#' Get Etags for locally saved v1 data files and save them into a RDS file in the inst/extdata folder.
#' @return Returns a tibble with the local path, local ETag and remote ETag.
#' @keywords internal
spod_store_etags <- function() {
  available_data <- spod_available_data(1, check_local_files = TRUE)
  available_data <- available_data |>
    dplyr::filter(.data$downloaded == TRUE)
  local_etags <- available_data$local_path |>
    purrr::map_chr(~ spod_compute_s3_etag(.x), .progress = TRUE)
  available_data <- available_data |>
    dplyr::mutate(local_etag = local_etags) |>
    dplyr::as_tibble()
  return(available_data)
}
# nocov end

#' Convert multiple formates of date arguments to a sequence of dates
#'
#' This function processes the date arguments provided to various functions in the package. It can handle single dates and arbitratry sequences (vectors) of dates in ISO (YYYY-MM-DD) and YYYYMMDD format. It can also handle date ranges in the format 'YYYY-MM-DD_YYYY-MM-DD' (or 'YYYYMMDD_YYYYMMDD'), date ranges in named vec and regular expressions to match dates in the format `YYYYMMDD`.
#'
#' @param dates A `character` or `Date` vector of dates to process. Kindly keep in mind that v1 and v2 data follow different data collection methodologies and may not be directly comparable. Therefore, do not try to request data from both versions for the same date range. If you need to compare data from both versions, please refer to the respective codebooks and methodology documents. The v1 data covers the period from 2020-02-14 to 2021-05-09, and the v2 data covers the period from 2022-01-01 to the present until further notice. The true dates range is checked against the available data for each version on every function run.
#'
#' The possible values can be any of the following:
#'
#'  * For the `spod_get()` and `spod_convert()` functions, the `dates` can be set to "cached_v1" or "cached_v2" to request data from cached (already previously downloaded) v1 (2020-2021) or v2 (2022 onwards) data. In this case, the function will identify and use all data files that have been downloaded and cached locally, (e.g. using an explicit run of `spod_download()`, or any data requests made using the `spod_get()` or `spod_convert()` functions).
#'
#'  * A single date in ISO (YYYY-MM-DD) or YYYYMMDD format. `character` or `Date` object.
#'
#' * A vector of dates in ISO (YYYY-MM-DD) or YYYYMMDD format. `character` or `Date` object. Can be any non-consecutive sequence of dates.
#'
#' * A date range
#'
#'   * eigher a `character` or `Date` object of length 2 with clearly named elements `start` and `end` in ISO (YYYY-MM-DD) or YYYYMMDD format. E.g. `c(start = "2020-02-15", end = "2020-02-17")`;
#'
#'   * or a `character` object of the form `YYYY-MM-DD_YYYY-MM-DD` or `YYYYMMDD_YYYYMMDD`. For example, `2020-02-15_2020-02-17` or `20200215_20200217`.
#'
#' * A regular expression to match dates in the format `YYYYMMDD`. `character` object. For example, `^202002` will match all dates in February 2020.
#'
#'
#' @return A `character` vector of dates in ISO format (YYYY-MM-DD).
#' @keywords internal
spod_dates_argument_to_dates_seq <- function(dates) {
  if (is.null(dates) || (!is.character(dates) && !inherits(dates, "Date"))) {
    stop(
      "Invalid date input format. Please provide a character vector or Date object."
    )
  }
  if (length(dates) == 1 && dates %in% c("cached_v1", "cached_v2")) {
    return(dates)
  }

  range_regex <- "^\\d{4}(-\\d{2}){2}_\\d{4}(-\\d{2}){2}$|^\\d{8}_\\d{8}$"
  single_date_regex <- "^(\\d{4}-\\d{2}-\\d{2}|\\d{8})$"
  # If dates is a vector of length one
  # Check if is single date, date range, or regex pattern
  if (length(dates) == 1) {
    # Check if date range
    # match both YYYY-MM-DD_YYYY-MM-DD and YYYYMMDD_YYYYMMDD
    if (grepl(range_regex, dates)) {
      date_parts <- strsplit(dates, "_")[[1]]
      date_parts <- lubridate::ymd(date_parts)
      dates <- seq.Date(date_parts[1], date_parts[2], by = "day")

      # if dates does not match the date range pattern
      # check if it is just a single day in YYYY-MM-DD or YYYYMMDD format
    } else if (grepl(single_date_regex, dates)) {
      dates <- lubridate::ymd(dates)

      # assume it is a regex pattern
    } else {
      dates <- spod_expand_dates_from_regex(dates)
      # since spod_expand_dates_from_regex already uses the metadata to generate valid dates we can skip any checks that are required for other date formats and only check for date overlaps between data versions
      if (isFALSE(spod_is_data_version_overlaps(dates))) {
        return(dates)
      }
    }

    # If dates if a vector of multiple values
  } else if (length(dates) > 1) {
    # Check if it is of length 2, then it may be a date range
    if (length(dates) == 2 & !is.null(names(dates))) {
      # if the vector is named with 'start' and 'end', we can assume it is a date range
      if (all(names(dates) %in% c("start", "end"))) {
        date_parts <- lubridate::ymd(dates)
        names(date_parts) <- names(dates)
        # check if start is before end
        if (date_parts["start"] > date_parts["end"]) {
          stop("Start date must be before end date.")
        }
        dates <- seq.Date(date_parts["start"], date_parts["end"], by = "day")
      }
    } else {
      # this is apparantly a sequence of dates
      dates <- lubridate::ymd(dates)
    }
  }

  # now that we have a clean sequence of dates, we can check for overlaps between data versions
  if (isFALSE(spod_is_data_version_overlaps(dates))) {
    return(dates)
  }
}


#' Check if specified dates span both data versions
#'
#' This function checks if the specified dates or date ranges span both v1 and v2 data versions.
#'
#' @param dates A `Dates` vector of dates to check.
#' @return `TRUE` if the dates span both data versions, `FALSE` otherwise.
#' @keywords internal
spod_is_data_version_overlaps <- function(dates) {
  all_dates_v1 <- spod_get_valid_dates(ver = 1)
  all_dates_v2 <- spod_get_valid_dates(ver = 2)

  if (any(dates %in% all_dates_v1) && any(dates %in% all_dates_v2)) {
    stop(paste0(
      "Dates found in both v1 and v2 data. The v1 and v2 data sets may not be comparable. Please see the respective codebooks and methodology documents: run `spod_codebook(1)` and `spod_codebook(2)`.\nThe valid dates range for v1 is: ",
      paste(spod_convert_dates_to_ranges(all_dates_v1), collapse = ", "),
      "\nThe valid dates range for v2 is: ",
      paste(spod_convert_dates_to_ranges(all_dates_v2), collapse = ", "),
      "\n To get the list of valid dates, you can use `spod_get_valid_dates(1)` and `spod_get_valid_dates(2)`."
    ))
  }
  return(FALSE)
}

#' Infer data version from dates
#'
#' @inheritParams spod_download
#' @inheritParams spod_dates_argument_to_dates_seq
#' @return An `integer` indicating the inferred data version.
#' @keywords internal
spod_infer_data_v_from_dates <- function(
  dates,
  ignore_missing_dates = FALSE
) {
  # check if user is requesting to just get all cached data
  if (length(dates) == 1) {
    if (as.character(dates) %in% c("cached_v1", "cached_v2")) {
      return(
        as.integer(stringr::str_extract(as.character(dates), "[0-9]$"))
      )
    }
  }
  cached_data_requested <- length(dates) == 1 &&
    all(as.character(dates) %in% c("cached_v1", "cached_v2"))

  # in case of overlap
  # will throw an error from the spod_is_data_version_overlaps
  if (spod_is_data_version_overlaps(dates)) {
    invisible(return(NULL))
  }

  # if no overlap identified above, compare with date ranges
  v1_dates <- spod_get_valid_dates(ver = 1)
  v2_dates <- spod_get_valid_dates(ver = 2)

  # Check if all dates are missing
  missing_dates <- dates[!dates %in% c(v1_dates, v2_dates)]
  if (length(missing_dates) == length(dates)) {
    if (isTRUE(ignore_missing_dates)) return(NULL)
    stop(paste0(
      "All requested dates are missing from the available data.\nThe valid dates range for v1 is: ",
      paste(spod_convert_dates_to_ranges(v1_dates), collapse = ", "),
      "\nThe valid dates range for v2 is: ",
      paste(spod_convert_dates_to_ranges(v2_dates), collapse = ", "),
      ".\nYou requested the following missing dates: ",
      paste0(missing_dates, collapse = ", "),
      "\nSome of these dates are missing 'naturally' because of the mobile network outages. You can check the up to date list of the missing dates on the source data page at https://www.transportes.gob.es/ministerio/proyectos-singulares/estudios-de-movilidad-con-big-data/opendata-movilidad",
      "\nTo get the list of valid dates, you can use `spod_get_valid_dates(1)` and `spod_get_valid_dates(2)`.",
      "\nConsider revising your date range or handling missing dates appropriately."
    ))
  }

  if (all(dates %in% v1_dates)) {
    return(1)
  } else if (all(dates %in% v2_dates)) {
    return(2)
  } else {
    # Handle missing dates based on ignore_missing_dates argument
    missing_dates <- dates[!dates %in% c(v1_dates, v2_dates)]
    if (length(missing_dates) > 0) {
      if (isTRUE(ignore_missing_dates)) {
        # Filter out missing dates and infer version based on the remaining ones
        valid_dates <- dates[dates %in% c(v1_dates, v2_dates)]
        if (all(valid_dates %in% v1_dates)) {
          return(1)
        } else if (all(valid_dates %in% v2_dates)) {
          return(2)
        } else {
          # If no valid dates remain, or none fully match a version
          return(NULL)
        }
      } else if (isFALSE(ignore_missing_dates)) {
        # Stop with an error if ignore_missing_dates is FALSE
        stop(paste0(
          "Some dates do not match the available data.\nThe valid dates range for v1 is: ",
          paste(spod_convert_dates_to_ranges(v1_dates), collapse = ", "),
          "\nThe valid dates range for v2 is: ",
          paste(spod_convert_dates_to_ranges(v2_dates), collapse = ", "),
          ".\nYou requested the following missing dates: ",
          paste0(missing_dates, collapse = ", "),
          "\nSome of these dates are missing 'naturally' because of the mobile network outages. You can check the up to date list of the missing dates on the source data page at https://www.transportes.gob.es/ministerio/proyectos-singulares/estudios-de-movilidad-con-big-data/opendata-movilidad",
          "\nTo get the list of valid dates, you can use `spod_get_valid_dates(1)` and `spod_get_valid_dates(2)`.",
          "\nYou can also set `ignore_missing_dates` argument when using `spod_get()`, `spod_convert()`, or `spod_download()` to TRUE to ignore missing dates in your requested date range. Use that with caution, as skipping the missing dates may lead to misleading results if you then intend to calculate monthly or weekly flows including the periods with missing data. Consider using imputation methods to fill in the missing data from the available dates."
        ))
      }
    }
  }
}

#' Function to expand dates from a regex
#'
#' This function generates a sequence of dates from a regular expression pattern based on the provided regular expression.
#'
#' @param date_regex A regular expression to match dates in the format 'yyyymmdd'.
#' @return A `character` vector of dates matching the regex.
#' @keywords internal
spod_expand_dates_from_regex <- function(date_regex) {
  all_dates_v1 <- spod_get_valid_dates(ver = 1)
  all_dates_v2 <- spod_get_valid_dates(ver = 2)

  # Filter dates matching the regex for both versions
  matching_dates_v1 <- all_dates_v1[grepl(
    date_regex,
    format(all_dates_v1, "%Y%m%d")
  )]
  matching_dates_v2 <- all_dates_v2[grepl(
    date_regex,
    format(all_dates_v2, "%Y%m%d")
  )]

  # if both vectors are empty, throw an error
  if (length(matching_dates_v1) == 0 && length(matching_dates_v2) == 0) {
    stop(paste0(
      "No matching dates found in the available data.",
      "\nThe valid dates range for v1 is: ",
      paste(spod_convert_dates_to_ranges(all_dates_v1), collapse = ", "),
      "\nThe valid dates range for v2 is: ",
      paste(spod_convert_dates_to_ranges(all_dates_v2), collapse = ", "),
      "."
    ))
  }
  # If checks above have passed, we can combine the matching dates as only one contains dates and the other is empty
  matching_dates <- sort(c(matching_dates_v1, matching_dates_v2))

  return(matching_dates)
}

#' @keywords internal
spod_get_valid_dates_impl <- function(ver = NULL) {
  # Validate input
  checkmate::assertIntegerish(ver, max.len = 1)
  if (!ver %in% c(1, 2)) {
    stop(
      "Invalid version number. Must be 1 (for v1 2020-2021 data) or 2 (for v2 2022 onwards)."
    )
  }

  if (ver == 1) {
    # available_data <- spod_available_data_v1(check_local_files = FALSE, quiet = TRUE)
    # all_dates <- unique(available_data[grepl("maestra1.*diarios", available_data$target_url),]$data_ymd, na.rm = TRUE)
    # perahps it is worth hardcoding at lest the v1 data range as it is unlikely to change at this point
    all_dates <- seq.Date(
      from = as.Date("2020-02-14"),
      to = as.Date("2021-05-09"),
      by = "day"
    )
  } else if (ver == 2) {
    available_data <- spod_available_data_v2(quiet = TRUE)
    all_dates <- unique(
      available_data[
        grepl("viajes.*diarios", available_data$target_url),
      ]$data_ymd,
      na.rm = TRUE
    )
  }
  all_dates <- sort(all_dates)
  return(all_dates)
}

spod_get_valid_dates_memoised <- memoise::memoise(spod_get_valid_dates_impl)

#' Get valid dates for the specified data version
#'
#' @description
#'
#' `r lifecycle::badge("stable")`
#'
#' Get all metadata for requested data version and identify all dates available for download.
#'
#' @inheritParams spod_available_data
#'
#' @return A vector of type `Date` with all possible valid dates for the specified data version (v1 for 2020-2021 and v2 for 2020 onwards).
#' @export
#' @examplesIf interactive()
#' \donttest{
#' # Get all valid dates for v1 (2020-2021) data
#' spod_get_valid_dates(ver = 1)
#'
#' # Get all valid dates for v2 (2020 onwards) data
#' spod_get_valid_dates(ver = 2)
#' }
#'
spod_get_valid_dates <- function(ver = NULL) {
  spod_get_valid_dates_memoised(ver)
}
# TODO: currently checks for date range for od data only. not all datasets may be available for all dates, so this function may need to be updated to check for the availability of the specific for the requested dates. spod_match_data_type() helper in the same file may be useful here.

#' Translate zone names from English to Spanish
#' @inheritParams spod_download
#' @return A `character` string with the translated zone name. Or `NULL` if the zone name is not recognized.
#' @keywords internal
spod_zone_names_en2es <- function(
  zones = c(
    "districts",
    "dist",
    "distr",
    "distritos",
    "municipalities",
    "muni",
    "municip",
    "municipios",
    "lua",
    "large_urban_areas",
    "gau",
    "grandes_areas_urbanas"
  )
) {
  zones <- tolower(zones)
  zones <- match.arg(zones)
  if (zones %in% c("districts", "dist", "distr", "distritos")) {
    return("distritos")
  } else if (zones %in% c("municipalities", "muni", "municip", "municipios")) {
    return("municipios")
  } else if (
    zones %in% c("gaus", "lua", "large_urban_areas", "gau", "grandes_areas_urbanas")
  ) {
    return("gau")
  }
}

#' Match data types to folders
#' @inheritParams spod_available_data
#' @return A `character` string with the folder name for the specified data type. Or `NULL` if the data type is not recognized.
#' @keywords internal
spod_match_data_type_for_local_folders <- function(
  type = c(
    "od",
    "origin-destination",
    "os",
    "overnight_stays",
    "nt",
    "number_of_trips"
  ),
  ver = c(1, 2)
) {
  if (!ver %in% c(1, 2)) {
    stop("Invalid version number. Must be 1 or 2.")
  }

  type <- tolower(type)
  type <- match.arg(type)

  if (ver == 1) {
    if (type %in% c("od", "origin-destination")) {
      return("maestra1")
    } else if (type %in% c("nt", "number_of_trips")) {
      return("maestra2")
    }
  }

  if (ver == 2) {
    if (type %in% c("od", "origin-destination")) {
      return("viajes")
    } else if (type %in% c("os", "overnight_stays")) {
      return("pernoctaciones")
    } else if (type %in% c("nt", "number_of_trips")) {
      return("personas")
    }
  }

  # need to add a warning here that the type is not recognized
  return(NULL)
}


#' Match data types for normalisation
#' @param type The type of data to match. Can be "od", "origin-destination", "os", "overnight_stays", or "nt", "number_of_trips".
#' @return A `character` string with the folder name for the specified data type. Or `NULL` if the type is not recognized.
#' @keywords internal
spod_match_data_type <- function(
  type = c(
    "od",
    "origin-destination",
    "viajes",
    "os",
    "overnight_stays",
    "pernoctaciones",
    "nt",
    "number_of_trips",
    "personas"
  )
) {
  type <- tolower(type)
  type <- match.arg(type)

  if (type %in% c("od", "origin-destination", "viajes")) {
    return("od")
  } else if (type %in% c("os", "overnight_stays", "pernoctaciones")) {
    return("os")
  } else if (type %in% c("nt", "number_of_trips", "personas")) {
    return("nt")
  }

  # need to add a warning here that the type is not recognized
  return(NULL)
}

#' Remove duplicate values in a semicolon-separated string
#'
#' @description
#' Remove duplicate IDs in a semicolon-separated string in a selected column in a data frame
#' @param column A `character` vector column in a data frame to remove duplicates from.
#'
#' @return A `character` vector with semicolon-separated unique IDs.
#' @keywords internal
spod_unique_separated_ids <- function(column) {
  purrr::map_chr(
    column,
    ~ {
      unique_ids <- unique(stringr::str_split(.x, ";\\s*")[[1]]) # Split by semicolon and remove duplicates
      stringr::str_c(unique_ids, collapse = "; ") # Join them back with semicolons
    }
  )
}

#' Convert dates to ranges
#'
#' This internal helper function reduces a vector of dates to a vector of date ranges to shorten the warning and error messages that mention the valid date ranges.
#' @param dates A `character` vector of dates.
#' @importFrom rlang .data
#' @return A `character` vector of date ranges.
#' @keywords internal
#'
spod_convert_dates_to_ranges <- function(dates) {
  dates <- as.Date(dates) # Ensure dates are in Date format
  ranges <- tibble::tibble(date = dates) |>
    dplyr::arrange(.data$date) |>
    dplyr::mutate(
      diff = c(0, diff(.data$date)), # Calculate differences
      group = cumsum(.data$diff != 1) # Create groups for consecutive ranges
    ) |>
    dplyr::group_by(.data$group) |>
    dplyr::summarise(
      start = dplyr::first(.data$date),
      end = dplyr::last(.data$date),
      .groups = "drop"
    )

  # Create a character vector of ranges
  range_strings <- ranges |>
    dplyr::mutate(range = paste(.data$start, "to", .data$end)) |>
    dplyr::pull(.data$range)

  return(range_strings)
}

#' Get valid dates from the GraphQL API
#' @return A `Date` vector of dates that are valid to request data with \code{spod_quick_get_od()}.
#' @keywords internal
spod_graphql_valid_dates <- function() {
  graphql_endpoint <- getOption("spanishoddata.graphql_api_endpoint")

  # Define the GraphQL query
  graphql_query <- list(
    query = "query {dates_find_available_dates_basic{startDate endDate } }",
    variables = structure(list(), .Names = character(0)) # Empty named list so that it serialises to {} not []
  )

  # Generate signature for the query
  session_token_and_signature <- spod_session_token_and_signature()
  x_session_token <- session_token_and_signature$x_session_token
  x_signature <- session_token_and_signature$x_signature

  # Send the POST request
  req <- httr2::request(graphql_endpoint) |>
    httr2::req_headers(
      "Content-Length" = spod_request_length(graphql_query),
      "Content-Type" = "application/json",
      "User-Agent" = getOption("spanishoddata.user_agent"),
      "X-Session-Token" = x_session_token,
      "X-Signature" = x_signature
    ) |>
    httr2::req_body_json(graphql_query) |>
    httr2::req_progress(type = "down")

  resp <- req |> spod_httr2_req_perform()

  # parse the response
  response_data <- spod_httr2_resp_body_json(resp, simplifyVector = TRUE)
  dates_table <- response_data$data$dates_find_available_dates_basic |>
    dplyr::rename(start_date = "startDate", end_date = "endDate") |>
    dplyr::mutate(
      end_date = dplyr::if_else(
        condition = .data$end_date == "2024-09-31",
        true = "2024-09-30",
        false = .data$end_date
      )
    )

  # Convert start_date and end_date columns to Date class
  dates_table$start_date <- as.Date(dates_table$start_date)
  dates_table$end_date <- as.Date(dates_table$end_date)

  # Generate a single vector of all dates in the intervals
  # Generate all dates for each interval
  date_sequences <- mapply(
    seq.Date,
    from = dates_table$start_date,
    to = dates_table$end_date,
    by = "day"
  )

  # Flatten the list of date sequences into a single vector and remove duplicates
  dates <- as.Date(as.vector(unique(unlist(date_sequences))), origin = "1970-01-01")

  return(dates)
}

#' @importFrom memoise memoise
spod_graphql_valid_dates_memoised <- memoise::memoise(spod_graphql_valid_dates)


spod_session_token_and_signature <- function() {
  # in-memory session token
  session_token <- getOption("spanishoddata.session_token")
  if (is.null(session_token)) {
    session_token <- paste0("rand-", floor(stats::runif(1, 1e6, 1e9)))
    options(spanishoddata.session_token = session_token)
  }

  # get HMAC secret
  hmac_secret <- spod_get_hmac_secret_memoised()

  # compute the raw HMAC-SHA256 bytes
  raw_sig <- digest::hmac(
    key = hmac_secret,
    object = session_token,
    algo = "sha256",
    serialize = FALSE,
    raw = TRUE # <= return raw bytes, not hex
  )

  # base64-encode those raw bytes
  signature_b64 <- openssl::base64_encode(raw_sig)

  x_session_token_and_signature <- list(
    x_session_token = session_token,
    x_signature = signature_b64
  )

  return(x_session_token_and_signature)
}

#' Get the length of the request payload
#' @param graphql_query Character. The GraphQL query.
#' @return Numeric. The length of the request payload.
#' @keywords internal
spod_request_length <- function(graphql_query) {
  payload_json <- jsonlite::toJSON(graphql_query, auto_unbox = TRUE)
  content_length <- nchar(payload_json, type = "bytes")

  return(content_length)
}

# reworked is_installed2 from https://github.com/dataheld/elf/blob/main/R/dependencies.R
#' Checks if a package is installed and *informs* the user if not
#'
#' This is wrapper around [rlang::check_installed];
#' instead of erroring out if the check fails it returns `FALSE`.
#' However, unlike [rlang::is_installed], it emits a message to the user.
#'
#' @inheritParams rlang::check_installed
#' @inheritDotParams rlang::check_installed
#' @keywords internal
spod_assert_package <- function(...) {
  if (rlang::is_installed(...)) {
    return(TRUE)
  }

  withRestarts(
    tryCatch(
      rlang::check_installed(...),
      error = function(cnd) {
        if (inherits(cnd, "rlib_error_package_not_found")) {
          message("The required package is not installed.")
          stop(cnd) # Re-throw the error
        }
      }
    ),
    abort = function(cnd) {
      message("The required package is not installed.")
      stop(cnd) # Re-throw the error
    }
  )

  rlang::is_installed(...)
}

test_that("spod_infer_data_v_from_dates handles ignore_missing_dates", {
  # Mock valid dates
  local_mocked_bindings(
    spod_get_valid_dates = function(ver) {
      if (ver == 1) as.Date("2020-01-01") else as.Date("2022-01-01")
    },
    .package = "spanishoddata"
  )

  # Case 1: All missing (error)
  expect_error(
    spod_infer_data_v_from_dates(as.Date("2030-01-01")),
    "All requested dates are missing"
  )

  # Case 2: Some missing (default error)
  expect_error(
    spod_infer_data_v_from_dates(as.Date(c("2020-01-01", "2030-01-01"))),
    "Some dates do not match"
  )

  # Case 3: Some missing (ignore = TRUE)
  res <- spod_infer_data_v_from_dates(
    as.Date(c("2020-01-01", "2030-01-01")),
    ignore_missing_dates = TRUE
  )
  expect_equal(res, 1)

  # Case 4: No valid remaining (ignore = TRUE) -> NULL
  res_null <- spod_infer_data_v_from_dates(
    as.Date("2030-01-01"),
    ignore_missing_dates = TRUE
  )
  expect_null(res_null)

  # Case 5: helper cached request
  expect_equal(
    spod_infer_data_v_from_dates("cached_v1"),
    1
  )
  expect_equal(
    spod_infer_data_v_from_dates("cached_v2"),
    2
  )
})

test_that("spod_graphql_valid_dates works (mocked)", {
  # Clear memoise cache to ensure we test the request logic
  memoise::forget(spod_get_hmac_secret_memoised)

  # Mock httr2::req_perform to return a valid response object
  local_mocked_bindings(
    req_perform = function(...) {
      structure(
        list(
          method = "GET",
          url = "https://mock",
          status_code = 200,
          headers = list(),
          body = charToRaw(
            '<html><body><script>var import_meta_env = {"HMAC_SECRET": "1234567890"};</script></body></html>'
          )
        ),
        class = "httr2_response"
      )
    },
    .package = "httr2"
  )

  # Mock httr2::resp_body_json
  local_mocked_bindings(
    resp_body_json = function(...) {
      list(
        data = list(
          dates_find_available_dates_basic = data.frame(
            startDate = "2024-01-01",
            endDate = "2024-01-02"
          )
        )
      )
    },
    .package = "httr2"
  )

  withr::local_options(spanishoddata.graphql_api_endpoint = "https://mock")

  dates <- spod_graphql_valid_dates()
  expect_equal(dates, as.Date(c("2024-01-01", "2024-01-02")))
})

test_that("spod_assert_package handles missing package", {
  # We should use a package name that definitely doesn't exist
  pkg_name <- "NonExistentPackageXYZ"

  expect_message(
    expect_error(
      spod_assert_package(pkg_name)
      # We don't check the error message string because it comes from rlang and varies
    ),
    "required package is not installed"
  )
})

test_that("spod_match_data_type handles NULL/Warning", {
  expect_error(spod_match_data_type("invalid_type"), "should be one of")
  expect_error(
    spod_match_data_type_for_local_folders("invalid_type", 1),
    "should be one of"
  )
})

test_that("spod_dates_argument_to_dates_seq processes inputs correctly", {
  # Mock valid dates to avoid network/cache calls
  # v1: 2020-01-01 to 2020-01-10
  # v2: 2022-01-01 to 2022-01-10
  mock_v1_dates <- seq.Date(
    as.Date("2020-01-01"),
    as.Date("2020-01-10"),
    by = "day"
  )
  mock_v2_dates <- seq.Date(
    as.Date("2022-01-01"),
    as.Date("2022-01-10"),
    by = "day"
  )

  # We need to mock spod_get_valid_dates using assignInNamespace since it's used internally
  # However, spod_dates_argument_to_dates_seq calls spod_expand_dates_from_regex which calls spod_get_valid_dates
  # And spod_dates_argument_to_dates_seq calls spod_is_data_version_overlaps which calls spod_get_valid_dates

  # Helper to mock the function
  local_mock_valid_dates <- function(env = parent.frame()) {
    mock_func <- function(ver = NULL) {
      if (ver == 1) {
        return(mock_v1_dates)
      }
      if (ver == 2) {
        return(mock_v2_dates)
      }
      stop("Invalid version")
    }

    # We use mockery stub if possible, but assignInNamespace is robust for internal calls
    # Check if we can use local_mocked_bindings (preferred if exported or visible)
    # Since spod_get_valid_dates is exported, we might be able to use it, but internal calls might bypass it.
    # Let's use assignInNamespace to be safe as per previous experience.

    orig <- get("spod_get_valid_dates", envir = asNamespace("spanishoddata"))
    assignInNamespace("spod_get_valid_dates", mock_func, ns = "spanishoddata")
    withr::defer(
      assignInNamespace("spod_get_valid_dates", orig, ns = "spanishoddata"),
      envir = env
    )
  }

  local_mock_valid_dates()

  # 1. Single Date
  expect_equal(
    spod_dates_argument_to_dates_seq("2020-01-01"),
    as.Date("2020-01-01")
  )
  expect_equal(
    spod_dates_argument_to_dates_seq("20200101"),
    as.Date("2020-01-01")
  )

  # 2. Vector
  dates_vec <- c("2020-01-01", "2020-01-03")
  expect_equal(spod_dates_argument_to_dates_seq(dates_vec), as.Date(dates_vec))

  # 3. Range String (underscore)
  expect_equal(
    spod_dates_argument_to_dates_seq("2020-01-01_2020-01-03"),
    as.Date(c("2020-01-01", "2020-01-02", "2020-01-03"))
  )
  expect_equal(
    spod_dates_argument_to_dates_seq("20200101_20200103"),
    as.Date(c("2020-01-01", "2020-01-02", "2020-01-03"))
  )

  # 4. Named Vector Range
  range_vec <- c(start = "2020-01-01", end = "2020-01-03")
  expect_equal(
    spod_dates_argument_to_dates_seq(range_vec),
    as.Date(c("2020-01-01", "2020-01-02", "2020-01-03"))
  )

  # 5. Regex
  # This relies on spod_expand_dates_from_regex which uses our mocked dates
  # Pattern to match 2020-01-0X
  expect_equal(
    spod_dates_argument_to_dates_seq("^2020010[1-2]$"),
    as.Date(c("2020-01-01", "2020-01-02"))
  )

  # 6. Cached keywords
  expect_equal(spod_dates_argument_to_dates_seq("cached_v1"), "cached_v1")
})

test_that("spod_infer_data_v_from_dates infers versions correctly", {
  # Mock valid dates again
  mock_v1_dates <- seq.Date(
    as.Date("2020-01-01"),
    as.Date("2020-01-10"),
    by = "day"
  )
  mock_v2_dates <- seq.Date(
    as.Date("2022-01-01"),
    as.Date("2022-01-10"),
    by = "day"
  )

  local_mock_valid_dates <- function(env = parent.frame()) {
    mock_func <- function(ver = NULL) {
      if (ver == 1) {
        return(mock_v1_dates)
      }
      if (ver == 2) {
        return(mock_v2_dates)
      }
      stop("Invalid version")
    }
    orig <- get("spod_get_valid_dates", envir = asNamespace("spanishoddata"))
    assignInNamespace("spod_get_valid_dates", mock_func, ns = "spanishoddata")
    withr::defer(
      assignInNamespace("spod_get_valid_dates", orig, ns = "spanishoddata"),
      envir = env
    )
  }
  local_mock_valid_dates()

  # v1 dates
  expect_equal(spod_infer_data_v_from_dates(as.Date("2020-01-01")), 1)
  expect_equal(
    spod_infer_data_v_from_dates(as.Date(c("2020-01-01", "2020-01-02"))),
    1
  )

  # v2 dates
  expect_equal(spod_infer_data_v_from_dates(as.Date("2022-01-01")), 2)

  # Cached keywords
  expect_equal(spod_infer_data_v_from_dates("cached_v1"), 1)
  expect_equal(spod_infer_data_v_from_dates("cached_v2"), 2)

  # Overlap/Mixed -> should error (returns NULL invisible if overlap check fails?)
  # logic: spod_infer calls spod_is_data_version_overlaps which STOPS if overlap.
  expect_error(
    spod_infer_data_v_from_dates(as.Date(c("2020-01-01", "2022-01-01"))),
    "Dates found in both v1 and v2 data"
  )

  # Missing dates
  # logic: stops if missing and ignore_missing_dates = FALSE
  expect_error(
    spod_infer_data_v_from_dates(
      as.Date("2025-01-01"),
      ignore_missing_dates = FALSE
    ),
    "All requested dates are missing"
  )

  # Ignore missing dates
  # Logic: if some valid, infer from them. If none valid, return NULL?
  # 2025-01-01 is missing. 2020-01-01 is v1.
  expect_equal(
    spod_infer_data_v_from_dates(
      as.Date(c("2020-01-01", "2025-01-01")),
      ignore_missing_dates = TRUE
    ),
    1
  )
})

test_that("spod_zone_names_en2es normalizes names", {
  # English to Spanish
  expect_equal(spod_zone_names_en2es("districts"), "distritos")
  expect_equal(spod_zone_names_en2es("muni"), "municipios")
  expect_equal(spod_zone_names_en2es("lua"), "gau")

  # Passthrough
  expect_equal(spod_zone_names_en2es("distritos"), "distritos")
  expect_equal(spod_zone_names_en2es("municipios"), "municipios")

  # Case insensitive
  expect_equal(spod_zone_names_en2es("Districts"), "distritos")
})

#' Get daily trip counts per origin-destionation municipality from 2022 onward
#'
#' @description
#'
#' `r lifecycle::badge("experimental")`
#'
#' **WARNING: this function may stop working at any time, as the API may change**. This function provides a quick way to get daily aggregated (no hourly data) trip counts per origin-destination municipality from v2 data (2022 onward). Compared to \code{\link[=spod_get]{spod_get()}}, which downloads large CSV files, this function downloads the data directly from the GraphQL API. An interactive web map with this data is available at [https://mapas-movilidad.transportes.gob.es/](https://mapas-movilidad.transportes.gob.es/). No data aggregation is performed on your computer (unlike in \code{\link[=spod_get]{spod_get()}}), so you do not need to worry about memory usage and do not have to use a powerful computer with multiple CPU cores just to get this simple data. Only about 1 MB of data is downloaded for a single day. The limitation of this function is that it can only retrieve data for a single day at a time and only with total number of trips and total km travelled. So it is not possible to get any of the extra variables available in the full dataset via \code{\link[=spod_get]{spod_get()}}.
#'
#' For detailed data descriptions, see package vignettes using [`spod_codebook(ver = 1)`][spod_codebook] and [`spod_codebook(ver = 2)`][spod_codebook] and official methodology documents in **References** section.
#'
#' @template references
#'
#' @param date A character or Date object specifying the date for which to retrieve the data. If date is a character, the date must be in "YYYY-MM-DD" or "YYYYMMDD" format.
#' @param min_trips A numeric value specifying the minimum number of journeys per origin-destination pair to retrieve. Defaults to 100 to reduce the amount of data returned. Can be set to 0 to retrieve all data.
#' @param distances A character vector specifying the distances to retrieve. Valid values are "500m-2km", "2-10km", "10-50km", and "50+km". Defaults to `c("500m-2km", "2-10km", "10-50km", "50+km")`. The resulting data will not have number of trips per category of distance. Therefore, if you want to retrieve the number of trips per distance category, you need to make 4 separate calls to this function or use \code{\link[=spod_get]{spod_get()}} instead to get the full data from source CSV files.
#' @param id_origin A character vector specifying the origin municipalities to retrieve. If not provided, all origin municipalities will be included. Valid municipality IDs can be found in the dataset returned by `spod_get_zones(zones = "muni", ver = 2)`.
#' @param id_destination A character vector specifying the target municipalities to retrieve. If not provided, all target municipalities will be included. Valid municipality IDs can be found in the dataset returned by `spod_get_zones(zones = "muni", ver = 2)`.
#' @return A `tibble` containing the flows for the specified date, minimum number of journeys, distances and origin-destination pairs if specified. The columns are:
#' \describe{
#'   \item{date}{The date of the trips.}
#'   \item{id_origin}{The origin municipality ID.}
#'   \item{id_destination}{The target municipality ID.}
#'   \item{n_trips}{The number of trips between the origin and target municipality.}
#'   \item{trips_total_length_km}{The total length of trips in kilometers.}
#' }
#'
#' @importFrom rlang .data
#'
#' @export
#'
#' @examplesIf interactive()
#' \donttest{
#' od_1000 <- spod_quick_get_od(
#'   date = "2022-01-01",
#'   min_trips = 1000
#' )
#' }
spod_quick_get_od <- function(
  date = NA,
  min_trips = 100,
  distances = c("500m-2km", "2-10km", "10-50km", "50+km"),
  id_origin = NA,
  id_destination = NA
) {
  # Validate inputs
  checkmate::assert_integerish(
    min_trips,
    lower = 0,
    null.ok = FALSE,
    max.len = 1
  )

  # Mapping user-friendly distances to GraphQL expected values
  distance_mapping <- c(
    "500m-2km" = "D_05_2",
    "2-10km" = "D_2_10",
    "10-50km" = "D_10_50",
    "50+km" = "D_50"
  )
  graphql_distances <- unname(distance_mapping[distances])
  if (any(is.na(graphql_distances))) {
    stop(
      "Invalid distance value. Allowed values are: ",
      paste(names(distance_mapping), collapse = ", ")
    )
  }

  checkmate::assert_character(id_origin, null.ok = TRUE)
  checkmate::assert_character(id_destination, null.ok = TRUE)

  # Convert the date into YYYYMMDD format
  if (is.character(date)) {
    if (grepl("^\\d{4}-\\d{2}-\\d{2}$", date)) {
      date <- as.Date(date)
    } else if (nchar(date) == 8 && grepl("^\\d{8}$", date)) {
      date <- as.Date(date, format = "%Y%m%d")
    } else {
      stop(
        "Invalid date format. Use 'YYYY-MM-DD', 'YYYYMMDD', or a Date object."
      )
    }
  }

  if (inherits(date, "Date")) {
    date_fmt <- format(date, "%Y%m%d")
  } else {
    stop(
      "Invalid date input. Must be a character in 'YYYY-MM-DD'/'YYYYMMDD' format or a Date object."
    )
  }

  muni_ref <- readRDS(
    system.file("extdata", "muni_v2_ref.rds", package = "spanishoddata")
  )
  validate_muni_ids <- function(ids, muni_ref) {
    if (is.null(ids) || length(ids) == 0 || all(is.na(ids))) {
      return(TRUE)
    }
    invalid <- setdiff(ids, muni_ref$id)
    if (length(invalid) > 0) {
      stop(
        "Invalid municipality IDs: ",
        paste(invalid, collapse = ", "),
        ". Use `spod_get_zones(zones='muni', ver=2)` for valid IDs."
      )
    }
    TRUE
  }
  if (!all(is.na(id_origin))) {
    validate_muni_ids(id_origin, muni_ref)
  }
  if (!all(is.na(id_destination))) {
    validate_muni_ids(id_destination, muni_ref)
  }

  # Check date is within API-supported range
  valid_dates <- spod_graphql_valid_dates_memoised()
  if (!lubridate::ymd(date_fmt) %in% valid_dates) {
    stop(
      "Invalid date. Must be within valid range: ",
      paste(spod_convert_dates_to_ranges(valid_dates), collapse = ", ")
    )
  }

  # build the GraphQL query
  full_query <- paste0(
    "query ($journeyDate:String!,$distances:[JourneysDistancesBasic!],",
    "$originMunicipality:[String!],$destinationMunicipality:[String!],",
    "$minJourneys:Float!){",
    "journeys_municipality_find_by_criteria_basic(criteria:{",
    "journeyDate:$journeyDate,",
    "distances:$distances,",
    "originMunicipality:$originMunicipality,",
    "destinationMunicipality:$destinationMunicipality,",
    "minJourneys:$minJourneys",
    "}){origin destination journeys journeysKm}}"
  )

  vars_list <- list(
    journeyDate = date_fmt,
    distances = graphql_distances,
    minJourneys = min_trips
  )

  if (!all(is.na(id_origin))) {
    vars_list$originMunicipality <- id_origin
  }

  if (!all(is.na(id_destination))) {
    vars_list$destinationMunicipality <- id_destination
  }

  # assemble final payload
  graphql_query <- list(
    query = full_query,
    variables = vars_list
  )

  # Query the API
  od <- spod_query_od_memoised(
    date_fmt = date_fmt,
    graphql_distances = graphql_distances,
    id_origin = id_origin,
    id_destination = id_destination,
    min_trips = min_trips,
    graphql_query = graphql_query
  )

  return(od)
}

#' Internal function to query the GraphQL API for origin-destination data
#' @inheritParams spod_quick_get_od
#' @return A `tibble` containing the flows for the specified date, minimum number of journeys, distances and origin-destination pairs if specified.
#' @keywords internal
spod_query_od_raw <- function(
  date_fmt,
  graphql_distances,
  id_origin,
  id_destination,
  min_trips,
  graphql_query
) {
  # in-memory session token
  session_token_and_signature <- spod_session_token_and_signature()
  x_session_token <- session_token_and_signature$x_session_token
  x_signature <- session_token_and_signature$x_signature

  graphql_endpoint <- getOption("spanishoddata.graphql_api_endpoint")
  req <- httr2::request(graphql_endpoint) |>
    httr2::req_headers(
      "Content-Type" = "application/json",
      "User-Agent" = getOption("spanishoddata.user_agent"),
      "X-Session-Token" = x_session_token,
      "X-Signature" = x_signature
    ) |>
    httr2::req_body_json(graphql_query) |>
    httr2::req_progress(type = "down")
  # req |> httr2::req_dry_run()
  resp <- req |>
    spod_httr2_req_perform() |>
    spod_httr2_resp_body_json(simplifyVector = TRUE)

  # Handle empty data
  if (length(resp$data[[1]]) == 0) {
    stop(
      "No data for ",
      date_fmt,
      ". The server reports the date as valid but returns no records."
    )
  }

  # Tidy and return
  od <- tibble::as_tibble(resp$data[[1]]) |>
    dplyr::select(
      id_origin = "origin",
      id_destination = "destination",
      n_trips = "journeys",
      trips_total_length_km = "journeysKm"
    ) |>
    dplyr::mutate(date = lubridate::ymd(date_fmt)) |>
    dplyr::relocate("date", .before = id_origin)

  return(od)
}

#' Cache the spod_query_od_raw function to avoid repeated requests
#' @importFrom memoise memoise
#' @keywords internal
spod_query_od_memoised <- memoise::memoise(spod_query_od_raw)

#' Get the municipalities geometries
#'
#' @description
#'
#' `r lifecycle::badge("experimental")`
#'
#' This function fetches the municipalities (for now this is the  only option) geometries from the mapas-movilidad website and returns a `sf` object with the municipalities geometries. This is intended for use with the flows data retrieved by the \code{\link[=spod_quick_get_od]{spod_quick_get_od()}} function. An interactive web map with this data is available at [https://mapas-movilidad.transportes.gob.es/](https://mapas-movilidad.transportes.gob.es/). These municipality geometries only include Spanish municipalities (and not the NUTS3 regions in Portugal and France) and do not contain extra columns that you can get with the \code{\link[=spod_get_zones]{spod_get_zones()}} function. The function caches the retrieved geometries in memory of the current R session to reduce the number of requests to the mapas-movilidad website.
#'
#' For detailed zone definitions and methodology, see
#' \insertRef{mitma_methodology_2020_v3}{spanishoddata} for v1 data and
#' \insertRef{mitms_methodology_2022_v8}{spanishoddata} for v2 data.
#'
#' @references
#' \insertRef{mitms_mobility_web}{spanishoddata}
#'
#' @param zones A character string specifying the zones to retrieve. Valid values are "municipalities", "muni", "municip", and "municipios". Defaults to "municipalities".
#' @return A `sf` object with the municipalities geometries to match with the data retrieved with \code{\link[=spod_quick_get_od]{spod_quick_get_od()}}.
#'
#' @export
#' @examplesIf interactive()
#' \donttest{
#' municipalities_sf <- spod_quick_get_zones()
#' }
#'
spod_quick_get_zones <- function(
  zones = "municipalities"
) {
  # Validate inputs
  checkmate::assert_choice(
    zones,
    choices = c(
      "municipalities",
      "muni",
      "municip",
      "municipios"
    )
  )
  zones <- spod_zone_names_en2es(zones)
  if (zones != "municipios") {
    stop("Only municipalities are available for now.")
  }
  municipalities_sf <- spod_fetch_municipalities_json_memoised()
  return(municipalities_sf)
}

#' Cache the municipalities geometries from the mapas-movilidad website
#' @importFrom memoise memoise
#' @keywords internal
spod_fetch_municipalities_json_memoised <- memoise::memoise(
  function() {
    municip_geometries_url <- "https://mapas-movilidad.transportes.gob.es/api/static/data/municipios60.json"
    municipalities_sf <- spod_sf_st_read(municip_geometries_url, quiet = TRUE) |>
      dplyr::rename(id = "ID") |>
      dplyr::mutate(
        population = as.numeric(dplyr::na_if(.data$population, "NA"))
      ) |>
      dplyr::select("id", "name", "population")
    return(municipalities_sf)
  }
)


#' Internal wrappers for httr2 and sf calls to enable mocking
#' @keywords internal
spod_httr2_req_perform <- function(req) {
  httr2::req_perform(req)
}

#' @keywords internal
spod_httr2_resp_body_string <- function(resp) {
  httr2::resp_body_string(resp)
}

#' @keywords internal
spod_httr2_resp_body_json <- function(resp, ...) {
  httr2::resp_body_json(resp, ...)
}

#' @keywords internal
spod_sf_st_read <- function(dsn, ...) {
  sf::st_read(dsn, ...)
}


#' Get the HMAC secret from the mapas-movilidad website
#' @param base_url The base URL of the mapas-movilidad website
#' @return Character vector with the HMAC secret.
#' @keywords internal
spod_get_hmac_secret <- function(
  base_url = "https://mapas-movilidad.transportes.gob.es"
) {
  # Fetch the homepage HTML
  homepage <- httr2::request(base_url) |>
    spod_httr2_req_perform() |>
    spod_httr2_resp_body_string()

  # Parse and find the inline <script> that mentions import_meta_env
  doc <- xml2::read_html(homepage)
  scripts <- xml2::xml_find_all(doc, "//script[not(@src)]")
  inline <- scripts[stringr::str_detect(
    xml2::xml_text(scripts),
    "import_meta_env"
  )]
  if (length(inline) != 1) {
    stop("Could not uniquely locate the import_meta_env script block.")
  }
  txt <- xml2::xml_text(inline)

  # Extract the real HMAC_SECRET from that block
  secret <- stringr::str_match(txt, '"HMAC_SECRET"\\s*:\\s*"([^"]+)"')[, 2]

  if (is.na(secret) || nchar(secret) < 10) {
    stop("Failed to parse HMAC_SECRET from the page.")
  }
  secret
}

#' Cache the HMAC secret to avoid repeated requests
#' @importFrom memoise memoise
#' @keywords internal
spod_get_hmac_secret_memoised <- memoise::memoise(spod_get_hmac_secret)

test_that("spod_quick_get_od fails out of range dates", {
  skip_on_ci()
  skip_on_cran()

  # Mock valid dates to return something that makes "2021-12-31" technically out of range but valid for checking
  # Actually, the error happens because the function calls spod_graphql_valid_dates which does a network call.
  # We should mock that helper to avoid network calls.

  local_mocked_bindings(
    spod_graphql_valid_dates_memoised = function() {
      return(as.Date("2022-01-01"))
    }
  )

  expect_error(
    spod_quick_get_od(
      date = "2021-12-31",
    ),
    ".*Must be within valid range.*"
  )
})


test_that("spod_quick_get_od fails on invalid date format", {
  expect_error(
    spod_quick_get_od(
      date = "202212-31"
    ),
    ".*Invalid date format.*"
  )
})

test_that("spod_quick_get_od fails on incorrect distances", {
  expect_error(
    spod_quick_get_od(
      date = "2022-01-01",
      distances = c("invalid", "0-200")
    ),
    ".*Invalid distance value.*"
  )
})

test_that("spod_quick_get_od fails on negative min_trips", {
  expect_error(
    spod_quick_get_od(
      date = "2022-01-02",
      min_trips = -1
    ),
    ".*Assertion.*failed.*"
  )
})

test_that("spod_quick_get_od fails on invalid municipality IDs", {
  expect_error(
    spod_quick_get_od(
      date = "2022-01-03",
      id_origin = "invalid"
    ),
    ".*Invalid municipality ID.*"
  )

  expect_error(
    spod_quick_get_od(
      date = "2022-01-04",
      id_destination = "invalid"
    ),
    ".*Invalid municipality ID.*"
  )

  expect_error(
    spod_quick_get_od(
      date = "2022-01-05",
      id_origin = "invalid",
      id_destination = "invalid"
    ),
    ".*Invalid municipality ID.*"
  )
})

test_that("spod_quick_get_od fetches data (mocked)", {
  # Mock valid dates
  orig_dates_func <- spod_graphql_valid_dates_memoised
  mock_dates_func <- function() {
    return(as.Date("2022-01-01"))
  }
  assignInNamespace(
    "spod_graphql_valid_dates_memoised",
    mock_dates_func,
    ns = "spanishoddata"
  )
  on.exit(
    assignInNamespace(
      "spod_graphql_valid_dates_memoised",
      orig_dates_func,
      ns = "spanishoddata"
    ),
    add = TRUE
  )

  # Mock query function
  orig_query_func <- spod_query_od_memoised
  mock_query_func <- function(date_fmt, id_origin, ...) {
    # Return dummy tibble matching structure
    dplyr::tibble(
      date = as.Date(date_fmt, format = "%Y%m%d"),
      id_origin = ifelse(all(is.na(id_origin)), "01001", id_origin[1]),
      id_destination = "01002",
      n_trips = 100,
      trips_total_length_km = 500
    )
  }
  assignInNamespace(
    "spod_query_od_memoised",
    mock_query_func,
    ns = "spanishoddata"
  )
  on.exit(
    assignInNamespace(
      "spod_query_od_memoised",
      orig_query_func,
      ns = "spanishoddata"
    ),
    add = TRUE
  )

  # Run test
  # Note: 01001 is a valid municipality code (Alegría-Dulantzi)
  res <- spod_quick_get_od(date = "2022-01-01", id_origin = "01001")

  expect_s3_class(res, "data.frame")
  expect_equal(nrow(res), 1)
  expect_equal(res$n_trips, 100)
  expect_equal(res$id_origin, "01001")
})

test_that("spod_quick_get_zones fetches geometries (mocked)", {
  # Mock geometry fetch
  orig_geo_func <- spod_fetch_municipalities_json_memoised
  mock_geo_func <- function() {
    # Create simple sf object
    geom <- sf::st_sfc(sf::st_point(c(0, 0)))
    sf::st_sf(
      id = "01001",
      name = "Mock Muni",
      population = 1000,
      geometry = geom
    )
  }
  assignInNamespace(
    "spod_fetch_municipalities_json_memoised",
    mock_geo_func,
    ns = "spanishoddata"
  )
  on.exit(
    assignInNamespace(
      "spod_fetch_municipalities_json_memoised",
      orig_geo_func,
      ns = "spanishoddata"
    ),
    add = TRUE
  )

  res <- spod_quick_get_zones(zones = "muni")

  expect_s3_class(res, "sf")
  expect_true("population" %in% names(res))
  expect_equal(res$id[1], "01001")
})

library(spanishoddata)

test_that("spod_get_hmac_secret parses secret from HTML", {
  mock_html <- '
    <html>
    <body>
    <script>
    window.import_meta_env = {"HMAC_SECRET":"this-is-a-test-secret-long-enough"};
    </script>
    </body>
    </html>
  '

  # Mock spod_httr2_req_perform
  m_resp <- structure(
    list(body = charToRaw(mock_html)),
    class = "httr2_response"
  )

  with_mocked_bindings(
    spod_httr2_req_perform = function(...) m_resp,
    spod_httr2_resp_body_string = function(...) mock_html,
    {
      secret <- spod_get_hmac_secret()
      expect_equal(secret, "this-is-a-test-secret-long-enough")
    }
  )
})

test_that("spod_get_hmac_secret fails if script missing", {
  mock_html <- "<html><body></body></html>"
  m_resp <- structure(
    list(body = charToRaw(mock_html)),
    class = "httr2_response"
  )

  with_mocked_bindings(
    spod_httr2_req_perform = function(...) m_resp,
    spod_httr2_resp_body_string = function(...) mock_html,
    {
      expect_error(spod_get_hmac_secret(), "Could not uniquely locate")
    }
  )
})

test_that("spod_query_od_raw handles API response", {
  mock_data <- list(
    data = list(
      journeys_municipality_find_by_criteria_basic = data.frame(
        origin = "01001",
        destination = "01002",
        journeys = 150,
        journeysKm = 300,
        stringsAsFactors = FALSE
      )
    )
  )

  with_mocked_bindings(
    spod_httr2_req_perform = function(...) {
      structure(list(), class = "httr2_response")
    },
    spod_httr2_resp_body_json = function(...) mock_data,
    spod_session_token_and_signature = function(...) {
      list(x_session_token = "token", x_signature = "sig")
    },
    {
      res <- spod_query_od_raw(
        date_fmt = "20220101",
        graphql_distances = "D_05_2",
        id_origin = NA,
        id_destination = NA,
        min_trips = 100,
        graphql_query = list()
      )

      expect_s3_class(res, "tbl_df")
      expect_equal(nrow(res), 1)
      expect_equal(res$n_trips, 150)
      expect_equal(res$date, as.Date("2022-01-01"))
    }
  )
})

test_that("spod_query_od_raw stops on empty data", {
  mock_data <- list(
    data = list(
      journeys_municipality_find_by_criteria_basic = list()
    )
  )

  with_mocked_bindings(
    spod_httr2_req_perform = function(...) {
      structure(list(), class = "httr2_response")
    },
    spod_httr2_resp_body_json = function(...) mock_data,
    spod_session_token_and_signature = function(...) {
      list(x_session_token = "token", x_signature = "sig")
    },
    {
      expect_error(
        spod_query_od_raw(
          date_fmt = "20220101",
          graphql_distances = "D_05_2",
          id_origin = NA,
          id_destination = NA,
          min_trips = 100,
          graphql_query = list()
        ),
        "No data for 20220101"
      )
    }
  )
})

test_that("spod_graphql_valid_dates parses date ranges", {
  mock_data <- list(
    data = list(
      dates_find_available_dates_basic = data.frame(
        startDate = "2022-01-01",
        endDate = "2022-01-02",
        stringsAsFactors = FALSE
      )
    )
  )

  with_mocked_bindings(
    spod_httr2_req_perform = function(...) {
      structure(list(), class = "httr2_response")
    },
    spod_httr2_resp_body_json = function(...) mock_data,
    spod_session_token_and_signature = function(...) {
      list(x_session_token = "token", x_signature = "sig")
    },
    {
      dates <- spod_graphql_valid_dates()
      expect_equal(length(dates), 2)
      expect_true(as.Date("2022-01-01") %in% dates)
      expect_true(as.Date("2022-01-02") %in% dates)
    }
  )
})

test_that("spod_fetch_municipalities_json_memoised logic works", {
  # The anonymous function inside memoise is hard to test directly
  # But we can test the effect by mocking sf::st_read

  geom <- sf::st_sfc(sf::st_point(c(0, 0)))
  mock_sf <- sf::st_sf(
    ID = "01001",
    name = "Test",
    population = "1000", # Note: it's a string in JSON usually, and then converted
    geometry = geom
  )

  # Forget current cache to ensure it runs
  memoise::forget(spod_fetch_municipalities_json_memoised)

  with_mocked_bindings(
    spod_sf_st_read = function(...) mock_sf,
    {
      res <- spod_fetch_municipalities_json_memoised()
      expect_s3_class(res, "sf")
      expect_equal(res$id, "01001")
      expect_equal(res$population, 1000)
    }
  )
})

#' Get available data list
#'
#' @description
#'
#' `r lifecycle::badge("stable")`
#'
#' Get a table with links to available data files for the specified data version. Optionally check (see arguments) the file size and availability of data files previously downloaded into the cache directory specified with SPANISH_OD_DATA_DIR environment variable (set by [spod_set_data_dir()]) or a custom path specified with `data_dir` argument. By default the data is fetched from Amazon S3 bucket where the data is stored. If that fails, the function falls back to downloading an XML file from the Spanish Ministry of Transport website. You can also control this behaviour with `use_s3` argument.
#'
#' For detailed data descriptions, see package vignettes using [`spod_codebook(ver = 1)`][spod_codebook] and [`spod_codebook(ver = 2)`][spod_codebook] and official methodology documents in **References** section.
#'
#' @template references
#'
#' @param ver Integer. Can be 1 or 2. The version of the data to use. v1 spans 2020-2021, v2 covers 2022 and onwards. See more details in codebooks with [spod_codebook()].
#' @param check_local_files Logical. Whether to check if the local files exist and get the file size. Defaults to `FALSE`.
#' @param data_dir The directory where the data is stored. Defaults to the value returned by `spod_get_data_dir()`.
#' @param use_s3 `r lifecycle::badge("experimental")` Logical. If `TRUE`, use Amazon S3 to get available data list, which does not require downloading the XML file and caching it locally, which may be a bit faster. If `FALSE`, use the XML file to get available data list.
#' @param force Logical. If `TRUE`, force re-download of metadata. For Amazon S3 this queries the S3 bucket for the XML file it re-downloads. If `FALSE`, only update the available data list if it is older than 1 day.
#' @inheritParams spod_available_data_v1
#' @inheritParams global_quiet_param
#' @return A tibble with links, release dates of files in the data, dates of data coverage, local paths to files, and the download status.
#' \describe{
#'   \item{target_url}{\code{character}. The URL link to the data file.}
#'   \item{pub_ts}{\code{POSIXct}. The timestamp of when the file was published.}
#'   \item{file_extension}{\code{character}. The file extension of the data file (e.g., 'tar', 'gz').}
#'   \item{data_ym}{\code{Date}. The year and month of the data coverage, if available.}
#'   \item{data_ymd}{\code{Date}. The specific date of the data coverage, if available.}
#' \item{study}{\code{factor}. Study category derived from the URL (e.g., 'basic', 'complete', 'routes').}
#'   \item{type}{\code{factor}. Data type category derived from the URL (e.g., 'number_of_trips', 'origin-destination', 'overnight_stays', 'data_quality', 'metadata').}
#'   \item{period}{\code{factor}. Temporal granularity category derived from the URL (e.g., 'day', 'month').}
#'   \item{zones}{\code{factor}. Geographic zone classification derived from the URL (e.g., 'districts', 'municipalities', 'large_urban_areas').}
#'   \item{local_path}{\code{character}. The local file path where the data is (or going to be) stored.}
#'   \item{downloaded}{\code{logical}. Indicator of whether the data file has been downloaded locally. This is only available if `check_local_files` is `TRUE`.}
#' }
#' @export
#' @examplesIf interactive()
#' \donttest{
#'
#' # Set data dir for file downloads
#' spod_set_data_dir(tempdir())
#'
#' # Get available data list for v1 (2020-2021) data
#' spod_available_data(ver = 1)
#'
#' # Get available data list for v2 (2022 onwards) data
#' spod_available_data(ver = 2)
#'
#' # Get available data list for v2 (2022 onwards) data
#' # while also checking for local files that are already downloaded
#' spod_available_data(ver = 2, check_local_files = TRUE)
#' }
#'
spod_available_data <- function(
  ver = 2,
  check_local_files = FALSE,
  quiet = FALSE,
  data_dir = spod_get_data_dir(),
  use_s3 = TRUE,
  force = FALSE
) {
  # Validate input
  checkmate::assertIntegerish(ver, max.len = 1)
  if (!ver %in% c(1, 2)) {
    stop(
      "Invalid version number. Must be 1 (for v1 2020-2021 data) or 2 (for v2 2022 onwards)."
    )
  }
  checkmate::assert_flag(check_local_files)
  checkmate::assert_flag(quiet)
  checkmate::assert_directory_exists(data_dir, access = "rw")

  if (ver == 1) {
    available_data <- spod_available_data_v1(
      data_dir = data_dir,
      check_local_files = check_local_files,
      quiet = quiet,
      use_s3 = use_s3,
      force = force
    )
  } else if (ver == 2) {
    available_data <- spod_available_data_v2(
      data_dir = data_dir,
      check_local_files = check_local_files,
      quiet = quiet,
      use_s3 = use_s3,
      force = force
    )
  }

  return(available_data)
}

#' Get latest file list from the XML for MITMA open mobility data v1 (2020-2021)
#'
#' @param data_dir The directory where the data is stored. Defaults to the value returned by `spod_get_data_dir()`.
#' @param xml_url The URL of the XML file to download. Defaults to "https://opendata-movilidad.mitma.es/RSS.xml".
#'
#' @return The path to the downloaded XML file.
#' @keywords internal
spod_get_latest_v1_file_list <- function(
  data_dir = spod_get_data_dir(),
  xml_url = "https://opendata-movilidad.mitma.es/RSS.xml",
  quiet = FALSE
) {
  if (!dir.exists(data_dir)) {
    fs::dir_create(data_dir)
  }

  current_date <- format(Sys.Date(), format = "%Y-%m-%d")
  current_filename <- fs::path(
    data_dir,
    spod_subfolder_metadata_cache(),
    glue::glue("data_links_v1_{current_date}.xml")
  )

  # ensure dir exists
  if (!dir.exists(dirname(current_filename))) {
    fs::dir_create(dirname(current_filename), recurse = TRUE)
  }

  if (!quiet) {
    message("Saving the file to: ", current_filename)
  }
  tryCatch(
    {
      utils::download.file(xml_url, current_filename, mode = "wb", quiet = quiet)
    },
    error = function(e) {
      if (!quiet) message("Failed to download the XML file: ", e$message)
    },
    warning = function(w) {
      if (!quiet) message("Warning during XML download: ", w$message)
    }
  )
  # disable curl::multi_download() for now
  # xml_requested <- curl::multi_download(
  #   urls = xml_url,
  #   destfiles = current_filename
  # )
  if (!fs::file_exists(current_filename)) {
    if (!quiet) message("Graceful exit: XML file could not be retrieved.")
    return(NULL)
  }
  return(current_filename)
}

#' Get the available v1 data list
#'
#' This function provides a table of the available data list of MITMA v1 (2020-2021), both remote and local.
#'
#' @inheritParams spod_available_data
#' @inheritParams global_quiet_param
#' @inherit spod_available_data return
#' @importFrom rlang .data
#' @keywords internal
spod_available_data_v1 <- function(
  data_dir = spod_get_data_dir(),
  # check_local_files (below) is FALSE by default to avoid excessive filesystem access, perhaps should be TRUE. Download functions use it to load the xml file, but we probably do not want the script to check all local cache directories every time we run a get data function. Perhaps it is better to offload this check to a separate function and have a csv file or some other way to keep track of the files that were downloaded and cached. An output of curl::multi_download() could be used for this purpose.
  check_local_files = FALSE,
  use_s3 = TRUE,
  force = FALSE,
  quiet = FALSE
) {
  metadata_folder <- fs::path(data_dir, spod_subfolder_metadata_cache())
  if (!dir.exists(metadata_folder)) {
    fs::dir_create(metadata_folder)
  }

  s3_successful <- FALSE

  if (use_s3) {
    files_table <- tryCatch(
      {
        files_table_s3 <- spod_available_data_s3(
          ver = 1,
          force = force,
          quiet = quiet
        )
        s3_successful <- TRUE
        files_table_s3
      },
      error = function(e) {
        if (!quiet) {
          message(
            "S3 fetch failed (",
            e$message,
            "); falling back to XML sequence."
          )
        }
        read_data_links_xml(
          metadata_folder = metadata_folder,
          data_dir = data_dir,
          force = force,
          quiet = quiet,
          latest_file_function = spod_get_latest_v1_file_list,
          ver = 1
        )
      }
    )
  } else {
    files_table <- read_data_links_xml(
      metadata_folder = metadata_folder,
      data_dir = data_dir,
      force = force,
      quiet = quiet,
      latest_file_function = spod_get_latest_v1_file_list,
      ver = 1
    )
  }

  files_table$file_extension <- tools::file_ext(files_table$target_url)
  files_table <- files_table[files_table$file_extension != "", ]

  files_table$data_ym <- lubridate::ym(stringr::str_extract(
    files_table$target_url,
    "[0-9]{4}-[0-9]{2}"
  ))
  files_table$data_ymd <- lubridate::ymd(stringr::str_extract(
    files_table$target_url,
    "[0-9]{8}"
  ))
  # order by pub_ts
  files_table <- files_table[order(files_table$pub_ts, decreasing = TRUE), ]
  files_table$local_path <- fs::path(
    data_dir,
    spod_subfolder_raw_data_cache(ver = 1),
    stringr::str_replace(
      files_table$target_url,
      ".*mitma.es/",
      ""
    )
  )

  files_table$local_path <- stringr::str_replace_all(
    files_table$local_path,
    "\\/\\/\\/|\\/\\/",
    "/"
  )

  # change path for daily data files to be in hive-style format
  files_table$local_path <- gsub(
    "([0-9]{4})-([0-9]{2})\\/[0-9]{6}([0-9]{2})_",
    "year=\\1\\/month=\\2\\/day=\\3\\/",
    files_table$local_path
  )

  # fix paths for files that are in '0000-referencia' folder
  files_table$local_path <- gsub(
    "0000-referencia\\/([0-9]{4})([0-9]{2})([0-9]{2})_",
    "year=\\1\\/month=\\2\\/day=\\3\\/",
    files_table$local_path
  )

  # replace 2 digit month with 1 digit month
  files_table$local_path <- gsub(
    "month=0([1-9])",
    "month=\\1",
    files_table$local_path
  )

  # replace 2 digit day with 1 digit day
  files_table$local_path <- gsub(
    "day=0([1-9])",
    "day=\\1",
    files_table$local_path
  )

  # change txt.gz to csv.gz
  files_table$local_path <- gsub(
    "\\.txt\\.gz",
    "\\.csv\\.gz",
    files_table$local_path
  )

  # replace all municipal data download links with districts links
  # this is to address the bugs described in detail in:
  # https://www.ekotov.pro/mitma-data-issues/issues/011-v1-tpp-mismatch-zone-ids-in-table-and-spatial-data.html
  # https://www.ekotov.pro/mitma-data-issues/issues/012-v1-tpp-district-files-in-municipality-folders.html
  # the decision was to use distrcit data and aggregate it to replicate municipal data
  files_table$target_url <- gsub(
    "mitma-municipios",
    "mitma-distritos",
    files_table$target_url
  )
  files_table$target_url <- gsub(
    "mitma_municipio",
    "mitma_distrito",
    files_table$target_url
  )

  files_table <- files_table |>
    dplyr::mutate(
      study = factor(
        dplyr::case_when(
          grepl("maestra", .data$target_url) ~ "basic",
          TRUE ~ NA_character_
        ),
        levels = c("basic")
      ),

      type = factor(
        dplyr::case_when(
          grepl("maestra2", .data$target_url) ~ "number_of_trips",
          grepl("maestra1", .data$target_url) ~ "origin-destination",
          grepl("RSS\\.xml", .data$target_url) ~ "metadata",
          grepl("zonificacion", .data$target_url) ~ "zones",
          grepl("relacion", .data$target_url) ~ "relations",
          grepl("index\\.html", .data$target_url) ~ "index",
          grepl("\\.pdf", .data$target_url) ~ "documentation",
          TRUE ~ NA_character_
        ),
        levels = c(
          "number_of_trips",
          "origin-destination",
          "metadata",
          "zones",
          "relations",
          "index",
          "documentation"
        )
      ),

      period = factor(
        dplyr::case_when(
          grepl("ficheros-diarios", .data$target_url) ~ "day",
          grepl("meses-completos|mensual", .data$target_url) ~ "month",
          TRUE ~ NA_character_
        ),
        levels = c("day", "month")
      ),

      zones = factor(
        dplyr::case_when(
          grepl("distrito", .data$target_url) ~ "districts",
          grepl("municipio", .data$target_url) ~ "municipalities",
          TRUE ~ NA_character_
        ),
        levels = c("districts", "municipalities")
      )
    )

  # add known file sizes from cached data
  if (s3_successful) {
    # replace remote file sizes for v1
    replacement_file_sizes_distr <- files_table |>
      dplyr::filter(grepl("mitma-distr", .data$local_path)) |>
      dplyr::select("target_url", "file_size_bytes")
    replaced_file_sizes_municip <- files_table |>
      dplyr::filter(grepl("mitma-municip", .data$local_path)) |>
      dplyr::select(-"file_size_bytes") |>
      dplyr::left_join(replacement_file_sizes_distr, by = "target_url")
    files_table_replaced_file_sizes <- files_table |>
      dplyr::filter(!grepl("mitma-municip", .data$local_path)) |>
      dplyr::bind_rows(replaced_file_sizes_municip) |>
      dplyr::arrange(dplyr::desc(.data$pub_ts))
    files_table <- files_table_replaced_file_sizes

    files_table$remote_file_size_mb <- round(
      files_table$file_size_bytes / 1024^2,
      2
    )

    file_sizes <- readRDS(
      system.file(
        "extdata",
        "available_data_v1.rds",
        package = "spanishoddata"
      )
    )
    files_table <- dplyr::left_join(
      files_table |> dplyr::select(-"file_size_bytes"),
      file_sizes |>
        dplyr::select(
          "target_url",
          "etag",
          "true_etag",
          file_size_bytes = "true_remote_file_size_bytes"
        ),
      by = c("target_url", "etag")
    ) |>
      dplyr::relocate("file_size_bytes", .after = "pub_ts") |>
      dplyr::mutate(
        etag = dplyr::if_else(
          condition = !is.na(.data$true_etag),
          true = .data$true_etag,
          false = .data$etag
        )
      ) |>
      dplyr::select(-"true_etag")
  } else {
    file_sizes <- readRDS(
      system.file(
        "extdata",
        "available_data_v1.rds",
        package = "spanishoddata"
      )
    )
    files_table <- dplyr::left_join(
      files_table |> dplyr::select(-"file_size_bytes"),
      file_sizes |>
        dplyr::select(
          "target_url",
          "etag",
          "true_etag",
          file_size_bytes = "true_remote_file_size_bytes"
        ),
      by = c("target_url", "etag")
    ) |>
      dplyr::relocate("file_size_bytes", .after = "pub_ts") |>
      dplyr::mutate(
        etag = dplyr::if_else(
          condition = !is.na(.data$true_etag),
          true = .data$true_etag,
          false = .data$etag
        )
      ) |>
      dplyr::select(-"true_etag")

    files_table$remote_file_size_mb <- round(
      files_table$file_size_bytes / 1024^2,
      2
    )

    # if there are files with missing sizes, impute them
    if (any(is.na(files_table$remote_file_size_mb))) {
      # impute uknown file sizes
      # primitive file categorisation
      # Extract file category from the target URL
      files_table <- files_table |>
        dplyr::mutate(
          file_category = stringr::str_extract(
            .data$target_url,
            "\\/maestra(\\d)-mitma-(distritos|municipios)\\/(ficheros-diarios|meses-completos)\\/"
          )
        )

      # Set other category for non-categorized files
      files_table$file_category[is.na(files_table$file_category)] <- "other"

      # Calculate mean file sizes by category
      size_by_file_category <- files_table |>
        dplyr::group_by(.data$file_category) |>
        dplyr::summarise(
          mean_file_size_mb = mean(.data$remote_file_size_mb, na.rm = TRUE)
        )

      # Impute missing file sizes
      files_table <- files_table |>
        dplyr::left_join(size_by_file_category, by = "file_category")
      files_table$remote_file_size_mb[is.na(
        files_table$remote_file_size_mb
      )] <- mean(files_table$mean_file_size_mb)

      # Update file_size_bytes if NA based on remote_file_size_mb
      files_table$file_size_bytes <- dplyr::if_else(
        condition = is.na(files_table$file_size_bytes),
        true = round(files_table$remote_file_size_mb * 1024^2),
        false = files_table$file_size_bytes
      )

      # Clean up temporary columns
      files_table <- files_table |>
        dplyr::select(-"mean_file_size_mb", -"file_category")
    }
  }

  # check file sizes
  if (check_local_files == TRUE) {
    files_table <- files_table |>
      dplyr::mutate(
        local_file_size = fs::file_size(.data$local_path)
      ) |>
      dplyr::mutate(
        downloaded = dplyr::if_else(
          condition = is.na(.data$local_file_size),
          true = FALSE,
          false = TRUE
        )
      )
  }

  return(files_table)
}

#' Get latest file list from the XML for MITMA open mobility data v2 (2022 onwards)
#'
#' @param data_dir The directory where the data is stored. Defaults to the value returned by `spod_get_data_dir()`.
#' @param xml_url The URL of the XML file to download. Defaults to "https://movilidad-opendata.mitma.es/RSS.xml".
#'
#' @return The path to the downloaded XML file.
#' @keywords internal
spod_get_latest_v2_file_list <- function(
  data_dir = spod_get_data_dir(),
  xml_url = "https://movilidad-opendata.mitma.es/RSS.xml",
  quiet = FALSE
) {
  if (!dir.exists(data_dir)) {
    fs::dir_create(data_dir)
  }

  current_date <- format(Sys.Date(), format = "%Y-%m-%d")
  current_filename <- fs::path(
    data_dir,
    spod_subfolder_metadata_cache(),
    glue::glue("data_links_v2_{current_date}.xml")
  )

  # ensure dir exists
  if (!dir.exists(dirname(current_filename))) {
    fs::dir_create(dirname(current_filename), recurse = TRUE)
  }

  if (!quiet) {
    message("Saving the file to: ", current_filename)
  }
  tryCatch(
    {
      utils::download.file(xml_url, current_filename, mode = "wb", quiet = quiet)
    },
    error = function(e) {
      if (!quiet) message("Failed to download the XML file: ", e$message)
    },
    warning = function(w) {
      if (!quiet) message("Warning during XML download: ", w$message)
    }
  )
  # disable curl::multi_download() for now
  # xml_requested <- curl::multi_download(
  #   urls = xml_url,
  #   destfiles = current_filename
  # )
  if (!fs::file_exists(current_filename)) {
    if (!quiet) message("Graceful exit: XML file could not be retrieved.")
    return(NULL)
  }

  return(current_filename)
}

#' Get the data dictionary
#'
#' This function retrieves the data dictionary for the specified data directory.
#'
#' @param data_dir The directory where the data is stored. Defaults to the value returned by `spod_get_data_dir()`.
#' @inheritParams spod_available_data_v1
#' @inheritParams global_quiet_param
#' @inherit spod_available_data return
#' @importFrom rlang .data
#' @keywords internal
spod_available_data_v2 <- function(
  data_dir = spod_get_data_dir(),
  check_local_files = FALSE,
  use_s3 = TRUE,
  force = FALSE,
  quiet = FALSE
) {
  metadata_folder <- fs::path(data_dir, spod_subfolder_metadata_cache())
  if (!dir.exists(metadata_folder)) {
    fs::dir_create(metadata_folder)
  }

  s3_successful <- FALSE

  if (use_s3) {
    files_table <- tryCatch(
      {
        files_table_s3 <- spod_available_data_s3(
          ver = 2,
          force = force,
          quiet = quiet
        )
        s3_successful <- TRUE
        files_table_s3
      },
      error = function(e) {
        if (!quiet) {
          message(
            "S3 fetch failed (",
            e$message,
            "); falling back to XML sequence."
          )
        }
        read_data_links_memoised(
          metadata_folder = metadata_folder,
          data_dir = data_dir,
          force = force,
          quiet = quiet,
          latest_file_function = spod_get_latest_v2_file_list,
          ver = 2
        )
      }
    )
  } else {
    files_table <- read_data_links_memoised(
      metadata_folder = metadata_folder,
      data_dir = data_dir,
      force = force,
      quiet = quiet,
      latest_file_function = spod_get_latest_v2_file_list,
      ver = 2
    )
  }

  files_table$file_extension <- tools::file_ext(files_table$target_url)
  files_table <- files_table[files_table$file_extension != "", ]

  files_table$data_ym <- lubridate::ym(stringr::str_extract(
    files_table$target_url,
    "[0-9]{4}-[0-9]{2}"
  ))
  files_table$data_ymd <- lubridate::ymd(stringr::str_extract(
    files_table$target_url,
    "[0-9]{8}"
  ))
  # order by pub_ts
  files_table <- files_table[order(files_table$pub_ts, decreasing = TRUE), ]
  files_table$local_path <- fs::path(
    data_dir,
    spod_subfolder_raw_data_cache(ver = 2),
    stringr::str_replace(
      files_table$target_url,
      ".*mitma.es/",
      ""
    )
  )
  files_table$local_path <- stringr::str_replace_all(
    files_table$local_path,
    "\\/\\/\\/|\\/\\/",
    "/"
  )

  # change path for daily data files to be in hive-style format
  # TODO: check if this is needed for estudios completo and rutas
  files_table$local_path <- gsub(
    "([0-9]{4})-([0-9]{2})\\/[0-9]{6}([0-9]{2})_",
    "year=\\1\\/month=\\2\\/day=\\3\\/",
    files_table$local_path
  )

  # replace 2 digit month with 1 digit month
  files_table$local_path <- gsub(
    "month=0([1-9])",
    "month=\\1",
    files_table$local_path
  )

  # replace 2 digit day with 1 digit day
  files_table$local_path <- gsub(
    "day=0([1-9])",
    "day=\\1",
    files_table$local_path
  )

  # lowercase GAU to avoid problems with case-sensitive matching
  files_table$local_path <- gsub("GAU", "gau", files_table$local_path)

  files_table <- files_table |>
    dplyr::mutate(
      study = factor(
        dplyr::case_when(
          grepl("estudios_basicos", .data$target_url) ~ "basic",
          grepl("estudios_completos", .data$target_url) ~ "complete",
          grepl("rutas", .data$target_url) ~ "routes",
          TRUE ~ NA_character_
        ),
        levels = c("basic", "complete", "routes")
      ),

      type = factor(
        dplyr::case_when(
          grepl("personas", .data$target_url) ~ "number_of_trips",
          grepl("viajes", .data$target_url) ~ "origin-destination",
          grepl("pernoctaciones", .data$target_url) ~ "overnight_stays",
          grepl("calidad", .data$target_url) ~ "data_quality",
          grepl("RSS\\.xml", .data$target_url) ~ "metadata",
          TRUE ~ NA_character_
        ),
        levels = c(
          "origin-destination",
          "number_of_trips",
          "overnight_stays",
          "data_quality",
          "metadata"
        )
      ),

      period = factor(
        dplyr::case_when(
          grepl("ficheros-diarios", .data$target_url) ~ "day",
          grepl("meses-completos|mensual", .data$target_url) ~ "month",
          TRUE ~ NA_character_
        ),
        levels = c("day", "month")
      ),

      zones = factor(
        dplyr::case_when(
          grepl("distritos", .data$target_url) ~ "districts",
          grepl("municipios", .data$target_url) ~ "municipalities",
          grepl("GAU", .data$target_url) ~ "large_urban_areas",
          TRUE ~ NA_character_
        ),
        levels = c("districts", "municipalities", "large_urban_areas")
      )
    )

  # add known file sizes from cached data
  if (s3_successful) {
    files_table$remote_file_size_mb <- round(
      files_table$file_size_bytes / 1024^2,
      2
    )
  } else {
    file_sizes <- readr::read_csv(
      system.file(
        "extdata",
        "url_file_sizes_v2.txt.gz",
        package = "spanishoddata"
      ),
      show_col_types = FALSE
    )
    files_table <- dplyr::left_join(files_table, file_sizes, by = "target_url")

    # if there are files with missing sizes, impute them
    if (any(is.na(files_table$remote_file_size_mb))) {
      # impute uknown file sizes
      # primitive file categorisation
      files_table <- files_table |>
        dplyr::mutate(
          cleaned_url = stringr::str_remove_all(
            .data$target_url,
            "/[0-9]{4}[-_][0-9]{2}[-_][0-9]{2}|/[0-9]{6,8}"
          ) |>
            stringr::str_remove("/[^/]+$"),
          file_category = dplyr::case_when(
            stringr::str_detect(.data$cleaned_url, "calidad") ~ "quality",
            stringr::str_detect(.data$cleaned_url, "rutas") ~ "routes",
            stringr::str_detect(.data$cleaned_url, "estudios_basicos") ~
              paste0(
                "basic_studies_",
                dplyr::case_when(
                  stringr::str_detect(.data$cleaned_url, "por-distritos") ~
                    "district_",
                  stringr::str_detect(.data$cleaned_url, "por-municipios") ~
                    "municipal_",
                  stringr::str_detect(.data$cleaned_url, "por-GAU") ~ "GAU_",
                  TRUE ~ "unknown_"
                ),
                dplyr::case_when(
                  stringr::str_detect(.data$cleaned_url, "viajes") ~ "trips_",
                  stringr::str_detect(.data$cleaned_url, "personas") ~
                    "people_",
                  stringr::str_detect(.data$cleaned_url, "pernoctaciones") ~
                    "overnight_",
                  TRUE ~ "unknown_"
                ),
                ifelse(
                  stringr::str_detect(.data$cleaned_url, "ficheros-diarios"),
                  "daily",
                  "monthly"
                )
              ),
            TRUE ~ "other"
          )
        ) |>
        dplyr::select(-"cleaned_url")

      # Calculate mean file sizes by category
      size_by_file_category <- files_table |>
        dplyr::group_by(.data$file_category) |>
        dplyr::summarise(
          mean_file_size_mb = mean(.data$remote_file_size_mb, na.rm = TRUE)
        )

      # Impute missing file sizes
      files_table <- dplyr::left_join(
        files_table,
        size_by_file_category,
        by = "file_category"
      )

      if (!"file_size_bytes" %in% names(files_table)) {
        files_table$file_size_bytes <- NA_real_
      }

      files_table <- files_table |>
        dplyr::mutate(
          size_imputed = ifelse(is.na(.data$remote_file_size_mb), TRUE, FALSE)
        )
      if (
        length(files_table$remote_file_size_mb[is.na(
          files_table$remote_file_size_mb
        )]) >
          0
      ) {
        files_table <- files_table |>
          dplyr::mutate(
            remote_file_size_mb = ifelse(
              is.na(.data$remote_file_size_mb),
              .data$mean_file_size_mb,
              .data$remote_file_size_mb
            ),
            # Update file_size_bytes if NA based on remote_file_size_mb
            file_size_bytes = dplyr::if_else(
              condition = is.na(.data$file_size_bytes),
              true = round(.data$remote_file_size_mb * 1024^2),
              false = .data$file_size_bytes
            )
          )
      }
      files_table$mean_file_size_mb <- NULL
      files_table$file_category <- NULL
    } else {
      files_table$size_imputed <- FALSE
    }
  }

  # check file sizes
  if (check_local_files == TRUE) {
    files_table <- files_table |>
      dplyr::mutate(
        local_file_size = fs::file_size(.data$local_path)
      ) |>
      dplyr::mutate(
        downloaded = dplyr::if_else(
          condition = is.na(.data$local_file_size),
          true = FALSE,
          false = TRUE
        )
      )
  }

  return(files_table)
}

read_data_links_xml <- function(
  metadata_folder,
  data_dir,
  force = FALSE,
  quiet = FALSE,
  latest_file_function,
  ver
) {
  xml_files_list <- fs::dir_ls(
    metadata_folder,
    type = "file",
    regexp = glue::glue("data_links_v{ver}")
  ) |>
    sort()
  latest_file <- utils::tail(xml_files_list, 1)

  needs_update <- isTRUE(force) ||
    length(xml_files_list) == 0 ||
    as.Date(
      stringr::str_extract(latest_file, "\\d{4}-\\d{2}-\\d{2}")
    ) <
      Sys.Date()

  if (needs_update) {
    if (!quiet) {
      message("Fetching latest data links xml")
    }
    latest_data_links_xml_path <- latest_file_function(
      data_dir = data_dir,
      quiet = quiet
    )
    if (is.null(latest_data_links_xml_path)) {
      return(tibble::tibble(
        target_url = character(0),
        pub_ts = as.POSIXct(numeric(0), origin = "1970-01-01", tz = "UTC")
      ))
    }
  } else {
    if (!quiet) {
      message("Using existing data links xml: ", latest_file)
    }
    latest_data_links_xml_path <- latest_file
  }

  x_xml <- xml2::read_xml(latest_data_links_xml_path)
  files_table <- tibble::tibble(
    target_url = xml2::xml_find_all(x_xml, "//link") |> xml2::xml_text(),
    pub_date = xml2::xml_find_all(x_xml, "//pubDate") |> xml2::xml_text()
  )
  files_table$pub_ts <- lubridate::dmy_hms(files_table$pub_date)
  files_table$pub_date <- NULL

  files_table
}

read_data_links_memoised <- memoise::memoise(read_data_links_xml)

# skip("Requires network access - skipped to prevent downloading RSS.xml")

test_that("spod_available_data falls back to XML on S3 error (v1)", {
  test_dir <- setup_test_data_dir()
  withr::defer(unlink(test_dir, recursive = TRUE))
  withr::local_envvar(c("SPANISH_OD_DATA_DIR" = test_dir))

  # Get a valid URL and size from the real package data
  real_data <- readRDS(system.file(
    "extdata",
    "available_data_v1.rds",
    package = "spanishoddata"
  ))
  target_entry <- real_data[1, ]

  # Mock S3 failure
  local_mocked_bindings(
    spod_available_data_s3 = function(...) stop("S3 bucket unavailable"),
    read_data_links_xml = function(...) {
      tibble::tibble(
        target_url = target_entry$target_url,
        pub_ts = as.POSIXct("2020-02-14", tz = "UTC"),
        file_size_bytes = 99999, # This should be ignored in favor of RDS
        etag = target_entry$etag
      )
    }
  )

  # Should warn about S3 failure or just message?
  msgs <- capture_messages(
    capture.output(
      res <- spod_available_data(ver = 1, quiet = FALSE),
      file = NULL
    )
  )
  expect_match(msgs, "S3 fetch failed", all = FALSE)

  expect_true(nrow(res) == 1)
  # Check that we got the size from RDS, not the 99999 from XML
  expected_mb <- round(target_entry$true_remote_file_size_bytes / 1024^2, 2)
  expect_equal(res$remote_file_size_mb, expected_mb)
})

test_that("spod_available_data v1 handles file size imputation", {
  test_dir <- setup_test_data_dir()
  withr::defer(unlink(test_dir, recursive = TRUE))
  withr::local_envvar(c("SPANISH_OD_DATA_DIR" = test_dir))

  # Mock S3 failure + XML return with missing file size
  local_mocked_bindings(
    spod_available_data_s3 = function(...) stop("S3 fail"),
    read_data_links_xml = function(...) {
      tibble::tibble(
        target_url = "https://opendata-movilidad.mitma.es/maestra1-mitma-distritos/ficheros-diarios/2020-02/20200215_maestra1_mitma_distrito.txt.gz",
        pub_ts = as.POSIXct("2020-02-15", tz = "UTC"),
        file_size_bytes = NA_real_,
        etag = "def"
      )
    }
  )

  # Note: The code calls readRDS(system.file(...)) to get known sizes.
  # We use a URL that is likely not in the bundled RDS, so left_join results in NA size.
  # This triggers the imputation logic.

  msgs <- capture_messages(
    capture.output(
      res <- spod_available_data(ver = 1, quiet = FALSE),
      file = NULL
    )
  )
  expect_match(msgs, "S3 fetch failed", all = FALSE)

  # Since we have only 1 file and it has NA size, and imputation takes mean of category.
  # But if all are NA, mean is NA.
  # Code says: files_table$remote_file_size_mb[is.na(...)] <- mean(...)
  # If mean is NaN (0/0), it assigns NaN.
  # Let's see if it crashes.
  expect_true(nrow(res) == 1)
})

test_that("check_local_files = TRUE works", {
  test_dir <- setup_test_data_dir()
  withr::defer(unlink(test_dir, recursive = TRUE))
  withr::local_envvar(c("SPANISH_OD_DATA_DIR" = test_dir))

  # Mock to prevent real RSS.xml download
  local_mocked_bindings(
    spod_get_latest_v2_file_list = function(...) NULL,
    spod_get_latest_v1_file_list = function(...) NULL
  )

  res <- spod_available_data(ver = 2, check_local_files = TRUE, quiet = TRUE)
  expect_true("downloaded" %in% names(res))
  expect_true("local_file_size" %in% names(res))
})

test_that("spod_available_data rejects invalid version number", {
  test_dir <- withr::local_tempdir()

  expect_error(
    spod_available_data(ver = 3, data_dir = test_dir),
    "Invalid version number.*Must be 1.*or 2"
  )

  expect_error(
    spod_available_data(ver = 0, data_dir = test_dir),
    "Invalid version number.*Must be 1.*or 2"
  )
})

test_that("spod_get_latest_v1_file_list handles download failure", {
  test_dir <- withr::local_tempdir()

  # Mock download.file to fail
  testthat::local_mocked_bindings(
    download.file = function(...) stop("Network error"),
    .package = "utils"
  )

  expect_null(
    spod_get_latest_v1_file_list(data_dir = test_dir, quiet = TRUE)
  )
})

test_that("spod_get_latest_v2_file_list handles download failure", {
  test_dir <- withr::local_tempdir()

  # Mock download.file to fail
  testthat::local_mocked_bindings(
    download.file = function(...) stop("Network error"),
    .package = "utils"
  )

  expect_null(
    spod_get_latest_v2_file_list(data_dir = test_dir, quiet = TRUE)
  )
})

test_that("spod_available_data_v2 handles file size imputation with categories", {
  test_dir <- setup_test_data_dir()
  withr::defer(unlink(test_dir, recursive = TRUE))
  withr::local_envvar(c("SPANISH_OD_DATA_DIR" = test_dir))

  # Mock S3 failure and XML with missing sizes
  testthat::local_mocked_bindings(
    spod_available_data_s3 = function(...) stop("S3 unavailable"),
    read_data_links_memoised = function(...) {
      tibble::tibble(
        target_url = c(
          "https://movilidad-opendata.mitma.es/estudios_basicos/por-distritos/viajes/ficheros-diarios/2022-01/20220101_personas_distritos.csv.gz",
          "https://movilidad-opendata.mitma.es/estudios_basicos/por-distritos/viajes/ficheros-diarios/2022-01/20220102_personas_distritos.csv.gz"
        ),
        pub_ts = as.POSIXct(c("2022-01-01", "2022-01-02"), tz = "UTC"),
        file_size_bytes = c(1000000, NA_real_),
        etag = c("abc", "def")
      )
    }
  )

  msgs <- capture_messages(
    capture.output(
      res <- spod_available_data_v2(
        data_dir = test_dir,
        quiet = FALSE,
        use_s3 = TRUE
      ),
      file = NULL
    )
  )
  expect_match(msgs, "S3 fetch failed", all = FALSE)

  # Check imputation happened
  expect_true(nrow(res) == 2)
  expect_true("size_imputed" %in% names(res))
})

test_that("spod_available_data with use_s3=FALSE uses XML directly", {
  test_dir <- setup_test_data_dir()
  withr::defer(unlink(test_dir, recursive = TRUE))
  withr::local_envvar(c("SPANISH_OD_DATA_DIR" = test_dir))

  # Mock XML reading
  mock_xml_called <- FALSE
  testthat::local_mocked_bindings(
    read_data_links_memoised = function(...) {
      mock_xml_called <<- TRUE
      tibble::tibble(
        target_url = "https://test.com/file.csv.gz",
        pub_ts = as.POSIXct("2022-01-01", tz = "UTC"),
        file_size_bytes = 1000000,
        etag = "abc"
      )
    }
  )

  res <- spod_available_data_v2(
    data_dir = test_dir,
    use_s3 = FALSE,
    quiet = TRUE
  )

  expect_true(mock_xml_called)
  expect_true(nrow(res) > 0)
})

test_that("read_data_links_xml with force=TRUE re-downloads", {
  test_dir <- withr::local_tempdir()
  metadata_folder <- file.path(test_dir, spod_subfolder_metadata_cache())
  dir.create(metadata_folder, recursive = TRUE)

  # Create an old XML file
  old_xml <- file.path(
    metadata_folder,
    paste0("data_links_v1_", Sys.Date() - 1, ".xml")
  )
  writeLines(
    "<xml><link>test</link><pubDate>Mon, 14 Feb 2020 00:00:00 +0000</pubDate></xml>",
    old_xml
  )

  # Mock the download function
  download_called <- FALSE
  mock_download <- function(...) {
    download_called <<- TRUE
    new_xml <- file.path(
      metadata_folder,
      paste0("data_links_v1_", Sys.Date(), ".xml")
    )
    writeLines(
      "<xml><link>new</link><pubDate>Mon, 14 Feb 2020 00:00:00 +0000</pubDate></xml>",
      new_xml
    )
    return(new_xml)
  }

  res <- read_data_links_xml(
    metadata_folder = metadata_folder,
    data_dir = test_dir,
    force = TRUE,
    quiet = TRUE,
    latest_file_function = mock_download,
    ver = 1
  )

  expect_true(download_called)
  expect_true(nrow(res) > 0)
})

test_that("read_data_links_xml uses cached file when not stale", {
  test_dir <- withr::local_tempdir()
  metadata_folder <- file.path(test_dir, spod_subfolder_metadata_cache())
  dir.create(metadata_folder, recursive = TRUE)

  # Create a fresh XML file (today's date)
  fresh_xml <- file.path(
    metadata_folder,
    paste0("data_links_v1_", Sys.Date(), ".xml")
  )
  writeLines(
    "<xml><link>cached</link><pubDate>Mon, 14 Feb 2020 00:00:00 +0000</pubDate></xml>",
    fresh_xml
  )

  # Mock the download function - should NOT be called
  download_called <- FALSE
  mock_download <- function(...) {
    download_called <<- TRUE
    stop("Should not download!")
  }

  res <- read_data_links_xml(
    metadata_folder = metadata_folder,
    data_dir = test_dir,
    force = FALSE,
    quiet = TRUE,
    latest_file_function = mock_download,
    ver = 1
  )

  expect_false(download_called)
  expect_true(nrow(res) > 0)
  expect_equal(res$target_url[1], "cached")
})

test_that("spod_available_data v2 falls back to XML on S3 error", {
  test_dir <- setup_test_data_dir()
  withr::defer(unlink(test_dir, recursive = TRUE))
  withr::local_envvar(c("SPANISH_OD_DATA_DIR" = test_dir))

  # Mock S3 to fail and prevent real XML download
  testthat::local_mocked_bindings(
    spod_available_data_s3 = function(...) stop("S3 error"),
    spod_get_latest_v2_file_list = function(...) {
      xml_file <- file.path(test_dir, "metadata/data_links_v2_2024-01-01.xml")
      dir.create(dirname(xml_file), recursive = TRUE, showWarnings = FALSE)
      writeLines(
        "<rss><channel><item><link>https://test.com/file.csv.gz</link><pubDate>Mon, 01 Jan 2024 00:00:00 +0000</pubDate></item></channel></rss>",
        xml_file
      )
      return(xml_file)
    }
  )

  # Should fall back to XML and show message
  msgs <- capture_messages(
    capture.output(
      res <- spod_available_data(ver = 2, use_s3 = TRUE, quiet = FALSE),
      file = NULL
    )
  )
  expect_match(msgs, "S3 fetch failed", all = FALSE)

  expect_s3_class(res, "data.frame")
})

test_that("spod_available_data v1 falls back to XML on S3 error", {
  test_dir <- setup_test_data_dir()
  withr::defer(unlink(test_dir, recursive = TRUE))
  withr::local_envvar(c("SPANISH_OD_DATA_DIR" = test_dir))

  # Mock S3 to fail
  testthat::local_mocked_bindings(
    spod_available_data_s3 = function(...) stop("S3 error"),
    read_data_links_xml = function(...) {
      tibble::tibble(
        target_url = "https://opendata-movilidad.mitma.es/maestra1-mitma-distritos/ficheros-diarios/2020-02/20200215_maestra1_mitma_distrito.txt.gz",
        pub_ts = as.POSIXct("2020-02-15", tz = "UTC"),
        file_size_bytes = 1000,
        etag = "mock_etag"
      )
    }
  )

  # Should fall back to XML and show message
  msgs <- capture_messages(
    capture.output(
      res <- spod_available_data(ver = 1, use_s3 = TRUE, quiet = FALSE),
      file = NULL
    )
  )
  expect_match(msgs, "S3 fetch failed", all = FALSE)

  expect_s3_class(res, "data.frame")
})

test_that("spod_get_latest_v1_file_list handles download failure with messages", {
  test_dir <- withr::local_tempdir()
  testthat::local_mocked_bindings(
    download.file = function(...) stop("Network error"),
    .package = "utils"
  )
  msgs <- capture_messages(
    res <- spod_get_latest_v1_file_list(data_dir = test_dir, quiet = FALSE)
  )
  expect_true(any(grepl("Failed to download the XML file", msgs)))
  expect_true(any(grepl("Graceful exit", msgs)))
  expect_null(res)
})

test_that("spod_get_latest_v1_file_list handles download warning with messages", {
  test_dir <- withr::local_tempdir()
  testthat::local_mocked_bindings(
    download.file = function(...) { warning("some warning"); return(1) },
    .package = "utils"
  )
  msgs <- capture_messages(
    res <- spod_get_latest_v1_file_list(data_dir = test_dir, quiet = FALSE)
  )
  expect_true(any(grepl("Warning during XML download", msgs)))
  expect_true(any(grepl("Graceful exit", msgs)))
  expect_null(res)
})

test_that("spod_get_latest_v2_file_list handles download failure with messages", {
  test_dir <- withr::local_tempdir()
  testthat::local_mocked_bindings(
    download.file = function(...) stop("Network error"),
    .package = "utils"
  )
  msgs <- capture_messages(
    res <- spod_get_latest_v2_file_list(data_dir = test_dir, quiet = FALSE)
  )
  expect_true(any(grepl("Failed to download the XML file", msgs)))
  expect_true(any(grepl("Graceful exit", msgs)))
  expect_null(res)
})

test_that("spod_get_latest_v2_file_list handles download warning with messages", {
  test_dir <- withr::local_tempdir()
  testthat::local_mocked_bindings(
    download.file = function(...) { warning("some warning"); return(1) },
    .package = "utils"
  )
  msgs <- capture_messages(
    res <- spod_get_latest_v2_file_list(data_dir = test_dir, quiet = FALSE)
  )
  expect_true(any(grepl("Warning during XML download", msgs)))
  expect_true(any(grepl("Graceful exit", msgs)))
  expect_null(res)
})


test_that("read_data_links_xml returns empty tibble on NULL path", {
  test_dir <- withr::local_tempdir()
  metadata_folder <- file.path(test_dir, spod_subfolder_metadata_cache())
  dir.create(metadata_folder, recursive = TRUE)
  
  expect_message(
    res <- read_data_links_xml(
      metadata_folder = metadata_folder,
      data_dir = test_dir,
      force = TRUE,
      quiet = FALSE,
      latest_file_function = function(...) NULL,
      ver = 1
    ),
    "Fetching latest data links xml"
  )
  expect_s3_class(res, "tbl_df")
  expect_equal(nrow(res), 0)
  expect_true(all(c("target_url", "pub_ts") %in% names(res)))
})

test_that("read_data_links_xml shows message when using existing XML", {
  test_dir <- withr::local_tempdir()
  metadata_folder <- file.path(test_dir, spod_subfolder_metadata_cache())
  dir.create(metadata_folder, recursive = TRUE)
  
  # Create fresh XML
  fresh_xml <- file.path(metadata_folder, paste0("data_links_v1_", Sys.Date(), ".xml"))
  writeLines("<xml><link>test</link><pubDate>Mon, 14 Feb 2020 00:00:00 +0000</pubDate></xml>", fresh_xml)
  
  expect_message(
    read_data_links_xml(
      metadata_folder = metadata_folder,
      data_dir = test_dir,
      force = FALSE,
      quiet = FALSE,
      latest_file_function = function(...) stop("Should not be called"),
      ver = 1
    ),
    "Using existing data links xml"
  )
})

test_that("spod_available_data handles complete failure (S3 and XML fail)", {
  test_dir <- setup_test_data_dir()
  withr::defer(unlink(test_dir, recursive = TRUE))
  withr::local_envvar(c("SPANISH_OD_DATA_DIR" = test_dir))

  # Mock S3 and XML to both fail
  testthat::local_mocked_bindings(
    spod_available_data_s3 = function(...) stop("S3 down"),
    spod_get_latest_v2_file_list = function(...) NULL, # Simulation of download failure
    .package = "spanishoddata"
  )

  # Should fall back to XML, then XML should fail, resulting in empty tibble
  msgs <- capture_messages(
    res <- suppressWarnings(spod_available_data(ver = 2, quiet = FALSE))
  )
  
  expect_match(msgs, "S3 fetch failed", all = FALSE)
  # Message "Graceful exit" won't show because we mocked the function to return NULL directly
  
  expect_s3_class(res, "tbl_df")
  expect_equal(nrow(res), 0)

  expect_true("target_url" %in% names(res))
})

---
title: "Download and convert mobility datasets"
vignette: >
  %\VignetteIndexEntry{Download and convert mobility datasets}
  %\VignetteEngine{quarto::html}
  %\VignetteEncoding{UTF-8}
bibliography: ../inst/REFERENCES.bib
number-sections: true
format:
  html:
    toc: true
    toc-depth: 2
    code-overflow: wrap
execute: 
  eval: false
---

## Introduction {#intro}

**TL;DR (too long, didn't read): For analysing more than 1 week of data, use `spod_convert()` to convert the data into `DuckDB` and `spod_connect()` to connect to it for analysis using `{dplyr}`. Skip to the [section about it](#duckdb).**

The main focus of this vignette is to show how to get long periods of origin-destination data for analysis. First, we describe and compare the two ways to get the mobility data using origin-destination data as an example. The package functions and overall approaches are the same for working with other types of data available through the package, such as the number of trips, overnight stays and any other data. Then we show how to get a few days of origin-destination data with `spod_get()`. Finally, we show how to download and convert multiple weeks, months or even years of origin-destination data into analysis-ready formats. See description of datasets in the [Codebook and cookbook for v1 (2020-2021) Spanish mobility data](v1-2020-2021-mitma-data-codebook.html) and in the [Codebook and cookbook for v2 (2022 onwards) Spanish mobility data](v2-2022-onwards-mitma-data-codebook.html).


## Two ways to get the data

There are two main ways to import the datasets:

1. as an in-memory object with `spod_get()`;

2. as a connection to DuckDB or Parquet files on disk with `spod_convert()` + `spod_connect()`. The latter is recommended for large datasets (more than 1 week), as it is much faster and more memory efficient, as we demonstarte below.

`spod_get()` returns objects that are only appropriate for small datasets representing a few days of the national origin-destination flows. We recommend converting the data into analysis-ready formats (`DuckDB` or `Parquet`) using `spod_convert()` + `spod_connect()`. This will allow you to work with much longer time periods (months and years) on a consumer laptop (with 8-16 GB of memory). See the section below for more details.

## Analysing large datasets {#analysing-large-datasets}

The mobility datasets available through `{spanishiddata}` are very large. Particularly the origin-destination data, which contains millions of rows. These data sets may not fit into the memory of your computer, especially if you plan to run the analysis over multiple days, weeks, months, or even years.

To work with these datasets, we highly recommend using `DuckDB` and `Parquet`. These are systems for efficiently processing larger-than-memory datasets, while being user-firendly by presenting the data in a familiar `data.frame`/`tibble` object (almost). For a great intoroduction to both, we recommend materials by Danielle Navarro, Jonathan Keane, and Stephanie Hazlitt: [website](https://arrow-user2022.netlify.app/){target="_blank"}, [slides](https://arrow-user2022.netlify.app/slides){target="_blank"}, and [the video tutorial](https://www.youtube.com/watch?v=YZMuFavEgA4){target="_blank"}. You can also find examples of aggregating origin-destination data for flows analysis and visualisation in our vignettes on [static](https://rOpenSpain.github.io/spanishoddata/articles/flowmaps-static.html) and [interactive](https://rOpenSpain.github.io/spanishoddata/articles/flowmaps-interactive.html) flows visualisation.

Learning to use `DuckDB` and `Parquet` is easy for anyone who have ever worked with `{dplyr}` functions such as `select()`, `filter()`, `mutate()`, `group_by()`, `summarise()`, etc. However, since there is some learning curve to master these new tools, we provide some helper functions for novices to get started and easily open the datasets from `DuckDB` and `Parquet`. Please read the relevant sections below, where we first show how to convert the data, and then how to use it.

### How to choose between DuckDB, Parquet, and CSV {#duckdb-vs-parquet-csv}

The main considerations to make when choosing between `DuckDB` and `Parquet` (that you can get with `spod_convert()` + `spod_connect()`), as well as `CSV.gz` (that you can get with `spod_get()`) are analysis speed, convenience of data analysis, and the specific approach you prefer when getting the data. We discuss all three below.

#### Analysis Speed {#speed-comparison}

The data format you choose may dramatically impact the speed of analysis (e.g. filtering by dates, calculating number of trips per hour, per week, per month, per origin-destination pair, and any other data aggregation or manipulation).

In our tests (see @fig-csv-duckdb-parquet-speed), we found that conducting an analysis using `DuckDB` database provided a significant speed advantage over using `Parquet` and, more importantly, raw `CSV.gz` files. Specifically, when comparing a query to determine the mean hourly trips over 18 months for each zone pair, we observed that using `DuckDB` database was **up to 5 times** faster than using `Parquet` files and **up to 8 times** faster than using `CSV.gz` files.

![Data processing speed comparison: DuckDB engine running on CSV.gz files vs DuckDB database vs a folder of Parquet files](media/duckdb-parquet-csv-speed-mean-hourly-v1.svg){#fig-csv-duckdb-parquet-speed}


<details><summary>You can see the query we used for measuring the speed here</summary>

For reference, here is a simple query we used for speed comparison in @fig-csv-duckdb-parquet-speed:

```r
# data represents either CSV files acquired from `spod_get()`, a `DuckDB` database or a folder of Parquet files connected with `spod_connect()`
data |>
  group_by(id_origin, id_destination, hour) |> 
  summarise(mean_hourly_trips = mean(n_trips, na.rm = TRUE),
    .groups = "drop")
```

</details>


@fig-csv-duckdb-parquet-speed also shows that `DuckDB` format will give you the best performance even on low-end systems with limited memory and number of processor cores, conditional on a fast SSD storage. Also note, that if you do choose to work with long time periods using CSV.gz files via `spod_get()`, you will need to balance the amount of memory and processor cores via the `max_n_cpu` and `max_mem_gb` arguments, otherwise the analysis may fail (see the grey area in the figure), when there are too many parallel processes running at the same time with limited memory.

#### Convenience of data analysis

Regardless of the data format (`DuckDB`, `Parquet`, or `CSV.gz`), the functions you will need for data manipulation and analysis are the same. This is because the analysis is actually performed by the `DuckDB` [@duckdb-r] engine, which presents the data as if it were a regular `data.frame`/`tibble` object in R (almost). So from that point of view, there is no difference between the data formats. You can manipulate the data using `{dplyr}` functions such as `select()`, `filter()`, `mutate()`, `group_by()`, `summarise()`, etc. In the end of any sequence of commands you will need to add `collect()` to execute the whole chain of data manipulations and load the results into memory in an R `data.frame`/`tibble`. We provide examples in the following sections. Please refer to the recommended external tutorials and our own vignettes in the [Analysing large datasets](#analysing-large-datasets) section.

#### Scenarios of getting the data

The choice between converting to `DuckDB` and `Parquet` could also be made based on how you plan to work with the data. Specifically whether you want to just download long periods or even all available data, or if you want to get the data gradually, as you progress through with the analysis.

- If you plan to work with long time periods, we recommend `DuckDB`, as it is one big file and it is easier to update it completely. For example you may be working with all 2020 data. Later you decide to add all of 2021 data. In this case it would be better to delete the database and create it from scratch.

- If you only want certain dates, analyse them and add additional dates later, `Parquet` may be better, as each day is saved in a separate file, just like the original CSV files. Therefore updating a folder of `Parquet` files is as easy as just creating a new file only for the missing date.

- If you only work with a few individual days, you may not notice the advantages of the `DuckDB` or `Parquet` formats. In this case, you can keep using the `CSV.gz` format for the analysis using the `spod_get()` function. This is also useful for quick tutorials, where you only need one or two days of data for demonstration purposes.

## Setup {#setup}

Make sure you have loaded the package:

```{r}
library(spanishoddata)
```

{{< include ../inst/vignette-include/setup-data-directory.qmd >}}

## Getting a single day with `spod_get()` {#spod-get}

As you might have seen in the codebooks for [v1](v1-2020-2021-mitma-data-codebook.html) and [v2](v2-2022-onwards-mitma-data-codebook.html) data, you can get a single day's worth of data as an in-memory object with `spod_get()`:

```{r}
dates <- c("2024-03-01")
d_1 <- spod_get(type = "od", zones = "distr", dates = dates)
class(d_1)
```

The output should look like this:

```
# Source:   table<od_csv_clean_filtered> [?? x 19]
# Database: DuckDB v1.0.0 [... 6.5.0-45-generic:R 4.4.1/:memory:]
   date       hour id_origin id_destination distance activity_origin  
   <date>         <int> <fct>     <fct>          <fct>    <fct>            
 1 2024-03-01        19 01009_AM  01001          0.5-2    frequent_activity
 2 2024-03-01        15 01002     01001          10-50    frequent_activity
```

Note that this is a lazily-evaluated in-memory object (note the `:memory:` in the database path). This means that the data is not loaded into memory until you call `collect()` on it. While this is useful for quick exploration of the data, we do not recommended this for large datasets, as we have demonstrated [above](#speed-comparison).




## Analysing the data using `DuckDB` database {#duckdb}

Please make sure you did all the steps in the [Setup](#setup) section above.

### Convert to `DuckDB` {#convert-to-duckdb}

You can download and convert the data into `DuckDB` database in two steps. For example, you select a few dates, and download the data manually (note: we use `dates_2` to refer to the fact that we are using v2 data):

```{r}
dates_2 <- c(start = "2023-02-14", end = "2023-02-17")
spod_download(type = "od", zones = "distr", dates = dates_2)
```

```{r}
#| echo: false
# Try for v1 data
dates_1 <- c(start = "2020-02-14", end = "2020-02-17")
spod_download(type = "od", zones = "distr", dates = dates_1)
```

```{r}
#| echo: false
db_1 <- spod_convert(type = "od", zones = "distr", dates = "cached_v1", save_format = "duckdb", overwrite = TRUE)
db_1 # check the path to the saved `DuckDB` database
```

```{r}
#| echo: false
# Test adding another day:
fs::file_size(db_1)
dates_1 <- c(start = "2020-02-14", end = "2020-02-2018")
spod_download(type = "od", zones = "distr", dates = dates_1)
db_1 <- spod_convert(type = "od", zones = "distr", dates = "cached_v1", save_format = "duckdb") # no overwrite, failed
db_1 <- spod_convert(type = "od", zones = "distr", dates = "cached_v1", save_format = "duckdb", overwrite = TRUE)
db_1 <- spod_convert(type = "od", zones = "distr", dates = "cached_v1", save_format = "duckdb", overwrite = 'update')
db_1 # check the path to the saved `DuckDB` database
```

After that, you can convert any downloaded data (including the files that might have been downloaded previosly by running `spod_get()` or `spod_download()` with other dates or date intervals) into `DuckDB` like so (`dates = "cached_v2"` means use *all* downloaded files):

```{r}
db_2 <- spod_convert(type = "od", zones = "distr", dates = "cached_v2", save_format = "duckdb", overwrite = TRUE)
db_2 # check the path to the saved `DuckDB` database
```

```{r}
#| echo: false
# Test adding another day:
fs::file_size(db_2)
dates_2 <- c(start = "2020-02-14", end = "2020-02-18")
spod_download(type = "od", zones = "distr", dates = dates_2)
db_2 <- spod_convert(type = "od", zones = "distr", dates = "cached_v2", save_format = "duckdb") # no overwrite, failed
db_2 <- spod_convert(type = "od", zones = "distr", dates = "cached_v2", save_format = "duckdb", overwrite = TRUE)
db_2 <- spod_convert(type = "od", zones = "distr", dates = "cached_v2", save_format = "duckdb", overwrite = 'update')
db_2 # check the path to the saved `DuckDB` database
```

The `dates = "cached_v2"` (which can also be `dates = "cached_v1"` for v1 data) argument instructs the function to only work with already-downloaded files. By default this resulting `DuckDB` database for v2 origin-destination data for districts will be saved in the `SPANISH_OD_DATA_DIR` directory under `v2/tabular/duckdb/` with filename `od_distritos.duckdb` (you can change this file path with the `save_path` argument). The function returns the full path to the database file, which we save into `db_2` variable.
You can also any desired save location with the `save_path` argument of `spod_convert()`.

```{r}
#| echo: false
# Test save_path argument:
db_2 <- spod_convert(type = "od", zones = "distr", dates = "cached_v2", save_format = "duckdb", save_path = "/tmp/test.duckdb")
```

You can also convert any dates range or dates list to `DuckDB`:

```{r}
dates_1 <- c(start = "2020-02-17", end = "2020-02-19")
db_2 <- spod_convert(type = "od", zones = "distr", dates = dates_1, overwrite = TRUE)
```

In this case, any missing data that has not yet been downloaded will be automatically downloaded, while 2020-02-17 will not be redownloaded, as we already requsted it when creating `db_1`. Then the requested dates will be converted into `DuckDB`, overwriting the file with `db_1`. Once again, we save the path to the output `DuckDB` database file into `db_2` variable.

### Load the converted `DuckDB` {#load-converted-duckdb}

You can read the introductory information on how to connect to `DuckDB` files [here](https://duckdb.org/docs/api/r){target="_blank"}, however to simplify things for you we created a helper function. So to connect to the data stored in at path `db_1` and `db_2` you can do the following:

```{r}
my_od_data_2 <- spod_connect(db_2)
```

Just like before, with `spod_get()` funciton that we used to download raw CSV.gz files and analyse them without any conversion, the resulting object `my_od_data_2` is also a `tbl_duckdb_connection`. So, you can treat it as regular `data.frame` or `tibble` and use `{dplyr}` functions such as `select()`, `filter()`, `mutate()`, `group_by()`, `summarise()`, etc. For analysis, please refer to the recommended external tutorials and our own vignettes in the [Analysing large datasets](#analysing-large-datasets) section.

After finishing working with `my_od_data_2` we advise that you "disconnect" this data using:

```{r}
spod_disconnect(my_od_data_2)
```

This is useful to free-up memory and is neccessary if you would like to run `spod_convert()` again and save the data to the same location. Otherwise, it is also helpful to avoid unnecessary possible warnings in terminal for garbage collected connections.

## Analysing the data using `Parquet` {#parquet}

Please make sure you did all the steps in the [Setup](#setup) section above.

### Convert to `Parquet` {#convert-to-parquet}

The process is exactly the same as for `DuckDB` above. The only difference is that the data is converted to `parquet` format and stored in `SPANISH_OD_DATA_DIR` under `v1/clean_data/tabular/parquet/` directory for v1 data (change this with the `save_path` argument), and the subfolders are in hive-style format like `year=2020/month=2/day=14` and inside each of these folders a single `parquet` file will be placed containing the data for that day.

The advantage of this format is that you can "update" it quickly. For example, if you first downloaded the data for March and April 2020, converted this period into `parquet` format, and then downloaded the data for May and June 2020, when you run the convertion function again, it will only convert the data for May and June 2020 and add it to the existing `parquet` files. So you will save time and will not have to wait for March and April 2020 to be converted again.

Let us convert a few dates to `parquet` format:

```{r}
type <- "od"
zones <- "distr"
dates <- c(start = "2020-02-14", end = "2020-02-17")
od_parquet <- spod_convert(type = type, zones = zones, dates = dates, save_format = "parquet")
```

If we now request additional dates that overlap with the already converted data like so and specifiy argument `overwrite = 'update'` we will update the existing `parquet` files with the new data:

```{r}
dates <- c(start = "2020-02-16", end = "2020-02-19")
od_parquet <- spod_convert(type = type, zones = zones, dates = dates, save_format = "parquet", overwrite = 'update')
```

That is, 16 and 17 Feboruary will not be converted again. Only the new data, that was not converted (18 and 19 February) will be converted, and these will be added to the existing folder structure of`parquet` files stored at the default `save_path` location, which is `<data_dir>/clean_data/v1/tabular/parquet/od_distritos`. Alternatively, you can set any other save location by setting the `save_path` argument.

```{r}
#| echo: false
# Test for v2 data
dates <- c(start = "2023-02-14", end = "2023-02-17")
od_parquet <- spod_convert(type = type, zones = zones, dates = dates, save_format = "parquet", save_path = file.path(tempdir(), "od_parquet"))
```

### Load the converted `Parquet` {#load-converted-parquet}

Working with these `parquet` files is exactly the same as with `DuckDB` and `Arrow` files. Just like before, you can use the same helper function `spod_connect()` to connect to the `parquet` files:

```{r}
my_od_data_3 <- spod_connect(od_parquet)
```

Mind you though, because we have first converted the data for a period between 14 and 17 February 2020, and then converted the data for a period between 16 and 19 February 2020 into the save default location, the `od_parquet` contains the path to all this data, and therefore `my_od_data_3` will connect you to all data.

You can check this like so:

```{r}
my_od_data_3 |> 
  dplyr::distinct(date) |>
  dplyr::arrange(date)
```

For analysis, please refer to the recommended external tutorials and our own vignettes in the [Analysing large datasets](#analysing-large-datasets) section.


## Download all available data {#all-dates}

To prepare origin-destination data v1 (2020-2021) for analysis over the whole period of data availability, please follow the steps below:

```{r}
dates_v1 <- spod_get_valid_dates(ver = 1)
dates_v2 <- spod_get_valid_dates(ver = 2)
```


{{< include ../inst/vignette-include/missing-dates-outages.qmd >}}

### Download all data

Here the example is for origin-destination on district level for v1 data. You can change the `type` to "number_of_trips" and the `zones` to "municipalities" for v1 data. For v2 data, just use `dates` starting with 2022-01-01 or the `dates_v2` from above. Use all other function arguments for v2 in the same way as shown for v1, but also consult the [v2 data codebook](v2-2022-onwards-mitma-data-codebook.html), as it has many more datasets in addition to "origin-destination" and "number_of_trips".

```{r}
type <- "origin-destination"
zones <- "districts"
spod_download(
  type = type,
  zones = zones,
  dates = dates_v1,
  return_local_file_paths = FALSE, # to avoid getting all downloaded file paths printed to console
  max_download_size_gb = 50 # in Gb, this should be well over the actual download size for v1 data
)
```

### Convert all data into analysis ready format

```{r}
save_format <- "duckdb"

analysis_data_storage <- spod_convert_data(
  type = type,
  zones = zones,
  dates = "cached_v1", # to just convert all data that was previously downloaded, no need to specify dates here
  save_format = save_format,
  overwrite = TRUE
)
```

This will convert all downloaded data to `DuckDB` format for lightning fast analysis. You can change the `save_format` to `parquet` if you want to save the data in `Parquet` format. For comparison overview of the two formats please see the [Converting the data to DuckDB/Parquet for faster analysis](https://ropenspain.github.io/spanishoddata/articles/convert.html).

By default, `spod_convert_data()` will save the converted data in the `SPANISH_OD_DATA_DIR` directory. You can change the `save_path` argument of `spod_convert_data()` if you want to save the data in a different location.

For this conversion, 4 GB or operating memory should be enough, the speed of the process depends on the number of processor cores and the speed of your disk storage. SSD is preferred. By default, the `spod_convert_data()` will use all except one processor cores on your computer. You can adjust this with the `max_n_cpu` argument of `spod_convert_data()`. You can also increase the maximum amount of memory used with the `max_mem_gb` argument, but this makes more difference during the analysis stage.

Finally, `analysis_data_storage` will simply store the path to the converted data. Either a path to the `DuckDB` database file or a path to the folder with `Parquet` files.

### Conversion speed

For reference, converting the whole v1 origin-destination data to `DuckDB` takes about 20 minutes with 4 GB of memory and 3 processor cores. The final size of the `DuckDB` database is about 18 GB, in `Parquet` format - 26 GB. The raw CSV files in gzip archives are about 20GB. v2 data is much larger, with origin-destination tables for 2022 - mid-2024 taking up 150+ GB in raw CSV.gz format.

### Connecting to and analysing the converted datasets

You can pass the `analysis_data_storage` path to `spod_connect()` function, whether it is `DuckDB` or `Parquet`. The function will determine the data type automatically and give you back a `tbl_duckdb_connection`[^1].

[^1]: For reference: this object also has classes: `tbl_dbi` ,`tbl_sql`, `tbl_lazy` ,and `tbl` .

```{r}
my_data <- spod_connect(
  data_path = analysis_data_storage, 
  max_mem_gb = 16
)
```

Here we set `max_mem_gb` to 16 GB. Generally, if you have more, feel free to increase it, but also consult the @fig-csv-duckdb-parquet-speed with our speed testing results in the [Speed](#speed-comparison) section. You can try other combinations of `max_mem_gb` and `max_n_cpu` arguments for your needs

Compared to conversion process, you might want to increase the available memory for the analysis step. The more, the better. You can control that with the `max_mem_gb` argument.

You can manipulate `my_data` using `{dplyr}` functions such as `select()`, `filter()`, `mutate()`, `group_by()`, `summarise()`, etc. In the end of any sequence of commands you will need to add `collect()` to execute the whole chain of data manipulations and load the results into memory in an R `data.frame`/`tibble`. For analysis, please refer to the recommended external tutorials and our own vignettes in the [Analysing large datasets](#analysing-large-datasets) section.

After finishing working with `my_data` we advise that you "disconnect" to free up memory:

```{r}
spod_disconnect(my_data)
```

---
execute: 
  eval: false
---


## Install the package {#install-package}

Install from CRAN:

```{r}
install.packages("spanishoddata")
```



<details><summary>Alternative installation and developemnt</summary>

You can also install the latest development version of the package from rOpenSpain R universe:

```{r}
install.packages("spanishoddata",
  repos = c("https://ropenspain.r-universe.dev",
    "https://cloud.r-project.org"))
```


Alternative way to install the development version from GitHub:

```{r}
if (!require("remotes")) install.packages("remotes")

remotes::install_github("rOpenSpain/spanishoddata",
  force = TRUE, dependencies = TRUE)
```

**For Developers**

To load the package locally, clone it and navigate to the root of the package in the terminal, e.g. with the following:

```bash
gh repo clone rOpenSpain/spanishoddata
code spanishoddata
# with rstudio:
rstudio spanishoddata/spanishoddata.Rproj
```

Then run the following command from the R console:

```{r}
#| eval: false
devtools::load_all()
```

</details>

You can also explore the package and the data in an interactive RStudio container right in your web browser thanks to Binder, just click [the link](https://mybinder.org/v2/gh/e-kotov/spanishoddata-playground/HEAD?urlpath=rstudio) or the button: [![Binder](https://mybinder.org/badge_logo.svg)](https://mybinder.org/v2/gh/e-kotov/spanishoddata-playground/HEAD?urlpath=rstudio). Note that the session will be limited by memory and you will only be able to work with one full day of data.

---
title: "Quicky get daily data"
vignette: >
  %\VignetteIndexEntry{Quicky get daily data}
  %\VignetteEngine{quarto::html}
  %\VignetteEncoding{UTF-8}
bibliography: ../inst/REFERENCES.bib
number-sections: true
format:
  html:
    toc: true
    toc-depth: 2
    code-overflow: wrap
execute: 
  eval: false
---

::: callout-warning

**This functionality may stop working at any time, as the API may change unexpectedly.**

:::

## Introduction {#intro}

This vignette demonstrates how to get minimal daily aggregated data on the number of trips between municipalities using the \code{\link[=spod_quick_get_od]{spod_quick_get_od()}} function. With this function, you only get total trips for a single day, and no additional variables that are available in the full [v2 (2022 onwards) data set](v2-2022-onwards-mitma-data-codebook.html). The advantage of this function is that it is much faster than downloading the full data from source CSV files using `spod_get()`, as each CSV file for a single day is about 200 MB in size. Also, this way of getting the data is much less demanding on your computer as you are only getting a small table from the internet (less than 1 MB), and no data processing (such as aggregation from more detailed hourly data with extra columns that is happening when you use `spod_get()` function) is required on your computer. The original interactive web map with this data is available at [https://mapas-movilidad.transportes.gob.es/](https://mapas-movilidad.transportes.gob.es/){target="_blank"}.

## Setup {#setup}

Make sure you have loaded the package:

```{r}
library(spanishoddata)
library(dplyr)
library(stringr)
```


{{< include ../inst/vignette-include/setup-data-directory.qmd >}}

::: callout-note
Setting a local data directory in this case is optional, as the data is downloaded directly from the web API and there is no caching on disk. However, the metadata will be downloaded to check the range of valid dates available at the time of the request. So metadata will be downloaded to a temporary location by default, or to the data directory, if you do set it.
:::


## Get the data {#get-data}

### Get all flows with at least 1000 trips

To get the data, use the \code{\link[=spod_quick_get_od]{spod_quick_get_od()}} function. There is no need to specify whether you need municipalities or districts, as the only municipal level data can be accessed with this function. The `min_trips` argument specifies the minimum number of trips to include in the data. If you set `min_trips` to 0, you will get all data for all origin-destination pairs for the specified date.

```{r}
od_1000 <- spod_quick_get_od(
  date = "2022-01-01",
  min_trips = 1000
)
```

The data is returned as a tibble and only contanes the requested date, the identifiers of the origin and destination municipalities, the number of trips, and the total length of trips in kilometers.

```{r}
glimpse(od_1000)
```

```
glimpse(od_1000)
Rows: 8,524
Columns: 5
$ date                  <date> 2022-01-01, 2022-01-01, 2022-01-01, 2022…
$ id_origin             <chr> "01001", "01002", "01002", "01002", "0100…
$ id_destination        <chr> "01059", "01036", "01002", "01054_AM", "0…
$ n_trips               <int> 2142, 1215, 8899, 1105, 2250, 4621, 1992,…
$ trips_total_length_km <int> 27130, 13743, 26700, 10603, 12228, 69999,…
```


```{r}
od_1000
```

```
# A tibble: 8,524 × 5
   date       id_origin id_destination n_trips trips_total_length_km
   <date>     <chr>     <chr>            <int>                 <int>
 1 2022-01-01 01001     01059             2142                 27130
 2 2022-01-01 01002     01036             1215                 13743
 3 2022-01-01 01002     01002             8899                 26700
 4 2022-01-01 01002     01054_AM          1105                 10603
 5 2022-01-01 01002     01010             2250                 12228
 6 2022-01-01 01009_AM  01059             4621                 69999
 7 2022-01-01 01009_AM  01009_AM          1992                 16395
 8 2022-01-01 01009_AM  01051             2680                 18554
 9 2022-01-01 01010     01002             2147                 11578
10 2022-01-01 01017_AM  01017_AM          1847                 12695
# ℹ 8,514 more rows
# ℹ Use `print(n = ...)` to see more rows
```

### Get only trips of certain length

To get only trips of a certain length, use the `distances` argument.

```{r}
od_long <- spod_quick_get_od(
  date = "2022-01-01",
  min_trips = 0,
  distances = c("10-50km", "50+km")
)
```


```{r}
glimpse(od_long)
```


```
Rows: 247,208
Columns: 5
$ date                  <date> 2022-01-01, 2022-01-01, 2022-01-01, 2022…
$ id_origin             <chr> "08015", "08015", "08015", "08015", "0801…
$ id_destination        <chr> "08285", "17902_AM", "43014", "08007", "0…
$ n_trips               <int> 5, 1, 5, 165, 210, 111, 1486, 39, 52, 166…
$ trips_total_length_km <int> 339, 161, 924, 5052, 2955, 2453, 29630, 1…
```


```{r}
od_long
```

```
# A tibble: 247,208 × 5
   date       id_origin id_destination n_trips trips_total_length_km
   <date>     <chr>     <chr>            <int>                 <int>
 1 2022-01-01 08015     08285                5                   339
 2 2022-01-01 08015     17902_AM             1                   161
 3 2022-01-01 08015     43014                5                   924
 4 2022-01-01 08015     08007              165                  5052
 5 2022-01-01 08015     08030              210                  2955
 6 2022-01-01 08015     08051              111                  2453
 7 2022-01-01 08015     08121             1486                 29630
 8 2022-01-01 08015     08122_AM            39                  1886
 9 2022-01-01 08015     08300_AM            52                  1301
10 2022-01-01 08015     08902              166                  2042
# ℹ 247,198 more rows
# ℹ Use `print(n = ...)` to see more rows
```

### Get only trips between certain municipalities


To get only trips between certain municipalities, use the `id_origin` and `id_destination` arguments.

You can get all valid munincipality identifiers with the `spod_get_zones("muni", ver = 2)` function. This function will need to download some spatial data, so it might take some time and you might want to [setup the data download folder](#setup) with `spod_setup_cache()` if you have not done so before.


```{r}
municipalities <- spod_quick_get_zones()

# if code above fails, you can also use the following:
# municipalities <- spod_get_zones("muni", ver = 2)
# head(municipalities)
```

#### All trips from Madrid

Let us select all locations with Madrid in the name:

```{r}
madrid_muni_ids <- municipalities |>
  filter(str_detect(name, "Madrid")) |>
  pull(id)

madrid_muni_ids
```

```
[1] "28073" "28079" "28127" "45087"
```

Now let use use these IDs as origins to gett all trips from Madrid to the rest of Spain:

```{r}
flows_from_Madrid <- spod_quick_get_od(
  date = "2022-01-01",
  min_trips = 0,
  id_origin = madrid_muni_ids
)
```


```{r}
glimpse(flows_from_Madrid)
```

```
Rows: 2,232
Columns: 5
$ date                  <date> 2022-01-01, 2022-01-01, 2022-01-01, 2022…
$ id_origin             <chr> "28073", "28073", "28073", "28073", "2807…
$ id_destination        <chr> "28007", "28066", "28079", "45081", "FR10…
$ n_trips               <int> 1239, 1505, 3730, 237, 2, 10, 11, 5, 9, 2…
$ trips_total_length_km <int> 11120, 7268, 75798, 3385, 82, 296, 1036, …
```

```
# A tibble: 2,232 × 5
   date       id_origin id_destination n_trips trips_total_length_km
   <date>     <chr>     <chr>            <int>                 <int>
 1 2022-01-01 28073     28007             1239                 11120
 2 2022-01-01 28073     28066             1505                  7268
 3 2022-01-01 28073     28079             3730                 75798
 4 2022-01-01 28073     45081              237                  3385
 5 2022-01-01 28073     FR102                2                    82
 6 2022-01-01 28073     28177               10                   296
 7 2022-01-01 28073     45165               11                  1036
 8 2022-01-01 28073     46011_AM             5                  1621
 9 2022-01-01 28073     21076_AM             9                  4102
10 2022-01-01 28073     46120_AM             2                   631
# ℹ 2,222 more rows
# ℹ Use `print(n = ...)` to see more rows
```

#### All trips from Madrid to Barcelona

Similarly, you can set limits on the destination municipalities:

```{r}
barcelona_muni_ids <- municipalities |>
  filter(str_detect(name, "Barcelona")) |>
  pull(id)
barcelona_muni_ids
```


```{r}
madrid_barcelona_od <- spod_quick_get_od(
  date = "2022-01-01",
  min_trips = 0,
  id_origin = madrid_muni_ids,
  id_destination = barcelona_muni_ids
)
```


```{r}
madrid_barcelona_od
```


```
# A tibble: 4 × 5
  date       id_origin id_destination n_trips trips_total_length_km
  <date>     <chr>     <chr>            <int>                 <int>
1 2022-01-01 28073     08019                2                  1334
2 2022-01-01 28079     08019             1661                838281
3 2022-01-01 28127     08019               26                 13699
4 2022-01-01 45087     08019                2                  1184
```

You can now proceed to analyse these flows or visualise them as in the [static](https://rOpenSpain.github.io/spanishoddata/articles/flowmaps-static.html) and [interactive](https://rOpenSpain.github.io/spanishoddata/articles/flowmaps-interactive.html) flow maps tutorials.

::: callout-note

As long as you use a table connection object created with `spod_get()` function, it is much quicker to filter the dates by the `year`, `month` and `day` variables, rather than by the `date` variable. This is because the data for each day is in a separate CSV file located in folders that look like `year=2020/month=2/day=14`. So when filtering by the `date` field, R will have to scan all CSV files comparing the specified date with what is stored inside each CSV file. However, if you query by `year`, `month` and `day` variables, R only needs to check these against the path to each CSV file, which is much quicker. This caveat is only relevant as long as you use `spod_get()` . If you convert (see the [relevant vignette](convert.qmd)) the downloaded data to a format that it optimized for quick analysis, you can use whichever field you want, it should not affect the performance.

:::

---
execute: 
  eval: false
---


::: callout-warning


Due to mobile network outages, the data on certain dates is missing. Kindly keep this in mind when calculating mean monthly or weekly flows.

\

Please check the [original data page](https://www.transportes.gob.es/ministerio/proyectos-singulares/estudios-de-movilidad-con-big-data/opendata-movilidad){target="_blank"} for currently known missing dates. At the time of writing, the following dates are missing: 26, 27, 30, 31 October; 1, 2 and 3 November 2023; 4, 18, 19 April 2024, 10 and 11 November 2024. You can use `spod_get_valid_dates()` function to get all available dates.


:::

---
title: "Codebook and cookbook for v1 (2020-2021) Spanish mobility data"
vignette: >
  %\VignetteIndexEntry{Codebook and cookbook for v1 (2020-2021) Spanish mobility data}
  %\VignetteEngine{quarto::html}
  %\VignetteEncoding{UTF-8}
bibliography: ../inst/REFERENCES.bib
format:
  html:
    toc: true
    toc-depth: 2
    code-overflow: wrap
execute: 
  eval: false
---

You can view this vignette any time by running:

```{r}
spod_codebook(ver = 1)
```

The mobility data v1 (2020-2021) was originally released by the Ministerio de Transportes, Movilidad y Agenda Urbana (MITMA) , now [Ministerio de Transportes y Movilidad Sostenible (MITMS)](https://www.transportes.gob.es/) [@mitma_methodology_2020_v3]. 

The dataset is produced by [Nommon](https://www.nommon.es/){target="_blank"} using the raw data from [Orange España](https://www.orange.es/){target="_blank"}. Even though the raw data is only from one mobile phone operator, the resulting flows and other counts of number of individuals in the data set are already resampled to be representative of the total population of Spain (see details in the official methodology).

The tables in the data set provide hourly flows between zones across Spain for every day of the observation period (2020-02-14 to 2021-05-09), and the number of individuals making trips for each zone. This document will introduce you to the available data and provide brief code snippets on how to access it using the `{spanishoddata}` R package.

Key sources for this codebook/cookbook include:

- [original data collection methodology in Spanish](https://cdn.mitma.gob.es/portal-web-drupal/covid-19/bigdata/mitma_-_estudio_movilidad_covid-19_informe_metodologico_v3.pdf){target="_blank"} + [automatically translated English version of methodology](https://rOpenSpain.github.io/spanishoddata/codebooks/mitma_-_estudio_movilidad_covid-19_informe_metodologico_v3_en.pdf){target="_blank"}

- [original data codebook in Spanish](https://opendata-movilidad.mitma.es/README%20-%20formato%20ficheros%20movilidad%20MITMA%2020201228.pdf){target="_blank"} + [automatically translated English version of codebook](https://rOpenSpain.github.io/spanishoddata/codebooks/README_-_formato_ficheros_movilidad_MITMA_20201228_en.pdf){target="_blank"}

- [original data download page](https://www.transportes.gob.es/ministerio/proyectos-singulares/estudios-de-movilidad-con-big-data/estudios-de-movilidad-anteriores/covid-19/opendata-movilidad){target="_blank"}

- [original data license](https://opendata-movilidad.mitma.es/LICENCIA%20de%20datos%20abiertos%20del%20MITMA%2020201203.pdf){target="_blank"}

- [homepage of the v1 study](https://www.transportes.gob.es/ministerio/proyectos-singulares/estudios-de-movilidad-con-big-data/estudios-de-movilidad-anteriores/covid-19){target="_blank"}

- [homepage of the open mobility data project of the Ministry of Transport and Sustainable Mobility of Spain](https://www.transportes.gob.es/ministerio/proyectos-singulares/estudio-de-movilidad-con-big-data){target="_blank"}


**Note: Kindly consult the documents above for any specific details on the methodology. The codebook here is only a simplified summary.**

To access the data we reference in this codebook, please follow these steps:

{{< include ../inst/vignette-include/install-package.qmd >}}


Load it as follows:

```{r}
library(spanishoddata)
```


Using the instructions below, set the data folder for the package to download the files into. You may need up to 30 GB to download all data and another 30 GB if you would like to convert the downloaded data into analysis ready format (a `DuckDB` database file, or a folder of `parquet` files). You can find more info on this conversion in the [Download and convert OD datasets](convert.html) vignette.

{{< include ../inst/vignette-include/setup-data-directory.qmd >}}

{{< include ../inst/vignette-include/overall-approach.qmd >}}

![The overview of package functions to get the data](../man/figures/package-functions-overview.svg){#fig-overall-flow width="78%"}

## 1. Spatial data with zoning boundaries

The boundary data is provided at two geographic levels: [`Districts`](#districts) and [`Municipalities`](#municipalities). It's important to note that these do not always align with the official Spanish census districts and municipalities. To comply with data protection regulations, certain aggregations had to be made to districts and municipalities".

### 1.1 `Districts` {#districts}

`Districts` correspond to official census districts in cities; however, in those with lower population density, they are grouped together. In rural areas, one district is often equal to a municipality, but municipalities with low population are combined into larger units to preserve privacy of individuals in the dataset. Therefore, there are 2850 'districts' compared to the 10494 official census districts on which they are based.

To access it:

```{r}
districts_v1 <- spod_get_zones("dist", ver = 1)
```

The `districts_v1` object is of class `sf` consisting of polygons.

Data structure:

| Variable Name | **Description** |
|------------------------------------|------------------------------------|
| `id` | District `id` assigned by the data provider. Matches with `id_origin`, `id_destination`, and `id` in district level [origin-destination data](#od-data) and [number of trips data](#ptc-data). |
| `census_districts` | A string with semicolon-separated list of census district semicolon-separated identifiers as classified by the Spanish Statistical Office (INE) that are spatially bound within polygons with `id` above. |
| `municipalities_mitma` | A string with semicolon-separated list of municipality identifiers as assigned by the data provider in municipality zones spatial dataset that correspond to a given district `id` . |
| `municipalities` | A string with semicolon-separated list of municipality identifiers as classified by the Spanish Statistical Office (INE) that correspond to polygons with `id` above. |
| `district_names_in_v2` | A string with semicolon-separated list of names of district polygons defined in the [v2 version of this data](v2-2022-onwards-mitma-data-codebook.html) that covers the year 2022 and onwards that correspond to polygons with `id` above. |
| `district_ids_in_v2` | A string with semicolon-separated list of identifiers of district polygons defined in the [v2 version of this data](v2-2022-onwards-mitma-data-codebook.html) that covers the year 2022 and onwards that correspond to polygons with `id` above. |

### 1.2 `Municipalities` {#municipalities}

`Municipalities` are made up of official municipalities in those of a certain size; however, they have also been aggregated in cases of lower population density. As a result, there are 2,205 municipalities compared to the 8,125 official municipalities on which they are based.

To access it:

```{r}
municipalities_v1 <- spod_get_zones("muni", ver = 1)
```

The resulting `municipalities_v1` object is type `sf` consisting of polygons.

Data structure:

| Variable Name | **Description** |
|------------------------------------|------------------------------------|
| `id` | District `id` assigned by the data provider. Matches with `id_origin`, `id_destination`, and `id` in municipality level [origin-destination data](#od-data) and [population by trip count](#pop-tc). |
| `municipalities` | A list of municipality identifiers as classified by the Spanish Statistical Office (INE) that correspond to polygons with `id` above. |
| `districts_mitma` | A list of district identifiers as assigned by the data provider in districts zones spatial dataset that correspond to a given municipality `id` . |
| `census_districts` | A list of census district identifiers as classified by the Spanish Statistical Office (INE) that are spatially bound within polygons with `id` above. |
| `municipality_names_in_v2` | A list of names of municipality polygons defined in the [v2 version of this data](v2-2022-onwards-mitma-data-codebook.html) that covers the year 2022 and onwards that correspond to polygons with `id` above. |
| `municipality_ids_in_v2` | A list of identifiers of municipality polygons defined in the [v2 version of this data](v2-2022-onwards-mitma-data-codebook.html) that covers the year 2022 and onwards that correspond to polygons with `id` above. |

The spatial data you get via `spanishoddata` package is downloaded directly from the source, the geometries of polygons are automatically fixed if there are any invalid geometries. The zone identifiers are stored in `id` column. Apart from that `id` column, the original zones files do not have any metadata. However, as seen above, using the `spanishoddata` package you get many additional columns that provide a semantic connection between official statistical zones used by the Spanish government and the zones you can get for the v2 data (for 2022 onward).

## 2. Mobility data

All mobility data is referenced via `id_origin`, `id_destination`, or other location identifiers (mostly labelled as `id`) with the two sets of zones described above.

### 2.1. Origin-Destination data {#od-data}

The origin-destination data contain the number of trips and distance travelled between `districts` or `municipalities` in Spain for every hour of every day between 2020-02-14 and 2021-05-09. Each flow also has attributes such as the trip purpose (composed of the type of activity (`home`/`work_or_study`/`other`) at both the origin and destination), province of residence of individuals making this trip, distance covered while making the trip. See the detailed attributes below in a table. @fig-flows-barcelona shows an example of total flows in the province of Barcelona on Feb 14th, 2020.

![Origin destination flows in Barcelona on 2020-02-14](media/flows_plot_barcelona.svg){#fig-flows-barcelona width="70%"}

Here are the variables you can find in both the `district` and `municipality` level origin-destination data:

| **English Variable Name** | **Original Variable Name** | **Type** | **Description** |
|----------------|----------------|----------------|------------------------|
| `date` | `fecha` | `Date` | The date of the recorded data, formatted as `YYYY-MM-DD`. |
| `hour` | `periodo` | `integer` | The time slot during which the trips occurred. |
| `id_origin` | `origen` | `factor` | The origin zone `id` of `district` or `municipalitity`. |
| `id_destination` | `destino` | `factor` | The destination zone `id` of `district` or `municipalitity`. |
| `distance` | `distancia` | `factor` | The distance range of the trip, categorized into specific intervals such as `0005-002` (500 m to 2 km), `002-005` (2-5 km), `005-010` (5-10km), `010-050` (10-50 km), `050-100` (50-100 km), and `100+` (more than 100 km). |
| `activity_origin` | `actividad_origen` | `factor` | The type of activity at the origin zone, recoded from `casa`, `otros`, `trabajo_estudio` to `home`, `other`, `work_or_study` respectively. |
| `activity_destination` | `actividad_destino` | `factor` | The type of activity at the destination zone, similarly recoded as for `activity_origin` above. |
| `residence_province_ine_code` | `residencia` | `factor` | The province code of residence if individuals who were making the trips in `n_trips`, encoded as province codes as classified by the Spanish Statistical Office (INE). |
| `residence_province_name` | Derived from `residencia` | `factor` | The full name of the residence province, derived from the province code above. |
| `n_trips` | `viajes` | `numeric` | The number of trips for that specific origin-destination pair and time slot. |
| `trips_total_length_km` | `viajes_km` | `numeric` | The total length of trips in kilometers, summing up all trips between the origin and destination zones. |
| `year` | `year` | `integer` | The year of the recorded data, extracted from the date. |
| `month` | `month` | `integer` | The month of the recorded data, extracted from the date. |
| `day` | `day` | `integer` | The day of the recorded data, extracted from the date. |

<details><summary>Data transformation note</summary>

The original data is stored in the [`maestra-2`](https://www.transportes.gob.es/ministerio/proyectos-singulares/estudios-de-movilidad-con-big-data/estudios-de-movilidad-anteriores/covid-19/opendata-movilidad){target="_blank"} folder with suffixes `distritos` (for district zoning) and `municipios` (for municipality zoning). We only use the `district` level data because of several data issues with the `municipality` data documented [here](https://www.ekotov.pro/mitma-data-issues/issues/011-v1-tpp-mismatch-zone-ids-in-table-and-spatial-data.html){target="_blank"} and [here](https://www.ekotov.pro/mitma-data-issues/issues/012-v1-tpp-district-files-in-municipality-folders.html){target="_blank"}, but also because the distric level data contains more columns with useful origin-destination flow characteristics. As a result, you get both the `district` level data and the `municipality` level data with the same columns. `Municipality` level data is simply a re-aggregation of `district` level data using the official relations file where `district` identifiers are mapped to `municipality` identifiers (orginal file is [`relaciones_distrito_mitma.csv`](https://opendata-movilidad.mitma.es/relaciones_distrito_mitma.csv){target="_blank"}).

</details>

**Getting the data**

To access the data, use the `spod_get()` function. In this example we will use a short interval of dates:

```{r}
dates <- c(start = "2020-02-14", end = "2020-02-17")
od_dist <- spod_get(type = "od", zones = "dist", dates = dates)
od_muni <- spod_get(type = "od", zones = "muni", dates = dates)
```

The data for the specified dates will be automatically downloaded and cached in the `SPANISH_OD_DATA_DIR` directory. Existing files will not be re-downloaded.

**Working with the data**

The resulting objects `od_dist` and `od_muni` are of class `tbl_duckdb_connection`[^1]. Basically, you can treat these as regular `data.frame`s or `tibble`s. One important difference is that the data is not actually loaded into memory, because if you requested more dates, e.g. a whole month or a year, all that data would most likely not fit into your computer's memory. A `tbl_duckdb_connection` is mapped to the downloaded CSV files that are cached on disk and the data is only loaded in small chunks as needed at the time of computation. You can manipulate `od_dist` and `od_muni` using `{dplyr}` functions such as `select()`, `filter()`, `mutate()`, `group_by()`, `summarise()`, etc. In the end of any sequence of commands you will need to add `collect()` to execute the whole chain of data manipulations and load the results into memory in an R `data.frame`/`tibble` like so:

[^1]: For reference: this object also has classes: `tbl_dbi` ,`tbl_sql`, `tbl_lazy` ,and `tbl` .

```{r}
library(dplyr)
od_mean_hourly_trips_over_the_4_days <- od_dist |>
  group_by(hour) |>
  summarise(
    mean_hourly_trips = mean(n_trips, na.rm = TRUE),
    .groups = "drop"
  ) |>
  collect()
od_mean_hourly_trips_over_the_4_days
```

```         
# A tibble: 24 × 2
   hour mean_hourly_trips
       <int>             <dbl>
 1        18              21.4
 2        10              19.3
 3         2              14.8
 4        15              19.8
 5        11              19.9
 6        16              19.6
 7        22              20.9
 8         0              18.6
 9        13              21.1
10        19              22.5
# ℹ 14 more rows
# ℹ Use `print(n = ...)` to see more rows
```

In this example above, we calculated mean hourly flows over the 4 days of the requested period. The full data for all 4 days was probably never loaded into memory all at once. Rather the available memory of the computer was used up to its maximum limit to make that calculation happen, without ever exceeding the available memory limit. If you were doing the same opearation on 100 or even more days, it would work in the same way and would be possible even with limited memory. This is done transparantly to the user with the help of [`DuckDB`](https://duckdb.org/){target="_blank"} (specifically, with [{duckdb} R package](https://r.duckdb.org/index.html){target="_blank"} @duckdb-r).

The same summary operation as provided in the example above can be done with the entire dataset for the full 18 month on a regular laptop with 8-16 GB memory. It will take a bit of time to complete, but it will be done. To speed things up, please also see the [vignette on converting the data](convert.qmd) into formats that will increase the analsysis performance.

{{< include ../inst/vignette-include/csv-date-filter-note.qmd >}}

### 2.2. Population by trip count data {#ptc-data}

The population by trip count data shows the number of individuals in each district or municipality, categorized by the trips they make: 0, 1, 2, or more than 2.

| **English Variable Name** | **Original Variable Name** | **Type** | **Description** |
|----------------|----------------|----------------|------------------------|
| `date` | `fecha` | `Date` | The date of the recorded data, formatted as `YYYY-MM-DD`. |
| `id` | `distrito` | `factor` | The identifier of the `district` or `municipality` zone. |
| `n_trips` | `numero_viajes` | `factor` | The number of individuals who made trips, categorized by `0`, `1`, `2`, or `2+` trips. |
| `n_persons` | `personas` | `factor` | The number of individuals making the trips from `district` or `municipality` with zone `id`. |
| `year` | `year` | `integer` | The year of the recorded data, extracted from the date. |
| `month` | `month` | `integer` | The month of the recorded data, extracted from the date. |
| `day` | `day` | `integer` | The day of the recorded data, extracted from the date. |


<details><summary>Data transformation note</summary>

The original data is stored in the [`maestra-2`](https://www.transportes.gob.es/ministerio/proyectos-singulares/estudios-de-movilidad-con-big-data/estudios-de-movilidad-anteriores/covid-19/opendata-movilidad){target="_blank"} folder with suffixes `distritos` (for district zoning) and `municipios` (for municipality zoning). We only use the `district` level data because of several data issues with the `municipality` data documented [here](https://www.ekotov.pro/mitma-data-issues/issues/011-v1-tpp-mismatch-zone-ids-in-table-and-spatial-data.html){target="_blank"} and [here](https://www.ekotov.pro/mitma-data-issues/issues/012-v1-tpp-district-files-in-municipality-folders.html){target="_blank"}, but also because the distric level data contains more columns with useful origin-destination flow characteristics. As a result, you get both the `district` level data and the `municipality` level data with the same columns. `Municipality` level data is simply a re-aggregation of `district` level data using the official relations file where `district` identifiers are mapped to `municipality` identifiers (orginal file is [`relaciones_distrito_mitma.csv`](https://opendata-movilidad.mitma.es/relaciones_distrito_mitma.csv){target="_blank"}).

</details>

**Getting the data**

To access it use `spod_get()` with `type` set to "number_of_trips", or just "nt". We can also set `dates` to the maximum possible date range `2020-02-14` to `2021-05-09` to get all the data, as this data is relatively small (under 200 Mb).

```{r}
dates <- c(start = "2020-02-14", end = "2021-05-09")
nt_dist <- spod_get(type = "number_of_trips", zones = "dist", dates = dates)
```

Because this data is small, we can actually load it completely into memory:

```{r}
nt_dist_tbl <- nt_dist |> dplyr::collect()
```

## Advanced use

For more advanced use, especially for analysing longer periods (months or even years), please see [Download and convert mobility datasets](convert.html).

---
execute: 
  eval: false
---


## Set the data directory {#set-data-folder}

Choose where `{spanishoddata}` should download (and convert) the data by setting the data directory following command:

```{r}
spod_set_data_dir(data_dir = "~/spanish_od_data")
```

The function above will also ensure that the directory is created and that you have sufficient permissions to write to it.

<details><summary>Setting data directory for advanced users</summary>

You can also set the data directory with an environment variable:

```{r}
Sys.setenv(SPANISH_OD_DATA_DIR = "~/spanish_od_data")
```

The package will create this directory if it does not exist on the first run of any function that downloads the data.

To permanently set the directory for all projects, you can specify the data directory globally by setting the `SPANISH_OD_DATA_DIR` environment variable, e.g. with the following command:

```{r}
#| eval: false
usethis::edit_r_environ()
# Then set the data directory globally, by typing this line in the file:
```

```
SPANISH_OD_DATA_DIR = "~/spanish_od_data"
```

You can also set the data directory locally, just for the current project. Set the 'envar' in the working directory by editing `.Renviron` file in the root of the project:

```{r}
#| eval: false
file.edit(".Renviron")
```


</details>

---
execute: 
  eval: false
---


## Overall approach to accessing the data

If you only need flows data aggregated by day at municipal level, you can use the `spod_quick_get_od()` function. This will download the data directly from the web API and let you analyse it in-memory. More on this in the [Quickly get daily data](https://ropenspain.github.io/spanishoddata/articles/quick-get.html) vignette.

If you only want to analyse the data for a few days, you can use the `spod_get()` function. It will download the raw data in CSV format and let you analyse it in-memory. This is what we cover in the steps on this page.

If you need longer periods (several months or years), you should use the `spod_convert()` and `spod_connect()` functions, which will convert the data into special format which is much faster for analysis, for this see the [Download and convert OD datasets](https://ropenspain.github.io/spanishoddata/articles/convert.html) vignette. `spod_get_zones()` will give you spatial data with zones that can be matched with the origin-destination flows from the functions above using zones 'id's. Please see a simple example below, and also consult the vignettes with detailed data description and instructions in the package vignettes with `spod_codebook(ver = 1)` and `spod_codebook(ver = 2)`, or simply visit the package website at [https://ropenspain.github.io/spanishoddata/](https://ropenspain.github.io/spanishoddata/). The @fig-overall-flow presents the overall approach to accessing the data in the `spanishoddata` package.

---
title: "Codebook and cookbook for v2 (2022 onwards) Spanish mobility data"
vignette: >
  %\VignetteIndexEntry{Codebook and cookbook for v2 (2022 onwards) Spanish mobility data}
  %\VignetteEngine{quarto::html}
  %\VignetteEncoding{UTF-8}
bibliography: ../inst/REFERENCES.bib
format:
  html:
    toc: true
    toc-depth: 2
    code-overflow: wrap
execute: 
  eval: false
---

You can view this vignette any time by running:

```{r}
spod_codebook(ver = 2)
```

The mobility data v2 (2022 onwards) was originally released by the Ministerio de Transportes, Movilidad y Agenda Urbana (MITMA) , now [Ministerio de Transportes y Movilidad Sostenible (MITMS)](https://www.transportes.gob.es/) [@mitms_methodology_2022_v8].

The dataset is produced by [Nommon](https://www.nommon.es/){target="_blank"} using the raw data from [Orange España](https://www.orange.es/){target="_blank"}. Even though the raw data is only from one mobile phone operator, the resulting flows and other counts of number of individuals in the data set are already resampled to be representative of the total population of Spain (see details in the official methodology).

The tables in the data set provide hourly flows between zones across Spain for every day of the observation period (2022-01-01 onwards), the number of individuals making trips for each zone, the number of individuals spending the nights in one location while regularly residing in a different location, and many more advanced datasets. This document will introduce you to the available data and provide brief code snippets on how to access it using the `{spanishoddata}` R package.

Compared to the [v1 data (2020-2021)](v1-2020-2021-mitma-data-codebook.qmd) this dataset has many additional variables, such as age, sex, and income, has better spatial resolution (the zones are more spatially granular) and covers a continuous period[^1] (2022-01-01 onward), rather than only a limited period (in v1 - 2020-02-14 to 2021-05-09).

[^1]: For reference: this object also has classes: `tbl_dbi` ,`tbl_sql`, `tbl_lazy` ,and `tbl` .

{{< include ../inst/vignette-include/missing-dates-outages.qmd >}}

In v2 data, the datasets comparable to the [v1 (2020-2021) data](v1-2020-2021-mitma-data-codebook.qmd) (hourly daily origin destination matrices and number of trips at each location) are called ["basic studies" ("estudios basicos")](https://www.transportes.gob.es/ministerio/proyectos-singulares/estudios-de-movilidad-con-big-data/estudio-basico-diario){target="_blank"}, however there are even more advanced datasets that are already available or will be made available soon: ["complete studies" ("estudios completos")](https://www.transportes.gob.es/ministerio/proyectos-singulares/estudios-de-movilidad-con-big-data/estudios-completos){target="_blank"} and ["road routes" ("rutas de carretera")](https://www.transportes.gob.es/ministerio/proyectos-singulares/estudios-de-movilidad-con-big-data/rutas-por-carretera){target="_blank"}. At the moment, the `{spanishoddata}` package only provides the interface to the "basic studies" datasets, but support for "complete studies" and "road routes" will be added in the future.

Key sources for this codebook/cookbook include:

-   [original data collection methodology and codebook in Spanish](https://www.transportes.gob.es/recursos_mfom/paginabasica/recursos/a3_informe_metodologico_estudio_movilidad_mitms_v8.pdf){target="_blank"} + [automatically translated English version of methodology and codebook](https://rOpenSpain.github.io/spanishoddata/codebooks/a3_informe_metodologico_estudio_movilidad_mitms_v8_en.pdf){target="_blank"}

- [original quality assessment](https://www.transportes.gob.es/recursos_mfom/paginabasica/recursos/20241024_validaciones_estudios_basicos_bigdata_v1.0.pdf){target="_blank"} + [automatically translated English version of quality assessment](https://rOpenSpain.github.io/spanishoddata/codebooks/20241024_validaciones_estudios_basicos_bigdata_v1.0_en.pdf){target="_blank"}

-   [original data download page](https://www.transportes.gob.es/ministerio/proyectos-singulares/estudios-de-movilidad-con-big-data/opendata-movilidad){target="_blank"}

-   [original data license](https://www.transportes.gob.es/el-ministerio/buen-gobierno/licencia_datos){target="_blank"}

-   [homepage of the v2 study and open mobility data project of the Ministry of Transport and Sustainable Mobility of Spain](https://www.transportes.gob.es/ministerio/proyectos-singulares/estudio-de-movilidad-con-big-data){target="_blank"}

::: callout-note
Kindly consult the documents above for any specific details on the methodology. The codebook here is only a simplified summary.
:::

To access the data we reference in this codebook, please follow these steps:

{{< include ../inst/vignette-include/install-package.qmd >}}

Load it as follows:

```{r}
library(spanishoddata)
```

Using the instructions below, set the data folder for the package to download the files into. You may need up to 400 GB to download all data and another 400 GB if you would like to convert the downloaded data into analysis ready format (a `DuckDB` database file, or a folder of `parquet` files). You can find more info on this conversion in the [Download and convert OD datasets](convert.html) vignette.

{{< include ../inst/vignette-include/setup-data-directory.qmd >}}

{{< include ../inst/vignette-include/overall-approach.qmd >}}

![The overview of package functions to get the data](../man/figures/package-functions-overview.svg){#fig-overall-flow width="78%"}

## 1. Spatial data with zoning boundaries

The boundary data is provided at three geographic levels: [`Distrtics`](#districts), [`Municipalities`](#municipalities), and [`Large Urban Areas`](#lua). It's important to note that these do not always align with the official Spanish census districts and municipalities. To comply with data protection regulations, certain aggregations had to be made to districts and municipalities".

### 1.1 `Districts` {#districts}

`Districts` correspond to official census districts in cities; however, in those with lower population density, they are grouped together. In rural areas, one district is often equal to a municipality, but municipalities with low population are combined into larger units to preserve privacy of individuals in the dataset. Therefore, there are 3792 'districts' compared to the 10494 official census districts on which they are based. There are also [NUTS3 statistical regions](https://ec.europa.eu/eurostat/web/nuts){target="_blank"} covering France (94 units) and Portugal (23 units). Therefore there is a total of 3909 zones in the `Districts` dataset.

```{r}
districts_v2 <- spod_get_zones("dist", ver = 2)
```

The `districts_v2` object is of class `sf` consisting of polygons.

Data structure:

| Variable Name | **Description** |
|------------------|------------------------------------------------------|
| `id` | District `id` assigned by the data provider. Matches with `id_origin`, `id_destination`, and `id` in district-level origin-destination and number of trips data. |
| `name` | Name of the district. |
| `population` | Number of individuals in the district according to [INE](https://ine.es/){target="_blank"}[^2]. |
| `census_sections` | Semicolon-separated list of census section identifiers that correspond to each district, classified by the Spanish Statistical Office (INE). |
| `census_districts` | Semicolon-separated list of census district identifiers corresponding to each district, as classified by the Spanish Statistical Office (INE). |
| `municipalities` | Semicolon-separated list of municipality identifiers corresponding to each district, as classified by the Spanish Statistical Office (INE). |
| `municipalities_mitma` | Semicolon-separated list of municipality identifiers as assigned by the data provider (MITMA). |
| `luas_mitma` | Semicolon-separated list of Large Urban Areas (LUAs) as assigned by the data provider, corresponding to each district. |
| `district_ids_in_v1` | Semicolon-separated district identifiers from v1 data corresponding to each district in v2. If no match exists, marked as `NA`. |
| `geometry` | Spatial geometry of each district stored as a `MULTIPOLYGON` object, projected in the ETRS89 / UTM zone 30N CRS with XY dimensions. |

[^2]: This is likely the population as of end of 2021 or start of 2022. Population for a few districts is missing. Instead of population, residence and overnight stays data may be used as a proxy with caution. Also, newer population figures may be obtained and joined with the provided zones using the reference tables that match the zones ids with official municipal and census district ids from INE.

### 1.2 `Municipalities` {#municipalities}

`Municipalities` are made up of official municipalities in those of a certain size; however, they have also been aggregated in cases of lower population density. As a result, there are 2618 municipalities compared to the 8,125 official municipalities on which they are based. There are also [NUTS3 statistical regions](https://ec.europa.eu/eurostat/web/nuts){target="_blank"} covering France (94 units) and Portugal (23 units). Therefore there is a total of 2735 zones in the `Districts` dataset.

```{r}
municipalities_v2 <- spod_get_zones("muni", ver = 2)
```

The resulting `municipalities_v2` object is type `sf` consisting of polygons.

Data structure:

| Variable Name | **Description** |
|------------------|------------------------------------------------------|
| `id` | District `id` assigned by the data provider. Matches with `id_origin`, `id_destination`, and `id` in district-level origin-destination and number of trips data. |
| `name` | Name of the district. |
| `population` | Number of individuals in the district according to [INE](https://ine.es/){target="_blank"}[^3]. |
| `census_sections` | Semicolon-separated list of census section identifiers that correspond to each district, classified by the Spanish Statistical Office (INE). |
| `census_districts` | Semicolon-separated list of census district identifiers corresponding to each district, as classified by the Spanish Statistical Office (INE). |
| `municipalities` | Semicolon-separated list of municipality identifiers corresponding to each district, as classified by the Spanish Statistical Office (INE). |
| `districts_mitma` | Semicolon-separated list of district identifiers as assigned by the data provider (MITMA). |
| `luas_mitma` | Semicolon-separated list of Large Urban Areas (LUAs) as assigned by the data provider, corresponding to each district. |
| `municipality_ids_in_v1` | Semicolon-separated district identifiers from v1 data corresponding to each district in v2. If no match exists, marked as `NA`. |
| `geometry` | Spatial geometry of each district stored as a `MULTIPOLYGON` object, projected in the ETRS89 / UTM zone 30N CRS with XY dimensions. |

[^3]: This is likely the population as of end of 2021 or start of 2022. Population for a few districts is missing. Instead of population, residence and overnight stays data may be used as a proxy with caution. Also, newer population figures may be obtained and joined with the provided zones using the reference tables that match the zones ids with official municipal and census district ids from INE.

### 1.3 `LUAs (Large Urban Areas)` {#luas}

`Large Urban Areas (LUAs)` has essentially the same spatial units as [`Municipalities`](#municipalities), but are not aggregated. Therefore, there are 2086 locations in the `LUAs` dataset. There are also [NUTS3 statistical regions](https://ec.europa.eu/eurostat/web/nuts){target="_blank"} covering France (94 units) and Portugal (23 units). Therefore there is a total of 2203 zones in the `LUAs` dataset.

```{r}
luas_v2 <- spod_get_zones("lua", ver = 2)
```

The resulting `luas_v2` object is type `sf` consisting of polygons.

Data structure:

| Variable Name | **Description** |
|------------------|------------------------------------------------------|
| `id` | District `id` assigned by the data provider. Matches with `id_origin`, `id_destination`, and `id` in district-level origin-destination and number of trips data. |
| `name` | Name of the district. |
| `population` | Number of individuals in the district according to [INE](https://ine.es/){target="_blank"}[^4]. |
| `census_sections` | Semicolon-separated list of census section identifiers that correspond to each district, classified by the Spanish Statistical Office (INE). |
| `census_districts` | Semicolon-separated list of census district identifiers corresponding to each district, as classified by the Spanish Statistical Office (INE). |
| `municipalities` | Semicolon-separated list of municipality identifiers corresponding to each district, as classified by the Spanish Statistical Office (INE). |
| `districts_mitma` | Semicolon-separated list of district identifiers as assigned by the data provider (MITMA). |
| `municipalities_mitma` | Semicolon-separated list of municipality identifiers as assigned by the data provider (MITMA). |
| `geometry` | Spatial geometry of each district stored as a `MULTIPOLYGON` object, projected in the ETRS89 / UTM zone 30N CRS with XY dimensions. |

[^4]: This is likely the population as of end of 2021 or start of 2022. Population for a few districts is missing. Instead of population, residence and overnight stays data may be used as a proxy with caution. Also, newer population figures may be obtained and joined with the provided zones using the reference tables that match the zones ids with official municipal and census district ids from INE.

## 2. Mobility data

All mobility data is referenced via `id_origin`, `id_destination`, or other location identifiers (mostly labelled as `id`) with the two sets of zones described above.

### 2.1. Origin-destination data {#od-data}

The origin-destination data contain the number of trips between `districts`, `municipalities`, or `large urban areas (LUAs)` in Spain for every hour of every day between 2022-02-01 and whichever currently available latest data (2024-06-30 at the time of writing). Each flow also has attributes such as the trip purpose (composed of the type of activity (`home`/`work_or_study`/`frequent_activity`/`infrequent_activity`) at both the origin and destination, but also age, sex, and income of each group of individuals traveling between the origin and destination), province of residence of individuals making this trip, distance covered while making the trip. See the detailed attributes below in a table.

Here are the variables you can find in the `district`, `municipality` and `large urban area` level data:

| **English Variable Name** | **Original Variable Name** | **Type** | **Description** |
|---------|---------|---------|---------------------------------------------|
| `date` | `fecha` | `Date` | The date of the recorded data, formatted as `YYYY-MM-DD`. |
| `hour` | `periodo` | `integer` | The time slot during which the trips occurred. |
| `id_origin` | `origen` | `factor` | The origin zone `id` of `district`, `municipality`, or `large urban area`. |
| `id_destination` | `destino` | `factor` | The destination zone `id` of `district`, `municipality`, or `large urban area`. |
| `distance` | `distancia` | `factor` | The distance range of the trip, categorized into specific intervals such as `0.5-2` (500 m to 2 km), `2-10` (2-10 km), `10-50` (10-50km), and `>50` (50 or more km). |
| `activity_origin` | `actividad_origen` | `factor` | The type of activity at the origin zone, recoded from `casa`, `trabajo_estudio`, `frecuente`, `no_frecuente` to `home`, `work_or_study`, `frequent_activity`, `infrequent_activity` respectively. |
| `activity_destination` | `actividad_destino` | `factor` | The type of activity at the destination zone, similarly recoded as for `activity_origin` above. |
| `study_possible_origin` | `estudio_origen_posible` | `logical` | `TRUE` if the activity at origin may be connected with study, and `FALSE` otherwise. |
| `study_possible_destination` | `estudio_destino_posible` | `logical` | `TRUE` if the activity at destination may be connected with study, and `FALSE` otherwise. |
| `residence_province_ine_code` | `residencia` | `factor` | The province code of residence of individuals making the trips in `n_trips`, encoded as province codes as classified by the Spanish Statistical Office (INE). |
| `residence_province_name` | Derived from `residencia` | `factor` | The full name of the residence province, derived from the province code above. |
| `income` | `renta` | `factor` | The income group of individuals making the trips in `n_trips`. Categorized into `<10`, `10-15`, and `>15` (thousands of euros per year). The income for each individual is assigned based on the mean census tract income per person (data source is [INE Household income distribution map](https://www.ine.es/dyngs/INEbase/en/operacion.htm?c=Estadistica_C&cid=1254736177088&menu=ultiDatos&idp=1254735976608)). |
| `age` | `edad` | `factor` | The age group of individuals making the trips in `n_trips`. Categorized into `0-25`, `25-45`, `45-65`, `65-100`, or `NA`. The data is partially imputed, for details see [this blogpost](https://www.nommon.es/blog/using-machine-learning-to-predict-sociodemographic-characteristics/) by Nommon. |
| `sex` | `sexo` | `factor` | The sex of individuals making the trips in `n_trips`. Categorized into `female`, `male`, or `NA`. The data is partially imputed, for details see [this blogpost](https://www.nommon.es/blog/using-machine-learning-to-predict-sociodemographic-characteristics/) by Nommon. |
| `n_trips` | `viajes` | `numeric` | The number of trips for that specific origin-destination pair and time slot. |
| `trips_total_length_km` | `viajes_km` | `numeric` | The total length of trips in kilometers, summing up all trips between the origin and destination zones. |
| `year` | `year` | `integer` | The year of the recorded data, extracted from the date. |
| `month` | `month` | `integer` | The month of the recorded data, extracted from the date. |
| `day` | `day` | `integer` | The day of the recorded data, extracted from the date. |

**Getting the data**

To access the data, use the `spod_get()` function. In this example we will use a short interval of dates:

```{r}
dates <- c(start = "2022-01-01", end = "2022-01-04")
od_dist <- spod_get(type = "od", zones = "dist", dates = dates)
od_muni <- spod_get(type = "od", zones = "muni", dates = dates)
```

The data for the specified dates will be automatically downloaded and cached in the `SPANISH_OD_DATA_DIR` directory. Existing files will not be re-downloaded.

**Working with the data**

The resulting objects `od_dist` and `od_muni` are of class `tbl_duckdb_connection`[^5]. Basically, you can treat these as regular `data.frame`s or `tibble`s. One important difference is that the data is not actually loaded into memory, because if you requested more dates, e.g. a whole month or a year, all that data would most likely not fit into your computer's memory. A `tbl_duckdb_connection` is mapped to the downloaded CSV files that are cached on disk and the data is only loaded in small chunks as needed at the time of computation. You can manipulate `od_dist` and `od_muni` using `{dplyr}` functions such as `select()`, `filter()`, `mutate()`, `group_by()`, `summarise()`, etc. In the end of any sequence of commands you will need to add `collect()` to execute the whole chain of data manipulations and load the results into memory in an R `data.frame`/`tibble` like so:

[^5]: For reference: this object also has classes: `tbl_dbi` ,`tbl_sql`, `tbl_lazy` ,and `tbl` .

```{r}
library(dplyr)
od_mean_trips_by_ses_over_the_4_days <- od_dist |>
  group_by(date, age, sex, income) |>
  summarise(
    n_trips = sum(n_trips, na.rm = TRUE),
    .groups = "drop"
  ) |>
  group_by(age, sex, income) |>
  summarise(
    daily_mean_n_trips = mean(n_trips, na.rm = TRUE),
    .groups = "drop"
  ) |>
  collect()
od_mean_trips_by_ses_over_the_4_days
```

```         
# A tibble: 39 × 4
   age   sex    income daily_mean_n_trips
   <fct> <fct>  <fct>               <dbl>
 1 NA    NA     <10              7002485.
 2 NA    NA     10-15           16551405.
 3 NA    NA     >15              2651481.
 4 0-25  NA     <10               539060.
 5 0-25  NA     10-15            1950892.
 6 0-25  NA     >15               401557.
 7 0-25  female <10              1484989.
 8 0-25  female 10-15            5357785.
 9 0-25  female >15              1764454.
10 0-25  male   <10              1558461.
# ℹ 29 more rows
# ℹ Use `print(n = ...)` to see more rows
```

In this example above, because the data is with hourly intervals within each day, we first summed the number of trips for each day by age, sex, and income groups. We then grouped the data again dropping the day variable and calculated the mean number of trips per day by age, sex, and income groups. The full data for all 4 days was probably never loaded into memory all at once. Rather the available memory of the computer was used up to its maximum limit to make that calculation happen, without ever exceeding the available memory limit. If you were doing the same opearation on 100 or even more days, it would work in the same way and would be possible even with limited memory. This is done transparently to the user with the help of [`DuckDB`](https://duckdb.org/){target="_blank"} (specifically, with [{duckdb} R package](https://r.duckdb.org/index.html){target="_blank"} @duckdb-r).

The same summary operation as provided in the example above can be done with the entire dataset for multiple years worth of data on a regular laptop with 8-16 GB memory. It will take a bit of time to complete, but it will be done. To speed things up, please also see the [vignette on converting the data](convert.qmd) into formats that will increase the analysis performance.

{{< include ../inst/vignette-include/csv-date-filter-note.qmd >}}

### 2.2. Population by trip count data {#ptc-data}

The population by trip count data shows the number of individuals in each district or municipality, categorized by the trips they make (0, 1, 2, or more than 2), age, and sex.

| **English Variable Name** | **Original Variable Name** | **Type** | **Description** |
|-----------------|-----------------|-----------------|----------------------|
| `date` | `fecha` | `Date` | The date of the recorded data, formatted as `YYYY-MM-DD`. |
| `id` | `distrito` | `factor` | The identifier of the `district` or `municipality` zone. |
| `age` | `edad` | `factor` | The age group of individuals making the trips in `n_trips`. Categorized into `0-25`, `25-45`, `45-65`, `65-100`, or `NA`. The data is partially imputed, for details see [this blogpost](https://www.nommon.es/blog/using-machine-learning-to-predict-sociodemographic-characteristics/) by Nommon. |
| `sex` | `sexo` | `factor` | The sex of individuals making the trips in `n_trips`. Categorized into `female`, `male`, or `NA`. The data is partially imputed, for details see [this blogpost](https://www.nommon.es/blog/using-machine-learning-to-predict-sociodemographic-characteristics/) by Nommon. |
| `n_trips` | `numero_viajes` | `factor` | The number of individuals who made trips, categorized by `0`, `1`, `2`, or `2+` trips. |
| `n_persons` | `personas` | `factor` | The number of persons making the trips from `district`, `municipality`, or `large urban area (LUA)` with zone `id`. |
| `year` | `year` | `integer` | The year of the recorded data, extracted from the date. |
| `month` | `month` | `integer` | The month of the recorded data, extracted from the date. |
| `day` | `day` | `integer` | The day of the recorded data, extracted from the date. |

**Getting the data**

To access it use `spod_get()` with `type` set to "number_of_trips", or just "nt".

```{r}
dates <- c(start = "2022-01-01", end = "2022-01-04")
nt_dist <- spod_get(type = "number_of_trips", zones = "dist", dates = dates)
```

Because this data is small, we can actually load it completely into memory:

```{r}
nt_dist_tbl <- nt_dist |> dplyr::collect()
```

### 2.3. Population by overnight stay data {#pos-data}

This dataset provides the number of people who spend the night in each location, also identifying their place of residence down to the census district level according to the [INE encoding](https://www.ine.es/en/daco/daco42/codmun/cod_provincia_en.htm){target="_blank"}.

Here are the variables you can find in the `district`, `municipality` and `large urban area` level data:

| **English Variable Name** | **Original Variable Name** | **Type** | **Description** |
|-----------------|-----------------|-----------------|----------------------|
| `date` | `fecha` | `Date` | The date of the recorded data, formatted as `YYYY-MM-DD`. |
| `id_residence` | `zona_residencia ` | `factor` | The identifier of the census district according to the [INE encoding](https://www.ine.es/en/daco/daco42/codmun/cod_provincia_en.htm){target="_blank"}. |
| `id_overnight_stay` | `zona_pernoctacion ` | `factor` | The identifier of the `district`, `municipality`, or `large urban area (LUA)` zone. |
| `n_persons` | `personas` | `factor` | The number of persons making the trips from `district`, `municipality`, or `large urban area` with zone `id`. |
| `year` | `year` | `integer` | The year of the recorded data, extracted from the date. |
| `month` | `month` | `integer` | The month of the recorded data, extracted from the date. |
| `day` | `day` | `integer` | The day of the recorded data, extracted from the date. |



**Getting the data**

To access it use `spod_get()` with `type` set to "number_of_trips", or just "nt".

```{r}
dates <- c(start = "2022-01-01", end = "2022-01-04")
os_dist <- spod_get(type = "overnight_stays", zones = "dist", dates = dates)
```

Because this data is small, we can actually load it completely into memory:

```{r}
os_dist_tbl <- os_dist |> dplyr::collect()
```


## Advanced use

For more advanced use, especially for analysing longer periods (months or even years), please see [Download and convert mobility datasets](convert.html).

# Codebook and cookbook for v2 (2022 onwards) Spanish mobility data


You can view this vignette any time by running:

``` r
spanishoddata::spod_codebook(ver = 2)
```

The mobility data v2 (2022 onwards) was originally released by the
Ministerio de Transportes, Movilidad y Agenda Urbana (MITMA) , now
[Ministerio de Transportes y Movilidad Sostenible
(MITMS)](https://www.transportes.gob.es/) (MITMS 2024).

The dataset is produced by
<a href="https://www.nommon.es/" target="_blank">Nommon</a> using the
raw data from
<a href="https://www.orange.es/" target="_blank">Orange España</a>. Even
though the raw data is only from one mobile phone operator, the
resulting flows and other counts of number of individuals in the data
set are already resampled to be representative of the total population
of Spain (see details in the official methodology).

The tables in the data set provide hourly flows between zones across
Spain for every day of the observation period (2022-01-01 onwards), the
number of individuals making trips for each zone, the number of
individuals spending the nights in one location while regularly residing
in a different location, and many more advanced datasets. This document
will introduce you to the available data and provide brief code snippets
on how to access it using the `{spanishoddata}` R package.

Compared to the [v1 data
(2020-2021)](v1-2020-2021-mitma-data-codebook.qmd) this dataset has many
additional variables, such as age, sex, and income, has better spatial
resolution (the zones are more spatially granular) and covers a
continuous period[1] (2022-01-01 onward), rather than only a limited
period (in v1 - 2020-02-14 to 2021-05-09).

> **Warning**
>
> Due to mobile network outages, the data on certain dates is missing.
> Kindly keep this in mind when calculating mean monthly or weekly
> flows.
>
>   
>
> Please check the <a
> href="https://www.transportes.gob.es/ministerio/proyectos-singulares/estudios-de-movilidad-con-big-data/opendata-movilidad"
> target="_blank">original data page</a> for currently known missing
> dates. At the time of writing, the following dates are missing: 26,
> 27, 30, 31 October; 1, 2 and 3 November 2023; 4, 18, 19 April 2024, 10
> and 11 November 2024. You can use `spod_get_valid_dates()` function to
> get all available dates.

In v2 data, the datasets comparable to the [v1 (2020-2021)
data](v1-2020-2021-mitma-data-codebook.qmd) (hourly daily origin
destination matrices and number of trips at each location) are called <a
href="https://www.transportes.gob.es/ministerio/proyectos-singulares/estudios-de-movilidad-con-big-data/estudio-basico-diario"
target="_blank">“basic studies” (“estudios basicos”)</a>, however there
are even more advanced datasets that are already available or will be
made available soon: <a
href="https://www.transportes.gob.es/ministerio/proyectos-singulares/estudios-de-movilidad-con-big-data/estudios-completos"
target="_blank">“complete studies” (“estudios completos”)</a> and <a
href="https://www.transportes.gob.es/ministerio/proyectos-singulares/estudios-de-movilidad-con-big-data/rutas-por-carretera"
target="_blank">“road routes” (“rutas de carretera”)</a>. At the moment,
the `{spanishoddata}` package only provides the interface to the “basic
studies” datasets, but support for “complete studies” and “road routes”
will be added in the future.

Key sources for this codebook/cookbook include:

-   <a
    href="https://www.transportes.gob.es/recursos_mfom/paginabasica/recursos/a3_informe_metodologico_estudio_movilidad_mitms_v8.pdf"
    target="_blank">original data collection methodology and codebook in
    Spanish</a> + <a
    href="https://rOpenSpain.github.io/spanishoddata/codebooks/a3_informe_metodologico_estudio_movilidad_mitms_v8_en.pdf"
    target="_blank">automatically translated English version of methodology
    and codebook</a>

-   <a
    href="https://www.transportes.gob.es/recursos_mfom/paginabasica/recursos/20241024_validaciones_estudios_basicos_bigdata_v1.0.pdf"
    target="_blank">original quality assessment</a> + <a
    href="https://rOpenSpain.github.io/spanishoddata/codebooks/20241024_validaciones_estudios_basicos_bigdata_v1.0_en.pdf"
    target="_blank">automatically translated English version of quality
    assessment</a>

-   <a
    href="https://www.transportes.gob.es/ministerio/proyectos-singulares/estudios-de-movilidad-con-big-data/opendata-movilidad"
    target="_blank">original data download page</a>

-   <a
    href="https://www.transportes.gob.es/el-ministerio/buen-gobierno/licencia_datos"
    target="_blank">original data license</a>

-   <a
    href="https://www.transportes.gob.es/ministerio/proyectos-singulares/estudio-de-movilidad-con-big-data"
    target="_blank">homepage of the v2 study and open mobility data project
    of the Ministry of Transport and Sustainable Mobility of Spain</a>

> **Note**
>
> Kindly consult the documents above for any specific details on the
> methodology. The codebook here is only a simplified summary.

To access the data we reference in this codebook, please follow these
steps:

## Install the package

The package is not yet available on CRAN.

You can install the latest version of the package from rOpenSpain R
universe:

``` r
install.packages("spanishoddata",
  repos = c("https://ropenspain.r-universe.dev",
    "https://cloud.r-project.org"))
```

<details>
<summary>
Alternative installation and developemnt
</summary>

Alternative way to install the package from GitHub:

``` r
if (!require("remotes")) install.packages("remotes")

remotes::install_github("rOpenSpain/spanishoddata",
  force = TRUE, dependencies = TRUE)
```

**For Developers**

To load the package locally, clone it and navigate to the root of the
package in the terminal, e.g. with the following:

``` bash
gh repo clone rOpenSpain/spanishoddata
code spanishoddata
# with rstudio:
rstudio spanishoddata/spanishoddata.Rproj
```

Then run the following command from the R console:

``` r
devtools::load_all()
```

</details>

Load it as follows:

``` r
library(spanishoddata)
```

Using the instructions below, set the data folder for the package to
download the files into. You may need up to 400 GB to download all data
and another 400 GB if you would like to convert the downloaded data into
analysis ready format (a `DuckDB` database file, or a folder of
`parquet` files). You can find more info on this conversion in the
[Download and convert OD datasets](convert.html) vignette.

## Set the data directory

Choose where `{spanishoddata}` should download (and convert) the data by
setting the data directory following command:

``` r
spod_set_data_dir(data_dir = "~/spanish_od_data")
```

The function above will also ensure that the directory is created and
that you have sufficient permissions to write to it.

<details>
<summary>
Setting data directory for advanced users
</summary>

You can also set the data directory with an environment variable:

``` r
Sys.setenv(SPANISH_OD_DATA_DIR = "~/spanish_od_data")
```

The package will create this directory if it does not exist on the first
run of any function that downloads the data.

To permanently set the directory for all projects, you can specify the
data directory globally by setting the `SPANISH_OD_DATA_DIR` environment
variable, e.g. with the following command:

``` r
usethis::edit_r_environ()
# Then set the data directory globally, by typing this line in the file:
```

    SPANISH_OD_DATA_DIR = "~/spanish_od_data"

You can also set the data directory locally, just for the current
project. Set the ‘envar’ in the working directory by editing `.Renviron`
file in the root of the project:

``` r
file.edit(".Renviron")
```

</details>

# Overall approach to accessing the data

If you only need flows data aggregated by day at municipal level, you
can use the `spod_quick_get_od()` function. This will download the data
directly from the web API and let you analyse it in-memory. More on this
in the [Quickly get daily
data](https://ropenspain.github.io/spanishoddata/articles/quick-get.html)
vignette.

If you only want to analyse the data for a few days, you can use the
`spod_get()` function. It will download the raw data in CSV format and
let you analyse it in-memory. This is what we cover in the steps on this
page.

If you need longer periods (several months or years), you should use the
`spod_convert()` and `spod_connect()` functions, which will convert the
data into special format which is much faster for analysis, for this see
the [Download and convert OD
datasets](https://ropenspain.github.io/spanishoddata/articles/convert.html)
vignette. `spod_get_zones()` will give you spatial data with zones that
can be matched with the origin-destination flows from the functions
above using zones ’id’s. Please see a simple example below, and also
consult the vignettes with detailed data description and instructions in
the package vignettes with `spod_codebook(ver = 1)` and
`spod_codebook(ver = 2)`, or simply visit the package website at
<https://ropenspain.github.io/spanishoddata/>. The
<a href="#fig-overall-flow" class="quarto-xref">Figure 1</a> presents
the overall approach to accessing the data in the `spanishoddata`
package.

<img src="../man/figures/package-functions-overview.svg"
style="width:78.0%" />

# 1. Spatial data with zoning boundaries

The boundary data is provided at three geographic levels:
[`Distrtics`](#districts), [`Municipalities`](#municipalities), and
[`Large Urban Areas`](#lua). It’s important to note that these do not
always align with the official Spanish census districts and
municipalities. To comply with data protection regulations, certain
aggregations had to be made to districts and municipalities”.

## 1.1 `Districts`

`Districts` correspond to official census districts in cities; however,
in those with lower population density, they are grouped together. In
rural areas, one district is often equal to a municipality, but
municipalities with low population are combined into larger units to
preserve privacy of individuals in the dataset. Therefore, there are
3792 ‘districts’ compared to the 10494 official census districts on
which they are based. There are also
<a href="https://ec.europa.eu/eurostat/web/nuts" target="_blank">NUTS3
statistical regions</a> covering France (94 units) and Portugal (23
units). Therefore there is a total of 3909 zones in the `Districts`
dataset.

``` r
districts_v2 <- spod_get_zones("dist", ver = 2)
```

The `districts_v2` object is of class `sf` consisting of polygons.

Data structure:

<table>
<colgroup>
<col style="width: 25%" />
<col style="width: 75%" />
</colgroup>
<thead>
<tr>
<th>Variable Name</th>
<th><strong>Description</strong></th>
</tr>
</thead>
<tbody>
<tr>
<td><code>id</code></td>
<td>District <code>id</code> assigned by the data provider. Matches with
<code>id_origin</code>, <code>id_destination</code>, and <code>id</code>
in district-level origin-destination and number of trips data.</td>
</tr>
<tr>
<td><code>name</code></td>
<td>Name of the district.</td>
</tr>
<tr>
<td><code>population</code></td>
<td>Number of individuals in the district according to <a
href="https://ine.es/" target="_blank">INE</a><a href="#fn1"
class="footnote-ref" id="fnref1"
role="doc-noteref"><sup>1</sup></a>.</td>
</tr>
<tr>
<td><code>census_sections</code></td>
<td>Semicolon-separated list of census section identifiers that
correspond to each district, classified by the Spanish Statistical
Office (INE).</td>
</tr>
<tr>
<td><code>census_districts</code></td>
<td>Semicolon-separated list of census district identifiers
corresponding to each district, as classified by the Spanish Statistical
Office (INE).</td>
</tr>
<tr>
<td><code>municipalities</code></td>
<td>Semicolon-separated list of municipality identifiers corresponding
to each district, as classified by the Spanish Statistical Office
(INE).</td>
</tr>
<tr>
<td><code>municipalities_mitma</code></td>
<td>Semicolon-separated list of municipality identifiers as assigned by
the data provider (MITMA).</td>
</tr>
<tr>
<td><code>luas_mitma</code></td>
<td>Semicolon-separated list of Large Urban Areas (LUAs) as assigned by
the data provider, corresponding to each district.</td>
</tr>
<tr>
<td><code>district_ids_in_v1</code></td>
<td>Semicolon-separated district identifiers from v1 data corresponding
to each district in v2. If no match exists, marked as
<code>NA</code>.</td>
</tr>
<tr>
<td><code>geometry</code></td>
<td>Spatial geometry of each district stored as a
<code>MULTIPOLYGON</code> object, projected in the ETRS89 / UTM zone 30N
CRS with XY dimensions.</td>
</tr>
</tbody>
</table>
<section id="footnotes" class="footnotes footnotes-end-of-document"
role="doc-endnotes">
<hr />
<ol>
<li id="fn1"><p>This is likely the population as of end of 2021 or start
of 2022. Population for a few districts is missing. Instead of
population, residence and overnight stays data may be used as a proxy
with caution. Also, newer population figures may be obtained and joined
with the provided zones using the reference tables that match the zones
ids with official municipal and census district ids from INE.<a
href="#fnref1" class="footnote-back" role="doc-backlink">↩︎</a></p></li>
</ol>
</section>

## 1.2 `Municipalities`

`Municipalities` are made up of official municipalities in those of a
certain size; however, they have also been aggregated in cases of lower
population density. As a result, there are 2618 municipalities compared
to the 8,125 official municipalities on which they are based. There are
also
<a href="https://ec.europa.eu/eurostat/web/nuts" target="_blank">NUTS3
statistical regions</a> covering France (94 units) and Portugal (23
units). Therefore there is a total of 2735 zones in the `Districts`
dataset.

``` r
municipalities_v2 <- spod_get_zones("muni", ver = 2)
```

The resulting `municipalities_v2` object is type `sf` consisting of
polygons.

Data structure:

<table>
<colgroup>
<col style="width: 25%" />
<col style="width: 75%" />
</colgroup>
<thead>
<tr>
<th>Variable Name</th>
<th><strong>Description</strong></th>
</tr>
</thead>
<tbody>
<tr>
<td><code>id</code></td>
<td>District <code>id</code> assigned by the data provider. Matches with
<code>id_origin</code>, <code>id_destination</code>, and <code>id</code>
in district-level origin-destination and number of trips data.</td>
</tr>
<tr>
<td><code>name</code></td>
<td>Name of the district.</td>
</tr>
<tr>
<td><code>population</code></td>
<td>Number of individuals in the district according to <a
href="https://ine.es/" target="_blank">INE</a><a href="#fn1"
class="footnote-ref" id="fnref1"
role="doc-noteref"><sup>1</sup></a>.</td>
</tr>
<tr>
<td><code>census_sections</code></td>
<td>Semicolon-separated list of census section identifiers that
correspond to each district, classified by the Spanish Statistical
Office (INE).</td>
</tr>
<tr>
<td><code>census_districts</code></td>
<td>Semicolon-separated list of census district identifiers
corresponding to each district, as classified by the Spanish Statistical
Office (INE).</td>
</tr>
<tr>
<td><code>municipalities</code></td>
<td>Semicolon-separated list of municipality identifiers corresponding
to each district, as classified by the Spanish Statistical Office
(INE).</td>
</tr>
<tr>
<td><code>districts_mitma</code></td>
<td>Semicolon-separated list of district identifiers as assigned by the
data provider (MITMA).</td>
</tr>
<tr>
<td><code>luas_mitma</code></td>
<td>Semicolon-separated list of Large Urban Areas (LUAs) as assigned by
the data provider, corresponding to each district.</td>
</tr>
<tr>
<td><code>municipality_ids_in_v1</code></td>
<td>Semicolon-separated district identifiers from v1 data corresponding
to each district in v2. If no match exists, marked as
<code>NA</code>.</td>
</tr>
<tr>
<td><code>geometry</code></td>
<td>Spatial geometry of each district stored as a
<code>MULTIPOLYGON</code> object, projected in the ETRS89 / UTM zone 30N
CRS with XY dimensions.</td>
</tr>
</tbody>
</table>
<section id="footnotes" class="footnotes footnotes-end-of-document"
role="doc-endnotes">
<hr />
<ol>
<li id="fn1"><p>This is likely the population as of end of 2021 or start
of 2022. Population for a few districts is missing. Instead of
population, residence and overnight stays data may be used as a proxy
with caution. Also, newer population figures may be obtained and joined
with the provided zones using the reference tables that match the zones
ids with official municipal and census district ids from INE.<a
href="#fnref1" class="footnote-back" role="doc-backlink">↩︎</a></p></li>
</ol>
</section>

## 1.3 `LUAs (Large Urban Areas)`

`Large Urban Areas (LUAs)` has essentially the same spatial units as
[`Municipalities`](#municipalities), but are not aggregated. Therefore,
there are 2086 locations in the `LUAs` dataset. There are also
<a href="https://ec.europa.eu/eurostat/web/nuts" target="_blank">NUTS3
statistical regions</a> covering France (94 units) and Portugal (23
units). Therefore there is a total of 2203 zones in the `LUAs` dataset.

``` r
luas_v2 <- spod_get_zones("lua", ver = 2)
```

The resulting `luas_v2` object is type `sf` consisting of polygons.

Data structure:

<table>
<colgroup>
<col style="width: 25%" />
<col style="width: 75%" />
</colgroup>
<thead>
<tr>
<th>Variable Name</th>
<th><strong>Description</strong></th>
</tr>
</thead>
<tbody>
<tr>
<td><code>id</code></td>
<td>District <code>id</code> assigned by the data provider. Matches with
<code>id_origin</code>, <code>id_destination</code>, and <code>id</code>
in district-level origin-destination and number of trips data.</td>
</tr>
<tr>
<td><code>name</code></td>
<td>Name of the district.</td>
</tr>
<tr>
<td><code>population</code></td>
<td>Number of individuals in the district according to <a
href="https://ine.es/" target="_blank">INE</a><a href="#fn1"
class="footnote-ref" id="fnref1"
role="doc-noteref"><sup>1</sup></a>.</td>
</tr>
<tr>
<td><code>census_sections</code></td>
<td>Semicolon-separated list of census section identifiers that
correspond to each district, classified by the Spanish Statistical
Office (INE).</td>
</tr>
<tr>
<td><code>census_districts</code></td>
<td>Semicolon-separated list of census district identifiers
corresponding to each district, as classified by the Spanish Statistical
Office (INE).</td>
</tr>
<tr>
<td><code>municipalities</code></td>
<td>Semicolon-separated list of municipality identifiers corresponding
to each district, as classified by the Spanish Statistical Office
(INE).</td>
</tr>
<tr>
<td><code>districts_mitma</code></td>
<td>Semicolon-separated list of district identifiers as assigned by the
data provider (MITMA).</td>
</tr>
<tr>
<td><code>municipalities_mitma</code></td>
<td>Semicolon-separated list of municipality identifiers as assigned by
the data provider (MITMA).</td>
</tr>
<tr>
<td><code>geometry</code></td>
<td>Spatial geometry of each district stored as a
<code>MULTIPOLYGON</code> object, projected in the ETRS89 / UTM zone 30N
CRS with XY dimensions.</td>
</tr>
</tbody>
</table>
<section id="footnotes" class="footnotes footnotes-end-of-document"
role="doc-endnotes">
<hr />
<ol>
<li id="fn1"><p>This is likely the population as of end of 2021 or start
of 2022. Population for a few districts is missing. Instead of
population, residence and overnight stays data may be used as a proxy
with caution. Also, newer population figures may be obtained and joined
with the provided zones using the reference tables that match the zones
ids with official municipal and census district ids from INE.<a
href="#fnref1" class="footnote-back" role="doc-backlink">↩︎</a></p></li>
</ol>
</section>

# 2. Mobility data

All mobility data is referenced via `id_origin`, `id_destination`, or
other location identifiers (mostly labelled as `id`) with the two sets
of zones described above.

## 2.1. Origin-destination data

The origin-destination data contain the number of trips between
`districts`, `municipalities`, or `large urban areas (LUAs)` in Spain
for every hour of every day between 2022-02-01 and whichever currently
available latest data (2024-06-30 at the time of writing). Each flow
also has attributes such as the trip purpose (composed of the type of
activity
(`home`/`work_or_study`/`frequent_activity`/`infrequent_activity`) at
both the origin and destination, but also age, sex, and income of each
group of individuals traveling between the origin and destination),
province of residence of individuals making this trip, distance covered
while making the trip. See the detailed attributes below in a table.

Here are the variables you can find in the `district`, `municipality`
and `large urban area` level data:

<table>
<colgroup>
<col style="width: 12%" />
<col style="width: 12%" />
<col style="width: 12%" />
<col style="width: 62%" />
</colgroup>
<thead>
<tr>
<th><strong>English Variable Name</strong></th>
<th><strong>Original Variable Name</strong></th>
<th><strong>Type</strong></th>
<th><strong>Description</strong></th>
</tr>
</thead>
<tbody>
<tr>
<td><code>date</code></td>
<td><code>fecha</code></td>
<td><code>Date</code></td>
<td>The date of the recorded data, formatted as
<code>YYYY-MM-DD</code>.</td>
</tr>
<tr>
<td><code>time_slot</code></td>
<td><code>periodo</code></td>
<td><code>integer</code></td>
<td>The time slot during which the trips occurred.</td>
</tr>
<tr>
<td><code>id_origin</code></td>
<td><code>origen</code></td>
<td><code>factor</code></td>
<td>The origin zone <code>id</code> of <code>district</code>,
<code>municipality</code>, or <code>large urban area</code>.</td>
</tr>
<tr>
<td><code>id_destination</code></td>
<td><code>destino</code></td>
<td><code>factor</code></td>
<td>The destination zone <code>id</code> of <code>district</code>,
<code>municipality</code>, or <code>large urban area</code>.</td>
</tr>
<tr>
<td><code>distance</code></td>
<td><code>distancia</code></td>
<td><code>factor</code></td>
<td>The distance range of the trip, categorized into specific intervals
such as <code>0.5-2</code> (500 m to 2 km), <code>2-10</code> (2-10 km),
<code>10-50</code> (10-50km), and <code>&gt;50</code> (50 or more
km).</td>
</tr>
<tr>
<td><code>activity_origin</code></td>
<td><code>actividad_origen</code></td>
<td><code>factor</code></td>
<td>The type of activity at the origin zone, recoded from
<code>casa</code>, <code>trabajo_estudio</code>, <code>frecuente</code>,
<code>no_frecuente</code> to <code>home</code>,
<code>work_or_study</code>, <code>frequent_activity</code>,
<code>infrequent_activity</code> respectively.</td>
</tr>
<tr>
<td><code>activity_destination</code></td>
<td><code>actividad_destino</code></td>
<td><code>factor</code></td>
<td>The type of activity at the destination zone, similarly recoded as
for <code>activity_origin</code> above.</td>
</tr>
<tr>
<td><code>study_possible_origin</code></td>
<td><code>estudio_origen_posible</code></td>
<td><code>logical</code></td>
<td><code>TRUE</code> if the activity at origin may be connected with
study, and <code>FALSE</code> otherwise.</td>
</tr>
<tr>
<td><code>study_possible_destination</code></td>
<td><code>estudio_destino_posible</code></td>
<td><code>logical</code></td>
<td><code>TRUE</code> if the activity at destination may be connected
with study, and <code>FALSE</code> otherwise.</td>
</tr>
<tr>
<td><code>residence_province_ine_code</code></td>
<td><code>residencia</code></td>
<td><code>factor</code></td>
<td>The province code of residence of individuals making the trips in
<code>n_trips</code>, encoded as province codes as classified by the
Spanish Statistical Office (INE).</td>
</tr>
<tr>
<td><code>residence_province_name</code></td>
<td>Derived from <code>residencia</code></td>
<td><code>factor</code></td>
<td>The full name of the residence province, derived from the province
code above.</td>
</tr>
<tr>
<td><code>income</code></td>
<td><code>renta</code></td>
<td><code>factor</code></td>
<td>The income group of individuals making the trips in
<code>n_trips</code>. Categorized into <code>&lt;10</code>,
<code>10-15</code>, and <code>&gt;15</code> (thousands of euros per
year). The income for each individual is assigned based on the mean
census tract income per person (data source is <a
href="https://www.ine.es/dyngs/INEbase/en/operacion.htm?c=Estadistica_C&amp;cid=1254736177088&amp;menu=ultiDatos&amp;idp=1254735976608">INE
Household income distribution map</a>).</td>
</tr>
<tr>
<td><code>age</code></td>
<td><code>edad</code></td>
<td><code>factor</code></td>
<td>The age group of individuals making the trips in
<code>n_trips</code>. Categorized into <code>0-25</code>,
<code>25-45</code>, <code>45-65</code>, <code>65-100</code>, or
<code>NA</code>. The data is partially imputed, for details see <a
href="https://www.nommon.es/blog/using-machine-learning-to-predict-sociodemographic-characteristics/">this
blogpost</a> by Nommon.</td>
</tr>
<tr>
<td><code>sex</code></td>
<td><code>sexo</code></td>
<td><code>factor</code></td>
<td>The sex of individuals making the trips in <code>n_trips</code>.
Categorized into <code>female</code>, <code>male</code>, or
<code>NA</code>. The data is partially imputed, for details see <a
href="https://www.nommon.es/blog/using-machine-learning-to-predict-sociodemographic-characteristics/">this
blogpost</a> by Nommon.</td>
</tr>
<tr>
<td><code>n_trips</code></td>
<td><code>viajes</code></td>
<td><code>numeric</code></td>
<td>The number of trips for that specific origin-destination pair and
time slot.</td>
</tr>
<tr>
<td><code>trips_total_length_km</code></td>
<td><code>viajes_km</code></td>
<td><code>numeric</code></td>
<td>The total length of trips in kilometers, summing up all trips
between the origin and destination zones.</td>
</tr>
<tr>
<td><code>year</code></td>
<td><code>year</code></td>
<td><code>integer</code></td>
<td>The year of the recorded data, extracted from the date.</td>
</tr>
<tr>
<td><code>month</code></td>
<td><code>month</code></td>
<td><code>integer</code></td>
<td>The month of the recorded data, extracted from the date.</td>
</tr>
<tr>
<td><code>day</code></td>
<td><code>day</code></td>
<td><code>integer</code></td>
<td>The day of the recorded data, extracted from the date.</td>
</tr>
</tbody>
</table>

**Getting the data**

To access the data, use the `spod_get()` function. In this example we
will use a short interval of dates:

``` r
dates <- c(start = "2022-01-01", end = "2022-01-04")
od_dist <- spod_get(type = "od", zones = "dist", dates = dates)
od_muni <- spod_get(type = "od", zones = "muni", dates = dates)
```

The data for the specified dates will be automatically downloaded and
cached in the `SPANISH_OD_DATA_DIR` directory. Existing files will not
be re-downloaded.

**Working with the data**

The resulting objects `od_dist` and `od_muni` are of class
`tbl_duckdb_connection`[2]. Basically, you can treat these as regular
`data.frame`s or `tibble`s. One important difference is that the data is
not actually loaded into memory, because if you requested more dates,
e.g. a whole month or a year, all that data would most likely not fit
into your computer’s memory. A `tbl_duckdb_connection` is mapped to the
downloaded CSV files that are cached on disk and the data is only loaded
in small chunks as needed at the time of computation. You can manipulate
`od_dist` and `od_muni` using `{dplyr}` functions such as `select()`,
`filter()`, `mutate()`, `group_by()`, `summarise()`, etc. In the end of
any sequence of commands you will need to add `collect()` to execute the
whole chain of data manipulations and load the results into memory in an
R `data.frame`/`tibble` like so:

``` r
library(dplyr)
od_mean_trips_by_ses_over_the_4_days <- od_dist |>
  group_by(date, age, sex, income) |>
  summarise(
    n_trips = sum(n_trips, na.rm = TRUE),
    .groups = "drop") |> 
  group_by(age, sex, income) |>
  summarise(
    daily_mean_n_trips = mean(n_trips, na.rm = TRUE),
    .groups = "drop") |> 
  collect()
od_mean_trips_by_ses_over_the_4_days
```

    # A tibble: 39 × 4
       age   sex    income daily_mean_n_trips
       <fct> <fct>  <fct>               <dbl>
     1 NA    NA     <10              7002485.
     2 NA    NA     10-15           16551405.
     3 NA    NA     >15              2651481.
     4 0-25  NA     <10               539060.
     5 0-25  NA     10-15            1950892.
     6 0-25  NA     >15               401557.
     7 0-25  female <10              1484989.
     8 0-25  female 10-15            5357785.
     9 0-25  female >15              1764454.
    10 0-25  male   <10              1558461.
    # ℹ 29 more rows
    # ℹ Use `print(n = ...)` to see more rows

In this example above, becaus the data is with hourly intervals within
each day, we first summed the number of trips for each day by age, sex,
and income groups. We then grouped the data again dropping the day
variable and calculated the mean number of trips per day by age, sex,
and income groups. The full data for all 4 days was probably never
loaded into memory all at once. Rather the available memory of the
computer was used up to its maximum limit to make that calculation
happen, without ever exceeding the available memory limit. If you were
doing the same opearation on 100 or even more days, it would work in the
same way and would be possible even with limited memory. This is done
transparantly to the user with the help of
<a href="https://duckdb.org/" target="_blank"><code>DuckDB</code></a>
(specifically, with
<a href="https://r.duckdb.org/index.html" target="_blank">{duckdb} R
package</a> Mühleisen and Raasveldt (2024)).

The same summary operation as provided in the example above can be done
with the entire dataset for multiple years worth of data on a regular
laptop with 8-16 GB memory. It will take a bit of time to complete, but
it will be done. To speed things up, please also see the [vignette on
converting the data](convert.qmd) into formats that will increase the
analsysis performance.

> **Note**
>
> As long as you use a table connection object created with `spod_get()`
> function, it is much quicker to filter the dates by the `year`,
> `month` and `day` variables, rather than by the `date` variable. This
> is because the data for each day is in a separate CSV file located in
> folders that look like `year=2020/month=2/day=14`. So when filtering
> by the `date` field, R will have to scan all CSV files comparing the
> specified date with what is stored inside each CSV file. However, if
> you query by `year`, `month` and `day` variables, R only needs to
> check these against the path to each CSV file, which is much quicker.
> This caveat is only relevant as long as you use `spod_get()` . If you
> convert (see the [relevant vignette](convert.qmd)) the downloaded data
> to a format that it optimized for quick analysis, you can use
> whichever field you want, it should not affect the performance.

## 2.2. Number of trips data

For each location, the “number of trips” data provides the number of
individuals who spent the night there, with breakdown by the number of
trips made, age, and sex.

<table style="width:100%;">
<colgroup>
<col style="width: 23%" />
<col style="width: 23%" />
<col style="width: 23%" />
<col style="width: 30%" />
</colgroup>
<thead>
<tr>
<th><strong>English Variable Name</strong></th>
<th><strong>Original Variable Name</strong></th>
<th><strong>Type</strong></th>
<th><strong>Description</strong></th>
</tr>
</thead>
<tbody>
<tr>
<td><code>date</code></td>
<td><code>fecha</code></td>
<td><code>Date</code></td>
<td>The date of the recorded data, formatted as
<code>YYYY-MM-DD</code>.</td>
</tr>
<tr>
<td><code>id</code></td>
<td><code>distrito</code></td>
<td><code>factor</code></td>
<td>The identifier of the <code>district</code> or
<code>municipality</code> zone.</td>
</tr>
<tr>
<td><code>age</code></td>
<td><code>edad</code></td>
<td><code>factor</code></td>
<td>The age group of individuals making the trips in
<code>n_trips</code>. Categorized into <code>0-25</code>,
<code>25-45</code>, <code>45-65</code>, <code>65-100</code>, or
<code>NA</code>. The data is partially imputed, for details see <a
href="https://www.nommon.es/blog/using-machine-learning-to-predict-sociodemographic-characteristics/">this
blogpost</a> by Nommon.</td>
</tr>
<tr>
<td><code>sex</code></td>
<td><code>sexo</code></td>
<td><code>factor</code></td>
<td>The sex of individuals making the trips in <code>n_trips</code>.
Categorized into <code>female</code>, <code>male</code>, or
<code>NA</code>. The data is partially imputed, for details see <a
href="https://www.nommon.es/blog/using-machine-learning-to-predict-sociodemographic-characteristics/">this
blogpost</a> by Nommon.</td>
</tr>
<tr>
<td><code>n_trips</code></td>
<td><code>numero_viajes</code></td>
<td><code>factor</code></td>
<td>The number of individuals who made trips, categorized by
<code>0</code>, <code>1</code>, <code>2</code>, or <code>2+</code>
trips.</td>
</tr>
<tr>
<td><code>n_persons</code></td>
<td><code>personas</code></td>
<td><code>factor</code></td>
<td>The number of persons making the trips from <code>district</code>,
<code>municipality</code>, or <code>large urban area (LUA)</code> with
zone <code>id</code>.</td>
</tr>
<tr>
<td><code>year</code></td>
<td><code>year</code></td>
<td><code>integer</code></td>
<td>The year of the recorded data, extracted from the date.</td>
</tr>
<tr>
<td><code>month</code></td>
<td><code>month</code></td>
<td><code>integer</code></td>
<td>The month of the recorded data, extracted from the date.</td>
</tr>
<tr>
<td><code>day</code></td>
<td><code>day</code></td>
<td><code>integer</code></td>
<td>The day of the recorded data, extracted from the date.</td>
</tr>
</tbody>
</table>

**Getting the data**

To access it use `spod_get()` with `type` set to “number_of_trips”, or
just “nt”.

``` r
dates <- c(start = "2022-01-01", end = "2022-01-04")
nt_dist <- spod_get(type = "number_of_trips", zones = "dist", dates = dates)
```

Because this data is small, we can actually load it completely into
memory:

``` r
nt_dist_tbl <- nt_dist |> dplyr::collect()
```

## 2.3. Overnight stays

This dataset provides the number of people who spend the night in each
location, also identifying their place of residence down to the census
district level according to the <a
href="https://www.ine.es/ss/Satellite?c=Page&amp;p=1259952026632&amp;pagename=ProductosYServicios%2FPYSLayout&amp;cid=1259952026632&amp;L=1"
target="_blank">INE encoding</a>.

Here are the variables you can find in the `district`, `municipality`
and `large urban area` level data:

<table style="width:100%;">
<colgroup>
<col style="width: 23%" />
<col style="width: 23%" />
<col style="width: 23%" />
<col style="width: 30%" />
</colgroup>
<thead>
<tr>
<th><strong>English Variable Name</strong></th>
<th><strong>Original Variable Name</strong></th>
<th><strong>Type</strong></th>
<th><strong>Description</strong></th>
</tr>
</thead>
<tbody>
<tr>
<td><code>date</code></td>
<td><code>fecha</code></td>
<td><code>Date</code></td>
<td>The date of the recorded data, formatted as
<code>YYYY-MM-DD</code>.</td>
</tr>
<tr>
<td><code>id_residence</code></td>
<td><code>zona_residencia</code></td>
<td><code>factor</code></td>
<td>The identifier of the census district according to the <a
href="https://www.ine.es/ss/Satellite?c=Page&amp;p=1259952026632&amp;pagename=ProductosYServicios%2FPYSLayout&amp;cid=1259952026632&amp;L=1"
target="_blank">INE encoding</a>.</td>
</tr>
<tr>
<td><code>id_overnight_stay</code></td>
<td><code>zona_pernoctacion</code></td>
<td><code>factor</code></td>
<td>The identifier of the <code>district</code>,
<code>municipality</code>, or <code>large urban area (LUA)</code>
zone.</td>
</tr>
<tr>
<td><code>n_persons</code></td>
<td><code>personas</code></td>
<td><code>factor</code></td>
<td>The number of persons making the trips from <code>district</code>,
<code>municipality</code>, or <code>large urban area</code> with zone
<code>id</code>.</td>
</tr>
<tr>
<td><code>year</code></td>
<td><code>year</code></td>
<td><code>integer</code></td>
<td>The year of the recorded data, extracted from the date.</td>
</tr>
<tr>
<td><code>month</code></td>
<td><code>month</code></td>
<td><code>integer</code></td>
<td>The month of the recorded data, extracted from the date.</td>
</tr>
<tr>
<td><code>day</code></td>
<td><code>day</code></td>
<td><code>integer</code></td>
<td>The day of the recorded data, extracted from the date.</td>
</tr>
</tbody>
</table>

**Getting the data**

To access it use `spod_get()` with `type` set to “number_of_trips”, or
just “nt”.

``` r
dates <- c(start = "2022-01-01", end = "2022-01-04")
os_dist <- spod_get(type = "overnight_stays", zones = "dist", dates = dates)
```

Because this data is small, we can actually load it completely into
memory:

``` r
os_dist_tbl <- os_dist |> dplyr::collect()
```

MITMS. 2024. “Estudio de Movilidad de Viajeros de Ámbito Nacional
Aplicando La Tecnología Big Data. Informe Metodológico (Study of
National Traveler Mobility Using Big Data Technology. Methodological
Report).” Secretaría de Estado de Transportes y Movilidad Sostenible;
Ministerio de Transportes, Movilidad y Agenda Urbana.
<https://www.transportes.gob.es/ministerio/proyectos-singulares/estudio-de-movilidad-con-big-data>.

Mühleisen, Hannes, and Mark Raasveldt. 2024. *Duckdb: DBI Package for
the DuckDB Database Management System*.
<https://doi.org/10.32614/CRAN.package.duckdb>.

[1] For reference: this object also has classes: `tbl_dbi` ,`tbl_sql`,
`tbl_lazy` ,and `tbl` .

[2] For reference: this object also has classes: `tbl_dbi` ,`tbl_sql`,
`tbl_lazy` ,and `tbl` .

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/disconnect.R
\name{spod_disconnect}
\alias{spod_disconnect}
\title{Safely disconnect from data and free memory}
\usage{
spod_disconnect(tbl_con, free_mem = TRUE)
}
\arguments{
\item{tbl_con}{A \code{tbl_duckdb_connection} connection object that you get from either \code{spod_get()} or \code{spod_connect()}.}

\item{free_mem}{A \code{logical}. Whether to free up memory by running \code{gc()}. Defaults to \code{TRUE}.}
}
\value{
No return value, called for side effect of disconnecting from the database and freeing up memory.
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#stable}{\figure{lifecycle-stable.svg}{options: alt='[Stable]'}}}{\strong{[Stable]}}

This function is to ensure that \code{DuckDB} connections to CSV.gz files (created via \code{spod_get()}), as well as to \code{DuckDB} files or folders of \code{parquet} files (created via \code{spod_convert()}) are closed properly to prevent conflicting connections. Essentially this is just a wrapper around \code{DBI::dbDisconnect()} that reaches out into the \code{.$src$con} object of the \code{tbl_duckdb_connection} connection object that is returned to the user via \code{spod_get()} and \code{spod_connect()}. After disonnecting the database, it also frees up memory by running \code{gc()}.
}
\examples{
\dontshow{if (interactive()) withAutoprint(\{ # examplesIf}
\donttest{
# Set data dir for file downloads
spod_set_data_dir(tempdir())

# basic example
# create a connection to the v1 data without converting
# this creates a duckdb database connection to CSV files
od_distr <- spod_get(
 "od",
 zones = "distr",
 dates = c("2020-03-01", "2020-03-02")
)
# disconnect from the database connection
spod_disconnect(od_distr)

# Advanced example
# download and convert data
dates_1 <- c(start = "2020-02-17", end = "2020-02-19")
db_2 <- spod_convert(
 type = "od",
 zones = "distr",
 dates = dates_1,
 overwrite = TRUE
)

# now connect to the converted data
my_od_data_2 <- spod_connect(db_2)

# disconnect from the database
spod_disconnect(my_od_data_2)
}
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/internal-utils.R
\name{spod_infer_data_v_from_dates}
\alias{spod_infer_data_v_from_dates}
\title{Infer data version from dates}
\usage{
spod_infer_data_v_from_dates(dates, ignore_missing_dates = FALSE)
}
\arguments{
\item{dates}{A \code{character} or \code{Date} vector of dates to process. Kindly keep in mind that v1 and v2 data follow different data collection methodologies and may not be directly comparable. Therefore, do not try to request data from both versions for the same date range. If you need to compare data from both versions, please refer to the respective codebooks and methodology documents. The v1 data covers the period from 2020-02-14 to 2021-05-09, and the v2 data covers the period from 2022-01-01 to the present until further notice. The true dates range is checked against the available data for each version on every function run.

The possible values can be any of the following:
\itemize{
\item For the \code{spod_get()} and \code{spod_convert()} functions, the \code{dates} can be set to "cached_v1" or "cached_v2" to request data from cached (already previously downloaded) v1 (2020-2021) or v2 (2022 onwards) data. In this case, the function will identify and use all data files that have been downloaded and cached locally, (e.g. using an explicit run of \code{spod_download()}, or any data requests made using the \code{spod_get()} or \code{spod_convert()} functions).
\item A single date in ISO (YYYY-MM-DD) or YYYYMMDD format. \code{character} or \code{Date} object.
\item A vector of dates in ISO (YYYY-MM-DD) or YYYYMMDD format. \code{character} or \code{Date} object. Can be any non-consecutive sequence of dates.
\item A date range
\itemize{
\item eigher a \code{character} or \code{Date} object of length 2 with clearly named elements \code{start} and \code{end} in ISO (YYYY-MM-DD) or YYYYMMDD format. E.g. \code{c(start = "2020-02-15", end = "2020-02-17")};
\item or a \code{character} object of the form \code{YYYY-MM-DD_YYYY-MM-DD} or \code{YYYYMMDD_YYYYMMDD}. For example, \verb{2020-02-15_2020-02-17} or \verb{20200215_20200217}.
}
\item A regular expression to match dates in the format \code{YYYYMMDD}. \code{character} object. For example, \verb{^202002} will match all dates in February 2020.
}}

\item{ignore_missing_dates}{Logical. If \code{TRUE}, the function will not raise an error if the some of the specified dates are missing. Any dates that are missing will be skipped, however the data for any valid dates will be acquired. Defaults to \code{FALSE}.}
}
\value{
An \code{integer} indicating the inferred data version.
}
\description{
Infer data version from dates
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data-dir.R
\name{spod_get_data_dir}
\alias{spod_get_data_dir}
\title{Get the data directory}
\usage{
spod_get_data_dir(quiet = FALSE)
}
\arguments{
\item{quiet}{A \code{logical} value indicating whether to suppress messages. Default is \code{FALSE}.}
}
\value{
A \code{character} vector of length 1 containing the path to the data directory where the package will download and convert the data.
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#stable}{\figure{lifecycle-stable.svg}{options: alt='[Stable]'}}}{\strong{[Stable]}}

This function retrieves the data directory from the environment variable SPANISH_OD_DATA_DIR (and previously set by \code{\link[=spod_set_data_dir]{spod_set_data_dir()}}).
If the environment variable is not set, it returns the temporary directory.
}
\examples{
spod_set_data_dir(tempdir())
spod_get_data_dir()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/available-data.R
\name{spod_available_data_v2}
\alias{spod_available_data_v2}
\title{Get the data dictionary}
\usage{
spod_available_data_v2(
  data_dir = spod_get_data_dir(),
  check_local_files = FALSE,
  use_s3 = TRUE,
  force = FALSE,
  quiet = FALSE
)
}
\arguments{
\item{data_dir}{The directory where the data is stored. Defaults to the value returned by \code{spod_get_data_dir()}.}

\item{check_local_files}{Logical. Whether to check if the local files exist and get the file size. Defaults to \code{FALSE}.}

\item{use_s3}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#experimental}{\figure{lifecycle-experimental.svg}{options: alt='[Experimental]'}}}{\strong{[Experimental]}} Logical. If \code{TRUE}, use Amazon S3 to get available data list, which does not require downloading the XML file and caching it locally, which may be a bit faster. If \code{FALSE}, use the XML file to get available data list.}

\item{force}{Logical. If \code{TRUE}, force re-download of metadata. For Amazon S3 this queries the S3 bucket for the XML file it re-downloads. If \code{FALSE}, only update the available data list if it is older than 1 day.}

\item{quiet}{A \code{logical} value indicating whether to suppress messages. Default is \code{FALSE}.}
}
\value{
A tibble with links, release dates of files in the data, dates of data coverage, local paths to files, and the download status.
\describe{
\item{target_url}{\code{character}. The URL link to the data file.}
\item{pub_ts}{\code{POSIXct}. The timestamp of when the file was published.}
\item{file_extension}{\code{character}. The file extension of the data file (e.g., 'tar', 'gz').}
\item{data_ym}{\code{Date}. The year and month of the data coverage, if available.}
\item{data_ymd}{\code{Date}. The specific date of the data coverage, if available.}
\item{study}{\code{factor}. Study category derived from the URL (e.g., 'basic', 'complete', 'routes').}
\item{type}{\code{factor}. Data type category derived from the URL (e.g., 'number_of_trips', 'origin-destination', 'overnight_stays', 'data_quality', 'metadata').}
\item{period}{\code{factor}. Temporal granularity category derived from the URL (e.g., 'day', 'month').}
\item{zones}{\code{factor}. Geographic zone classification derived from the URL (e.g., 'districts', 'municipalities', 'large_urban_areas').}
\item{local_path}{\code{character}. The local file path where the data is (or going to be) stored.}
\item{downloaded}{\code{logical}. Indicator of whether the data file has been downloaded locally. This is only available if \code{check_local_files} is \code{TRUE}.}
}
}
\description{
This function retrieves the data dictionary for the specified data directory.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/folders.R
\name{spod_subfolder_metadata_cache}
\alias{spod_subfolder_metadata_cache}
\title{Get metadata cache subfolder name}
\usage{
spod_subfolder_metadata_cache()
}
\value{
A \code{character} string with the subfolder name for the raw data cache.
}
\description{
Change subfolder name in the code of this function for metadata cache here to apply globally, as all functions in the package should use this function to get the metadata cache path.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/available-data.R
\name{spod_available_data}
\alias{spod_available_data}
\title{Get available data list}
\usage{
spod_available_data(
  ver = 2,
  check_local_files = FALSE,
  quiet = FALSE,
  data_dir = spod_get_data_dir(),
  use_s3 = TRUE,
  force = FALSE
)
}
\arguments{
\item{ver}{Integer. Can be 1 or 2. The version of the data to use. v1 spans 2020-2021, v2 covers 2022 and onwards. See more details in codebooks with \code{\link[=spod_codebook]{spod_codebook()}}.}

\item{check_local_files}{Logical. Whether to check if the local files exist and get the file size. Defaults to \code{FALSE}.}

\item{quiet}{A \code{logical} value indicating whether to suppress messages. Default is \code{FALSE}.}

\item{data_dir}{The directory where the data is stored. Defaults to the value returned by \code{spod_get_data_dir()}.}

\item{use_s3}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#experimental}{\figure{lifecycle-experimental.svg}{options: alt='[Experimental]'}}}{\strong{[Experimental]}} Logical. If \code{TRUE}, use Amazon S3 to get available data list, which does not require downloading the XML file and caching it locally, which may be a bit faster. If \code{FALSE}, use the XML file to get available data list.}

\item{force}{Logical. If \code{TRUE}, force re-download of metadata. For Amazon S3 this queries the S3 bucket for the XML file it re-downloads. If \code{FALSE}, only update the available data list if it is older than 1 day.}
}
\value{
A tibble with links, release dates of files in the data, dates of data coverage, local paths to files, and the download status.
\describe{
\item{target_url}{\code{character}. The URL link to the data file.}
\item{pub_ts}{\code{POSIXct}. The timestamp of when the file was published.}
\item{file_extension}{\code{character}. The file extension of the data file (e.g., 'tar', 'gz').}
\item{data_ym}{\code{Date}. The year and month of the data coverage, if available.}
\item{data_ymd}{\code{Date}. The specific date of the data coverage, if available.}
\item{study}{\code{factor}. Study category derived from the URL (e.g., 'basic', 'complete', 'routes').}
\item{type}{\code{factor}. Data type category derived from the URL (e.g., 'number_of_trips', 'origin-destination', 'overnight_stays', 'data_quality', 'metadata').}
\item{period}{\code{factor}. Temporal granularity category derived from the URL (e.g., 'day', 'month').}
\item{zones}{\code{factor}. Geographic zone classification derived from the URL (e.g., 'districts', 'municipalities', 'large_urban_areas').}
\item{local_path}{\code{character}. The local file path where the data is (or going to be) stored.}
\item{downloaded}{\code{logical}. Indicator of whether the data file has been downloaded locally. This is only available if \code{check_local_files} is \code{TRUE}.}
}
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#stable}{\figure{lifecycle-stable.svg}{options: alt='[Stable]'}}}{\strong{[Stable]}}

Get a table with links to available data files for the specified data version. Optionally check (see arguments) the file size and availability of data files previously downloaded into the cache directory specified with SPANISH_OD_DATA_DIR environment variable (set by \code{\link[=spod_set_data_dir]{spod_set_data_dir()}}) or a custom path specified with \code{data_dir} argument. By default the data is fetched from Amazon S3 bucket where the data is stored. If that fails, the function falls back to downloading an XML file from the Spanish Ministry of Transport website. You can also control this behaviour with \code{use_s3} argument.

For detailed data descriptions, see package vignettes using \code{\link[=spod_codebook]{spod_codebook(ver = 1)}} and \code{\link[=spod_codebook]{spod_codebook(ver = 2)}} and official methodology documents in \strong{References} section.
}
\examples{
\dontshow{if (interactive()) withAutoprint(\{ # examplesIf}
\donttest{

# Set data dir for file downloads
spod_set_data_dir(tempdir())

# Get available data list for v1 (2020-2021) data
spod_available_data(ver = 1)

# Get available data list for v2 (2022 onwards) data
spod_available_data(ver = 2)

# Get available data list for v2 (2022 onwards) data
# while also checking for local files that are already downloaded
spod_available_data(ver = 2, check_local_files = TRUE)
}
\dontshow{\}) # examplesIf}
}
\references{
\itemize{
\item \strong{For the official website of the mobility study}: \insertRef{mitms_mobility_web}{spanishoddata}
\item \strong{For v1 data methodology}: \insertRef{mitma_methodology_2020_v3}{spanishoddata}
\item \strong{For v2 data methodology}: \insertRef{mitms_methodology_2022_v8}{spanishoddata}
\item \strong{For the \code{spanishoddata} R package}: \insertRef{spanishoddata-r-pkg}{spanishoddata}
}

Use \code{\link[=spod_cite]{spod_cite()}} to cite the package and the data with correct plain text, markdown, or BibTeX formats.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get-zones.R
\name{spod_get_zones}
\alias{spod_get_zones}
\title{Get zones}
\usage{
spod_get_zones(
  zones = c("districts", "dist", "distr", "distritos", "municipalities", "muni",
    "municip", "municipios", "lua", "large_urban_areas", "gau", "grandes_areas_urbanas"),
  ver = NULL,
  data_dir = spod_get_data_dir(),
  quiet = FALSE
)
}
\arguments{
\item{zones}{The zones for which to download the data. Can be \code{"districts"} (or \code{"dist"}, \code{"distr"}, or the original Spanish \code{"distritos"}) or \code{"municipalities"} (or \code{"muni"}, \code{"municip"}, or the original Spanish \code{"municipios"}) for both data versions. Additionaly, these can be \code{"large_urban_areas"} (or \code{"lua"}, or the original Spanish \code{"grandes_areas_urbanas"}, or \code{"gau"}) for v2 data (2022 onwards).}

\item{ver}{Integer. Can be 1 or 2. The version of the data to use. v1 spans 2020-2021, v2 covers 2022 and onwards. See more details in codebooks with \code{\link[=spod_codebook]{spod_codebook()}}.}

\item{data_dir}{The directory where the data is stored. Defaults to the value returned by \code{spod_get_data_dir()} which returns the value of the environment variable \code{SPANISH_OD_DATA_DIR} or a temporary directory if the variable is not set. To set the data directory, use \link{spod_set_data_dir}.}

\item{quiet}{A \code{logical} value indicating whether to suppress messages. Default is \code{FALSE}.}
}
\value{
An \code{sf} object (Simple Feature collection).

The columns for v1 (2020-2021) data include:
\describe{
\item{id}{A character vector containing the unique identifier for each district, assigned by the data provider. This \code{id} matches the \code{id_origin}, \code{id_destination}, and \code{id} in district-level origin-destination and number of trips data.}
\item{census_districts}{A string with semicolon-separated identifiers of census districts classified by the Spanish Statistical Office (INE) that are spatially bound within the polygons for each \code{id}.}
\item{municipalities_mitma}{A string with semicolon-separated municipality identifiers (as assigned by the data provider) corresponding to each district \code{id}.}
\item{municipalities}{A string with semicolon-separated municipality identifiers classified by the Spanish Statistical Office (INE) corresponding to each \code{id}.}
\item{district_names_in_v2/municipality_names_in_v2}{A string with semicolon-separated district names (from the v2 version of this data) corresponding to each district \code{id} in v1.}
\item{district_ids_in_v2/municipality_ids_in_v2}{A string with semicolon-separated district identifiers (from the v2 version of this data) corresponding to each district \code{id} in v1.}
\item{geometry}{A \code{MULTIPOLYGON} column containing the spatial geometry of each district, stored as an sf object. The geometry is projected in the ETRS89 / UTM zone 30N coordinate reference system (CRS), with XY dimensions.}
}

The columns for v2 (2022 onwards) data include:
\describe{
\item{id}{A character vector containing the unique identifier for each zone, assigned by the data provider.}
\item{name}{A character vector with the name of each district.}
\item{population}{A numeric vector representing the population of each district (as of 2022).}
\item{census_sections}{A string with semicolon-separated identifiers of census sections corresponding to each district.}
\item{census_districts}{A string with semicolon-separated identifiers of census districts as classified by the Spanish Statistical Office (INE) corresponding to each district.}
\item{municipalities}{A string with semicolon-separated identifiers of municipalities classified by the Spanish Statistical Office (INE) corresponding to each district.}
\item{municipalities_mitma}{A string with semicolon-separated identifiers of municipalities, as assigned by the data provider, that correspond to each district.}
\item{luas_mitma}{A string with semicolon-separated identifiers of LUAs (Local Urban Areas) from the provider, associated with each district.}
\item{district_ids_in_v1/municipality_ids_in_v1}{A string with semicolon-separated district identifiers from v1 data corresponding to each district in v2. If no match exists, it is marked as \code{NA}.}
\item{geometry}{A \code{MULTIPOLYGON} column containing the spatial geometry of each district, stored as an sf object. The geometry is projected in the ETRS89 / UTM zone 30N coordinate reference system (CRS), with XY dimensions.}
}
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#stable}{\figure{lifecycle-stable.svg}{options: alt='[Stable]'}}}{\strong{[Stable]}}

Get spatial zones for the specified data version. Supports both v1 (2020-2021) and v2 (2022 onwards) data.

For detailed data descriptions, see package vignettes using \code{\link[=spod_codebook]{spod_codebook(ver = 1)}} and \code{\link[=spod_codebook]{spod_codebook(ver = 2)}} and official methodology documents in \strong{References} section.
}
\examples{
\dontshow{if (interactive()) withAutoprint(\{ # examplesIf}
\donttest{
# get polygons for municipalities for the v2 data
municip_v2 <- spod_get_zones(zones = "municipalities", ver = 2)

# get polygons for the districts for the v1 data
distr_v1 <- spod_get_zones(zones = "districts", ver = 1)
}
\dontshow{\}) # examplesIf}
}
\references{
\itemize{
\item \strong{For the official website of the mobility study}: \insertRef{mitms_mobility_web}{spanishoddata}
\item \strong{For v1 data methodology}: \insertRef{mitma_methodology_2020_v3}{spanishoddata}
\item \strong{For v2 data methodology}: \insertRef{mitms_methodology_2022_v8}{spanishoddata}
\item \strong{For the \code{spanishoddata} R package}: \insertRef{spanishoddata-r-pkg}{spanishoddata}
}

Use \code{\link[=spod_cite]{spod_cite()}} to cite the package and the data with correct plain text, markdown, or BibTeX formats.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/internal-utils.R
\name{spod_convert_dates_to_ranges}
\alias{spod_convert_dates_to_ranges}
\title{Convert dates to ranges}
\usage{
spod_convert_dates_to_ranges(dates)
}
\arguments{
\item{dates}{A \code{character} vector of dates.}
}
\value{
A \code{character} vector of date ranges.
}
\description{
This internal helper function reduces a vector of dates to a vector of date ranges to shorten the warning and error messages that mention the valid date ranges.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/quick-get.R
\name{spod_fetch_municipalities_json_memoised}
\alias{spod_fetch_municipalities_json_memoised}
\title{Cache the municipalities geometries from the mapas-movilidad website}
\usage{
spod_fetch_municipalities_json_memoised()
}
\description{
Cache the municipalities geometries from the mapas-movilidad website
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/duckdb-helpers.R
\name{spod_duckdb_overnight_stays}
\alias{spod_duckdb_overnight_stays}
\title{Create a duckdb overnight stays table}
\usage{
spod_duckdb_overnight_stays(
  con = DBI::dbConnect(duckdb::duckdb(), dbdir = ":memory:", read_only = FALSE),
  zones = c("districts", "dist", "distr", "distritos", "municipalities", "muni",
    "municip", "municipios", "lua", "large_urban_areas", "gau", "grandes_areas_urbanas"),
  ver = NULL,
  data_dir = spod_get_data_dir()
)
}
\arguments{
\item{con}{A duckdb connection object. If not specified, a new in-memory connection will be created.}

\item{zones}{The zones for which to download the data. Can be \code{"districts"} (or \code{"dist"}, \code{"distr"}, or the original Spanish \code{"distritos"}) or \code{"municipalities"} (or \code{"muni"}, \code{"municip"}, or the original Spanish \code{"municipios"}) for both data versions. Additionaly, these can be \code{"large_urban_areas"} (or \code{"lua"}, or the original Spanish \code{"grandes_areas_urbanas"}, or \code{"gau"}) for v2 data (2022 onwards).}

\item{ver}{Integer. Can be 1 or 2. The version of the data to use. v1 spans 2020-2021, v2 covers 2022 and onwards. See more details in codebooks with \code{\link[=spod_codebook]{spod_codebook()}}.}

\item{data_dir}{The directory where the data is stored. Defaults to the value returned by \code{spod_get_data_dir()}.}
}
\value{
A \code{duckdb} connection object with 2 views:
\itemize{
\item \code{od_csv_raw} - a raw table view of all cached CSV files with the origin-destination data that has been previously cached in $SPANISH_OD_DATA_DIR
\item \code{od_csv_clean} - a cleaned-up table view of \code{od_csv_raw} with column names and values translated and mapped to English. This still includes all cached data.
}
}
\description{
This function creates a duckdb connection to the overnight stays data stored in a folder of CSV.gz files.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/folders.R
\name{spod_subfolder_clean_data_cache}
\alias{spod_subfolder_clean_data_cache}
\title{Get clean data subfolder name}
\usage{
spod_subfolder_clean_data_cache(ver = 1)
}
\arguments{
\item{ver}{Integer. Can be 1 or 2. The version of the data to use. v1 spans 2020-2021, v2 covers 2022 and onwards. See more details in codebooks with \code{\link[=spod_codebook]{spod_codebook()}}.}
}
\value{
A \code{character} string with the subfolder name for the clean data cache.
}
\description{
Change subfolder name in the code of this function for clean data cache here to apply globally, as all functions in the package should use this function to get the clean data cache path.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/duckdb-helpers.R
\name{spod_duckdb_set_temp}
\alias{spod_duckdb_set_temp}
\title{Set temp file for DuckDB connection}
\usage{
spod_duckdb_set_temp(con, temp_path = spod_get_temp_dir())
}
\arguments{
\item{con}{A duckdb connection}

\item{temp_path}{The path to the temp folder for DuckDB for \href{https://duckdb.org/2024/07/09/memory-management.html#intermediate-spilling}{intermediate spilling} in case the set memory limit and/or physical memory of the computer is too low to perform the query. By default this is set to the \code{temp} directory in the data folder defined by SPANISH_OD_DATA_DIR environment variable (set by \code{\link[=spod_set_data_dir]{spod_set_data_dir()}})). Otherwise, for queries on folders of CSV files or parquet files, the temporary path would be set to the current R working directory, which probably is undesirable, as the current working directory can be on a slow storage, or storage that may have limited space, compared to the data folder.}
}
\value{
A \code{duckdb} connection.
}
\description{
Set temp file for DuckDB connection
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/duckdb-helpers.R
\name{spod_duckdb_create_province_enum}
\alias{spod_duckdb_create_province_enum}
\title{Create province names ENUM in a duckdb connection}
\usage{
spod_duckdb_create_province_enum(con)
}
\arguments{
\item{con}{A \code{duckdb} connection.}
}
\value{
A \code{duckdb} connection with \code{INE_PROV_NAME_ENUM} and \code{INE_PROV_CODE_ENUM} created.
}
\description{
Create province names ENUM in a duckdb connection
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/folders.R
\name{spod_get_temp_dir}
\alias{spod_get_temp_dir}
\title{Get temporary directory for DuckDB intermediate spilling}
\usage{
spod_get_temp_dir(data_dir = spod_get_data_dir())
}
\arguments{
\item{data_dir}{The directory where the data is stored. Defaults to the value returned by \code{spod_get_data_dir()}.}
}
\value{
A \code{character} string with the path to the temp folder for \code{DuckDB} for \href{https://duckdb.org/2024/07/09/memory-management.html#intermediate-spilling}{intermediate spilling}.
}
\description{
Get the The path to the temp folder for DuckDB for \href{https://duckdb.org/2024/07/09/memory-management.html#intermediate-spilling}{intermediate spilling} in case the set memory limit and/or physical memory of the computer is too low to perform the query.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/global-params.R
\name{global_quiet_param}
\alias{global_quiet_param}
\title{Global Quiet Parameter}
\usage{
global_quiet_param(quiet = FALSE)
}
\arguments{
\item{quiet}{A \code{logical} value indicating whether to suppress messages. Default is \code{FALSE}.}
}
\value{
Nothing. This function is just a placeholder for global quiet parameter.
}
\description{
Documentation for the \code{quiet} parameter, used globally.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/internal-utils.R
\name{spod_graphql_valid_dates}
\alias{spod_graphql_valid_dates}
\title{Get valid dates from the GraphQL API}
\usage{
spod_graphql_valid_dates()
}
\value{
A \code{Date} vector of dates that are valid to request data with \code{spod_quick_get_od()}.
}
\description{
Get valid dates from the GraphQL API
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get-zones.R
\name{spod_download_zones_v1}
\alias{spod_download_zones_v1}
\title{Downloads and extracts the raw v1 zones data}
\usage{
spod_download_zones_v1(
  zones = c("districts", "dist", "distr", "distritos", "municipalities", "muni",
    "municip", "municipios"),
  data_dir = spod_get_data_dir(),
  quiet = FALSE,
  metadata = NULL
)
}
\arguments{
\item{zones}{The zones for which to download the data. Can be \code{"districts"} (or \code{"dist"}, \code{"distr"}, or the original Spanish \code{"distritos"}) or \code{"municipalities"} (or \code{"muni"}, \code{"municip"}, or the original Spanish \code{"municipios"}).}

\item{data_dir}{The directory where the data is stored.}

\item{quiet}{Boolean flag to control the display of messages.}

\item{metadata}{Optional metadata table returned by \code{spod_available_data()}.}
}
\value{
A \code{character} string containing the path to the downloaded and extracted file.
}
\description{
This function ensures that the necessary v1 raw data for zones files are downloaded and extracted from the specified data directory.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get-zones.R
\name{spod_get_zones_v2}
\alias{spod_get_zones_v2}
\title{Retrieves the zones v2 data}
\usage{
spod_get_zones_v2(
  zones = c("districts", "dist", "distr", "distritos", "municipalities", "muni",
    "municip", "municipios", "lua", "large_urban_areas", "gau", "grandes_areas_urbanas"),
  data_dir = spod_get_data_dir(),
  quiet = FALSE
)
}
\arguments{
\item{zones}{The zones for which to download the data. Can be \code{"districts"} (or \code{"dist"}, \code{"distr"}, or the original Spanish \code{"distritos"}) or \code{"municipalities"} (or \code{"muni"}, \code{"municip"}, or the original Spanish \code{"municipios"}).}

\item{data_dir}{The directory where the data is stored.}

\item{quiet}{A \code{logical} value indicating whether to suppress messages. Default is \code{FALSE}.}
}
\value{
An \code{sf} object (Simple Feature collection) with 4 fields:
\describe{
\item{id}{A character vector containing the unique identifier for each zone, to be matched with identifiers in the tabular data.}
\item{name}{A character vector with the name of the zone.}
\item{population}{A numeric vector representing the population of each zone (as of 2022).}
\item{geometry}{A \code{MULTIPOLYGON} column containing the spatial geometry of each zone, stored as an sf object.
The geometry is projected in the ETRS89 / UTM zone 30N coordinate reference system (CRS), with XY dimensions.}
}
}
\description{
This function retrieves the zones data from the specified data directory.
It can retrieve either "distritos" or "municipios" zones data.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/duckdb-helpers.R
\name{spod_duckdb_filter_by_dates}
\alias{spod_duckdb_filter_by_dates}
\title{Filter a duckdb conenction by dates}
\usage{
spod_duckdb_filter_by_dates(con, source_view_name, new_view_name, dates)
}
\arguments{
\item{con}{A duckdb connection}

\item{source_view_name}{The name of the source duckdb "view" (the virtual table, in the context of current package likely connected to a folder of CSV files)}

\item{new_view_name}{The name of the new duckdb "view" (the virtual table, in the context of current package likely connected to a folder of CSV files).}

\item{dates}{A \code{character} or \code{Date} vector of dates to process. Kindly keep in mind that v1 and v2 data follow different data collection methodologies and may not be directly comparable. Therefore, do not try to request data from both versions for the same date range. If you need to compare data from both versions, please refer to the respective codebooks and methodology documents. The v1 data covers the period from 2020-02-14 to 2021-05-09, and the v2 data covers the period from 2022-01-01 to the present until further notice. The true dates range is checked against the available data for each version on every function run.

The possible values can be any of the following:
\itemize{
\item For the \code{spod_get()} and \code{spod_convert()} functions, the \code{dates} can be set to "cached_v1" or "cached_v2" to request data from cached (already previously downloaded) v1 (2020-2021) or v2 (2022 onwards) data. In this case, the function will identify and use all data files that have been downloaded and cached locally, (e.g. using an explicit run of \code{spod_download()}, or any data requests made using the \code{spod_get()} or \code{spod_convert()} functions).
\item A single date in ISO (YYYY-MM-DD) or YYYYMMDD format. \code{character} or \code{Date} object.
\item A vector of dates in ISO (YYYY-MM-DD) or YYYYMMDD format. \code{character} or \code{Date} object. Can be any non-consecutive sequence of dates.
\item A date range
\itemize{
\item eigher a \code{character} or \code{Date} object of length 2 with clearly named elements \code{start} and \code{end} in ISO (YYYY-MM-DD) or YYYYMMDD format. E.g. \code{c(start = "2020-02-15", end = "2020-02-17")};
\item or a \code{character} object of the form \code{YYYY-MM-DD_YYYY-MM-DD} or \code{YYYYMMDD_YYYYMMDD}. For example, \verb{2020-02-15_2020-02-17} or \verb{20200215_20200217}.
}
\item A regular expression to match dates in the format \code{YYYYMMDD}. \code{character} object. For example, \verb{^202002} will match all dates in February 2020.
}}
}
\value{
A \code{duckdb} connection with original views and a new filtered view.
}
\description{
IMPORTANT: This function assumes that the table or view that is being filtered has separate \code{year}, \code{month} and \code{day} columns with integer values. This is done so that the filtering is faster on CSV files that are stored in a folder structure with hive-style \verb{/year=2020/month=2/day=14/}.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cite.R
\name{spod_cite}
\alias{spod_cite}
\title{Cite the package and the data}
\usage{
spod_cite(what = "all", format = "all")
}
\arguments{
\item{what}{Character vector specifying what to cite.
Can include "package", "data", "methodology_v1", "methodology_v2", or "all".
Default is "all".}

\item{format}{Character vector specifying output format(s).
Can include "text", "markdown", "bibtex", or "all".
Default is "all".}
}
\value{
Nothing. Prints citation in plain text, markdown, BibTeX, or all formats at once to console.
}
\description{
Cite the package and the data
}
\examples{
# Cite everything in all formats
\dontrun{
spod_cite()
}

# Cite just the package in BibTeX format
\dontrun{
spod_cite(what = "package", format = "bibtex")
}

# Cite both methodologies in plain text
\dontrun{
spod_cite(what = c("methodology_v1", "methodology_v2"), format = "text")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/spanishoddata-package.R
\docType{package}
\name{spanishoddata-package}
\alias{spanishoddata}
\alias{spanishoddata-package}
\title{spanishoddata: Get Spanish Origin-Destination Data}
\description{
\if{html}{\figure{logo.png}{options: style='float: right' alt='logo' width='120'}}

Gain seamless access to origin-destination (OD) data from the Spanish Ministry of Transport, hosted at \url{https://www.transportes.gob.es/ministerio/proyectos-singulares/estudios-de-movilidad-con-big-data/opendata-movilidad}. This package simplifies the management of these large datasets by providing tools to download zone boundaries, handle associated origin-destination data, and process it efficiently with the 'duckdb' database interface. Local caching minimizes repeated downloads, streamlining workflows for researchers and analysts. Methods described in Kotov et al. (2026) \doi{10.1177/23998083251415040}. Extensive documentation is available at \url{https://ropenspain.github.io/spanishoddata/index.html}, offering guides on creating static and dynamic mobility flow visualizations and transforming large datasets into analysis-ready formats.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://rOpenSpain.github.io/spanishoddata/}
  \item \url{https://github.com/rOpenSpain/spanishoddata}
  \item \doi{10.1177/23998083251415040}
  \item Report bugs at \url{https://github.com/rOpenSpain/spanishoddata/issues}
}

}
\author{
\strong{Maintainer}: Egor Kotov \email{kotov.egor@gmail.com} (\href{https://orcid.org/0000-0001-6690-5345}{ORCID})

Authors:
\itemize{
  \item Robin Lovelace \email{rob00x@gmail.com} (\href{https://orcid.org/0000-0001-5679-6536}{ORCID})
}

Other contributors:
\itemize{
  \item Eugeni Vidal-Tortosa (\href{https://orcid.org/0000-0001-5199-4103}{ORCID}) [contributor]
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/quick-get.R
\name{spod_httr2_req_perform}
\alias{spod_httr2_req_perform}
\title{Internal wrappers for httr2 and sf calls to enable mocking}
\usage{
spod_httr2_req_perform(req)
}
\description{
Internal wrappers for httr2 and sf calls to enable mocking
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/duckdb-helpers.R
\name{spod_sql_where_dates}
\alias{spod_sql_where_dates}
\title{Generate a WHERE part of an SQL query from a sequence of dates}
\usage{
spod_sql_where_dates(dates)
}
\arguments{
\item{dates}{A Dates vector of dates to process.}
}
\value{
A character vector of the SQL query.
}
\description{
Generate a WHERE part of an SQL query from a sequence of dates
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/available-data-s3.R
\name{spod_available_data_s3}
\alias{spod_available_data_s3}
\title{Get available data list from Amazon S3 storage}
\usage{
spod_available_data_s3(
  ver = c(1, 2),
  force = FALSE,
  quiet = FALSE,
  data_dir = spod_get_data_dir()
)
}
\arguments{
\item{ver}{Integer. Can be 1 or 2. The version of the data to use. v1 spans 2020-2021, v2 covers 2022 and onwards. See more details in codebooks with \code{\link[=spod_codebook]{spod_codebook()}}.}

\item{force}{Logical. If \code{TRUE}, force re-download of metadata. For Amazon S3 this queries the S3 bucket for the XML file it re-downloads. If \code{FALSE}, only update the available data list if it is older than 1 day.}

\item{quiet}{A \code{logical} value indicating whether to suppress messages. Default is \code{FALSE}.}

\item{data_dir}{The directory where the data is stored. Defaults to the value returned by \code{spod_get_data_dir()}.}
}
\value{
A tibble with links, release dates of files in the data, dates of data coverage, local paths to files, and the download status.
}
\description{
Get a table with links to available data files for the specified data version from Amazon S3 storage.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/internal-utils.R
\name{spod_match_data_type}
\alias{spod_match_data_type}
\title{Match data types for normalisation}
\usage{
spod_match_data_type(
  type = c("od", "origin-destination", "viajes", "os", "overnight_stays",
    "pernoctaciones", "nt", "number_of_trips", "personas")
)
}
\arguments{
\item{type}{The type of data to match. Can be "od", "origin-destination", "os", "overnight_stays", or "nt", "number_of_trips".}
}
\value{
A \code{character} string with the folder name for the specified data type. Or \code{NULL} if the type is not recognized.
}
\description{
Match data types for normalisation
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/quick-get.R
\name{spod_query_od_memoised}
\alias{spod_query_od_memoised}
\title{Cache the spod_query_od_raw function to avoid repeated requests}
\usage{
spod_query_od_memoised(
  date_fmt,
  graphql_distances,
  id_origin,
  id_destination,
  min_trips,
  graphql_query
)
}
\description{
Cache the spod_query_od_raw function to avoid repeated requests
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/check-files.R
\name{spod_check_files}
\alias{spod_check_files}
\title{Check cached files consistency against checksums from S3}
\usage{
spod_check_files(
  type = c("od", "origin-destination", "os", "overnight_stays", "nt", "number_of_trips"),
  zones = c("districts", "dist", "distr", "distritos", "municipalities", "muni",
    "municip", "municipios", "lua", "large_urban_areas", "gau", "grandes_areas_urbanas"),
  dates = NULL,
  data_dir = spod_get_data_dir(),
  quiet = FALSE,
  ignore_missing_dates = FALSE,
  n_threads = 1
)
}
\arguments{
\item{type}{The type of data to download. Can be \code{"origin-destination"} (or ust \code{"od"}), or \code{"number_of_trips"} (or just \code{"nt"}) for v1 data. For v2 data \code{"overnight_stays"} (or just \code{"os"}) is also available. More data types to be supported in the future. See codebooks for v1 and v2 data in vignettes with \code{spod_codebook(1)} and \code{spod_codebook(2)} (\link{spod_codebook}).}

\item{zones}{The zones for which to download the data. Can be \code{"districts"} (or \code{"dist"}, \code{"distr"}, or the original Spanish \code{"distritos"}) or \code{"municipalities"} (or \code{"muni"}, \code{"municip"}, or the original Spanish \code{"municipios"}) for both data versions. Additionaly, these can be \code{"large_urban_areas"} (or \code{"lua"}, or the original Spanish \code{"grandes_areas_urbanas"}, or \code{"gau"}) for v2 data (2022 onwards).}

\item{dates}{A \code{character} or \code{Date} vector of dates to process. Kindly keep in mind that v1 and v2 data follow different data collection methodologies and may not be directly comparable. Therefore, do not try to request data from both versions for the same date range. If you need to compare data from both versions, please refer to the respective codebooks and methodology documents. The v1 data covers the period from 2020-02-14 to 2021-05-09, and the v2 data covers the period from 2022-01-01 to the present until further notice. The true dates range is checked against the available data for each version on every function run.

The possible values can be any of the following:
\itemize{
\item For the \code{spod_get()} and \code{spod_convert()} functions, the \code{dates} can be set to "cached_v1" or "cached_v2" to request data from cached (already previously downloaded) v1 (2020-2021) or v2 (2022 onwards) data. In this case, the function will identify and use all data files that have been downloaded and cached locally, (e.g. using an explicit run of \code{spod_download()}, or any data requests made using the \code{spod_get()} or \code{spod_convert()} functions).
\item A single date in ISO (YYYY-MM-DD) or YYYYMMDD format. \code{character} or \code{Date} object.
\item A vector of dates in ISO (YYYY-MM-DD) or YYYYMMDD format. \code{character} or \code{Date} object. Can be any non-consecutive sequence of dates.
\item A date range
\itemize{
\item eigher a \code{character} or \code{Date} object of length 2 with clearly named elements \code{start} and \code{end} in ISO (YYYY-MM-DD) or YYYYMMDD format. E.g. \code{c(start = "2020-02-15", end = "2020-02-17")};
\item or a \code{character} object of the form \code{YYYY-MM-DD_YYYY-MM-DD} or \code{YYYYMMDD_YYYYMMDD}. For example, \verb{2020-02-15_2020-02-17} or \verb{20200215_20200217}.
}
\item A regular expression to match dates in the format \code{YYYYMMDD}. \code{character} object. For example, \verb{^202002} will match all dates in February 2020.
}}

\item{data_dir}{The directory where the data is stored. Defaults to the value returned by \code{spod_get_data_dir()} which returns the value of the environment variable \code{SPANISH_OD_DATA_DIR} or a temporary directory if the variable is not set. To set the data directory, use \link{spod_set_data_dir}.}

\item{quiet}{A \code{logical} value indicating whether to suppress messages. Default is \code{FALSE}.}

\item{ignore_missing_dates}{Logical. If \code{TRUE}, the function will not raise an error if the some of the specified dates are missing. Any dates that are missing will be skipped, however the data for any valid dates will be acquired. Defaults to \code{FALSE}.}

\item{n_threads}{Numeric. Number of threads to use for file verificaiton. Defaults to 1. When set to 2 or more threads, uses \code{future.mirai} as a backend for parallelization, resulting in significant (~4x) speedup, unless disk read speed is a bottleneck.}
}
\value{
A tibble similar to the output of \code{spod_available_data}, but with an extra column \code{local_file_consistent}, where \code{TRUE} indicates that the file cheksum matches the expected checksums in Amazon S3. Note: some v1 (2020-2021) files were not stored correctly on S3 and their ETag checksums are incorrectly reported by Amazon S3, so their true file sizes and ETag checksums were cached inside the \code{spanishoddata} package.
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#experimental}{\figure{lifecycle-experimental.svg}{options: alt='[Experimental]'}}}{\strong{[Experimental]}}

\strong{WARNING: The checks may fail for May 2022 data and for some 2025 data, as the remote cheksums that are used for checking the file consistency are incorrect. We are working on solving this in future updates, for now, kindly rely on the built-in file size checks of \code{\link{spod_download}}, \code{\link{spod_get}}, and \code{\link{spod_convert}}.} This function checks downloaded data files whether they are consistent with their checksums in Amazon S3 by computing ETag for each file. This involves computing MD5 for each part of the file and concatenating them and computing MD5 again on the resulting concatenated MD5s. This may take very long time if you check all files, so use with caution.
}
\examples{
\dontshow{if (interactive()) withAutoprint(\{ # examplesIf}
\donttest{
spod_set_data_dir(tempdir())
spod_download(
 type = "number_of_trips",
 zones = "distr",
 dates = "2020-03-14"
)

# now check the consistency
check_results <- spod_check_files(
  type = "number_of_trips",
  zones = "distr",
  dates = "2020-03-14"
)
all(check_results$local_file_consistent)
}
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get.R
\name{spod_get}
\alias{spod_get}
\title{Get tabular mobility data}
\usage{
spod_get(
  type = c("od", "origin-destination", "os", "overnight_stays", "nt", "number_of_trips"),
  zones = c("districts", "dist", "distr", "distritos", "municipalities", "muni",
    "municip", "municipios", "lua", "large_urban_areas", "gau", "grandes_areas_urbanas"),
  dates = NULL,
  data_dir = spod_get_data_dir(),
  quiet = FALSE,
  max_mem_gb = NULL,
  max_n_cpu = max(1, parallelly::availableCores() - 1),
  max_download_size_gb = 1,
  duckdb_target = ":memory:",
  temp_path = spod_get_temp_dir(),
  ignore_missing_dates = FALSE
)
}
\arguments{
\item{type}{The type of data to download. Can be \code{"origin-destination"} (or ust \code{"od"}), or \code{"number_of_trips"} (or just \code{"nt"}) for v1 data. For v2 data \code{"overnight_stays"} (or just \code{"os"}) is also available. More data types to be supported in the future. See codebooks for v1 and v2 data in vignettes with \code{spod_codebook(1)} and \code{spod_codebook(2)} (\link{spod_codebook}).}

\item{zones}{The zones for which to download the data. Can be \code{"districts"} (or \code{"dist"}, \code{"distr"}, or the original Spanish \code{"distritos"}) or \code{"municipalities"} (or \code{"muni"}, \code{"municip"}, or the original Spanish \code{"municipios"}) for both data versions. Additionaly, these can be \code{"large_urban_areas"} (or \code{"lua"}, or the original Spanish \code{"grandes_areas_urbanas"}, or \code{"gau"}) for v2 data (2022 onwards).}

\item{dates}{A \code{character} or \code{Date} vector of dates to process. Kindly keep in mind that v1 and v2 data follow different data collection methodologies and may not be directly comparable. Therefore, do not try to request data from both versions for the same date range. If you need to compare data from both versions, please refer to the respective codebooks and methodology documents. The v1 data covers the period from 2020-02-14 to 2021-05-09, and the v2 data covers the period from 2022-01-01 to the present until further notice. The true dates range is checked against the available data for each version on every function run.

The possible values can be any of the following:
\itemize{
\item For the \code{spod_get()} and \code{spod_convert()} functions, the \code{dates} can be set to "cached_v1" or "cached_v2" to request data from cached (already previously downloaded) v1 (2020-2021) or v2 (2022 onwards) data. In this case, the function will identify and use all data files that have been downloaded and cached locally, (e.g. using an explicit run of \code{spod_download()}, or any data requests made using the \code{spod_get()} or \code{spod_convert()} functions).
\item A single date in ISO (YYYY-MM-DD) or YYYYMMDD format. \code{character} or \code{Date} object.
\item A vector of dates in ISO (YYYY-MM-DD) or YYYYMMDD format. \code{character} or \code{Date} object. Can be any non-consecutive sequence of dates.
\item A date range
\itemize{
\item eigher a \code{character} or \code{Date} object of length 2 with clearly named elements \code{start} and \code{end} in ISO (YYYY-MM-DD) or YYYYMMDD format. E.g. \code{c(start = "2020-02-15", end = "2020-02-17")};
\item or a \code{character} object of the form \code{YYYY-MM-DD_YYYY-MM-DD} or \code{YYYYMMDD_YYYYMMDD}. For example, \verb{2020-02-15_2020-02-17} or \verb{20200215_20200217}.
}
\item A regular expression to match dates in the format \code{YYYYMMDD}. \code{character} object. For example, \verb{^202002} will match all dates in February 2020.
}}

\item{data_dir}{The directory where the data is stored. Defaults to the value returned by \code{spod_get_data_dir()} which returns the value of the environment variable \code{SPANISH_OD_DATA_DIR} or a temporary directory if the variable is not set. To set the data directory, use \link{spod_set_data_dir}.}

\item{quiet}{A \code{logical} value indicating whether to suppress messages. Default is \code{FALSE}.}

\item{max_mem_gb}{\code{integer} value of the maximum operating memory to use in GB. \code{NULL} by default, delegates the choice to the \code{DuckDB} engine which usually sets it to 80\% of available memory. Caution, in HPC use, the amount of memory available to your job may be determined incorrectly by the \code{DuckDB} engine, so it is recommended to set this parameter explicitly according to your job's memory limits.}

\item{max_n_cpu}{The maximum number of threads to use. Defaults to the number of available cores minus 1.}

\item{max_download_size_gb}{The maximum download size in gigabytes. Defaults to 1.}

\item{duckdb_target}{(Optional) The path to the duckdb file to save the data to, if a convertation from CSV is reuqested by the \code{spod_convert} function. If not specified, it will be set to ":memory:" and the data will be stored in memory.}

\item{temp_path}{The path to the temp folder for DuckDB for \href{https://duckdb.org/2024/07/09/memory-management.html#intermediate-spilling}{intermediate spilling} in case the set memory limit and/or physical memory of the computer is too low to perform the query. By default this is set to the \code{temp} directory in the data folder defined by SPANISH_OD_DATA_DIR environment variable (set by \code{\link[=spod_set_data_dir]{spod_set_data_dir()}})). Otherwise, for queries on folders of CSV files or parquet files, the temporary path would be set to the current R working directory, which probably is undesirable, as the current working directory can be on a slow storage, or storage that may have limited space, compared to the data folder.}

\item{ignore_missing_dates}{Logical. If \code{TRUE}, the function will not raise an error if the some of the specified dates are missing. Any dates that are missing will be skipped, however the data for any valid dates will be acquired. Defaults to \code{FALSE}.}
}
\value{
A DuckDB lazy table connection object of class \code{tbl_duckdb_connection}.
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#stable}{\figure{lifecycle-stable.svg}{options: alt='[Stable]'}}}{\strong{[Stable]}}

This function creates a DuckDB lazy table connection object from the specified type and zones. It checks for missing data and downloads it if necessary. The connnection is made to the raw CSV files in gzip archives, so analysing the data through this connection may be slow if you select more than a few days. You can manipulate this object using \code{dplyr} functions such as \code{\link[dplyr]{select}}, \code{\link[dplyr]{filter}}, \code{\link[dplyr]{mutate}}, \code{\link[dplyr]{group_by}}, \code{\link[dplyr]{summarise}}, etc. In the end of any sequence of commands you will need to add \code{\link[dplyr]{collect}} to execute the whole chain of data manipulations and load the results into memory in an R \code{data.frame}/\code{tibble}. See codebooks for v1 and v2 data in vignettes with \code{\link{spod_codebook}(1)} and \code{\link{spod_codebook}(2)}.

If you want to analyse longer periods of time (especiially several months or even the whole data over several years), consider using the \code{\link{spod_convert}} and then \code{\link{spod_connect}}.

If you want to quickly get the origin-destination data with flows aggregated for a single day at municipal level and without any extra socio-economic variables, consider using the \code{\link{spod_quick_get_od}} function.

For detailed data descriptions, see package vignettes using \code{\link[=spod_codebook]{spod_codebook(ver = 1)}} and \code{\link[=spod_codebook]{spod_codebook(ver = 2)}} and official methodology documents in \strong{References} section.
}
\examples{
\dontshow{if (interactive()) withAutoprint(\{ # examplesIf}
\donttest{

# create a connection to the v1 data
spod_set_data_dir(tempdir())
dates <- c("2020-02-14", "2020-03-14", "2021-02-14", "2021-02-14", "2021-02-15")
nt_dist <- spod_get(type = "number_of_trips", zones = "distr", dates = dates)

# nt_dist is a table view filtered to the specified dates

# for advanced users only
# access the source connection with all dates
# list tables
DBI::dbListTables(nt_dist$src$con)

# disconnect
spod_disconnect(nt_dist)
}
\dontshow{\}) # examplesIf}
}
\references{
\itemize{
\item \strong{For the official website of the mobility study}: \insertRef{mitms_mobility_web}{spanishoddata}
\item \strong{For v1 data methodology}: \insertRef{mitma_methodology_2020_v3}{spanishoddata}
\item \strong{For v2 data methodology}: \insertRef{mitms_methodology_2022_v8}{spanishoddata}
\item \strong{For the \code{spanishoddata} R package}: \insertRef{spanishoddata-r-pkg}{spanishoddata}
}

Use \code{\link[=spod_cite]{spod_cite()}} to cite the package and the data with correct plain text, markdown, or BibTeX formats.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/quick-get.R
\name{spod_query_od_raw}
\alias{spod_query_od_raw}
\title{Internal function to query the GraphQL API for origin-destination data}
\usage{
spod_query_od_raw(
  date_fmt,
  graphql_distances,
  id_origin,
  id_destination,
  min_trips,
  graphql_query
)
}
\arguments{
\item{id_origin}{A character vector specifying the origin municipalities to retrieve. If not provided, all origin municipalities will be included. Valid municipality IDs can be found in the dataset returned by \code{spod_get_zones(zones = "muni", ver = 2)}.}

\item{id_destination}{A character vector specifying the target municipalities to retrieve. If not provided, all target municipalities will be included. Valid municipality IDs can be found in the dataset returned by \code{spod_get_zones(zones = "muni", ver = 2)}.}

\item{min_trips}{A numeric value specifying the minimum number of journeys per origin-destination pair to retrieve. Defaults to 100 to reduce the amount of data returned. Can be set to 0 to retrieve all data.}
}
\value{
A \code{tibble} containing the flows for the specified date, minimum number of journeys, distances and origin-destination pairs if specified.
}
\description{
Internal function to query the GraphQL API for origin-destination data
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/quick-get.R
\name{spod_quick_get_zones}
\alias{spod_quick_get_zones}
\title{Get the municipalities geometries}
\usage{
spod_quick_get_zones(zones = "municipalities")
}
\arguments{
\item{zones}{A character string specifying the zones to retrieve. Valid values are "municipalities", "muni", "municip", and "municipios". Defaults to "municipalities".}
}
\value{
A \code{sf} object with the municipalities geometries to match with the data retrieved with \code{\link[=spod_quick_get_od]{spod_quick_get_od()}}.
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#experimental}{\figure{lifecycle-experimental.svg}{options: alt='[Experimental]'}}}{\strong{[Experimental]}}

This function fetches the municipalities (for now this is the  only option) geometries from the mapas-movilidad website and returns a \code{sf} object with the municipalities geometries. This is intended for use with the flows data retrieved by the \code{\link[=spod_quick_get_od]{spod_quick_get_od()}} function. An interactive web map with this data is available at \url{https://mapas-movilidad.transportes.gob.es/}. These municipality geometries only include Spanish municipalities (and not the NUTS3 regions in Portugal and France) and do not contain extra columns that you can get with the \code{\link[=spod_get_zones]{spod_get_zones()}} function. The function caches the retrieved geometries in memory of the current R session to reduce the number of requests to the mapas-movilidad website.

For detailed zone definitions and methodology, see
\insertRef{mitma_methodology_2020_v3}{spanishoddata} for v1 data and
\insertRef{mitms_methodology_2022_v8}{spanishoddata} for v2 data.
}
\examples{
\dontshow{if (interactive()) withAutoprint(\{ # examplesIf}
\donttest{
municipalities_sf <- spod_quick_get_zones()
}
\dontshow{\}) # examplesIf}
}
\references{
\insertRef{mitms_mobility_web}{spanishoddata}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data-dir.R
\name{spod_set_data_dir}
\alias{spod_set_data_dir}
\title{Set the data directory}
\usage{
spod_set_data_dir(data_dir, quiet = FALSE)
}
\arguments{
\item{data_dir}{The data directory to set.}

\item{quiet}{A \code{logical} value indicating whether to suppress messages. Default is \code{FALSE}.}
}
\value{
Nothing. If quiet is \code{FALSE}, prints a message with the path and confirmation that the path exists.
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#stable}{\figure{lifecycle-stable.svg}{options: alt='[Stable]'}}}{\strong{[Stable]}}

This function sets the data directory in the environment variable SPANISH_OD_DATA_DIR, so that all other functions in the package can access the data. It also creates the directory if it doesn't exist.
}
\examples{
spod_set_data_dir(tempdir())

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get-zones.R
\name{spod_clean_zones_v2}
\alias{spod_clean_zones_v2}
\title{Fixes common issues in the zones data and cleans up variable names}
\usage{
spod_clean_zones_v2(zones_path)
}
\arguments{
\item{zones_path}{The path to the zones spatial data file.}
}
\value{
A spatial `sf`` object containing the cleaned zones data.
}
\description{
This function fixes any invalid geometries in the zones data and renames the "ID" column to "id". It also attacches the population counts and zone names provided in the csv files supplied by the original data provider.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/internal-utils.R
\name{spod_match_data_type_for_local_folders}
\alias{spod_match_data_type_for_local_folders}
\title{Match data types to folders}
\usage{
spod_match_data_type_for_local_folders(
  type = c("od", "origin-destination", "os", "overnight_stays", "nt", "number_of_trips"),
  ver = c(1, 2)
)
}
\arguments{
\item{ver}{Integer. Can be 1 or 2. The version of the data to use. v1 spans 2020-2021, v2 covers 2022 and onwards. See more details in codebooks with \code{\link[=spod_codebook]{spod_codebook()}}.}
}
\value{
A \code{character} string with the folder name for the specified data type. Or \code{NULL} if the data type is not recognized.
}
\description{
Match data types to folders
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/internal-utils.R
\name{spod_zone_names_en2es}
\alias{spod_zone_names_en2es}
\title{Translate zone names from English to Spanish}
\usage{
spod_zone_names_en2es(
  zones = c("districts", "dist", "distr", "distritos", "municipalities", "muni",
    "municip", "municipios", "lua", "large_urban_areas", "gau", "grandes_areas_urbanas")
)
}
\arguments{
\item{zones}{The zones for which to download the data. Can be \code{"districts"} (or \code{"dist"}, \code{"distr"}, or the original Spanish \code{"distritos"}) or \code{"municipalities"} (or \code{"muni"}, \code{"municip"}, or the original Spanish \code{"municipios"}) for both data versions. Additionaly, these can be \code{"large_urban_areas"} (or \code{"lua"}, or the original Spanish \code{"grandes_areas_urbanas"}, or \code{"gau"}) for v2 data (2022 onwards).}
}
\value{
A \code{character} string with the translated zone name. Or \code{NULL} if the zone name is not recognized.
}
\description{
Translate zone names from English to Spanish
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/download_data.R
\name{spod_download}
\alias{spod_download}
\title{Download the data files of specified type, zones, and dates}
\usage{
spod_download(
  type = c("od", "origin-destination", "os", "overnight_stays", "nt", "number_of_trips"),
  zones = c("districts", "dist", "distr", "distritos", "municipalities", "muni",
    "municip", "municipios", "lua", "large_urban_areas", "gau", "grandes_areas_urbanas"),
  dates = NULL,
  max_download_size_gb = 1,
  data_dir = spod_get_data_dir(),
  quiet = FALSE,
  return_local_file_paths = FALSE,
  ignore_missing_dates = FALSE,
  check_local_files = TRUE
)
}
\arguments{
\item{type}{The type of data to download. Can be \code{"origin-destination"} (or ust \code{"od"}), or \code{"number_of_trips"} (or just \code{"nt"}) for v1 data. For v2 data \code{"overnight_stays"} (or just \code{"os"}) is also available. More data types to be supported in the future. See codebooks for v1 and v2 data in vignettes with \code{spod_codebook(1)} and \code{spod_codebook(2)} (\link{spod_codebook}).}

\item{zones}{The zones for which to download the data. Can be \code{"districts"} (or \code{"dist"}, \code{"distr"}, or the original Spanish \code{"distritos"}) or \code{"municipalities"} (or \code{"muni"}, \code{"municip"}, or the original Spanish \code{"municipios"}) for both data versions. Additionaly, these can be \code{"large_urban_areas"} (or \code{"lua"}, or the original Spanish \code{"grandes_areas_urbanas"}, or \code{"gau"}) for v2 data (2022 onwards).}

\item{dates}{A \code{character} or \code{Date} vector of dates to process. Kindly keep in mind that v1 and v2 data follow different data collection methodologies and may not be directly comparable. Therefore, do not try to request data from both versions for the same date range. If you need to compare data from both versions, please refer to the respective codebooks and methodology documents. The v1 data covers the period from 2020-02-14 to 2021-05-09, and the v2 data covers the period from 2022-01-01 to the present until further notice. The true dates range is checked against the available data for each version on every function run.

The possible values can be any of the following:
\itemize{
\item For the \code{spod_get()} and \code{spod_convert()} functions, the \code{dates} can be set to "cached_v1" or "cached_v2" to request data from cached (already previously downloaded) v1 (2020-2021) or v2 (2022 onwards) data. In this case, the function will identify and use all data files that have been downloaded and cached locally, (e.g. using an explicit run of \code{spod_download()}, or any data requests made using the \code{spod_get()} or \code{spod_convert()} functions).
\item A single date in ISO (YYYY-MM-DD) or YYYYMMDD format. \code{character} or \code{Date} object.
\item A vector of dates in ISO (YYYY-MM-DD) or YYYYMMDD format. \code{character} or \code{Date} object. Can be any non-consecutive sequence of dates.
\item A date range
\itemize{
\item eigher a \code{character} or \code{Date} object of length 2 with clearly named elements \code{start} and \code{end} in ISO (YYYY-MM-DD) or YYYYMMDD format. E.g. \code{c(start = "2020-02-15", end = "2020-02-17")};
\item or a \code{character} object of the form \code{YYYY-MM-DD_YYYY-MM-DD} or \code{YYYYMMDD_YYYYMMDD}. For example, \verb{2020-02-15_2020-02-17} or \verb{20200215_20200217}.
}
\item A regular expression to match dates in the format \code{YYYYMMDD}. \code{character} object. For example, \verb{^202002} will match all dates in February 2020.
}}

\item{max_download_size_gb}{The maximum download size in gigabytes. Defaults to 1.}

\item{data_dir}{The directory where the data is stored. Defaults to the value returned by \code{spod_get_data_dir()} which returns the value of the environment variable \code{SPANISH_OD_DATA_DIR} or a temporary directory if the variable is not set. To set the data directory, use \link{spod_set_data_dir}.}

\item{quiet}{A \code{logical} value indicating whether to suppress messages. Default is \code{FALSE}.}

\item{return_local_file_paths}{Logical. If \code{TRUE}, the function returns a character vector of the paths to the downloaded files. If \code{FALSE}, the function returns \code{NULL}.}

\item{ignore_missing_dates}{Logical. If \code{TRUE}, the function will not raise an error if the some of the specified dates are missing. Any dates that are missing will be skipped, however the data for any valid dates will be acquired. Defaults to \code{FALSE}.}

\item{check_local_files}{Logical. Whether to check the file size of local files against known remote file sizes on the Amazon S3 storage. Defaults to \code{TRUE}, which fetches the metadata from Amazon S3. This setting ensures your downloaded files are not broken, so it is recommended to keep it \code{TRUE}.}
}
\value{
Nothing. If \code{return_local_file_paths = TRUE}, a \code{character} vector of the paths to the downloaded files.
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#stable}{\figure{lifecycle-stable.svg}{options: alt='[Stable]'}}}{\strong{[Stable]}}

This function downloads the data files of the specified type, zones, dates and data version.

For detailed data descriptions, see package vignettes using \code{\link[=spod_codebook]{spod_codebook(ver = 1)}} and \code{\link[=spod_codebook]{spod_codebook(ver = 2)}} and official methodology documents in \strong{References} section.
}
\details{
Download the data files of specified type, zones, and dates
}
\examples{
\dontshow{if (interactive()) withAutoprint(\{ # examplesIf}
\donttest{

# Set data dir for file downloads
spod_set_data_dir(tempdir())

# Download the number of trips on district level for the a date range in March 2020
spod_download(
  type = "number_of_trips", zones = "districts",
  dates = c(start = "2020-03-20", end = "2020-03-21")
)

# Download the number of trips on district level for select dates in 2020 and 2021
spod_download(
  type = "number_of_trips", zones = "dist",
  dates = c("2020-03-20", "2020-03-24", "2021-03-20", "2021-03-24")
)

# Download the number of trips on municipality level using regex for a date range in March 2020
# (the regex will capture the dates 2020-03-20 to 2020-03-24)
spod_download(
  type = "number_of_trips", zones = "municip",
  dates = "2020032[0-4]"
)
}
\dontshow{\}) # examplesIf}
}
\references{
\itemize{
\item \strong{For the official website of the mobility study}: \insertRef{mitms_mobility_web}{spanishoddata}
\item \strong{For v1 data methodology}: \insertRef{mitma_methodology_2020_v3}{spanishoddata}
\item \strong{For v2 data methodology}: \insertRef{mitms_methodology_2022_v8}{spanishoddata}
\item \strong{For the \code{spanishoddata} R package}: \insertRef{spanishoddata-r-pkg}{spanishoddata}
}

Use \code{\link[=spod_cite]{spod_cite()}} to cite the package and the data with correct plain text, markdown, or BibTeX formats.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/check-files.R
\name{spod_compute_s3_etag}
\alias{spod_compute_s3_etag}
\title{Compute ETag for a file}
\usage{
spod_compute_s3_etag(file_path, part_size = 8 * 1024^2)
}
\arguments{
\item{file_path}{Character. The path to the file.}

\item{part_size}{Numeric. The size of each part in bytes. Do not change, as this is a default for S3 Etag.}
}
\value{
Character. The ETag for the file.
}
\description{
Compute ETag for a file
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/convert.R
\name{spod_convert}
\alias{spod_convert}
\title{Convert data from plain text to duckdb or parquet format}
\usage{
spod_convert(
  type = c("od", "origin-destination", "os", "overnight_stays", "nt", "number_of_trips"),
  zones = c("districts", "dist", "distr", "distritos", "municipalities", "muni",
    "municip", "municipios"),
  dates = NULL,
  save_format = "duckdb",
  save_path = NULL,
  overwrite = FALSE,
  data_dir = spod_get_data_dir(),
  quiet = FALSE,
  max_mem_gb = NULL,
  max_n_cpu = max(1, parallelly::availableCores() - 1),
  max_download_size_gb = 1,
  ignore_missing_dates = FALSE
)
}
\arguments{
\item{type}{The type of data to download. Can be \code{"origin-destination"} (or ust \code{"od"}), or \code{"number_of_trips"} (or just \code{"nt"}) for v1 data. For v2 data \code{"overnight_stays"} (or just \code{"os"}) is also available. More data types to be supported in the future. See codebooks for v1 and v2 data in vignettes with \code{spod_codebook(1)} and \code{spod_codebook(2)} (\link{spod_codebook}).}

\item{zones}{The zones for which to download the data. Can be \code{"districts"} (or \code{"dist"}, \code{"distr"}, or the original Spanish \code{"distritos"}) or \code{"municipalities"} (or \code{"muni"}, \code{"municip"}, or the original Spanish \code{"municipios"}) for both data versions. Additionaly, these can be \code{"large_urban_areas"} (or \code{"lua"}, or the original Spanish \code{"grandes_areas_urbanas"}, or \code{"gau"}) for v2 data (2022 onwards).}

\item{dates}{A \code{character} or \code{Date} vector of dates to process. Kindly keep in mind that v1 and v2 data follow different data collection methodologies and may not be directly comparable. Therefore, do not try to request data from both versions for the same date range. If you need to compare data from both versions, please refer to the respective codebooks and methodology documents. The v1 data covers the period from 2020-02-14 to 2021-05-09, and the v2 data covers the period from 2022-01-01 to the present until further notice. The true dates range is checked against the available data for each version on every function run.

The possible values can be any of the following:
\itemize{
\item For the \code{spod_get()} and \code{spod_convert()} functions, the \code{dates} can be set to "cached_v1" or "cached_v2" to request data from cached (already previously downloaded) v1 (2020-2021) or v2 (2022 onwards) data. In this case, the function will identify and use all data files that have been downloaded and cached locally, (e.g. using an explicit run of \code{spod_download()}, or any data requests made using the \code{spod_get()} or \code{spod_convert()} functions).
\item A single date in ISO (YYYY-MM-DD) or YYYYMMDD format. \code{character} or \code{Date} object.
\item A vector of dates in ISO (YYYY-MM-DD) or YYYYMMDD format. \code{character} or \code{Date} object. Can be any non-consecutive sequence of dates.
\item A date range
\itemize{
\item eigher a \code{character} or \code{Date} object of length 2 with clearly named elements \code{start} and \code{end} in ISO (YYYY-MM-DD) or YYYYMMDD format. E.g. \code{c(start = "2020-02-15", end = "2020-02-17")};
\item or a \code{character} object of the form \code{YYYY-MM-DD_YYYY-MM-DD} or \code{YYYYMMDD_YYYYMMDD}. For example, \verb{2020-02-15_2020-02-17} or \verb{20200215_20200217}.
}
\item A regular expression to match dates in the format \code{YYYYMMDD}. \code{character} object. For example, \verb{^202002} will match all dates in February 2020.
}}

\item{save_format}{A \code{character} vector of length 1 with values "duckdb" or "parquet". Defaults to "duckdb". If \code{NULL} automatically inferred from the \code{save_path} argument. If only \code{save_format} is provided, \code{save_path} will be set to the default location set in \code{SPANISH_OD_DATA_DIR} environment variable using \link{spod_set_data_dir}\code{(path = 'path/to/your/cache/dir')}. So for v1 data that path would be \verb{<data_dir>/clean_data/v1/tabular/duckdb/} or \verb{<data_dir>/clean_data/v1/tabular/parquet/}.

You can also set \code{save_path}. If it ends with ".duckdb", will save to \code{DuckDB} database format, if \code{save_path} does not end with ".duckdb", will save to \code{parquet} format and will treat the \code{save_path} as a path to a folder, not a file, will create necessary hive-style subdirectories in that folder. Hive style looks like \code{year=2020/month=2/day=14} and inside each such directory there will be a \code{data_0.parquet} file that contains the data for that day.}

\item{save_path}{A \code{character} vector of length 1. The full (not relative) path to a \code{DuckDB} database file or \code{parquet} folder.
\itemize{
\item If \code{save_path} ends with \code{.duckdb}, it will be saved as a DuckDB database file. The format argument will be automatically set to \code{save_format='duckdb'}.
\item If \code{save_path} ends with a folder name (e.g. \verb{/data_dir/clean_data/v1/tabular/parquet/od_distr} for origin-destination data for district level), the data will be saved as a collection of \code{parquet} files in a hive-style directory structure. So the subfolders of \code{od_distr} will be \code{year=2020/month=2/day=14} and inside each of these folders a single \code{parquet} file will be placed containing the data for that day.
\item If \code{NULL}, uses the default location in \code{data_dir} (set by the \code{SPANISH_OD_DATA_DIR} environment variable using \link{spod_set_data_dir}\code{(path = 'path/to/your/cache/dir')}. Therefore, the default relative path for \code{DuckDB} is \verb{<data_dir>/clean_data/v1/tabular/duckdb/<type>_<zones>.duckdb} and for \code{parquet} files is \verb{<data_dir>/clean_data/v1/tabular/parquet/<type>_<zones>/}, where \code{type} is the type of data (e.g. 'od', 'os', 'nt', that correspoind to 'origin-destination', 'overnight-stays', 'number-of-trips', etc.) and \code{zones} is the name of the geographic zones (e.g. 'distr', 'muni', etc.). See the details below in the function arguments description.
}}

\item{overwrite}{A \code{logical} or a \code{character} vector of length 1. If \code{TRUE}, overwrites existing \code{DuckDB} or \code{parquet} files. Defaults to \code{FALSE}. For parquet files can also be set to 'update', so that only parquet files are only created for the dates that have not yet been converted.}

\item{data_dir}{The directory where the data is stored. Defaults to the value returned by \code{spod_get_data_dir()} which returns the value of the environment variable \code{SPANISH_OD_DATA_DIR} or a temporary directory if the variable is not set. To set the data directory, use \link{spod_set_data_dir}.}

\item{quiet}{A \code{logical} value indicating whether to suppress messages. Default is \code{FALSE}.}

\item{max_mem_gb}{\code{integer} value of the maximum operating memory to use in GB. \code{NULL} by default, delegates the choice to the \code{DuckDB} engine which usually sets it to 80\% of available memory. Caution, in HPC use, the amount of memory available to your job may be determined incorrectly by the \code{DuckDB} engine, so it is recommended to set this parameter explicitly according to your job's memory limits.}

\item{max_n_cpu}{The maximum number of threads to use. Defaults to the number of available cores minus 1.}

\item{max_download_size_gb}{The maximum download size in gigabytes. Defaults to 1.}

\item{ignore_missing_dates}{Logical. If \code{TRUE}, the function will not raise an error if the some of the specified dates are missing. Any dates that are missing will be skipped, however the data for any valid dates will be acquired. Defaults to \code{FALSE}.}
}
\value{
Path to saved \code{DuckDB} database file or to a folder with \code{parquet} files in hive-style directory structure.
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#stable}{\figure{lifecycle-stable.svg}{options: alt='[Stable]'}}}{\strong{[Stable]}}

Converts data for faster analysis into either \code{DuckDB} file or into \code{parquet} files in a hive-style directory structure. Running analysis on these files is sometimes 100x times faster than working with raw CSV files, espetially when these are in gzip archives. To connect to converted data, please use 'mydata <- \link{spod_connect}(data_path = path_returned_by_spod_convert)' passing the path to where the data was saved. The connected \code{mydata} can be analysed using \code{dplyr} functions such as \link[dplyr]{select}, \link[dplyr]{filter}, \link[dplyr]{mutate}, \link[dplyr]{group_by}, \link[dplyr]{summarise}, etc. In the end of any sequence of commands you will need to add \link[dplyr]{collect} to execute the whole chain of data manipulations and load the results into memory in an R \code{data.frame}/\code{tibble}. For more in-depth usage of such data, please refer to DuckDB documentation and examples at \url{https://duckdb.org/docs/api/r#dbplyr} . Some more useful examples can be found here \url{https://arrow-user2022.netlify.app/data-wrangling#combining-arrow-with-duckdb} . You may also use \code{arrow} package to work with parquet files \url{https://arrow.apache.org/docs/r/}.

For detailed data descriptions, see package vignettes using \code{\link[=spod_codebook]{spod_codebook(ver = 1)}} and \code{\link[=spod_codebook]{spod_codebook(ver = 2)}} and official methodology documents in \strong{References} section.
}
\examples{
\dontshow{if (interactive()) withAutoprint(\{ # examplesIf}
\donttest{
# Set data dir for file downloads
spod_set_data_dir(tempdir())

# download and convert data
dates_1 <- c(start = "2020-02-17", end = "2020-02-18")
db_2 <- spod_convert(
  type = "number_of_trips",
  zones = "distr",
  dates = dates_1,
  overwrite = TRUE
)

# now connect to the converted data
my_od_data_2 <- spod_connect(db_2)

# disconnect from the database
spod_disconnect(my_od_data_2)
}
\dontshow{\}) # examplesIf}
}
\references{
\itemize{
\item \strong{For the official website of the mobility study}: \insertRef{mitms_mobility_web}{spanishoddata}
\item \strong{For v1 data methodology}: \insertRef{mitma_methodology_2020_v3}{spanishoddata}
\item \strong{For v2 data methodology}: \insertRef{mitms_methodology_2022_v8}{spanishoddata}
\item \strong{For the \code{spanishoddata} R package}: \insertRef{spanishoddata-r-pkg}{spanishoddata}
}

Use \code{\link[=spod_cite]{spod_cite()}} to cite the package and the data with correct plain text, markdown, or BibTeX formats.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/internal-utils.R
\name{spod_get_valid_dates}
\alias{spod_get_valid_dates}
\title{Get valid dates for the specified data version}
\usage{
spod_get_valid_dates(ver = NULL)
}
\arguments{
\item{ver}{Integer. Can be 1 or 2. The version of the data to use. v1 spans 2020-2021, v2 covers 2022 and onwards. See more details in codebooks with \code{\link[=spod_codebook]{spod_codebook()}}.}
}
\value{
A vector of type \code{Date} with all possible valid dates for the specified data version (v1 for 2020-2021 and v2 for 2020 onwards).
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#stable}{\figure{lifecycle-stable.svg}{options: alt='[Stable]'}}}{\strong{[Stable]}}

Get all metadata for requested data version and identify all dates available for download.
}
\examples{
\dontshow{if (interactive()) withAutoprint(\{ # examplesIf}
\donttest{
# Get all valid dates for v1 (2020-2021) data
spod_get_valid_dates(ver = 1)

# Get all valid dates for v2 (2020 onwards) data
spod_get_valid_dates(ver = 2)
}
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dev-tools.R
\name{spod_get_file_size_from_url}
\alias{spod_get_file_size_from_url}
\title{Get file size from URL}
\usage{
spod_get_file_size_from_url(x_url)
}
\arguments{
\item{x_url}{URL}
}
\value{
File size in MB
}
\description{
Get file size from URL
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/internal-utils.R
\name{spod_request_length}
\alias{spod_request_length}
\title{Get the length of the request payload}
\usage{
spod_request_length(graphql_query)
}
\arguments{
\item{graphql_query}{Character. The GraphQL query.}
}
\value{
Numeric. The length of the request payload.
}
\description{
Get the length of the request payload
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/duckdb-helpers.R
\name{spod_duckdb_od}
\alias{spod_duckdb_od}
\title{Creates a duckdb connection to origin-destination data}
\usage{
spod_duckdb_od(
  con = DBI::dbConnect(duckdb::duckdb(), dbdir = ":memory:", read_only = FALSE),
  zones = c("districts", "dist", "distr", "distritos", "municipalities", "muni",
    "municip", "municipios", "lua", "large_urban_areas", "gau", "grandes_areas_urbanas"),
  ver = NULL,
  data_dir = spod_get_data_dir()
)
}
\arguments{
\item{con}{A duckdb connection object. If not specified, a new in-memory connection will be created.}

\item{zones}{The zones for which to download the data. Can be \code{"districts"} (or \code{"dist"}, \code{"distr"}, or the original Spanish \code{"distritos"}) or \code{"municipalities"} (or \code{"muni"}, \code{"municip"}, or the original Spanish \code{"municipios"}) for both data versions. Additionaly, these can be \code{"large_urban_areas"} (or \code{"lua"}, or the original Spanish \code{"grandes_areas_urbanas"}, or \code{"gau"}) for v2 data (2022 onwards).}

\item{ver}{Integer. Can be 1 or 2. The version of the data to use. v1 spans 2020-2021, v2 covers 2022 and onwards. See more details in codebooks with \code{\link[=spod_codebook]{spod_codebook()}}.}

\item{data_dir}{The directory where the data is stored. Defaults to the value returned by \code{spod_get_data_dir()}.}
}
\value{
A \code{duckdb} connection object with 2 views:
\itemize{
\item \code{od_csv_raw} - a raw table view of all cached CSV files with the origin-destination data that has been previously cached in $SPANISH_OD_DATA_DIR
\item \code{od_csv_clean} - a cleaned-up table view of \code{od_csv_raw} with column names and values translated and mapped to English. This still includes all cached data.
}

The structure of the cleaned-up views \code{od_csv_clean} is as follows:

\describe{
\item{date}{\code{Date}. The full date of the trip, including year, month, and day.}
\item{id_origin}{\code{factor}. The identifier for the origin location of the trip, formatted as a code (e.g., '01001_AM').}
\item{id_destination}{\code{factor}. The identifier for the destination location of the trip, formatted as a code (e.g., '01001_AM').}
\item{activity_origin}{\code{factor}. The type of activity at the origin location (e.g., 'home', 'work'). \strong{Note:} Only available for district level data.}
\item{activity_destination}{\code{factor}. The type of activity at the destination location (e.g., 'home', 'other'). \strong{Note:} Only available for district level data.}
\item{residence_province_ine_code}{\code{factor}. The province of residence for the group of individual making the trip, encoded according to the INE classification. \strong{Note:} Only available for district level data.}
\item{residence_province_name}{\code{factor}. The province of residence for the group of individuals making the trip (e.g., 'Cuenca', 'Girona'). \strong{Note:} Only available for district level data.}
\item{hour}{\code{integer}. The time slot (the hour of the day) during which the trip started, represented as an integer (e.g., 0, 1, 2).}
\item{distance}{\code{factor}. The distance category of the trip, represented as a code (e.g., '002-005' for 2-5 km).}
\item{n_trips}{\code{double}. The number of trips taken within the specified time slot and distance.}
\item{trips_total_length_km}{\code{double}. The total length of all trips in kilometers for the specified time slot and distance.}
\item{year}{\code{double}. The year of the trip.}
\item{month}{\code{double}. The month of the trip.}
\item{day}{\code{double}. The day of the trip.}
}

The structure of the original data in \code{od_csv_raw} is as follows:

\describe{
\item{fecha}{\code{Date}. The date of the trip, including year, month, and day.}
\item{origen}{\code{character}. The identifier for the origin location of the trip, formatted as a character string (e.g., '01001_AM').}
\item{destino}{\code{character}. The identifier for the destination location of the trip, formatted as a character string (e.g., '01001_AM').}
\item{actividad_origen}{\code{character}. The type of activity at the origin location (e.g., 'casa', 'trabajo').}
\item{actividad_destino}{\code{character}. The type of activity at the destination location (e.g., 'otros', 'trabajo').}
\item{residencia}{\code{character}. The code representing the residence of the individual making the trip (e.g., '01') according to the official INE classification.}
\item{edad}{\code{character}. The age of the individual making the trip. This data is actaully filled with 'NA' values, which is why this column is removed in the cleaned-up and translated view described above.}
\item{periodo}{\code{integer}. The time period during which the trip started, represented as an integer (e.g., 0, 1, 2).}
\item{distancia}{\code{character}. The distance category of the trip, represented as a character string (e.g., '002-005' for 2-5 km).}
\item{viajes}{\code{double}. The number of trips taken within the specified time period and distance.}
\item{viajes_km}{\code{double}. The total length of all trips in kilometers for the specified time period and distance.}
\item{day}{\code{double}. The day of the trip.}
\item{month}{\code{double}. The month of the trip.}
\item{year}{\code{double}. The year of the trip.}
}
}
\description{
This function creates a duckdb connection to the origin-destination data stored in CSV.gz files.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/quick-get.R
\name{spod_get_hmac_secret_memoised}
\alias{spod_get_hmac_secret_memoised}
\title{Cache the HMAC secret to avoid repeated requests}
\usage{
spod_get_hmac_secret_memoised(
  base_url = "https://mapas-movilidad.transportes.gob.es"
)
}
\description{
Cache the HMAC secret to avoid repeated requests
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/quick-get.R
\name{spod_quick_get_od}
\alias{spod_quick_get_od}
\title{Get daily trip counts per origin-destionation municipality from 2022 onward}
\usage{
spod_quick_get_od(
  date = NA,
  min_trips = 100,
  distances = c("500m-2km", "2-10km", "10-50km", "50+km"),
  id_origin = NA,
  id_destination = NA
)
}
\arguments{
\item{date}{A character or Date object specifying the date for which to retrieve the data. If date is a character, the date must be in "YYYY-MM-DD" or "YYYYMMDD" format.}

\item{min_trips}{A numeric value specifying the minimum number of journeys per origin-destination pair to retrieve. Defaults to 100 to reduce the amount of data returned. Can be set to 0 to retrieve all data.}

\item{distances}{A character vector specifying the distances to retrieve. Valid values are "500m-2km", "2-10km", "10-50km", and "50+km". Defaults to \code{c("500m-2km", "2-10km", "10-50km", "50+km")}. The resulting data will not have number of trips per category of distance. Therefore, if you want to retrieve the number of trips per distance category, you need to make 4 separate calls to this function or use \code{\link[=spod_get]{spod_get()}} instead to get the full data from source CSV files.}

\item{id_origin}{A character vector specifying the origin municipalities to retrieve. If not provided, all origin municipalities will be included. Valid municipality IDs can be found in the dataset returned by \code{spod_get_zones(zones = "muni", ver = 2)}.}

\item{id_destination}{A character vector specifying the target municipalities to retrieve. If not provided, all target municipalities will be included. Valid municipality IDs can be found in the dataset returned by \code{spod_get_zones(zones = "muni", ver = 2)}.}
}
\value{
A \code{tibble} containing the flows for the specified date, minimum number of journeys, distances and origin-destination pairs if specified. The columns are:
\describe{
\item{date}{The date of the trips.}
\item{id_origin}{The origin municipality ID.}
\item{id_destination}{The target municipality ID.}
\item{n_trips}{The number of trips between the origin and target municipality.}
\item{trips_total_length_km}{The total length of trips in kilometers.}
}
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#experimental}{\figure{lifecycle-experimental.svg}{options: alt='[Experimental]'}}}{\strong{[Experimental]}}

\strong{WARNING: this function may stop working at any time, as the API may change}. This function provides a quick way to get daily aggregated (no hourly data) trip counts per origin-destination municipality from v2 data (2022 onward). Compared to \code{\link[=spod_get]{spod_get()}}, which downloads large CSV files, this function downloads the data directly from the GraphQL API. An interactive web map with this data is available at \url{https://mapas-movilidad.transportes.gob.es/}. No data aggregation is performed on your computer (unlike in \code{\link[=spod_get]{spod_get()}}), so you do not need to worry about memory usage and do not have to use a powerful computer with multiple CPU cores just to get this simple data. Only about 1 MB of data is downloaded for a single day. The limitation of this function is that it can only retrieve data for a single day at a time and only with total number of trips and total km travelled. So it is not possible to get any of the extra variables available in the full dataset via \code{\link[=spod_get]{spod_get()}}.

For detailed data descriptions, see package vignettes using \code{\link[=spod_codebook]{spod_codebook(ver = 1)}} and \code{\link[=spod_codebook]{spod_codebook(ver = 2)}} and official methodology documents in \strong{References} section.
}
\examples{
\dontshow{if (interactive()) withAutoprint(\{ # examplesIf}
\donttest{
od_1000 <- spod_quick_get_od(
  date = "2022-01-01",
  min_trips = 1000
)
}
\dontshow{\}) # examplesIf}
}
\references{
\itemize{
\item \strong{For the official website of the mobility study}: \insertRef{mitms_mobility_web}{spanishoddata}
\item \strong{For v1 data methodology}: \insertRef{mitma_methodology_2020_v3}{spanishoddata}
\item \strong{For v2 data methodology}: \insertRef{mitms_methodology_2022_v8}{spanishoddata}
\item \strong{For the \code{spanishoddata} R package}: \insertRef{spanishoddata-r-pkg}{spanishoddata}
}

Use \code{\link[=spod_cite]{spod_cite()}} to cite the package and the data with correct plain text, markdown, or BibTeX formats.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/duckdb-helpers.R
\name{spod_duckdb_limit_resources}
\alias{spod_duckdb_limit_resources}
\title{Set maximum memory and number of threads for a \code{DuckDB} connection}
\usage{
spod_duckdb_limit_resources(
  con,
  max_mem_gb = NULL,
  max_n_cpu = max(1, parallelly::availableCores() - 1)
)
}
\arguments{
\item{con}{A \code{duckdb} connection}

\item{max_mem_gb}{\code{integer} value of the maximum operating memory to use in GB. \code{NULL} by default, delegates the choice to the \code{DuckDB} engine which usually sets it to 80\% of available memory. Caution, in HPC use, the amount of memory available to your job may be determined incorrectly by the \code{DuckDB} engine, so it is recommended to set this parameter explicitly according to your job's memory limits.}

\item{max_n_cpu}{The maximum number of threads to use. Defaults to the number of available cores minus 1.}
}
\value{
A \code{duckdb} connection.
}
\description{
Set maximum memory and number of threads for a \code{DuckDB} connection
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/download_data.R
\name{spod_download_in_batches}
\alias{spod_download_in_batches}
\title{Download multiple files with progress bar in parallel}
\usage{
spod_download_in_batches(
  files_to_download,
  batch_size = 5,
  bar_width = 20,
  chunk_size = 1024 * 1024,
  test_size = 10 * 1024 * 1024,
  max_retries = 3L,
  timeout = 900,
  show_progress = spod_interactive() && !isTRUE(getOption("knitr.in.progress"))
)
}
\arguments{
\item{files_to_download}{A data frame with columns \code{target_url}, \code{local_path} and \code{file_size_bytes}.}

\item{batch_size}{Numeric. Number of files to download at a time.}

\item{bar_width}{Numeric. Width of the progress bar.}

\item{chunk_size}{Numeric. Number of bytes to download at a time for speed test.}

\item{max_retries}{Integer. Maximum number of retries for failed downloads.}

\item{timeout}{Numeric. Timeout in seconds for each download.}

\item{show_progress}{Logical. Whether to show the progress bar.}
}
\value{
A data frame with columns \code{target_url}, \code{local_path}, \code{file_size_bytes} and \code{local_file_size}.
}
\description{
Download multiple files with a progress bar. Retries failed downloads up to 3 times. Downloads are in parallel and in batches to show progress. First 10 Mb of a file is downloaded to check the speed.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get-zones.R
\name{spod_clean_zones_v1}
\alias{spod_clean_zones_v1}
\title{Fixes common issues in the zones data and cleans up variable names}
\usage{
spod_clean_zones_v1(zones_path, zones, quiet = FALSE)
}
\arguments{
\item{zones_path}{The path to the zones spatial data file.}

\item{zones}{The zones for which to download the data. Can be \code{"districts"} (or \code{"dist"}, \code{"distr"}, or the original Spanish \code{"distritos"}) or \code{"municipalities"} (or \code{"muni"}, \code{"municip"}, or the original Spanish \code{"municipios"}) for both data versions. Additionaly, these can be \code{"large_urban_areas"} (or \code{"lua"}, or the original Spanish \code{"grandes_areas_urbanas"}, or \code{"gau"}) for v2 data (2022 onwards).}

\item{quiet}{A \code{logical} value indicating whether to suppress messages. Default is \code{FALSE}.}
}
\value{
A spatial object containing the cleaned zones data.
}
\description{
This function fixes any invalid geometries in the zones data and renames the "ID" column to "id".
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/internal-utils.R
\name{spod_expand_dates_from_regex}
\alias{spod_expand_dates_from_regex}
\title{Function to expand dates from a regex}
\usage{
spod_expand_dates_from_regex(date_regex)
}
\arguments{
\item{date_regex}{A regular expression to match dates in the format 'yyyymmdd'.}
}
\value{
A \code{character} vector of dates matching the regex.
}
\description{
This function generates a sequence of dates from a regular expression pattern based on the provided regular expression.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/connect.R
\name{spod_connect}
\alias{spod_connect}
\title{Connect to data converted to \code{DuckDB} or hive-style \code{parquet} files}
\usage{
spod_connect(
  data_path,
  target_table_name = NULL,
  quiet = FALSE,
  max_mem_gb = NULL,
  max_n_cpu = max(1, parallelly::availableCores() - 1),
  temp_path = spod_get_temp_dir()
)
}
\arguments{
\item{data_path}{a path to the \code{DuckDB} database file with '.duckdb' extension, or a path to the folder with \code{parquet} files. Eigher one should have been created with the \link{spod_convert} function.}

\item{target_table_name}{Default is \code{NULL}. When connecting to a folder of \code{parquet} files, this argument is ignored. When connecting to a \code{DuckDB} database, a \code{character} vector of length 1 with the table name to open from the database file. If not specified, it will be guessed from the \code{data_path} argument and from table names that are available in the database. If you have not manually interfered with the database, this should be guessed automatically and you do not need to specify it.}

\item{quiet}{A \code{logical} value indicating whether to suppress messages. Default is \code{FALSE}.}

\item{max_mem_gb}{\code{integer} value of the maximum operating memory to use in GB. \code{NULL} by default, delegates the choice to the \code{DuckDB} engine which usually sets it to 80\% of available memory. Caution, in HPC use, the amount of memory available to your job may be determined incorrectly by the \code{DuckDB} engine, so it is recommended to set this parameter explicitly according to your job's memory limits.}

\item{max_n_cpu}{The maximum number of threads to use. Defaults to the number of available cores minus 1.}

\item{temp_path}{The path to the temp folder for DuckDB for \href{https://duckdb.org/2024/07/09/memory-management.html#intermediate-spilling}{intermediate spilling} in case the set memory limit and/or physical memory of the computer is too low to perform the query. By default this is set to the \code{temp} directory in the data folder defined by SPANISH_OD_DATA_DIR environment variable (set by \code{\link[=spod_set_data_dir]{spod_set_data_dir()}})). Otherwise, for queries on folders of CSV files or parquet files, the temporary path would be set to the current R working directory, which probably is undesirable, as the current working directory can be on a slow storage, or storage that may have limited space, compared to the data folder.}
}
\value{
a \code{DuckDB} table connection object.
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#stable}{\figure{lifecycle-stable.svg}{options: alt='[Stable]'}}}{\strong{[Stable]}}

This function allows the user to quickly connect to the data converted to DuckDB with the \link{spod_convert} function. This function simplifies the connection process. The user is free to use the \code{DBI} and \code{DuckDB} packages to connect to the data manually, or to use the \code{arrow} package to connect to the \code{parquet} files folder.
}
\examples{
\dontshow{if (interactive()) withAutoprint(\{ # examplesIf}
\donttest{
# Set data dir for file downloads
spod_set_data_dir(tempdir())

# download and convert data
dates_1 <- c(start = "2020-02-17", end = "2020-02-18")
db_2 <- spod_convert(
 type = "number_of_trips",
 zones = "distr",
 dates = dates_1,
 overwrite = TRUE
)

# now connect to the converted data
my_od_data_2 <- spod_connect(db_2)

# disconnect from the database
spod_disconnect(my_od_data_2)
}
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/codebook.R
\name{spod_codebook}
\alias{spod_codebook}
\title{View codebooks for v1 and v2 open mobility data}
\usage{
spod_codebook(ver = 1)
}
\arguments{
\item{ver}{An \code{integer} or \code{numeric} value. The version of the data. Defaults to 1. Can be \code{1} for v1 (2020-2021) data and 2 for v2 (2022 onwards) data.}
}
\value{
Nothing, opens vignette if it is installed. If vignette is missing, prints a message with a link to a webpage with the codebook.
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#stable}{\figure{lifecycle-stable.svg}{options: alt='[Stable]'}}}{\strong{[Stable]}}

Opens relevant vignette with a codebook for v1 (2020-2021) and v2 (2022 onwards) data or provide a webpage if vignette is missing.
}
\examples{

# View codebook for v1 (2020-2021) data
spod_codebook(ver = 1)

# View codebook for v2 (2022 onwards) data
spod_codebook(ver = 2)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/quick-get.R
\name{spod_get_hmac_secret}
\alias{spod_get_hmac_secret}
\title{Get the HMAC secret from the mapas-movilidad website}
\usage{
spod_get_hmac_secret(base_url = "https://mapas-movilidad.transportes.gob.es")
}
\arguments{
\item{base_url}{The base URL of the mapas-movilidad website}
}
\value{
Character vector with the HMAC secret.
}
\description{
Get the HMAC secret from the mapas-movilidad website
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/available-data.R
\name{spod_get_latest_v2_file_list}
\alias{spod_get_latest_v2_file_list}
\title{Get latest file list from the XML for MITMA open mobility data v2 (2022 onwards)}
\usage{
spod_get_latest_v2_file_list(
  data_dir = spod_get_data_dir(),
  xml_url = "https://movilidad-opendata.mitma.es/RSS.xml",
  quiet = FALSE
)
}
\arguments{
\item{data_dir}{The directory where the data is stored. Defaults to the value returned by \code{spod_get_data_dir()}.}

\item{xml_url}{The URL of the XML file to download. Defaults to "https://movilidad-opendata.mitma.es/RSS.xml".}
}
\value{
The path to the downloaded XML file.
}
\description{
Get latest file list from the XML for MITMA open mobility data v2 (2022 onwards)
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/internal-utils.R
\name{spod_is_data_version_overlaps}
\alias{spod_is_data_version_overlaps}
\title{Check if specified dates span both data versions}
\usage{
spod_is_data_version_overlaps(dates)
}
\arguments{
\item{dates}{A \code{Dates} vector of dates to check.}
}
\value{
\code{TRUE} if the dates span both data versions, \code{FALSE} otherwise.
}
\description{
This function checks if the specified dates or date ranges span both v1 and v2 data versions.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dev-tools.R
\name{spod_files_sizes}
\alias{spod_files_sizes}
\title{Get files sizes for remote files of v1 and v2 data and save them into a csv.gz file in the inst/extdata folder.}
\usage{
spod_files_sizes(ver = 2)
}
\arguments{
\item{ver}{The version of the data (1 or 2). Can be both. Defaults to 2, as v1 data is not being updated since 2021.}
}
\value{
Nothing. Only saves a csv.gz file with up to date file sizes in the inst/extdata folder.
}
\description{
Get files sizes for remote files of v1 and v2 data and save them into a csv.gz file in the inst/extdata folder.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/duckdb-helpers.R
\name{spod_read_sql}
\alias{spod_read_sql}
\title{Load an SQL query, glue it, dplyr::sql it}
\usage{
spod_read_sql(sql_file_name)
}
\arguments{
\item{sql_file_name}{The name of the SQL file to load from the package installation directory.}
}
\value{
Text of the SQL query of class \code{sql}/\code{character}.
}
\description{
Load an SQL query from a specified file in package installation directory, glue::collapse it, glue::glue it in case of any variables that need to be replaced, and dplyr::sql it for additional safety.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/check-files.R
\name{spod_filter_available_files}
\alias{spod_filter_available_files}
\title{Filter available data by version, zones, type and dates}
\usage{
spod_filter_available_files(
  available_data,
  ver,
  zones,
  matched_type,
  dates_to_use
)
}
\arguments{
\item{available_data}{Tibble from spod_available_data}

\item{ver}{Integer. Data version (1 or 2)}

\item{zones}{Character. Zone name in Spanish (e.g., "distritos")}

\item{matched_type}{Character. Data type in Spanish (e.g., "viajes")}

\item{dates_to_use}{Date vector. Dates to filter by}
}
\value{
Filtered tibble
}
\description{
Internal helper to filter available_data tibble based on version-specific
grepl patterns. Extracted for testability.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dev-tools.R
\name{spod_store_etags}
\alias{spod_store_etags}
\title{Get Etags for locally saved v1 data files and save them into a RDS file in the inst/extdata folder.}
\usage{
spod_store_etags()
}
\value{
Returns a tibble with the local path, local ETag and remote ETag.
}
\description{
Get Etags for locally saved v1 data files and save them into a RDS file in the inst/extdata folder.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/internal-utils.R
\name{spod_assert_package}
\alias{spod_assert_package}
\title{Checks if a package is installed and \emph{informs} the user if not}
\usage{
spod_assert_package(...)
}
\arguments{
\item{...}{
  Arguments passed on to \code{\link[rlang:is_installed]{rlang::check_installed}}
  \describe{
    \item{\code{pkg}}{The package names. Can include version requirements,
e.g. \code{"pkg (>= 1.0.0)"}.}
    \item{\code{version}}{Minimum versions for \code{pkg}. If supplied, must be the
same length as \code{pkg}. \code{NA} elements stand for any versions.}
    \item{\code{compare}}{A character vector of comparison operators to use
for \code{version}. If supplied, must be the same length as
\code{version}. If \code{NULL}, \code{>=} is used as default for all
elements. \code{NA} elements in \code{compare} are also set to \code{>=} by
default.}
    \item{\code{reason}}{Optional string indicating why is \code{pkg} needed.
Appears in error messages (if non-interactive) and user prompts
(if interactive).}
    \item{\code{action}}{An optional function taking \code{pkg} and \code{...}
arguments. It is called by \code{check_installed()} when the user
chooses to update outdated packages. The function is passed the
missing and outdated packages as a character vector of names.}
    \item{\code{call}}{The execution environment of a currently
running function, e.g. \code{caller_env()}. The function will be
mentioned in error messages as the source of the error. See the
\code{call} argument of \code{\link[rlang:abort]{abort()}} for more information.}
  }}
}
\description{
This is wrapper around \link[rlang:is_installed]{rlang::check_installed};
instead of erroring out if the check fails it returns \code{FALSE}.
However, unlike \link[rlang:is_installed]{rlang::is_installed}, it emits a message to the user.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/internal-utils.R
\name{spod_unique_separated_ids}
\alias{spod_unique_separated_ids}
\title{Remove duplicate values in a semicolon-separated string}
\usage{
spod_unique_separated_ids(column)
}
\arguments{
\item{column}{A \code{character} vector column in a data frame to remove duplicates from.}
}
\value{
A \code{character} vector with semicolon-separated unique IDs.
}
\description{
Remove duplicate IDs in a semicolon-separated string in a selected column in a data frame
}
\keyword{internal}
